unit AIEditor;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ToolWin, ExtCtrls, Grids, IniFiles, Menus,
  Buttons, Math,

  ARUtils, AInterpreter, HMemo, FindText, ReplaceText;

type

  TEditorDialog = class(TCustomEditorDialog)
    ToolBar1: TToolBar;
    StatusBar: TStatusBar;
    ToolButton2: TToolButton;
    OpenDlg: TOpenDialog;
    SaveDlg: TSaveDialog;
    MainMenu1: TMainMenu;
    N1: TMenuItem;
    N2: TMenuItem;
    N3: TMenuItem;
    N4: TMenuItem;
    N5: TMenuItem;
    N6: TMenuItem;
    N7: TMenuItem;
    NewBtn: TSpeedButton;
    OpenBtn: TSpeedButton;
    SaveBtn: TSpeedButton;
    SaveAsBtn: TSpeedButton;
    ToolButton1: TToolButton;
    RunBtn: TSpeedButton;
    StepBtn: TSpeedButton;
    Run2: TMenuItem;
    ToolButton3: TToolButton;
    FontDialog: TFontDialog;
    PC: TPageControl;
    Saveall1: TMenuItem;
    PCPopupMenu: TPopupMenu;
    Closepage1: TMenuItem;
    Nextpage1: TMenuItem;
    Priorpage1: TMenuItem;
    N8: TMenuItem;
    N9: TMenuItem;
    Find1: TMenuItem;
    Findmore1: TMenuItem;
    Replace1: TMenuItem;
    ShowErrorCB: TCheckBox;
    CompileWithDebugInfoCB: TCheckBox;
    Step1: TMenuItem;
    RuntoCursor1: TMenuItem;
    Reset1: TMenuItem;
    Splitter1: TSplitter;
    View1: TMenuItem;
    WatchList1: TMenuItem;
    Splitter2: TSplitter;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    MsgPanel: TPanel;
    ToolBar2: TToolBar;
    ClearAllBtn: TSpeedButton;
    InfoBtn: TSpeedButton;
    MsgMemo: TMemo;
    WatchPanel: TPanel;
    WL: TListView;
    ResetBtn: TSpeedButton;
    ShowWatchBtn: TSpeedButton;
    ShowMsgBtn: TSpeedButton;
    Messages1: TMenuItem;
    procedure RunBtnClick(Sender: TObject);
    procedure NewBtnClick(Sender: TObject);
    procedure OpenBtnClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);override;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);override;
    procedure FormShow(Sender: TObject);override;
    procedure SaveBtnClick(Sender: TObject);
    procedure SaveAsBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);override;
    procedure StepBtnClick(Sender: TObject);
    procedure PCChange(Sender: TObject);
    procedure Saveall1Click(Sender: TObject);
    procedure Closepage1Click(Sender: TObject);
    procedure Priorpage1Click(Sender: TObject);
    procedure Nextpage1Click(Sender: TObject);
    procedure Find1Click(Sender: TObject);
    procedure Findmore1Click(Sender: TObject);
    procedure Replace1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Reset1Click(Sender: TObject);
    procedure ShowWatchBtnClick(Sender: TObject);
    procedure ShowMsgBtnClick(Sender: TObject);
    procedure ClearAllBtnClick(Sender: TObject);
    procedure InfoBtnClick(Sender: TObject);
  private
    FFileName   :string;
    FileList    :TAIStringList;

    FindTextDlg  :TFindTextDlg;
    ReplaceTextDlg:TReplaceTextDlg;

    function    CloseAllText:boolean;
    function    ClosePage(AFileName:string):boolean;
    procedure   SetModified(status:boolean);

    function    SaveText(QueryToSave:boolean):boolean;
    function    SaveAllText(QueryToSave:boolean):boolean;

    procedure   ShowPage(AFileName:string);
    procedure   PMChange(Sender: TObject);
    procedure   PMChangeStatus(Sender: TObject);
    procedure   ShowModified;
    procedure   SetFileName(AFileName:string);
    function    GetPM:THMemo;
    procedure   IntShowMsg(msg:string);
  protected
    FFormStyle   :TFormStyle;
    procedure    CreateParams(var Params: TCreateParams);override;
  public
    Interpret   :TAInterpreter;
    property    PM:THMemo    read  GetPM;

    property    FileName:string read FFileName write SetFileName;

    procedure   ShowErrorPos(AFileName:string;APos:integer);override;
    procedure   ShowStopPos(AFileName:string;APos:integer);override;
    procedure   ClearMsg;override;
    procedure   ShowMessage(msg:string);override;
    procedure   ShowError(msg:string);override;
    procedure   UpdateWatches;override;

    procedure   LoadFile(AFileName:string);override;
    procedure   NewText;
    procedure   RunScript;

    constructor Create(AOwner:TComponent; AInterpret:TAInterpreter);reintroduce;
  end;

implementation

{$R *.DFM}

constructor TEditorDialog.Create(AOwner:TComponent; AInterpret:TAInterpreter);
begin
  if (AOwner is TForm) and (TForm(AOwner).FormStyle <> fsMDIForm) then begin
    FFormStyle:=fsNormal;
  end else begin
    FFormStyle:=fsMDIChild;
  end;

  inherited Create(AOwner);

  Interpret :=AInterpret;
  FileName  :='No name';
end;

procedure   TEditorDialog.CreateParams(var Params: TCreateParams);
begin
  inherited;
  FormStyle:=FFormStyle;
end;

function    TEditorDialog.GetPM:THMemo;
var idx:integer;
begin
  idx:=FileList.IndexOf(FileName);
  if idx <> -1 then begin
    Result:=THMemo(FileList.Objects[idx]);
  end else begin
    Result:=nil;
  end;
end;

procedure   TEditorDialog.SetFileName(AFileName:string);
begin
  FFileName:=AFileName;

  if FileName = '' then begin
    Caption:='No_name';
  end else begin
    Caption:=FFileName;
  end;
  if Assigned(PM) then TTabSheet(PM.Owner).Caption:=Caption;
end;

procedure   TEditorDialog.LoadFile(AFileName:string);
begin
  ShowPage(AFileName);
end;

procedure   TEditorDialog.NewText;
begin
  ShowPage('No name');
end;

procedure   TEditorDialog.ShowPage(AFileName:string);
var ps:TTabSheet;
    newpm:THMemo;
    i:integer;
begin
  FileName:=AFileName;

  if not Assigned(PM) then begin
    ps:=TTabSheet.Create(PC);
    ps.PageControl:=PC;
    ps.Caption:=AFileName;
    PC.ActivePage:=ps;

    newpm:=THMemo.Create(ps);
    newpm.Parent:=ps;
    newpm.Align:=alClient;
    newpm.ScrollBars:=ssBoth;

    newpm.Font.Name:='Courier';
    newpm.Font.Size:=10;
    newpm.OnChange:=PMChange;
    newpm.OnChangeStatus:=PMChangeStatus;
    newpm.TabStops:=2;

    for i:=0 to Interpret.Words.KeyWordList.Count - 1 do begin
      newpm.Keywords.Add(Interpret.Words.KeyWordList[i]);
    end;

    FileList.AddObject(FileName, newpm);
    try
      try
        newpm.Lines.LoadFromStream(Interpret.FileSystem.CreateReadStream(AFileName));
        newpm.Modified:=false;
        SetModified(false);
      finally
        Interpret.FileSystem.CloseReadStream;
      end;
    except
    end;
  end;
  PC.ActivePage:=TTabSheet(PM.Owner);
  PM.SetFocus;
end;

procedure   TEditorDialog.SetModified(status:boolean);
begin
  if Assigned(PM) then begin
    PM.Modified:=status;
    ShowModified;
  end;
end;

function    TEditorDialog.SaveText(QueryToSave:boolean):boolean;
var idx,res:integer;
begin
  Result:=true;
  if CT(FileName, 'No name') then begin
    case MessageDlg('Save changes to ' + FileName + ' ?',mtConfirmation, [mbYes, mbNo, mbCancel], 0) of
      mrYes:  if SaveDlg.Execute then begin
                idx:=FileList.IndexOf(FileName);
                FileList[idx]:=SaveDlg.FileName;
                FileName:=SaveDlg.FileName;
                try
                  PM.Lines.SaveToStream(Interpret.FileSystem.CreateWriteStream(FileName));
                  SetModified(false);
                finally
                  Interpret.FileSystem.CloseWriteStream;
                end;
              end else begin
                Result:=false;
              end;
      mrCancel: begin
                  Result:=false;
                  exit;
                end;
    end;
  end else begin
    if Assigned(PM) then begin
      if PM.Modified then begin
        if QueryToSave then begin
          res:=MessageDlg('Save changes to ' + FileName + ' ?',mtConfirmation, [mbYes, mbNo, mbCancel],0);
          case res of
            mrYes   : begin
                        try
                          PM.Lines.SaveToStream(Interpret.FileSystem.CreateWriteStream(FileName));
                          SetModified(false);
                        finally
                          Interpret.FileSystem.CloseWriteStream;
                        end;
                      end;
            mrCancel: begin
                        Result:=false;
                        exit;
                      end;
          end;
        end else begin
          try
            PM.Lines.SaveToStream(Interpret.FileSystem.CreateWriteStream(FileName));
            SetModified(false);
          finally
            Interpret.FileSystem.CloseWriteStream;
          end;
        end;
      end;
    end;
  end;
end;

function    TEditorDialog.ClosePage(AFileName:string):boolean;
var idx:integer;
begin
  FileName:=AFileName;
  if SaveText(true) then begin
    PM.Owner.Free;
    idx:=FileList.IndexOf(FileName);
    if idx <> -1 then begin
      FileList.Delete(idx);
      if FileList.Count > 0 then begin
        FileName:=FileList[Min(idx, FileList.Count - 1)];
        ShowPage(FileName);
      end else begin
        Close;
      end;
    end;
    Result:=true;
  end else begin
    Result:=false;
  end;
end;

function    TEditorDialog.SaveAllText(QueryToSave:boolean):boolean;
var res:integer;
    i:integer;
    allclose:boolean;
    oldfilename:string;
begin
  allclose:=false;
  Result:=true;

  oldfilename:=FileName;
  try
    for i:=0 to FileList.Count - 1 do begin
      FileName:=FileList[i];
      if PM.Modified then begin
        if (not allclose) and QueryToSave then begin
          res:=MessageDlg('Save changes to ' + FileName + ' ?',mtConfirmation, [mbYes, mbNo, mbAll, mbCancel],0);
        end else begin
          res:=mrYes;
        end;
        case res of
          mrYes   : begin
                      SaveText(false);
                    end;
          mrAll   : begin
                      SaveText(false);
                      allclose:=true;
                    end;
          mrCancel: begin
                      Result:=false;
                      exit;
                    end;
        end;
      end;
    end;
  finally
    FileName:=oldfilename;
  end;
end;

function    TEditorDialog.CloseAllText:boolean;
begin
  if SaveAllText(true) then begin
    while FileList.Count > 0 do ClosePage(FileList[0]);
    Result:=true;
  end else begin
    Result:=false;
  end;
end;

procedure   TEditorDialog.NewBtnClick(Sender: TObject);
begin
  NewText;
end;

procedure   TEditorDialog.OpenBtnClick(Sender: TObject);
var i:integer;
begin
//  Interpret.ShowLoadFileDialog(fname, 'Scripts (*.for)|*.for|Text files (*.txt)|*.txt|All files (*.*)|*.*');
  Interpret.FileSystem.FileName :=FileName;
  Interpret.FileSystem.Options:=Interpret.FileSystem.Options + [ofAllowMultiSelect];
  Interpret.FileSystem.Filter   :='Scripts (*.for)|*.for|Text files (*.txt)|*.txt|All files (*.*)|*.*';

  if Interpret.FileSystem.OpenDialogExecute then begin
    for i:=0 to Interpret.FileSystem.Files.Count - 1 do begin
      LoadFile(Interpret.FileSystem.Files[i]);
    end;
  end;
end;

procedure   TEditorDialog.SaveBtnClick(Sender: TObject);
begin
  SaveText(false);
end;

procedure   TEditorDialog.SaveAsBtnClick(Sender: TObject);
var idx:integer;
begin
  Interpret.FileSystem.FileName :=FileName;
  Interpret.FileSystem.Filter   :='Scripts (*.for)|*.for|Text files (*.txt)|*.txt|All files (*.*)|*.*';

  if Interpret.FileSystem.SaveDialogExecute then begin
    idx:=FileList.IndexOf(FileName);
    if idx <> -1 then begin
      FileList[idx]:=Interpret.FileSystem.FileName;
    end;
    FileName:=Interpret.FileSystem.FileName;

//    Interpret.SaveFile(FileName, PM.Lines);
    try
      PM.Lines.SaveToStream(Interpret.FileSystem.CreateWriteStream(FileName));
      SetModified(false);
    finally
      Interpret.FileSystem.CloseWriteStream;
    end;
  end;
end;

procedure   TEditorDialog.PMChange(Sender: TObject);
begin
  SetModified(true);
end;

procedure   TEditorDialog.PMChangeStatus(Sender: TObject);
begin
  StatusBar.Panels[0].Text:=IntToStr(PM.CurX + 1);
  StatusBar.Panels[1].Text:=IntToStr(PM.CurY + 1);
end;

procedure   TEditorDialog.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose:=CloseAllText;
end;

procedure   TEditorDialog.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action:=caFree;
end;

procedure   TEditorDialog.FormShow(Sender: TObject);
begin
  FileList:=TAIStringList.Create('TEditorDialog.FileList');

  RestorePlace(Self, Self.ClassName);
end;

procedure   TEditorDialog.FormDestroy(Sender: TObject);
begin
  SavePlace(Self, Self.ClassName);
  Interpret.EditorDialog:=nil;
  FileList.Free;
end;

procedure   TEditorDialog.Saveall1Click(Sender: TObject);
begin
  SaveAllText(false);
end;
                                                      
procedure   TEditorDialog.ShowModified;
begin
  if PM.Modified then begin
    StatusBar.Panels[2].Text:='Modified';
  end else begin
    StatusBar.Panels[2].Text:='';
  end;
end;

procedure   TEditorDialog.PCChange(Sender: TObject);
begin
  FileName:=PC.ActivePage.Caption;
  ShowModified;
end;

procedure   TEditorDialog.Closepage1Click(Sender: TObject);
begin
  ClosePage(FileName);
end;

procedure   TEditorDialog.Priorpage1Click(Sender: TObject);
begin
  PC.SelectNextPage(true);
end;

procedure   TEditorDialog.Nextpage1Click(Sender: TObject);
begin
  PC.SelectNextPage(false);
end;

procedure   TEditorDialog.Find1Click(Sender: TObject);
var oldx,oldy:integer;
begin
  FindTextDlg.TextToFind:=PM.GetWordAt(PM.CurX, PM.CurY);

  if FindTextDlg.Execute then begin
    oldx:=PM.CurX;
    oldy:=PM.CurY;
    PM.CurX:=0;
    PM.CurY:=0;
    if not PM.FindText(FindTextDlg.TextToFind, FindTextDlg.MatchCase, FindTextDlg.WholeWord) then begin
      PM.CurX:=oldx;
      PM.CurY:=oldy;
    end;
  end;
end;

procedure   TEditorDialog.Findmore1Click(Sender: TObject);
begin
  if Length(FindTextDlg.TextToFind) > 0 then begin
    PM.FindText(FindTextDlg.TextToFind, FindTextDlg.MatchCase, FindTextDlg.WholeWord);
  end;
end;

procedure   TEditorDialog.Replace1Click(Sender: TObject);
var QueryToReplace:boolean;
    page:integer;
    lastfile:string;
    replacecnt:integer;
begin
  QueryToReplace:=true;
  ReplaceTextDlg.TextToFind:=PM.GetWordAt(PM.CurX, PM.CurY);

  if ReplaceTextDlg.Execute then begin
    PM.CurX:=0; PM.CurY:=0;

    case ReplaceTextDlg.ModalResult of
      mrOK:  begin
              if PM.FindText(ReplaceTextDlg.TextToFind, ReplaceTextDlg.MatchCase, ReplaceTextDlg.WholeWord) then begin
                PM.SelText:=ReplaceTextDlg.TextToReplace;
              end;
            end;
      mrAll:begin
              if not ReplaceTextDlg.AllFiles then begin
                while PM.FindText(ReplaceTextDlg.TextToFind, ReplaceTextDlg.MatchCase, ReplaceTextDlg.WholeWord) do begin
                  if QueryToReplace and ReplaceTextDlg.Prompt then begin
                    case MessageDlg('Replace this occurence of "' + ReplaceTextDlg.TextToFind + '"',
                                    mtConfirmation,
                                    [mbYes, mbNo, mbCancel, mbAll],
                                    0) of
                      mrYes:    PM.SelText:=ReplaceTextDlg.TextToReplace;
                      mrNo:      ;
                      mrCancel:  break;
                      mrAll:    begin
                                  QueryToReplace:=false;
                                  PM.SelText:=ReplaceTextDlg.TextToReplace;
                                end;

                    end;
                  end else begin
                    PM.SelText:=ReplaceTextDlg.TextToReplace;
                  end;
                end;
              end else begin
                replacecnt:=0;
                lastfile:=FileName;
                for page:=0 to FileList.Count - 1 do begin
                  ShowPage(FileList[page]);
                  while PM.FindText(ReplaceTextDlg.TextToFind, ReplaceTextDlg.MatchCase, ReplaceTextDlg.WholeWord) do begin
                    PM.SelText:=ReplaceTextDlg.TextToReplace;
                    inc(replacecnt);
                  end;
                end;
                ShowPage(lastfile);
                if replacecnt > 0 then begin
                  MessageDlg(Format('To be %d replaces', [replacecnt]), mtInformation, [mbOK], 0);
                end;
              end;
            end;
    end;
  end;
end;

procedure   TEditorDialog.FormCreate(Sender: TObject);
begin
  FindTextDlg    :=TFindTextDlg.Create(Self);
  ReplaceTextDlg:=TReplaceTextDlg.Create(Self);
end;

procedure   TEditorDialog.RunBtnClick(Sender: TObject);
begin
  Self.Interpret.CompileWithDebugInfo:=CompileWithDebugInfoCB.Checked;
  Self.Interpret.StepOnRun:=false;

  SaveAllText(false);
  RunScript;
end;

procedure   TEditorDialog.StepBtnClick(Sender: TObject);
begin
  Self.Interpret.CompileWithDebugInfo:=CompileWithDebugInfoCB.Checked;
  Self.Interpret.StepOnRun:=true;

  SaveAllText(false);
  RunScript;
end;

procedure   TEditorDialog.RunScript;
begin
  Self.Interpret.RunScript(nil, PChar(PM.Text));
end;

procedure   TEditorDialog.ShowErrorPos(AFileName:string;APos:integer);
begin
  if ShowErrorCB.Checked then begin
    ShowPage(AFileName);

    if APos >= 0 then begin
      PM.SelStart:=APos;
    end else begin
      PM.CurX:=0;
      PM.CurY:=-APos;
    end;
    PM.SetColorLine(PM.CurY, clRed);
  end;
end;

procedure TEditorDialog.Reset1Click(Sender: TObject);
begin
  Self.Interpret.Reset;
end;

procedure   TEditorDialog.ShowStopPos(AFileName:string;APos:integer);
begin
  PM.CurX:=0;
  PM.CurY:=APos;
  PM.SetColorLine(APos, clBlue);
  UpdateWatches;
end;

procedure   TEditorDialog.ClearMsg;
begin
  MsgMemo.Clear;
end;

procedure   TEditorDialog.IntShowMsg(msg:string);
begin
  MsgMemo.Lines.Add(msg);
  while MsgMemo.Lines.Count > 100000 do begin
    MsgMemo.Lines.Delete(0);
    if (MsgMemo.Lines.Count mod 10000) = 0 then Application.ProcessMessages;
  end;
  Application.ProcessMessages;
end;

procedure   TEditorDialog.ShowMessage(msg:string);
begin
  if Visible then begin
    IntShowMsg(msg);
  end;
end;
procedure   TEditorDialog.ShowError(msg:string);
begin
  if Visible then begin
    IntShowMsg(msg);
  end;
end;

procedure TEditorDialog.ShowWatchBtnClick(Sender: TObject);
begin
  WatchPanel.Visible:=not WatchPanel.Visible;
  ShowWatchBtn.Down:=WatchPanel.Visible;
end;

procedure TEditorDialog.ShowMsgBtnClick(Sender: TObject);
begin
  MsgPanel.Visible:=not MsgPanel.Visible;
  ShowMsgBtn.Down:=MsgPanel.Visible;
end;

procedure   TEditorDialog.UpdateWatches;
var i:integer;
    li:TListItem;
    vl:TAIStringList;
    PropList:TAIStringList;
    v:variant;
begin
  WL.Items.Clear;
  PropList:=TAIStringList.Create('editor update watches');
  try
    if Interpret.CurrContext.LocWS.Count > 0 then begin
      vl:=TAIStringList(Interpret.CurrContext.LocWS[Interpret.CurrContext.LocWS.Count - 1]);
      for i:=0 to vl.Count - 1 do begin
        li:=WL.Items.Add;
        li.Caption:=vl[i];

        PropList.Clear;
        v:='';
        TData(vl.Objects[i]).PropList:=PropList;
        v:=TData(vl.Objects[i]).Value;

        if not TData(vl.Objects[i]).ProcessedProp then begin
          li.SubItems.Add('???');
        end else begin
          li.SubItems.Add( v);
        end;

      end;
    end;
  finally
    PropList.Free;
  end;
end;

procedure TEditorDialog.ClearAllBtnClick(Sender: TObject);
begin
  MsgMemo.Clear;
end;

procedure TEditorDialog.InfoBtnClick(Sender: TObject);
begin
  MsgMemo.Lines.Add(GetAIListCounts);
end;

end.
