unit AIEvents;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes, SysUtils, TypInfo, Forms,

  AInterpreter, ErrorMsg;

type
  TAIEventHandler = class
    class procedure SetEvent(obj:TObject; EventName:string; HOnEvent:TScriptEventHandler);
    procedure       OnClick(Sender: TObject);
    procedure       OnDblClick(Sender: TObject);
    procedure       OnClose(Sender: TObject; var Action: TCloseAction);
    procedure       OnChange(Sender: TObject);
    procedure       OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  end;

implementation

uses  AIObject;

var
  EventList:TStringList;

class procedure     TAIEventHandler.SetEvent(obj:TObject; EventName:string; HOnEvent:TScriptEventHandler);
var idx:integer;
    mtd:TMethod;
begin
  idx:=EventList.IndexOf(EventName);
  if idx = -1 then begin
    Error(errUndeclaredEvent, [obj.ClassName, EventName]);
  end;

  mtd.Code:=EventList.Objects[idx];
  mtd.Data:=pointer(HOnEvent);
  SetMethodProp(obj, EventName, mtd);
end;

procedure     TAIEventHandler.OnChange(Sender: TObject);
begin
  AI.ExecProcTData(TScriptEventHandler(Self), [TDataObject.Create(Sender)]);
end;

procedure     TAIEventHandler.OnClick(Sender: TObject);
begin
  AI.ExecProcTData(TScriptEventHandler(Self), [TDataObject.Create(Sender)]);
end;

procedure     TAIEventHandler.OnDblClick(Sender: TObject);
begin
  AI.ExecProcTData(TScriptEventHandler(Self), [TDataObject.Create(Sender)]);
end;

procedure     TAIEventHandler.OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  AI.ExecProcTData(TScriptEventHandler(Self), [ TDataObject.Create(Sender),
//                                                TVariant.Create(Ord(Shift)),
                                                TVariant.Create(X),
                                                TVariant.Create(Y)
                                              ]
                  );
end;

procedure     TAIEventHandler.OnClose(Sender: TObject; var Action: TCloseAction);
var obj:TDataObject;
    v:TVariant;
begin
  obj:=TDataObject.Create;
  obj.link:=TAIObject.Create(nil);
  obj.link.obj:=Sender;
  inc(obj.UseCount);
  inc(obj.link.UseCount);

  v:=TVariant.Create(Ord(Action));
  inc(v.UseCount);

  try
    AI.ExecProcTData(TScriptEventHandler(Self), [obj, v]);
    Action:=v.Value;
    dec(v.UseCount);
  finally
    v.Free;
    obj.Free;
  end;
end;

initialization
  EventList:=TStringList.Create;
  EventList.Sorted:=true;
  EventList.AddObject('OnClick',      pointer(@TAIEventHandler.OnClick));
  EventList.AddObject('OnDblClick',   pointer(@TAIEventHandler.OnDblClick));
  EventList.AddObject('OnClose',      pointer(@TAIEventHandler.OnClose));
  EventList.AddObject('OnChange',     pointer(@TAIEventHandler.OnChange));
  EventList.AddObject('OnMouseMove',  pointer(@TAIEventHandler.OnMouseMove));

finalization
  EventList.Free;
end.
