unit AIHandler;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes,

  ARUtils, AInterpreter;

type

  PCallMethod = ^TCallMethod;
  TCallMethod = procedure(obj:TObject; PropList:TAIStringList; Ex:TContext) of object;

  PGetValue   = ^TGetValue;
  TGetValue   = function(obj:TObject; PropList:TAIStringList):variant of object;

  PAssign     = ^TAssign;
  TAssign     = procedure(obj:TObject; PropList:TAIStringList; D:TData) of object;

  TProp = class
  public
    GetValue   :TGetValue;
    Assign     :TAssign;
    Lenght     :integer;
    constructor Create(AGetValue:TGetValue; AAssign:TAssign; ALenght:integer);
  end;

  TAIHandler = class
  private
    MethodList    :TAIStringList;
    PropertyList  :TAIStringList;
  public
    constructor   Create;virtual;
    destructor    Destroy;override;

    procedure     AddMethod (MethodName:string; ACallMethod:TCallMethod);
    procedure     AddProp   (PropName:string;   AGetValue:TGetValue; AAssign:TAssign; ALenght:integer);

    function      CallMethod    (obj:TObject; PropList:TAIStringList; Ex:TContext):boolean;
    function      GetValue      (obj:TObject; PropList:TAIStringList):variant;
    function      Assign        (obj:TObject; PropList:TAIStringList; D:TData):boolean;
    function      GetPropLenght (obj:TObject; PropList:TAIStringList):integer;
  end;

//procedure RegisterHandler(ClassName:string; h:TAIHandler);
procedure RegisterHandler(refClass:TPersistentClass; h:TAIHandler);

function  CallMethod(obj:TObject; PropList:TAIStringList; Ex:TContext):boolean;
function  GetValue  (obj:TObject; PropList:TAIStringList):variant;
function  Assign    (obj:TObject; PropList:TAIStringList; D:TData):boolean;
function  PropLenght(obj:TObject; PropList:TAIStringList):integer;

implementation
uses AIVCL;

constructor TProp.Create(AGetValue:TGetValue; AAssign:TAssign; ALenght:integer);
begin
  GetValue   :=AGetValue;
  Assign     :=AAssign;
  Lenght     :=ALenght;
end;

/////////////// TAIHandler /////////////////
constructor   TAIHandler.Create;
begin
  MethodList  :=TAIStringList.Create('TAIHandler.MethodList');
  MethodList.Sorted:=true;
  PropertyList:=TAIStringList.Create('TAIHandler.PropList');
  PropertyList.Sorted:=true;
end;

destructor    TAIHandler.Destroy;
begin
  MethodList.Free;

  while PropertyList.Count > 0 do begin
    PropertyList.Objects[0].Free;
    PropertyList.Delete(0);
  end;
  PropertyList.Free;

  inherited;
end;

procedure     TAIHandler.AddMethod (MethodName:string; ACallMethod:TCallMethod);
var idx:integer;
begin
  idx:=MethodList.IndexOf(MethodName);
  if idx <> -1 then begin
    MethodList.Objects[idx]:=pointer(@ACallMethod);
  end else begin
    MethodList.AddObject(MethodName, pointer(@ACallMethod));
  end;
end;

procedure     TAIHandler.AddProp   (PropName:string;   AGetValue:TGetValue; AAssign:TAssign; ALenght:integer);
var idx:integer;
begin
  idx:=PropertyList.IndexOf(PropName);
  if idx <> -1 then begin
    TProp(PropertyList.Objects[idx]).Free;
    PropertyList.Objects[idx]:=TProp.Create(AGetValue, AAssign, ALenght);
  end else begin
    PropertyList.AddObject(PropName, TProp.Create(AGetValue, AAssign, ALenght));
  end;
end;

function      TAIHandler.CallMethod(obj:TObject; PropList:TAIStringList; Ex:TContext):boolean;
var idx:integer;
    p:TCallMethod;
begin
  Result:=false;
  idx:=MethodList.IndexOf(PropList[0]);
  if idx <> -1 then begin
    Result:=true;
    p:=nil;
    @p:=MethodList.Objects[idx];
    p(obj, PropList, Ex);
  end;
end;


function      TAIHandler.Assign    (obj:TObject; PropList:TAIStringList; D:TData):boolean;
var idx:integer;
begin
  Result:=false;
  idx:=PropertyList.IndexOf(PropList[0]);
  if (idx <> -1) and Assigned(TProp(PropertyList.Objects[idx]).Assign) then begin
    Result:=true;
    TProp(PropertyList.Objects[idx]).Assign(obj, PropList, D);
  end;
end;

function      TAIHandler.GetValue  (obj:TObject; PropList:TAIStringList):variant;
var idx:integer;
begin
  VarClear(Result);
  idx:=PropertyList.IndexOf(PropList[0]);
  if (idx <> -1) and Assigned(TProp(PropertyList.Objects[idx]).GetValue) then begin
    Result:=TProp(PropertyList.Objects[idx]).GetValue(obj, PropList);
  end;
end;

function      TAIHandler.GetPropLenght (obj:TObject; PropList:TAIStringList):integer;
var idx:integer;
begin
  idx:=PropertyList.IndexOf(PropList[0]);
  if idx = -1 then begin
    Result:=0;
  end else begin
    Result:=TProp(PropertyList.Objects[idx]).Lenght;
  end;
end;

/////////////////////////////////////////////

var
  HandlerList:TAIStringList;

//procedure RegisterHandler(rClassName:string; h:TAIHandler);
procedure RegisterHandler(refClass:TPersistentClass; h:TAIHandler);
var idx:integer;
begin
  if not Assigned(HandlerList) then begin
    HandlerList:=TAIStringList.Create('HandlerList');
    HandlerList.Sorted :=true;
  end;

  RegisterClass(refClass);

  idx:=HandlerList.IndexOf(refClass.ClassName);
  if idx = -1 then begin
    HandlerList.AddObject(refClass.ClassName, h);
  end else begin
    HandlerList.Objects[idx].Free;
    HandlerList.Objects[idx]:=h;
  end;
end;

{
function  CallMethod(obj:TObject; PropList:TAIStringList; Ex:TContext):boolean;
var idx:integer;
    ClassRef: TClass;
begin
  ClassRef := objSender.ClassType;
  while ClassRef <> nil do
  begin
    ListBox1.Items.Add(ClassRef.ClassName);
    ClassRef := ClassRef.ClassParent;
  end;


  idx:=HandlerList.IndexOf(obj.ClassName);
  if idx = -1 then begin
    Result:=false;
  end else begin
    Result:=TAIHandler(HandlerList.Objects[idx]).CallMethod(obj, PropList, Ex);
  end;
  if Result then PropList.Delete(0);
end;
}

function  CallMethod(obj:TObject; PropList:TAIStringList; Ex:TContext):boolean;
var idx:integer;
    ClassRef: TClass;
    s:string;
begin
  ClassRef:=obj.ClassType;
  Result:=false;
  while ClassRef <> nil do begin
    s:=ClassRef.ClassName;
    idx:=HandlerList.IndexOf(s);
    if idx <> -1 then begin
      Result:=TAIHandler(HandlerList.Objects[idx]).CallMethod(obj, PropList, Ex);
      if Result then begin
        PropList.Delete(0);
        break;
      end;
    end;
    ClassRef := ClassRef.ClassParent;
  end;
end;

function  Assign    (obj:TObject; PropList:TAIStringList; D:TData):boolean;
var idx, i:integer;
    ClassRef: TClass;
    s:string;
begin
  ClassRef:=obj.ClassType;
  Result:=false;
  while ClassRef <> nil do begin
    s:=ClassRef.ClassName;

    idx:=HandlerList.IndexOf(s);
    if idx = -1 then begin
      Result:=false;
    end else begin
      Result:=TAIHandler(HandlerList.Objects[idx]).Assign(obj, PropList, D);
    end;
    if Result then begin
      for i:=0 to TAIHandler(HandlerList.Objects[idx]).GetPropLenght(obj, PropList) - 1 do begin
        PropList.Delete(0);
      end;
      break;
    end;
    ClassRef := ClassRef.ClassParent;
  end;
end;

function  GetValue  (obj:TObject; PropList:TAIStringList):variant;
var idx, i:integer;
    ClassRef: TClass;
    s:string;
begin
  ClassRef:=obj.ClassType;
  Result:=false;
  while ClassRef <> nil do begin
    s:=ClassRef.ClassName;

    idx:=HandlerList.IndexOf(s);
    if idx = -1 then begin
      Result:=false;
    end else begin
      Result:=TAIHandler(HandlerList.Objects[idx]).GetValue(obj, PropList);
      for i:=0 to TAIHandler(HandlerList.Objects[idx]).GetPropLenght(obj, PropList) - 1 do begin
        PropList.Delete(0);
      end;
      break;
    end;
    ClassRef := ClassRef.ClassParent;
  end;
end;

function  PropLenght(obj:TObject; PropList:TAIStringList):integer;
var idx:integer;
    ClassRef: TClass;
    s:string;
begin
  Result:=0;

  ClassRef:=obj.ClassType;
  while ClassRef <> nil do begin
    s:=ClassRef.ClassName;

    idx:=HandlerList.IndexOf(s);
    if idx <> -1 then begin
      Result:=TAIHandler(HandlerList.Objects[idx]).GetPropLenght(obj, PropList);
      break;
    end;
    ClassRef := ClassRef.ClassParent;
  end;
end;

/////////////////////////////////////////////
initialization

finalization
  while HandlerList.Count > 0 do begin
    HandlerList.Objects[0].Free;
    HandlerList.Delete(0);
  end;

  HandlerList.Free;

end.
