unit AIList;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  SysUtils, Controls, Classes,

  ARUtils, ErrorMsg, AInterpreter;

procedure RegisterWords(Words:TWords);

type
  TDataStringList = class(TAIStringList)
  public
    UseCount  :integer;
  end;

  TDataList = class(TData)
  private
    SL          :TDataStringList;
    pos         :integer;
  protected
    function    GetValue:variant;override;
    procedure   SetItems(AItems:TStringList);
    function    GetItems:TStringList;
  public
    constructor Create;
    destructor  Destroy;override;
    procedure   CallMethod(Ex:TContext);override;
    procedure   Assign(D:TData);override;

    procedure   IncUse;
    procedure   DecUse;

    property    Items:TStringList read  GetItems  write SetItems;
  end;

procedure GetFileList(RootDir:string; FileMask:string; SL:TStringList);

implementation
///////////////////////////////////////////////////////////////////////////////////////////
constructor  TDataList.Create;
begin
  inherited Create;

  SL:=TDataStringList.Create('AI data string list');
  IncUse;
  pos:=0;
end;

destructor  TDataList.Destroy;
begin
  DecUse;

  inherited;
end;

procedure   TDataList.IncUse;
begin
  inc(SL.UseCount);
end;

procedure   TDataList.DecUse;
begin
  if Assigned(SL) then begin
    dec(SL.UseCount);
    if SL.UseCount = 0 then begin
      SL.Free;
      SL:=nil;
    end;
  end;
end;

procedure   TDataList.SetItems(AItems:TStringList);
begin
  SL.Assign(AItems);
end;

function    TDataList.GetItems:TStringList;
begin
  Result:=SL;
end;

procedure    TDataList.CallMethod(Ex:TContext);
var val,id,idx,FileName, RootDir, FileMask:TData;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if   CT(PropList[0], 'CLEAR') then begin
      SL.Clear;
      pos:=0;
    end else
    if   CT(PropList[0], 'SORT') then begin
      SL.Sort;
      pos:=0;
    end else
    if   CT(PropList[0], 'FIRST') then begin
      pos:=0;
    end else
    if  CT(PropList[0], 'NEXT')  then begin
      inc(pos);
    end else
    if  CT(PropList[0], 'LAST')  then begin
      pos:=SL.Count - 1;
    end else
    if  CT(PropList[0], 'PRIOR')  then begin
      dec(pos);
    end else
    if  CT(PropList[0], 'INSERTID')  then begin
      id  :=Ex.PopDS;
      val :=Ex.PopDS;
      idx :=Ex.PopDS;
      try
        SL.InsertObject(idx.value, val.value, pointer(integer(id.value)));
      finally
        val.Free;
        id.Free;
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'INSERT')  then begin
      val :=Ex.PopDS;
      idx :=Ex.PopDS;
      try
        SL.Insert(idx.value, val.value);
      finally
        val.Free;
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'DELETE')  then begin
      idx :=Ex.PopDS;
      try
        SL.Delete(idx.value);
      finally
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'INDEXOF')  then begin
      idx :=Ex.PopDS;
      try
        Ex.PushDS(TVariant.Create(SL.IndexOf(idx.value)));
      finally
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'INDEXOFID')  then begin
      idx :=Ex.PopDS;
      try
        Ex.PushDS(TVariant.Create(SL.IndexOfObject(pointer(integer(idx.value)))));
      finally
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'LOAD')  then begin
      FileName:=Ex.PopDS;
      try
        SL.LoadFromFile(FileName.Value);
      finally
        pos:=0;
        FileName.Free;
      end;
    end else
    if  CT(PropList[0], 'SAVE')  then begin
      FileName:=Ex.PopDS;
      try
        SL.SaveToFile(FileName.Value);
      finally
        FileName.Free;
      end;
    end else
    if  CT(PropList[0], 'ADDID')  then begin
      id:=Ex.PopDS;
      val:=Ex.PopDS;
      try
        SL.AddObject(val.value, pointer(integer(id.value)));
      finally
        val.Free;
        id.Free;
      end;
    end else
    if  CT(PropList[0], 'ADD')  then begin
      val:=Ex.PopDS;
      try
        SL.Add(val.value);
      finally
        val.Free;
      end;
    end else
    if  CT(PropList[0], 'ADDLIST')  then begin
      val:=Ex.PopDS;
      try
        if val is TDataList then begin
          SL.AddStrings(TDataList(val).SL);
        end else begin
          Error(errListAddList, [0]);
        end;
      finally
        val.Free;
      end;
    end else
    if  CT(PropList[0], 'SET')  then begin
      val:=Ex.PopDS;
      idx:=Ex.PopDS;
      try
        SL[idx.value]:=val.value;
      finally
        val.Free;
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'SETID')  then begin
      id  :=Ex.PopDS;
      val :=Ex.PopDS;
      idx :=Ex.PopDS;
      try
        SL[idx.value]:=val.value; SL.Objects[idx.value]:=pointer(integer(id.value));
      finally
        val.Free;
        id.Free;
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'RECURSEDIR')  then begin // RootDir, FileMask
      FileMask :=Ex.PopDS;
      RootDir  :=Ex.PopDS;
      try
        SL.Clear;
        GetFileList(RootDir.Value, FileMask.Value, SL);
        SL.Sort;
      finally
        FileMask.Free();
        RootDir.Free()
      end;
    end else
    ProcessedProp:=false;
  end;
end;

function    TDataList.GetValue:variant;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if CT(PropList[0], 'STR')       then if (pos >= SL.Count) or (pos < 0) then Result:='' else Result:=SL[pos] else
    if CT(PropList[0], 'ID')        then if (pos >= SL.Count) or (pos < 0) then Result:=-1 else Result:=integer(SL.Objects[pos]) else
    if CT(PropList[0], 'COUNT')     then Result:=SL.Count else
    if CT(PropList[0], 'EOF')       then Result:=integer((pos > (SL.Count - 1)) or (pos < 0)) else
    if CT(PropList[0], 'POS')       then Result:=pos else
    if CT(PropList[0], 'TEXT')      then Result:=SL.Text else
    if CT(PropList[0], 'COMMATEXT') then Result:=SL.CommaText else
    if CT(PropList[0], 'SORTED')    then Result:=SL.Sorted else
    if CT(PropList[0], 'NAMES')     then Result:=SL.Names[StrToInt(PropList[1])] else
    if CT(PropList[0], 'VALUE')     then Result:=SL.Values[PropList[1]] else
    ProcessedProp:=false;
  end;
end;

procedure    TDataList.Assign(D:TData);
var sr:TSearchRec;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if CT(PropList[0], 'STR')       then SL[pos]                :=SafeVarToStr(D.Value)     else
    if CT(PropList[0], 'ID')        then SL.Objects[pos]        :=pointer(integer(D.Value)) else
    if CT(PropList[0], 'POS')       then pos                    :=D.Value                   else
    if CT(PropList[0], 'TEXT')      then SL.Text                :=SafeVarToStr(D.Value)     else
    if CT(PropList[0], 'COMMATEXT') then SL.CommaText           :=SafeVarToStr(D.Value)     else
    if CT(PropList[0], 'SORTED')    then SL.Sorted              :=D.Value                   else
    if CT(PropList[0], 'VALUE')     then SL.Values[PropList[1]] :=SafeVarToStr(D.Value)     else
    if CT(PropList[0], 'FILEMASK')  then begin
      SL.Clear;
      try
        if FindFirst(D.Value, 0, sr) = 0 then begin
          SL.Add(sr.Name);
          while FindNext(sr) = 0 do begin
            SL.Add(sr.Name);
          end;
        end;
      finally
        FindClose(sr);
      end;
      SL.Sort;
    end else
    ProcessedProp:=false;
  end else begin
    if D.ClassType = Self.ClassType then begin
      dec(SL.UseCount);
      if SL.UseCount = 0 then begin
        SL.Free;
        SL:=nil;
      end;
      inc(TDataList(D).SL.UseCount);
      SL:=TDataList(D).SL;
    end else begin
      ProcessedProp:=false;
    end;
  end;
end;

procedure GetFileList(RootDir:string; FileMask:string; SL:TStringList);
  procedure GetFileListInt(RootDir:string; FileMask:string; SL:TStringList; prefixdir:string);
  var sr:TSearchRec;
  begin
    if FindFirst(RootDir + '\' + FileMask, faDirectory, sr) = 0 then begin
      try
        if (sr.Name <> '.') and (sr.Name <> '..') then begin
          if (sr.Attr and faDirectory) = faDirectory then begin
            GetFileListInt(RootDir + '\' + sr.Name, FileMask, SL, prefixdir + '\' + sr.Name);
          end else begin
            SL.Add(prefixdir + '\' + sr.Name);
          end;
        end;
        while FindNext(sr) = 0 do begin
          if (sr.Name <> '.') and (sr.Name <> '..') then begin
            if (sr.Attr and faDirectory) = faDirectory then begin
              GetFileListInt(RootDir + '\' + sr.Name, FileMask, SL, prefixdir + '\' + sr.Name);
            end else begin
              SL.Add(prefixdir + '\' + sr.Name);
            end;
          end;
        end;
      finally
        FindClose(sr);
      end;
    end;
  end;
begin
  SL.Clear;
  GetFileListInt(RootDir, FileMask, SL, '');
  SL.Sort;
end;

procedure BW_CreateList(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TDataList.Create);
  inc(Ex.IP);
end;

procedure BW_List(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateList'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, TDataList.Create);
  end;
end;

procedure BW_ReturnList(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateList'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

///////////////////////////////////////

procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('SYSTEM', 'List',           BWP(BW_List,        atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', '_CreateList',    BWP(BW_CreateList,  0));
  Words.AddWord('SYSTEM', '_ReturnList',    BWP(BW_ReturnList,  atIsImm));
end;

end.


