unit AIObject;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses  Controls, Classes, TypInfo, StdCtrls, Forms, SysUtils,

      ErrorMsg, ARUtils, AInterpreter, AIEvents, AIHandler, AIVCL;

type
  IAIObject = interface(IUnknown)
    function  CallMethod(PropName:string; Ex:TContext):boolean;
    function  GetValue(PropName:string; Ex:TContext):variant;
    function  Assign(PropName:string; D:TData):boolean;
  end;

  TAIObject = class
  private
    destructor    Destroy;override;
  public
    obj           :TObject;
    UseCount      :integer;

    constructor   Create(cls:TClass);
    procedure     Free;
  end;

  TDataObject = class(TData)
  public
    link          :TAIObject;
    constructor   Create(Obj:TObject);overload;
    constructor   Create(const AClassName:string);overload;
    destructor    Destroy;override;

    procedure     CallMethod(Ex:TContext);override;
    function      GetValue:variant;override;
    procedure     Assign(D:TData);override;
  end;

procedure RegisterWords(Words:TWords);
procedure ProcessUndeclaredWord(Ex:TContext; const term:string; var Processed:Boolean);

implementation
///////////////////////////////////////////////////////////////////////////////////////////
constructor   TAIObject.Create(cls:TClass);
begin
  if Assigned(cls) then begin
    if cls.InheritsFrom(TForm) then begin
      obj:=TComponentClass(cls).Create(Application.MainForm);
    end else begin
      if cls.InheritsFrom(TComponent) then begin
        obj:=TComponentClass(cls).Create(Application);
      end else begin
        obj:=cls.Create;
      end;
    end;
  end;
end;

procedure     TAIObject.Free;
begin
  dec(UseCount);
  if UseCount = 0 then begin
    if not ((obj is TComponent) and Assigned(TComponent(obj).Owner)) then begin
      UseCount:=0;
      obj.Free;
      obj:=nil;
    end;
    Destroy;
  end;
end;

destructor    TAIObject.Destroy;
begin
  inherited;
end;

///////////////////////////////////////////////////////////////////////////////////////////
constructor   TDataObject.Create(Obj:TObject);
begin
  inherited Create;

  if Assigned(Obj) then begin
    link:=TAIObject.Create(nil);
    link.obj:=Obj;
    inc(link.UseCount);
  end;
end;

constructor   TDataObject.Create(const AClassName:string);
var cls:TClass;
begin
  inherited Create;

  cls:=GetClass(AClassName);
  if not Assigned(cls) then begin
    cls:=GetClass('T' + AClassName);
    if not Assigned(cls) then begin
      Error(errUnknownClass, [AClassName]);
    end;
  end;
  link:=TAIObject.Create(cls);
  link.UseCount:=1;
  ProcessedProp:=true;
end;

destructor    TDataObject.Destroy;
begin
  if Assigned(link) then link.Free;
  inherited;
end;

function  InternalSetPropValue(obj:TObject; PropName:string; const Value:variant):boolean;
var pi:PPropInfo;
    s:string;
begin
  Result:=false;
  pi:=GetPropInfo(obj, PropName);
  if Assigned(pi) then begin
    case pi^.PropType^.Kind of
      tkChar, tkWChar, tkEnumeration, tkInteger:
        begin
          SetOrdProp(obj, PropName, Value);
          Result:=true;
        end;

      tkFloat:
        begin
          SetFloatProp(obj, PropName, Value);
          Result:=true;
        end;

      tkString, tkLString, tkWString:
        begin
          s:=VarAsType(Value, varString);
          SetStrProp(obj, PropName, s);
          Result:=true;
        end;
      tkVariant:
        begin
          SetVariantProp(obj, PropName, Value);
          Result:=true;
        end;

      tkMethod:
        begin
          TAIEventHandler.SetEvent(obj, PropName, Value);
          Result:=true;
        end;

    else
      Result:=false;
    end;
  end;
end;

function  InternalPropLenght(obj:TObject; PropList:TAIStringList):integer;
var pi:PPropInfo;
    s:string;
begin
  s:=obj.Classname;
  // native published property
  pi:=GetPropInfo(obj, PropList[0]);
  if Assigned(pi) then begin
    Result:=1;
  end else begin
    if CT(PropList[0], 'OWNER') or CT(PropList[0], 'PARENT') then begin
      Result:=1;
    end else begin
      Result:=AIHandler.PropLenght(obj, PropList);
    end;
  end;
end;

function  InternalGetNativeProp(obj:TObject; PropList:TAIStringList):variant;
var pi:PPropInfo;
begin
  VarClear(Result);
  // native published property
  pi:=GetPropInfo(obj, PropList[0]);
  if Assigned(pi) then begin
    case pi^.PropType^.Kind of
      tkChar, tkWChar, tkEnumeration, tkInteger:  Result:=GetOrdProp(obj, PropList[0]);
      tkFloat:                                    Result:=GetFloatProp(obj, PropList[0]);
      tkString, tkLString, tkWString:             Result:=GetStrProp(obj, PropList[0]);
      tkVariant:                                  Result:=GetVariantProp(obj, PropList[0]);
      tkClass:                                    Result:=ObjAsVar(GetObjectProp(obj, PropList[0]));
    end;
    PropList.Delete(0);
  end else
  if obj is TComponent then begin
    if CT(PropList[0], 'OWNER') then begin
      Result:=ObjAsVar(TComponent(obj).Owner);
      PropList.Delete(0);
    end else begin
      Result:=ObjAsVar(TComponent(obj).FindComponent(PropList[0]));
      if Assigned(VarAsObj(Result)) then begin
        PropList.Delete(0);
      end else begin
        if obj is TWinControl then begin
          if CT(PropList[0], 'PARENT') then begin
            Result:=ObjAsVar(TWinControl(obj).Parent);
            PropList.Delete(0);
          end else begin
            Result:=ObjAsVar(TWinControl(obj).FindChildControl(PropList[0]));
            if Assigned(VarAsObj(Result)) then begin
              PropList.Delete(0);
            end;
          end;
        end;
      end;
    end;
  end;
  if VarIsEmpty(Result) or (IsObject(Result) and (not Assigned(VarAsObj(Result)))) then begin
    Result:=AIHandler.GetValue(obj, PropList);
  end;
end;

function TDataObject.GetValue:variant;
begin
  ProcessedProp:=false;
  VarClear(Result);

  if (not Assigned(link)) or (not Assigned(link.obj)) then exit;

  if PropListCount = 0 then begin
    Result:=ObjAsVar(link.obj);
    ProcessedProp:=true;
  end else begin
    Result:=ObjAsVar(link.obj);
    while PropList.Count > 0 do begin
      Result:=InternalGetNativeProp(VarAsObj(Result), PropList);

      if IsObject(Result) then begin
        if not Assigned(VarAsObj(Result)) then break;
      end else begin
        break;
      end;
    end;

    if PropList.Count = 0 then begin
      ProcessedProp:=true;
    end;
  end;
end;

{
function TDataObject.GetValue:variant;
var i:integer;
    obj, obj2:TObject;
    v:variant;
begin
  ProcessedProp:=false;
  VarClear(Result);

  if (not Assigned(link)) or (not Assigned(link.obj)) then exit;

  if PropListCount = 0 then begin
    Result:=ObjAsVar(link.obj);
    ProcessedProp:=true;
  end else begin
    obj:=link.obj;
    while PropList.Count > 0 do begin
      v:=InternalGetNativeProp(obj, PropList);
      obj:=nil;

      if IsObject(v) then begin
        obj:=VarAsObj(v);
        if not Assigned(obj) then break;
      end else begin
        break;
      end;
    end;

    if PropList.Count = 0 then begin
      ProcessedProp:=true;
    end;
  end;
end;

}
procedure     TDataObject.CallMethod(Ex:TContext);
var f:TData;
    obj, obj2:TObject;
    cls:TClass;
    v:variant;
begin
  ProcessedProp:=false;

  obj:=nil;
  
  //search last object at the prop list
  if Assigned(link) and Assigned(link.obj) then begin
    obj:=link.obj;
    while PropList.Count > 0 do begin
      v:=InternalGetNativeProp(obj, PropList);
      obj2:=obj;

      if IsObject(v) then begin
        obj:=VarAsObj(v);
        if not Assigned(obj) then begin
          obj:=obj2;
          break;
        end;
      end else begin
        break;
      end;
    end;
  end;

  if PropListCount = 1 then begin
    if CT(PropList[0], 'CREATE') then begin
      f:=Ex.PopDS;
      try
        if Assigned(link) then link.Free;
        cls:=GetClass(f.Value);
        if not Assigned(cls) then begin
          cls:=GetClass('T' + f.Value);
          if not Assigned(cls) then begin
            Error(errUnknownClass, [f.Value]);
          end;
        end;
        link:=TAIObject.Create(cls);
        link.UseCount:=1;
        ProcessedProp:=true;
      finally
        f.Free;
      end;
    end else
    if CT(PropList[0], 'FREE')  then begin
      if Assigned(link) then link.Free;
      link:=nil;
      ProcessedProp:=true;
    end else begin
      if Assigned(obj) then begin
        ProcessedProp:=AIHandler.CallMethod(obj, PropList, Ex);
      end else begin
        ProcessedProp:=false;
      end;
    end;
  end;
end;

procedure    TDataObject.Assign(D:TData);
var obj, obj2:TObject;
    v:variant;
begin
  ProcessedProp:=false;

  if PropListCount = 0 then begin
//    if CT(D.ClassName, Self.ClassName) then begin
    if D.InheritsFrom(Self.ClassType) then begin
      if Assigned(link) then link.Free;
      link:=TDataObject(D).link;
      if Assigned(link) then inc(link.UseCount);
      ProcessedProp:=true;
    end;
  end else begin
    if Assigned(link) and Assigned(link.obj) then begin
      obj:=link.obj;
      while PropList.Count > 0 do begin
        if InternalPropLenght(obj, PropList) >= PropList.Count then break;

        v:=InternalGetNativeProp(obj, PropList);
        obj:=nil;

        if IsObject(v) then begin
          obj:=VarAsObj(v);
          if not Assigned(obj) then break;
        end else begin
          break;
        end;
      end;

      if (PropListCount > 0) and Assigned(obj) then begin
        v:=D.Value;

        if CT(PropList[0], 'OWNER') then begin
          if obj is TComponent then begin
            if IsObject(v) or (v = 0) then begin
              obj2:=VarAsObj(v);
              if (not Assigned(obj2)) or (obj2 is TComponent) then begin
                TComponent(obj).Owner.RemoveComponent(TComponent(obj));
                TComponent(obj2).InsertComponent(TComponent(obj));
                ProcessedProp:=true;
              end;
            end;
          end;
        end else
        if CT(PropList[0], 'PARENT') then begin
          if obj is TControl then begin
            if IsObject(v) or (v = 0) then begin
              obj2:=VarAsObj(v);
              if (not Assigned(obj2)) or (obj2 is TWinControl) then begin
                TControl(obj).Parent:=TWinControl(obj2);
                ProcessedProp:=true;
              end;
            end;
          end;
        end else
        if InternalSetPropValue(obj, PropList[0], v) then begin
          ProcessedProp:=true;
        end else begin
          ProcessedProp:=AIHandler.Assign(obj, PropList, D);
        end;
      end;
    end;
  end;

  if not ProcessedProp then begin
    Error(errPropWritingToType, [link.obj.ClassName, PropListToStr(PropList)]);
  end;
end;

type
  TMethodParam = record
    Flags     :TParamFlags;
    ParamName :string;
    TypeName  :string;
  end;

  TMethodParams = array[0..30] of TMethodParam;

var MethodParams:TMethodParams;

procedure SetMethodParams(cnt:integer; PL:array of Char);
var p, i, j:integer;
begin
  p:=0;
  for i:=0 to cnt - 1 do begin
    MethodParams[i].Flags:=TParamFlags(PL[p]);
    inc(p);

    MethodParams[i].ParamName:='';
    for j:=0 to byte(PL[p]) - 1 do begin
      MethodParams[i].ParamName:=MethodParams[i].ParamName + PL[p + j + 1];
    end;
    inc(p, byte(PL[p]) + 1);

    MethodParams[i].TypeName:='';
    for j:=0 to byte(PL[p]) - 1 do begin
      MethodParams[i].TypeName:=MethodParams[i].TypeName + PL[p + j + 1];
    end;
    inc(p, byte(PL[p]) + 1);
  end;
end;

procedure ProcessUndeclaredWord(Ex:TContext; const term:string; var Processed:Boolean);
var VarName:string;
    cls:TClass;
begin
  cls:=GetClass(term);
  if not Assigned(cls) then begin
    cls:=GetClass('T' + term);
  end;
  if Assigned(cls) then begin
    if Ex.CompileMode then begin
      VarName:=Upper(Ex.GetNextTerm.Value);
      Ex.LocalWords.Add(VarName);
      Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateObjectEx'));
      Ex.CompileData(TVariant.Create(term));
      Ex.CompileData(TVariant.Create(VarName));
    end else begin
      Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, Ex.PushDS(TDataObject.Create(term)));
    end;
    Processed:=true;
  end;
end;


procedure BW_CreateObject(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TDataObject.Create);
  inc(Ex.IP);
end;

procedure BW_CreateObjectEx(Ex:TContext);
var s:string;
begin
  s:=TVariant(Ex.CS[Ex.IP]).Value;
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP + 1]).Value, TDataObject.Create(s));
  inc(Ex.IP);
  inc(Ex.IP);
end;

procedure BW_Object(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateObject'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, Ex.PushDS(TDataObject.Create));
  end;
end;

procedure BW_ReturnObject(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateObject'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

///////////////////////////////////////
procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('SYSTEM', 'OBJECT',           BWP(BW_Object,              atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', '_CreateObject',    BWP(BW_CreateObject,        0));
  Words.AddWord('SYSTEM', '_CreateObjectEx',  BWP(BW_CreateObjectEx,      0));
  Words.AddWord('SYSTEM', '_ReturnObject',    BWP(BW_ReturnObject,        atIsImm));
end;

end.


