unit AIRecord;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, StdCtrls,

  ErrorMsg, ARUtils, AIWords, AInterpreter;

type
  TAIRecord = class(TAIStringList)
  public
    UseCount        :integer;
    destructor      Destroy;override;
  end;

  TDataRecord = class(TData)
  private
    control         :TAIRecord;
  protected
    function        GetValue:variant;override;
  public
    constructor     Create;
    destructor      Destroy;override;
    procedure       CallMethod(Ex:TContext);override;
    procedure       Assign(D:TData);override;
    function        FindVar(VarName:string):TData;
  end;
///////////////////////////////////////

procedure RegisterWords(Words:TWords);

implementation

///////////////////////////////////////
destructor  TAIRecord.Destroy;
var i:integer;
begin
  for i:=0 to Count - 1 do begin
    dec(TData(Objects[i]).UseCount);
    TData(Objects[i]).Free;
  end;

  inherited;
end;
///////////////////////////////////////
constructor  TDataRecord.Create;
begin
  inherited Create;

  control:=TAIRecord.Create('AI record');
  control.UseCount:=1;
end;

destructor  TDataRecord.Destroy;
begin
  dec(control.UseCount);
  if control.UseCount = 0 then control.Free;

  inherited;
end;

function    TDataRecord.FindVar(VarName:string):TData;
var idx:integer;
begin
  Result:=nil;

  idx:=control.IndexOf(VarName);
  if idx <> -1 then begin
    Result:=TData(control.Objects[idx]);
  end;
end;

procedure    TDataRecord.CallMethod(Ex:TContext);
var v,vname:TData;
begin
  ProcessedProp:=true;
  if PropListCount > 0 then begin
    if   CT(PropList[0], 'ADD')   then begin
      vname  :=Ex.PopDS;  v:=Ex.PopDS;

      control.AddObject(vname.Value, v);
      inc(v.UseCount);

      v.Free;  vname.Free;
    end else
    if   CT(PropList[0], 'ITEMS')   then begin
      v:=Ex.PopDS;
      try
        Ex.PushDS(TData(control.Objects[v.value]));
      finally
        v.Free;
      end;
    end else begin
      ProcessedProp:=false;
    end;
  end;
end;

function    TDataRecord.GetValue:variant;
var v:TData;
begin
  ProcessedProp:=true;
  if PropListCount > 0 then begin
    if CT(PropList[0], 'COUNT')  then Result:=control.Count else
    begin
      v:=FindVar(PropList[0]);
      if Assigned(v) then begin
        PropList.Delete(0);
        v.PropList:=PropList;
        Result:=v.Value;
//        v.PropList:=nil;
      end else begin
        ProcessedProp:=false;
      end;
    end;
  end;
end;

procedure    TDataRecord.Assign(D:TData);
var v:TData;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    v:=FindVar(PropList[0]);
    if Assigned(v) then begin
      PropList.Delete(0);
      v.PropList:=PropList;
      v.Assign(D);
    end else begin
      ProcessedProp:=false;
    end;
  end else begin
    if D.ClassType = Self.ClassType then begin
      dec(control.UseCount);
      if control.UseCount = 0 then control.Free;

      inc(TDataRecord(D).control.UseCount);
      control:=TDataRecord(D).control;
    end else begin
      ProcessedProp:=false;
    end;
  end;
end;
///////////////////////////////////////////////////////////////////////////////

procedure BW_CreateRecord(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TDataRecord.Create);
  inc(Ex.IP);
end;

procedure BW_Record(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateRecord'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, TDataRecord.Create);
  end;
end;

procedure BW_ReturnRecord(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateRecord'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('SYSTEM', 'Record',         BWP(BW_Record,         atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', '_CreateRecord',  BWP(BW_CreateRecord,   0));
  Words.AddWord('SYSTEM', '_ReturnRecord',  BWP(BW_ReturnRecord,   atIsImm));
end;

end.

