unit AIVCL;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface
uses
  Classes, Forms, StdCtrls, Grids, ExtCtrls, Controls, ComCtrls,
  SysUtils,

  ARUtils, AInterpreter, AIHandler;

type
  TAIControlHandler = class(TAIHandler)
  protected
    class procedure CM_SetBounds(obj:TObject; PropList:TAIStringList; Ex:TContext);
  public
    constructor     Create;override;
  end;

  TAIWinControlHandler = class(TAIHandler)
  protected
    class procedure CM_SetFocus(obj:TObject; PropList:TAIStringList; Ex:TContext);
  public
    constructor     Create;override;
  end;

  TAIFormHandler = class(TAIHandler)
  protected
    class procedure CM_Show(obj:TObject; PropList:TAIStringList; Ex:TContext);
    class procedure CM_Close(obj:TObject; PropList:TAIStringList; Ex:TContext);
  public
    constructor     Create;override;
  end;

  TAIProgressBarHandler = class(TAIHandler)
  protected
    class procedure CM_StepIt(obj:TObject; PropList:TAIStringList; Ex:TContext);
  public
    constructor     Create;override;
  end;

  TAIStringsHandler = class(TAIHandler)
  protected
    class procedure CM_Add          (obj:TObject; PropList:TAIStringList; Ex:TContext);
    class procedure CM_AddObject    (obj:TObject; PropList:TAIStringList; Ex:TContext);
    class procedure CM_Clear        (obj:TObject; PropList:TAIStringList; Ex:TContext);
    class procedure CM_LoadFromFile (obj:TObject; PropList:TAIStringList; Ex:TContext);

    class procedure SetCommaText  (obj:TObject; PropList:TAIStringList; D:TData);
    class procedure SetText       (obj:TObject; PropList:TAIStringList; D:TData);
    class procedure SetStrings    (obj:TObject; PropList:TAIStringList; D:TData);

    class function  GetCommaText  (obj:TObject; PropList:TAIStringList):variant;
    class function  GetText       (obj:TObject; PropList:TAIStringList):variant;
    class function  GetCount      (obj:TObject; PropList:TAIStringList):variant;
    class function  GetObjects    (obj:TObject; PropList:TAIStringList):variant;
    class function  GetStrings    (obj:TObject; PropList:TAIStringList):variant;
  public
    constructor     Create;override;
  end;

implementation
//  TControl  //////////////////////////////////////////////////////////////////
class procedure TAIControlHandler.CM_SetBounds(obj:TObject; PropList:TAIStringList; Ex:TContext);
var Left, Top, Width, Height:TData;
begin
  Height:=Ex.PopDS;
  Width:=Ex.PopDS;
  Top:=Ex.PopDS;
  Left:=Ex.PopDS;
  try
    TControl(obj).SetBounds(Left.Value, Top.Value, Width.Value, Height.Value);
  finally
    Left.Free;
    Top.Free;
    Width.Free;
    Height.Free;
  end;
end;

constructor     TAIControlHandler.Create;
begin
  inherited;
  AddMethod('SetBounds',  CM_SetBounds);
end;

//  TWinControl  ///////////////////////////////////////////////////////////////
class procedure TAIWinControlHandler.CM_SetFocus(obj:TObject; PropList:TAIStringList; Ex:TContext);
begin
  TWinControl(obj).SetFocus();
end;

constructor     TAIWinControlHandler.Create;
begin
  inherited;
  AddMethod('SetFocus',  CM_SetFocus);
end;

//  TForm  /////////////////////////////////////////////////////////////////////
class procedure TAIFormHandler.CM_Show(obj:TObject; PropList:TAIStringList; Ex:TContext);
begin
  TForm(obj).Show();
end;

class procedure TAIFormHandler.CM_Close(obj:TObject; PropList:TAIStringList; Ex:TContext);
begin
  TForm(obj).Close();
end;

constructor     TAIFormHandler.Create;
begin
  inherited;

  AddMethod('Show',       CM_Show);
  AddMethod('Close',      CM_Close);
end;

//  TProgressBar  //////////////////////////////////////////////////////////////
class procedure TAIProgressBarHandler.CM_StepIt(obj:TObject; PropList:TAIStringList; Ex:TContext);
begin
  TProgressBar(obj).StepIt();
end;

constructor     TAIProgressBarHandler.Create;
begin
  inherited;

  AddMethod('StepIt',       CM_StepIt);
end;

//  TAIStringsHandler  /////////////////////////////////////////////////////////
class procedure TAIStringsHandler.CM_Add(obj:TObject; PropList:TAIStringList; Ex:TContext);
var s:TData;
begin
  s:=Ex.PopDS;
  try
    TStrings(obj).Add(s.Value);
  finally
    s.Free;
  end;
end;
class procedure TAIStringsHandler.CM_AddObject(obj:TObject; PropList:TAIStringList; Ex:TContext);
var s:TData;
    id:TData;
    i:integer;
begin
  id:=Ex.PopDS();
  s :=Ex.PopDS();
  try
    i:=id.Value;
    TStrings(obj).AddObject(s.Value, pointer(i));
  finally
    id.Free();
    s.Free();
  end;
end;
class procedure TAIStringsHandler.CM_Clear(obj:TObject; PropList:TAIStringList; Ex:TContext);
begin
  TStrings(obj).Clear();
end;
class procedure TAIStringsHandler.CM_LoadFromFile(obj:TObject; PropList:TAIStringList; Ex:TContext);
var fname:TData;
begin
  fname:=Ex.PopDS();
  try
    TStrings(obj).LoadFromFile(fname.Value);
  finally
    fname.Free();
  end;
end;
class procedure TAIStringsHandler.SetCommaText(obj:TObject; PropList:TAIStringList; D:TData);
begin
  TStrings(obj).CommaText:=D.Value;
end;
class procedure TAIStringsHandler.SetText(obj:TObject; PropList:TAIStringList; D:TData);
begin
  TStrings(obj).Text:=D.Value;
end;
class function  TAIStringsHandler.GetCommaText(obj:TObject; PropList:TAIStringList):variant;
begin
  Result:=TStrings(obj).CommaText;
end;
class function  TAIStringsHandler.GetText(obj:TObject; PropList:TAIStringList):variant;
begin
  Result:=TStrings(obj).Text;
end;
class function  TAIStringsHandler.GetCount(obj:TObject; PropList:TAIStringList):variant;
begin
  Result:=TStrings(obj).Count;
end;
class function  TAIStringsHandler.GetObjects(obj:TObject; PropList:TAIStringList):variant;
begin
  Result:=integer(TStrings(obj).Objects[StrToInt(PropList[1])]);
end;
class function TAIStringsHandler.GetStrings(obj: TObject; PropList: TAIStringList): variant;
begin
  Result:=TStrings(obj).Strings[StrToInt(PropList[1])];
end;
class procedure TAIStringsHandler.SetStrings(obj: TObject; PropList: TAIStringList; D: TData);
begin
  TStrings(obj).Strings[StrToInt(PropList[1])]:=D.Value;
end;
constructor     TAIStringsHandler.Create;
begin
  inherited;

  AddMethod ('Add',           CM_Add);
  AddMethod ('Clear',         CM_Clear);
  AddMethod ('AddObject',     CM_AddObject);
  AddMethod ('LoadFromFile',  CM_LoadFromFile);

  AddProp   ('Text',          GetText,      SetText,      1);
  AddProp   ('CommaText',     GetCommaText, SetCommaText, 1);
  AddProp   ('Count',         GetCount,     nil,          1);
  AddProp   ('Objects',       GetObjects,   nil,          2);
  AddProp   ('Strings',       GetStrings,   SetStrings,   2);
end;


////////////////////////////////////////////////////////////////////////////////
initialization
  RegisterHandler(TControl,         TAIControlHandler.Create());
  RegisterHandler(TWinControl,      TAIWinControlHandler.Create());
  RegisterHandler(TForm,            TAIFormHandler.Create());
  RegisterHandler(TProgressBar,     TAIProgressBarHandler.Create());
  RegisterHandler(TStrings,         TAIStringsHandler.Create());
end.

