unit AIVarList;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  SysUtils, Controls, Classes,

  ARUtils, ErrorMsg, AInterpreter;

procedure RegisterWords(Words:TWords);

type
  TAIVarList = class(TAIStringList)
  public
    UseCount        :integer;

    procedure       Clear;override;
    procedure       Insert(index:integer; s:string; v:TData);reintroduce;
    procedure       Delete(index:integer);override;
    procedure       Add(s:string; v:TData);reintroduce;

    constructor     Create;
    destructor      Destroy;override;
  end;

  TDataVarList = class(TData)
  private
    SL              :TAIVarList;
    pos             :integer;
  protected
    function        GetValue:variant;override;
  public
    constructor     Create;
    destructor      Destroy;override;
    procedure       CallMethod(Ex:TContext);override;
    procedure       Assign(D:TData);override;
  end;

implementation
///////////////////////////////////////////////////////////////////////////////////////////
constructor     TAIVarList.Create;
begin
  inherited Create('AI VAR data string list');
end;

destructor      TAIVarList.Destroy;
begin
  Clear;

  inherited;
end;

procedure       TAIVarList.Clear;
var i:integer;
begin
  for i:=0 to Count - 1 do begin
    dec(TData(Objects[i]).UseCount);
    TData(Objects[i]).Free;
  end;
  inherited Clear;
end;

procedure       TAIVarList.Insert(index:integer; s:string; v:TData);
begin
  inc(v.UseCount);
  InsertObject(index, s, pointer(v));
end;

procedure       TAIVarList.Delete(index:integer);
begin
  dec(TData(Objects[index]).UseCount);
  TData(Objects[index]).Free;
  inherited Delete(index);
end;

procedure       TAIVarList.Add(s:string; v:TData);
begin
  inc(v.UseCount);
  AddObject(s, pointer(v));
end;

//////////////////////////////////////////////////////////////////////////////////////////
constructor  TDataVarList.Create;
begin
  inherited Create;

  SL:=TAIVarList.Create;
  SL.UseCount:=1;
  pos:=0;
end;

destructor  TDataVarList.Destroy;
begin
  dec(SL.UseCount);
  if SL.UseCount = 0 then begin
    SL.Free;
  end;

  inherited;
end;

procedure    TDataVarList.CallMethod(Ex:TContext);
var val, s, idx:TData;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if   CT(PropList[0], 'CLEAR') then begin
      SL.Clear;
      pos:=0;
    end else
    if   CT(PropList[0], 'SORT') then begin
      SL.Sort;
      pos:=0;
    end else
    if   CT(PropList[0], 'FIRST') then begin
      pos:=0;
    end else
    if  CT(PropList[0], 'NEXT')  then begin
      inc(pos);
    end else
    if  CT(PropList[0], 'LAST')  then begin
      pos:=SL.Count - 1;
    end else
    if  CT(PropList[0], 'PRIOR')  then begin
      dec(pos);
    end else
    if  CT(PropList[0], 'INSERT')  then begin
      val :=Ex.PopDS;
      s   :=Ex.PopDS;
      idx :=Ex.PopDS;
      try
        SL.Insert(idx.Value, s.Value, val);
        pos:=idx.Value;
      finally
        val.Free;
        s.Free;
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'DELETE')  then begin
      idx :=Ex.PopDS;
      try
        SL.Delete(idx.value);
      finally
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'INDEXOF')  then begin
      idx :=Ex.PopDS;
      try
        Ex.PushDS(TVariant.Create(SL.IndexOf(idx.Value)));
      finally
        idx.Free;
      end;
    end else
    if  CT(PropList[0], 'ADD')  then begin
      val :=Ex.PopDS;
      s   :=Ex.PopDS;
      try
        SL.Add(s.Value, val);
        pos:=SL.Count - 1;
      finally
        val.Free;
        s.Free;
      end;
    end else
    ProcessedProp:=false;
  end;
end;

function    TDataVarList.GetValue:variant;
var v:TData;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if CT(PropList[0], 'STR')       then if (pos >= SL.Count) or (pos < 0) then Result:='' else Result:=SL[pos] else
    if CT(PropList[0], 'COUNT')     then Result:=SL.Count                 else
    if CT(PropList[0], 'EOF')       then Result:=integer((pos > (SL.Count - 1)) or (pos < 0))  else
    if CT(PropList[0], 'POS')       then Result:=pos else
    if CT(PropList[0], 'TEXT')      then Result:=SL.Text else
    if CT(PropList[0], 'COMMATEXT') then Result:=SL.CommaText else
    if CT(PropList[0], 'SORTED')    then Result:=SL.Sorted
    else begin
      v:=TData(SL.Objects[pos]);
      if Assigned(v) then begin
        v.PropList:=PropList;
        Result:=v.Value;
      end else begin
        ProcessedProp:=false;
      end;
    end;
  end;
end;

procedure    TDataVarList.Assign(D:TData);
var v:TData;
begin
  ProcessedProp:=true;

  if PropListCount > 0 then begin
    if CT(PropList[0], 'STR')       then SL[pos]        :=D.Value else
    if CT(PropList[0], 'POS')       then pos            :=D.Value else
    if CT(PropList[0], 'SORTED')    then SL.Sorted      :=D.Value else
    begin
      v:=TData(SL.Objects[pos]);
      if Assigned(v) then begin
        v.PropList:=PropList;
        v.Assign(D);
      end else begin
        ProcessedProp:=false;
      end;
    end;
  end else begin
    if D.ClassType = Self.ClassType then begin
      dec(SL.UseCount);
      if SL.UseCount = 0 then begin
        SL.Free;
      end;
      inc(TDataVarList(D).SL.UseCount);
      SL:=TDataVarList(D).SL;
    end else begin
      ProcessedProp:=false;
    end;
  end;
end;

procedure BW_CreateVarList(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TDataVarList.Create);
  inc(Ex.IP);
end;

procedure BW_VarList(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateVarList'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, TDataVarList.Create);
  end;
end;

procedure BW_ReturnVarList(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateVarList'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

///////////////////////////////////////

procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('SYSTEM', 'VarList',           BWP(BW_VarList,        atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', '_CreateVarList',    BWP(BW_CreateVarList,  0));
  Words.AddWord('SYSTEM', '_ReturnVarList',    BWP(BW_ReturnVarList,  atIsImm));
end;

end.


