unit AIWords;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

{$INCLUDE DEFINES.INC}

interface
uses
  Windows, SysUtils, Classes, Dialogs, Controls, Forms, FileCtrl,
  Math, ComCtrls, ShellAPI,

  ErrorMsg, ARUtils, AInterpreter;

type
  TRegisterWordsProc = procedure (Words:TWords);

procedure RegisterWords(Words:TWords);

procedure BW_ApplyProp(Ex:TContext);

implementation

//////////////////////////////////////////////////////////////////////////////////////////
procedure BW_Unit(Ex:TContext);
begin
  Ex.CurrUnit:=Upper(Ex.GetNextTerm.Value);
  Ex.UnitList.Clear;
  Ex.UnitList.Add(Ex.CurrUnit);
  Ex.UnitList.Add('SYSTEM');
  Ex.Words.DropUnit(Ex.CurrUnit);
  Ex.Words.AddUnit(Ex.CurrUnit);
end;

procedure BW_Uses(Ex:TContext);
var Term:TTerm;
    idx:integer;
begin
  Ex.UnitList.Clear;
  Ex.UnitList.Add(Ex.CurrUnit);
  Ex.UnitList.Add('SYSTEM');

  Term:=Ex.GetNextTerm;

  while Term.Value <> ';' do begin
    if Term.Value = ',' then begin
      Term:=Ex.GetNextTerm;
      continue;
    end;
    case Term.Attr of
      taString,
      taNumber,
      taError:  Error(errExpectedTerm, [';']);
      taWord:   begin
        if not Ex.Words.FindUnit(Term.Value) then begin
          idx:=Ex.ProjectList.IndexOfName(Term.Value);
          if idx <> -1 then begin
            Ex.RunFile(Ex.ProjectList.Values[Term.Value]);
          end else begin
            Error(errUnknownUnit, [Term.Value]);
          end;
        end;
        Ex.UnitList.Add(Upper(Term.Value));
      end;
    end;
    Term:=Ex.GetNextTerm;
  end;
end;

procedure BW_ProcDefinition(Ex:TContext);
var Term:TTerm;
    i:integer;
    addr:TAddr;
begin
  Ex.CompileMode:=true;

  Term:=Ex.GetNextTerm;

  case Term.Attr of
    taString,
    taNumber,
    taError:  Error(errRedeclareWord, [Term.Value]);
    taWord:
      begin
        Ex.Words.DropWord(Ex.CurrUnit, Upper(Term.Value));
        Ex.CS:=TAIList.Create('CS proc');              // creates proc CS
        Ex.Words.AddWord(Ex.CurrUnit, Term.Value, TAddr.Create(Ex.CS, atIsFunction));
        Ex.LocalWords.Clear;
      end;
  end;

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateLocalStack'));

  Term:=Ex.GetNextTerm;
  if Term.Value <> '(' then begin
    Error(errExpectedTerm, ['(']);
  end;

  Term:=Ex.GetNextTerm;
  while Term.Value <> ')' do begin
    if Term.Value = ',' then begin
      Term:=Ex.GetNextTerm;
      continue;
    end;

    case Term.Attr of
      taString,
      taNumber,
      taError:  Error(errExpectedTerm, [')']);
      taWord:   begin
        addr:=TAddr(Ex.Words.FindWord(Ex.UnitList, Term.Value));

        if Assigned(addr) then begin
          TBaseProc(TAddr(addr).Addr)(Ex);  //compile: var creation and adds it to local var list
          Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', 'DROP'));
        end else begin
          Error(errExpectedTerm, [')']);
        end;
      end;
    end;

    Term:=Ex.GetNextTerm;
  end;
  // compile: call local vars and assigns proc params
  for i:=Ex.LocalWords.Count - 1 downto 0 do begin
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
    Ex.CompileData(TVariant.Create(i));
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', 'SWAP'));
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '='));
  end;

  Ex.StatementList.Add('PROC'); // mark current block as PROC

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_AddDSSection'));  // compile: creates new DS section to correct work ";"

  Ex.AddOpSection;  // to expression compile
end;

procedure BW_FuncDefinition(Ex:TContext);
var Term:TTerm;
    addr:TAddr;
begin
  BW_ProcDefinition(Ex);
  // mark current block as FUNC
  Ex.DropTopStatement;
  Ex.StatementList.Add('FUNC');

  Term:=Ex.GetNextTerm;
  if Term.Value <> ':' then begin
    Error(errExpectedTerm, [':']);
  end;

  Term:=Ex.GetNextTerm;
  case Term.Attr of
    taString,
    taNumber,
    taError:  Error(errExpectedTypeName, [0]);
    taWord:   begin
      addr:=TAddr(Ex.Words.FindWord(Ex.UnitList, '_Return' + Term.Value));
      if Assigned(addr) then begin
        TBaseProc(TAddr(addr).Addr)(Ex);  //compile: var creation and adds it to local var list
      end else begin
        Error(errExpectedTypeName, [0]);
      end;
    end;
  end;
end;

procedure  BW_AddDSSection(Ex:TContext);
begin
  Ex.AddDSSection;
end;

procedure  BW_DropDSSection(Ex:TContext);
begin
  Ex.DropDSSection;
end;

procedure BW_CreateVar(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TVariant.Create(0));
  inc(Ex.IP);
end;

// VAR v - simple var definition (VARIANT)
procedure BW_Var(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateVar'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, Ex.PushDS(TVariant.Create(0)));
  end;
end;

procedure BW_ReturnVar(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_CreateVar'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

procedure BW_CreateLocalStack(Ex:TContext);
begin
  Ex.CreateLocalStack;
end;

procedure BW_FreeLocalStack(Ex:TContext);
begin
  Ex.FreeLocalStack;
end;

procedure BW_EndDefinition(Ex:TContext);
begin
  if not (CT(Ex.TopStatement, 'PROC') or CT(Ex.TopStatement, 'FUNC')) then Error(errENDPROCwithoutPROC, [0]);
  Ex.DropTopStatement;

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_DropDSSection'));

  // clears local word stack
  while Ex.LocalWords.Count > 0 do begin
    // if loc. var name = RESULT, then compile: leave RESULT value on DS
    if CT(Ex.LocalWords[Ex.LocalWords.Count - 1], 'RESULT') then begin
      Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
      Ex.CompileData(TVariant.Create(Ex.LocalWords.Count - 1));
    end;

    Ex.LocalWords.Delete(Ex.LocalWords.Count - 1);
  end;
  Ex.CompileMode:=false;

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_FreeLocalStack'));
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_Ret'));

  Ex.DropOpSection;  // to expression compile
end;

// on called proc: leave var value from loc. vars list by order N
// compiled follow this word,
// i.e. CS -  ...
//            <_PushLocalWord>
//            <N local variable>
//            ...
procedure BW_PushLocalWord(Ex:TContext);
var n:integer;
begin
  n:=Ex.GetTopCSAsInt;
  Ex.PushDS(Ex.LocalVar[n]);
  inc(Ex.IP);
end;

procedure BW_Ret(Ex:TContext);
begin
  Ex.PopRS;
end;

procedure BW_Swap(Ex:TContext);
var a,b:TData;
begin
  a:=Ex.PopDS;
  b:=Ex.PopDS;
  Ex.PushDS(a);
  Ex.PushDS(b);
end;

procedure BW_Drop(Ex:TContext);      // <value> -->
begin
  Ex.PopDS.Free;
end;

procedure BW_Print(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
//    ShowLine(a.Text);
    Ex.ShowMessage(a.Value);
  finally
    a.Free;
  end;
end;

procedure BW_Assign(Ex:TContext); // <var a>, <value b> --> <>
var a,b:TData;
begin
  b:=Ex.PopDS;  a:=Ex.PopDS;
  try
    a.Assign(b);
//    Ex.PushDS(a);
  finally
    b.Free;  a.Free;
  end;
end;

procedure BW_Sum(Ex:TContext);
var a,b:TData;
    s1,s2:string;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if (VarType(a.Value) = varString) or (VarType(b.Value) = varString) then begin
        s1:=a.Value;
        s2:=b.Value;
        Ex.PushDS(TVariant.Create(s1+s2));
      end else begin
        Ex.PushDS(TVariant.Create(a.Value + b.Value));
      end;
    except
      Error(errOperatorNotApplicable, ['+', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Mul(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(TVariant(a).Value * TVariant(b).Value));
    except
      Error(errOperatorNotApplicable, ['*', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Sub(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(TVariant(a).Value - TVariant(b).Value));
    except
      Error(errOperatorNotApplicable, ['-', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Div(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(TVariant(a).Value / TVariant(b).Value));
    except
      Error(errOperatorNotApplicable, ['/', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_SumAssign(Ex:TContext); // <var a>, <value b> --> <>
var a, b:TData;
begin
  b:=Ex.PopDS;
  a:=Ex.PopDS;
  try
    Ex.PushDS(a); //  a
    Ex.PushDS(a); //  a a
    Ex.PushDS(b); //  a a b
    BW_Sum(Ex);   //  a a+b
    BW_Assign(Ex);//  <>
  finally
    b.Free();  a.Free();
  end;
end;

procedure BW_SubAssign(Ex:TContext); // <var a>, <value b> --> <>
var a, b:TData;
begin
  b:=Ex.PopDS;
  a:=Ex.PopDS;
  try
    Ex.PushDS(a); //  a
    Ex.PushDS(a); //  a a
    Ex.PushDS(b); //  a a b
    BW_Sub(Ex);   //  a a-b
    BW_Assign(Ex);//  <>
  finally
    b.Free();  a.Free();
  end;
end;

procedure BW_MulAssign(Ex:TContext); // <var a>, <value b> --> <>
var a, b:TData;
begin
  b:=Ex.PopDS;
  a:=Ex.PopDS;
  try
    Ex.PushDS(a); //  a
    Ex.PushDS(a); //  a a
    Ex.PushDS(b); //  a a b
    BW_Mul(Ex);   //  a a*b
    BW_Assign(Ex);//  <>
  finally
    b.Free();  a.Free();
  end;
end;

procedure BW_DivAssign(Ex:TContext); // <var a>, <value b> --> <>
var a, b:TData;
begin
  b:=Ex.PopDS;
  a:=Ex.PopDS;
  try
    Ex.PushDS(a); //  a
    Ex.PushDS(a); //  a a
    Ex.PushDS(b); //  a a b
    BW_Div(Ex);   //  a a/b
    BW_Assign(Ex);//  <>
  finally
    b.Free();  a.Free();
  end;
end;

procedure BW_Greate(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if a.Value > b.Value then begin
        Ex.PushDS(TVariant.Create(1));
      end else begin
        Ex.PushDS(TVariant.Create(0));
      end;
    except
      Error(errOperatorNotApplicable, ['>', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Less(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if a.Value < b.Value then begin
        Ex.PushDS(TVariant.Create(1));
      end else begin
        Ex.PushDS(TVariant.Create(0));
      end;
    except
      Error(errOperatorNotApplicable, ['<', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_NotEQ(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if a.Value <> b.Value then begin
        Ex.PushDS(TVariant.Create(1));
      end else begin
        Ex.PushDS(TVariant.Create(0));
      end;
    except
      Error(errOperatorNotApplicable, ['<>', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_EQ(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if VarType(a.Value) = VarType(b.Value) then begin
        if a.Value = b.Value then begin
          Ex.PushDS(TVariant.Create(1));
        end else begin
          Ex.PushDS(TVariant.Create(0));
        end;
      end else begin
        if VarAsType(a.Value, varString) = VarAsType(b.Value, varString) then begin
          Ex.PushDS(TVariant.Create(1));
        end else begin
          Ex.PushDS(TVariant.Create(0));
        end;
      end;
    except
//      Error(errOperatorNotApplicable, ['==', a.ClassName + ', ' + b.ClassName]);
      Error(errOperatorNotApplicable, ['==', VarAsType(a.Value, varString) + ', ' + VarAsType(b.Value, varString)]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_GreateEQ(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if a.Value >= b.Value then begin
        Ex.PushDS(TVariant.Create(1));
      end else begin
        Ex.PushDS(TVariant.Create(0));
      end;
    except
      Error(errOperatorNotApplicable, ['>=', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_LessEQ(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      if a.Value <= b.Value then begin
        Ex.PushDS(TVariant.Create(1));
      end else begin
        Ex.PushDS(TVariant.Create(0));
      end;
    except
      Error(errOperatorNotApplicable, ['<=', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Mod(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(a.Value mod b.Value));
    except
      Error(errOperatorNotApplicable, ['MOD', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_IntDiv(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(a.Value div b.Value));
    except
      Error(errOperatorNotApplicable, ['DIV', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

{
procedure BW_ApplyProp(Ex:TContext);
var variable,prop:TData;
begin
  prop:=Ex.PopDS;
  variable:=Ex.PopDS;

  try
    if prop is TVariant then begin
      Ex.PushDS(TDataRef.Create(prop.Value, variable));
    end else begin
      Error(errInternalPropIsNotString, [prop.Text]);
    end;
  finally
    variable.Free;
    prop.Free;
  end;
end;
}
procedure BW_ApplyProp(Ex:TContext);
var variable,prop:TData;
begin
  prop:=Ex.PopDS;
  variable:=Ex.PopDS;

  try
    try
      Ex.PushDS(TDataRef.Create(prop.Value, variable));
    except
      Error(errInternalPropIsNotString, [prop.Text]);
    end;
  finally
    variable.Free;
    prop.Free;
  end;
end;

procedure BW_CallMethod(Ex:TContext);
var variable,prop:TData;
    d:TDataRef;
begin
  prop:=Ex.PopDS;
  variable:=Ex.PopDS;

  try
    if prop is TVariant then begin
      d:=TDataRef.Create(prop.Value, variable);
      try
        d.CallMethod(Ex);
      finally
        d.Free;
      end;
    end else begin
      Error(errInternalPropIsNotString, [prop.Text]);
    end;
  finally
    variable.Free;
    prop.Free;
  end;
end;

procedure BW_ResumeExpression(Ex:TContext; pos:integer);
var op:TBO;
begin
  while Ex.Op.Count > pos do begin
    op:=Ex.Op[Ex.Op.Count-1];
    Ex.Op.Delete(Ex.Op.Count-1);

    Ex.ResumeOperand(op);
  end;
end;

procedure BW_Semicolon(Ex:TContext);      //
begin
  BW_ResumeExpression(Ex, 0);

  while Ex.DS.Count > 0 do begin
    Ex.PopDS.Free;
  end;
end;

procedure BW_Comma(Ex:TContext);      //
begin
  BW_ResumeExpression(Ex, 1);
end;

procedure BW_ResumeDSSection(Ex:TContext);
var a:TData;
begin
  if Ex.DS.Count > 0 then begin
    a:=Ex.PopDS;
    Ex.DropDSSection;
    Ex.PushDS(a);
  end else begin
    Ex.DropDSSection;
  end;
end;

procedure BW_DebugInfo(Ex:TContext);
begin
  Ex.CurrExecLine:=Ex.GetTopCSAsInt;
  inc(Ex.IP);

  if Ex.Owner.StepOnRun then begin
    raise EStepOnRun.Create(Ex);
  end;
end;

procedure BW_OpenRound(Ex:TContext);
begin
  Ex.AddOpSection;
end;

procedure BW_CloseRound(Ex:TContext);
begin
  BW_ResumeExpression(Ex, 0);
  Ex.DropOpSection;
end;

procedure  BW_GOTO(Ex:TContext);// go to CS[IP] : <GOTO><GOTO ADDRESS>
begin
  Ex.IP:=Ex.GetTopCSAsInt;
end;

procedure  BW_IFZEROGOTO(Ex:TContext); // if DS = 0 then go to CS[IP] else inc(IP), , i.e. CS : <IF><goto address>
var a:TData;
begin
  a:=Ex.PopDS;

  try
    if a.Value = 0 then begin
      Ex.IP:=Ex.GetTopCSAsInt;
    end else begin
      inc(Ex.IP);
    end;
  finally
    a.Free;
  end;
end;

procedure  BW_IF(Ex:TContext); // if DS = 0 then goto to CS[IP] else inc(IP), , i.e. CS : <IF><goto address>
begin
  Ex.StatementList.Add('IF');

  Ex.AddOpSection;
end;

procedure BW_THEN(Ex:TContext);
begin
  if not CT(Ex.TopStatement, 'IF') then Error(errTHENWithoutIF, [0]);
  Ex.DropTopStatement;
  Ex.StatementList.Add('THEN');

  BW_CloseRound(Ex); // close IF expression

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_IFZEROGOTO'));
  Ex.PushJamp(Ex.CS.Count);      //  current address
  Ex.CompileAddr(nil);                //  reserve place to goto address
end;

procedure  BW_ELSE(Ex:TContext);
var a:integer;
begin
  if not CT(Ex.TopStatement, 'THEN') then Error(errELSEWithoutTHEN, [0]);

  a:=Ex.PopJamp;
  if Ex.CS[a] <> nil then Error(errInternalWriteToNonEmptyCell, ['ELSE']);
  Ex.CompileDataAt(a, TVariant.Create(Ex.CS.Count+2));

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_GOTO'));  // goto ENDIF
  Ex.PushJamp(Ex.CS.Count);      //  current address
  Ex.CompileAddr(nil);                 //  reserve place to goto address
end;

procedure  BW_ENDIF(Ex:TContext);
var a:integer;
begin
  if not CT(Ex.TopStatement, 'THEN') then Error(errENDIFWithoutTHEN, [0]);
  Ex.DropTopStatement;

  a:=Ex.PopJamp;
  if Ex.CS[a] <> nil then Error(errInternalWriteToNonEmptyCell, ['ENDIF']);
  Ex.CompileDataAt(a, TVariant.Create(Ex.CS.Count));
end;

{
  WHILE
    <EXPR>    A: <EXPR>   <-- FROM ENDWHILE
  DO          B: _IFZEROGOTO
              C: <G>
    <EXPR>    D: <EXPR>
  ENDWHILE    E: _GOTO
              F: <A>
              G:          <-- BREAK FROM WHILE
}
procedure  BW_WHILE(Ex:TContext);
begin
  Ex.StatementList.Add('WHILE');

  Ex.PushJamp(Ex.CS.Count);

  Ex.AddOpSection;
end;

procedure BW_DO(Ex:TContext);
begin
  if not CT(Ex.TopStatement, 'WHILE') then Error(errDOWithoutWHILE, [0]);

  BW_CloseRound(Ex); // resume WHILE expression

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_IFZEROGOTO'));

  Ex.PushJamp(Ex.CS.Count);      //  current address
  Ex.CompileAddr(nil);                 //  reserve place to goto address to ENDWHILE
end;

procedure  BW_ENDWHILE(Ex:TContext);
var a:integer;
begin
  if not CT(Ex.TopStatement, 'WHILE') then Error(errENDWHILEWithoutWHILE, [0]);
  Ex.DropTopStatement;

  a:=Ex.PopJamp;
  if Ex.CS[a] <> nil then Error(errInternalWriteToNonEmptyCell, ['ENDWHILE']);
  Ex.CompileDataAt(a, TVariant.Create(Ex.CS.Count+2));    // break from WHILE

  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_GOTO'));  // goto WHILE
  Ex.CompileData(TVariant.Create(Ex.PopJamp));
end;

procedure  BW_ProcessMessages(Ex:TContext);
begin
  Application.ProcessMessages;
end;

procedure  BW_GetAddrProc(Ex:TContext);
var Term:TTerm;
    o:TBO;
begin
  Term:=Ex.GetNextTerm;
  if Term.Value <> '(' then begin
    Error(errExpectedTerm, ['(']);
  end;

  Term:=Ex.GetNextTerm;
  o:=Ex.Words.FindWord(Ex.UnitList, Term.Value);
  if Assigned(o) and (o is TAddr) then begin
    Ex.CompileData(TVariant.Create(TAddr(o).Value));
  end else begin
    Error(errUndeclaredWord, [Term.Value]);
  end;

  Term:=Ex.GetNextTerm;
  if Term.Value <> ')' then begin
    Error(errExpectedTerm, [')']);
  end;
end;

procedure  BW_True(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create(integer(1)));
end;

procedure  BW_False(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create(integer(0)));
end;

procedure  BW_Now(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create((Now)));
end;

procedure  BW_Trunc(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(Int(a.Value)));
    except
      Error(errOperatorNotApplicable, ['Trunc', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure  BW_Frac(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(Frac(a.Value)));
    except
      Error(errOperatorNotApplicable, ['Frac', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure  BW_StrCopy(Ex:TContext);// DS: S: string; Index, Count: Integer, see Delphi Copy
var S, Index, Count: TData;
begin
  Count:= Ex.PopDS;
  Index:= Ex.PopDS;
  S    := Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(System.Copy(S.Value, Trunc(Index.Value), Trunc(Count.Value))));
    except
			Error(errIllegalParameters, ['StrCopy']);
    end;
  finally
    Count.Free;
    Index.Free;
    S.Free;
  end;
end;

procedure  BW_StrPos(Ex:TContext);// DS: string SubStr, string S -> integer pos // see Delphi Pos
var SubStr, S:TData;
begin
  S     := Ex.PopDS;
  SubStr:= Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(System.Pos(SubStr.Value, S.Value)));
    except
			Error(errIllegalParameters, ['StrPos']);
    end;
  finally
    SubStr.Free;
    S.Free;
  end;
end;

procedure  BW_StrDelete(Ex:TContext);// DS: string S, integer Index, Count -> // see Delphi Delete
var S, Index, Count: TData;
    ss:string;
begin
  Count:= Ex.PopDS;
  Index:= Ex.PopDS;
  S    := Ex.PopDS;
  try
    try
      ss:=S.Value;
      System.Delete(ss, Trunc(Index.Value), Trunc(Count.Value));
      Ex.PushDS(TVariant.Create(ss));
    except
			Error(errIllegalParameters, ['StrDelete']);
    end;
  finally
    Count.Free;
    Index.Free;
    S.Free;
  end;
end;

procedure  BW_Length(Ex:TContext);// DS: string S -> Length(s) // see Delphi Length
var S:TData;
begin
  S:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(Length(S.Value)));
    except
			Error(errIllegalParameters, ['Length']);
    end;
  finally
    S.Free;
  end;
end;

procedure  BW_Upper(Ex:TContext);// DS: string S -> Length(s) // see Delphi Length
var S:TData;
begin
  S:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(Upper(S.Value)));
    except
			Error(errIllegalParameters, ['Upper']);
    end;
  finally
    S.Free;
  end;
end;

procedure  BW_Trim(Ex:TContext);// DS: string S -> Length(s) // see Delphi Length
var S:TData;
begin
  S:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(Trim(S.Value)));
    except
			Error(errIllegalParameters, ['Trim']);
    end;
  finally
    S.Free;
  end;
end;

procedure  BW_ParamStr(Ex:TContext);// DS: Index -> ParamStr(Index) see Delphi ParamStr function
var Index:TData;
begin
  Index:= Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(System.ParamStr(Trunc(Index.Value))));
    except
			Error(errIllegalParameters, ['ParamStr']);
    end;
  finally
    Index.Free;
  end;
end;

procedure  BW_ParamCount(Ex:TContext);// DS: -> ParamCount() see Delphi ParamCount function
begin
  Ex.PushDS(TVariant.Create(System.ParamCount));
end;

procedure BW_Round(Ex:TContext);// DS: N, prcision -> rounded N
var n,prec:TData;
		m:currency;
const	AD:array[-8..8] of extended= (
  0.00000001,
  0.0000001,
  0.000001,
  0.00001,
  0.0001,
  0.001,
  0.01,
  0.1,
  1,
  10,
  100,
  1000,
  10000,
  100000,
  1000000,
  10000000,
  100000000);
begin
	prec:=Ex.PopDS;
	n:=Ex.PopDS;
	try
		try
			m:=(Int(Abs(n.Value) * AD[Trunc(prec.Value)] + 0.5) / AD[Trunc(prec.Value)]);
			if n.Value < 0 then m:=-m;

			Ex.PushDS(TVariant.Create(m));
		except
			Error(errIllegalParameters, ['Round']);
		end;
	finally
		prec.Free;
		n.Free;
	end;
end;

procedure BW_Not(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(a.Value = 0));
    except
      Error(errOperatorNotApplicable, ['NOT', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure BW_AND(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create((a.Value <> 0) and (b.Value <> 0)));
    except
      Error(errOperatorNotApplicable, ['AND', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_OR(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create((a.Value <> 0) or (b.Value <> 0)));
    except
      Error(errOperatorNotApplicable, ['OR', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_XOR(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create((a.Value <> 0) xor (b.Value <> 0)));
    except
      Error(errOperatorNotApplicable, ['XOR', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Max(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
//      if a.Value > b.Value then begin
//        Ex.PushDS(TVariant.Create(a.Value));
//      end else begin
//        Ex.PushDS(TVariant.Create(b.Value));
//      end;
      Ex.PushDS(TVariant.Create(Max(a.Value, b.Value)));
    except
      Error(errOperatorNotApplicable, ['MAX', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

procedure BW_Min(Ex:TContext);
var a,b:TData;
begin
  b:=Ex.PopDS;a:=Ex.PopDS;
  try
    try
//      if a.Value < b.Value then begin
//        Ex.PushDS(TVariant.Create(a.Value));
//      end else begin
//        Ex.PushDS(TVariant.Create(b.Value));
//      end;
      Ex.PushDS(TVariant.Create(Min(a.Value, b.Value)));
    except
      Error(errOperatorNotApplicable, ['MIN', a.ClassName + ', ' + b.ClassName]);
    end;
  finally
    b.Free;a.Free;
  end;
end;

{
procedure BW_GetInt(Ex:TContext); // path, defval :integer
var path,defval:TData;
begin
  defval:=Ex.PopDS; path:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(GetInt(path.Value, defval.Value)));
    except
      Error(errOperatorNotApplicable, ['GetInt', path.ClassName + ', ' + defval.ClassName]);
    end;
  finally
    defval.Free;path.Free;
  end;
end;

procedure BW_GetIntIni(Ex:TContext); //Section, Param, DefNumVal :integer
var Section, Param, DefNumVal:TData;
begin
  DefNumVal:=Ex.PopDS; Param:=Ex.PopDS; Section:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(GetIntIni(Section.Value, Param.Value, DefNumVal.Value)));
    except
      Error(errOperatorNotApplicable, ['GetIntIni', Section.ClassName + ',' + Param.ClassName + ',' + DefNumVal.ClassName]);
    end;
  finally
    Section.Free; Param.Free; DefNumVal.Free;
  end;
end;
}

procedure BW_Str(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(VarToStr(a.Value)));
    except
      Error(errOperatorNotApplicable, ['STR', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure BW_DateToStr(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(DateToStr(a.Value)));
    except
      Error(errOperatorNotApplicable, ['DateToStr', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure BW_StrToDate(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(StrToDate(a.Value)));
    except
      Error(errOperatorNotApplicable, ['StrToDate', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure BW_GetTickCount(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create(integer(GetTickCount)));
end;

procedure BW_Curr(Ex:TContext);
var a:TData;
    v:variant;
begin
  a:=Ex.PopDS;
  try
    try
      VarCast(v, a.Value, varCurrency);
      Ex.PushDS(TVariant.Create(v));
    except
      Error(errOperatorNotApplicable, ['CURR', a.ClassName]);
    end;
  finally
    a.Free;
  end;
end;

procedure BW_ShowUnits(Ex:TContext);
var i:integer;
begin
  for i:=0 to Ex.Words.UnitList.Count - 1 do begin
    Ex.ShowMessage(Ex.Words.UnitList[i]);
  end;
end;

procedure BW_ShowUnit(Ex:TContext);  // DS UnitName ->
var u:TData;
    i,idx:integer;
begin
  u:=Ex.PopDS;
  try
    idx:=Ex.Words.UnitList.IndexOf(Upper(u.Value));
    if idx <> -1 then begin
      for i:=0 to TAIStringList(Ex.Words.UnitList.Objects[idx]).Count - 1 do begin
        Ex.ShowMessage(TAIStringList(Ex.Words.UnitList.Objects[idx])[i]);
      end;
    end;
  finally
    u.Free;
  end;
end;

procedure BW_CLS(Ex:TContext);  // DS UnitName ->
begin
  Ex.Owner.ClearMsg;
end;

type
  AIUserException = class(Exception)
  end;

procedure BW_Error(Ex:TContext);  // DS errMsg ->  ; raise an user exception
var msg:TData;
begin
  msg:=Ex.PopDS;
  try
    raise AIUserException.Create(msg.Value);
  finally
    msg.Free;
  end;
end;

//////////////////////////////////////////////////////////////////////////////////////////////////////
procedure  BW_FilePath(Ex:TContext);// DS: FileName -> FilePath see Delphi ExtractFilePath function
var FileName:TData;
begin
  FileName:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(ExtractFilePath(FileName.Value)));
    except
			Error(errIllegalParameters, ['FilePath']);
    end;
  finally
    FileName.Free;
  end;
end;

procedure  BW_FileName(Ex:TContext);// DS: FullFileName -> FileName see Delphi ExtractFileName function
var FileName:TData;
begin
  FileName:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(ExtractFileName(FileName.Value)));
    except
			Error(errIllegalParameters, ['FileName']);
    end;
  finally
    FileName.Free;
  end;
end;

procedure  BW_FileNameOnly(Ex:TContext);// DS: FullFileName -> FileName without extention Delphi ExtractFileName function
var FileName:TData;
    fname:string;
    p:integer;
begin
  FileName:=Ex.PopDS;
  try
    try
      fname:=ExtractFileName(FileName.Value);

      for p:=Length(fname) downto 1 do begin
        if fname[p] = '.' then begin
          System.Delete(fname, p, Length(fname));
          break;
        end;
      end;

      Ex.PushDS(TVariant.Create(fname));
    except
			Error(errIllegalParameters, ['FileNameOnly']);
    end;
  finally
    FileName.Free;
  end;
end;

procedure  BW_FileExists(Ex:TContext);// DS: FileName -> IsFileExists(FileNameFilePath see Delphi ExtractFilePath function
var FileName:TData;
begin
  FileName:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(integer(FileExists(FileName.Value))));
    except
			Error(errIllegalParameters, ['FileExists']);
    end;
  finally
    FileName.Free;
  end;
end;

procedure  BW_DeleteFile(Ex:TContext);// DS: FileName ->
var FileName:TData;
begin
  FileName:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(integer(DeleteFile(FileName.Value))));
    except
			Error(errIllegalParameters, ['DeleteFile']);
    end;
  finally
    FileName.Free;
  end;
end;

procedure  BW_CopyFile(Ex:TContext);// DS: ExistingFileName, NewFileName, FailIfExists(bool) -> success
var ExistingFileName, NewFileName, FailIfExists:TData;
    zExistingFileName, zNewFileName:array[0..256] of Char;
begin
  FailIfExists    :=Ex.PopDS;
  NewFileName     :=Ex.PopDS;
  ExistingFileName:=Ex.PopDS;
  try
    try
      StrPCopy(zExistingFileName, ExistingFileName.Value);
      StrPCopy(zNewFileName,      NewFileName.Value);

      Ex.PushDS(TVariant.Create(integer(CopyFile( zExistingFileName, zNewFileName, FailIfExists.Value))));
    except
			Error(errIllegalParameters, ['CopyFile']);
    end;
  finally
    FailIfExists.Free;
    ExistingFileName.Free;
    NewFileName.Free;
  end;
end;

procedure  BW_CopyFiles(Ex:TContext);// DS: ExistingFileName, NewFileName, FailIfExists(bool) -> success
var ExistingFileName, NewFileName, FailIfExists:TData;
    zExistingFileName, zNewFileName:array[0..256] of Char;
begin
  FailIfExists    :=Ex.PopDS;
  NewFileName     :=Ex.PopDS;
  ExistingFileName:=Ex.PopDS;
  try
    try
      Ex.PushDS(TVariant.Create(integer(CopyFile( StrPCopy(zExistingFileName, ExistingFileName.Value),
                                                  StrPCopy(zNewFileName,      NewFileName.Value),
                                                  FailIfExists.Value
                                                )
                                       )
                               )
               );

    except
			Error(errIllegalParameters, ['CopyFile']);
    end;
  finally
    FailIfExists.Free;
    ExistingFileName.Free;
    NewFileName.Free;
  end;
end;

procedure  BW_MoveFile(Ex:TContext);// DS: ExistingFileName, NewFileName -> success
var ExistingFileName, NewFileName:TData;
    zExistingFileName, zNewFileName:array[0..256] of Char;
begin
  NewFileName     :=Ex.PopDS;
  ExistingFileName:=Ex.PopDS;
  try
    try
      StrPCopy(zExistingFileName, ExistingFileName.Value);
      StrPCopy(zNewFileName,      NewFileName.Value);

      Ex.PushDS(TVariant.Create(integer(MoveFileEx( zExistingFileName, zNewFileName, MOVEFILE_COPY_ALLOWED or MOVEFILE_REPLACE_EXISTING))));
    except
			Error(errIllegalParameters, ['MoveFile']);
    end;
  finally
    ExistingFileName.Free;
    NewFileName.Free;
  end;
end;

procedure  BW_MakeDir(Ex:TContext);// DS: NewPath -> success
var dir:TData;
begin
  dir:=Ex.PopDS;
  try
    try
      if not DirectoryExists(dir.Value) then begin
        ForceDirectories(dir.Value);
        Ex.PushDS(TVariant.Create(integer(true)));
      end;
    except
      Ex.PushDS(TVariant.Create(integer(false)));
    end;
  finally
    dir.Free;
  end;
end;

//////////////////////////////////////////////////////////////////////////////////////////////////////
procedure  BW_Random(Ex:TContext);// DS: Range -> random(0..range - 1)
var range:TData;
    res:integer;
begin
  range:=Ex.PopDS;
  try
    try
      res:=range.Value;
      res:=random(res);
      Ex.PushDS(TVariant.Create(res));
    except
			Error(errIllegalParameters, ['Random']);
    end;
  finally
    range.Free;
  end;
end;

procedure  BW_Sleep(Ex:TContext);// DS: Sleep ms ->
var ms:TData;
begin
  ms:=Ex.PopDS;
  try
    try
      Sleep(ms.Value);
    except
			Error(errIllegalParameters, ['Sleep']);
    end;
  finally
    ms.Free;
  end;
end;

procedure  BW_MessageDlg(Ex:TContext);// DS: msg ->
var msg:TData;
begin
  msg:=Ex.PopDS;
  try
    try
      MessageDlg(msg.Value, mtInformation, [mbOK], 0);
    except
			Error(errIllegalParameters, ['MessageDlg']);
    end;
  finally
    msg.Free;
  end;
end;

procedure  BW_GetComCtlVersion(Ex:TContext);// DS:  -> ComCtlVersion
begin
  Ex.PushDS(TVariant.Create(GetComCtlVersion()));
end;

procedure  BW_LoadLibrary(Ex:TContext);// DS: LibName -> HInstance
var LibName:TData;
    s:string;
    h:THandle;
    p:TRegisterWordsProc;
begin
  LibName:=Ex.PopDS;
  try
    try
      s:=LibName.Value;
      h:=LoadLibrary(PChar(s));

      p:=GetProcAddress(h, 'RegisterWords');
      if not Assigned(p) then begin
  			Error(errIllegalLibrary, [s]);
      end;

      p(Ex.Words);

      Ex.PushDS(TVariant.Create(integer(h)));
    except
			Error(errIllegalParameters, ['LoadLibrary']);
    end;
  finally
    LibName.Free;
  end;
end;

procedure  BW_FreeLibrary(Ex:TContext);// DS: HInstance ->
var LibHandle:TData;
    p:TRegisterWordsProc;
begin
  LibHandle:=Ex.PopDS;
  try
    try
      p:=GetProcAddress(LibHandle.Value, 'UnRegisterWords');
      if not Assigned(p) then begin
  			Error(errIllegalParameters, ['FreeLibrary']);
      end;

      p(Ex.Words);

      FreeLibrary(LibHandle.Value);
    except
			Error(errIllegalParameters, ['FreeLibrary']);
    end;
  finally
    LibHandle.Free;
  end;
end;
//////////////////////////////////////////////////////////////////////////////////////////////////////

procedure BW_OpenSquareBrackets(Ex:TContext);
begin
  Ex.AddOpSection;
end;

procedure BW_CloseSquareBrackets(Ex:TContext);
begin
  BW_ResumeExpression(Ex, 0);
  Ex.DropOpSection;
  Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
end;

//////////////////////////////////////////////////////////////////////////////////////////////////////
procedure BW_DropUnit(Ex:TContext);
var UnitName:TData;
begin
  unitname:=Ex.PopDS;
  try
    try
      Ex.Owner.Words.DropUnit(unitname.Value);
    except
			Error(errIllegalParameters, ['DropUnit']);
    end;
  finally
    UnitName.Free;
  end;
end;

procedure BW_ScreenHeight(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create(Screen.Height));
end;

procedure BW_ScreenWidth(Ex:TContext);
begin
  Ex.PushDS(TVariant.Create(Screen.Width));
end;

procedure BW_Terminate(Ex:TContext);
begin
  Application.Terminate();
end;

function ExecuteFile(const FileName, Params, DefaultDir: string; ShowCmd: Integer):boolean;
var zFileName, zParams, zDir:PChar;
    pi:TProcessInformation;
    si:TStartupInfo;
    res:cardinal;
begin
  with si do begin
    cb              :=Sizeof(si);
    lpReserved      :=nil;
    lpDesktop       :=nil;
    lpTitle         :=nil;
    dwX             :=0;
    dwY             :=0;
    dwXSize         :=0;
    dwYSize         :=0;
    dwXCountChars   :=0;
    dwYCountChars   :=0;
    dwFillAttribute :=0;
    dwFlags         :=0;
    wShowWindow     :=0;
    cbReserved2     :=0;
    lpReserved2     :=nil;
    hStdInput       :=0;
    hStdOutput      :=0;
    hStdError       :=0;
  end;

  GetMem(zFileName,   Length(FileName) + 2);
  GetMem(zParams,     Length(Params) + 2);
  GetMem(zDir,        Length(DefaultDir) + 2);
  try
    StrPCopy(zFileName, FileName);
    StrPCopy(zParams,   Params);
    StrPCopy(zDir,      DefaultDir);

    Result:=CreateProcess(
        zFileName,                      // pointer to name of executable module
        zParams,                        // pointer to command line string
        nil,	                          // pointer to process security attributes
        nil,	                          // pointer to thread security attributes
        true, 	                        // handle inheritance flag
        0,	                            // creation flags
        nil,	                          // pointer to new environment block
        zDir,                           // pointer to current directory name
        si,	                            // pointer to STARTUPINFO
        pi 	                            // pointer to PROCESS_INFORMATION
    );
  finally
    FreeMem(zFileName);
    FreeMem(zParams);
    FreeMem(zDir);
  end;

  if Result then begin
    res:=WaitForSingleObject(pi.hProcess,  INFINITE);
    case res of
      WAIT_FAILED,
      WAIT_ABANDONED,
      WAIT_TIMEOUT:   Result:=false;

      WAIT_OBJECT_0:  ;
    end;
  end;
end;

procedure BW_ExecCMD(Ex:TContext); // DS FileName, Params, DefaultDir: string; ShowCmd: Integer -> success
var FileName, Params, DefaultDir, ShowCmd:TData;
begin
  ShowCmd   :=Ex.PopDS();
  DefaultDir:=Ex.PopDS();
  Params    :=Ex.PopDS();
  FileName  :=Ex.PopDS();
  try
    try
      Ex.PushDS(TVariant.Create(ExecuteFile(FileName.Value, '_ ' + Params.Value, DefaultDir.Value, ShowCmd.Value)));
    except
			Error(errIllegalParameters, ['ExecCMD']);
    end;
  finally
    ShowCmd.Free();
    DefaultDir.Free();
    Params.Free();
    FileName.Free();
  end;
end;

procedure BW_ShellExecute(Ex:TContext); // DS Operation, FileName, Parameters, Directory, ShowCmd -> success
var Operation   :TData;	// string that specifies operation to perform
    FileName    :TData;	// filename or folder name string
    Parameters  :TData;	// string that specifies executable-file parameters
    Directory   :TData;	// string that specifies default directory
    ShowCmd     :TData; // whether file is shown when opened

    lpOperation :PChar;	// pointer to string that specifies operation to perform
    lpFileName  :PChar;	// pointer to filename or folder name string
    lpParameters:PChar;	// pointer to string that specifies executable-file parameters
    lpDirectory :PChar;	// pointer to string that specifies default directory

    res:cardinal;
    s:string;
begin
  ShowCmd     :=nil;
  Directory   :=nil;
  Parameters  :=nil;
  FileName    :=nil;
  Operation   :=nil;

  lpOperation :=nil;
  lpFileName  :=nil;
  lpParameters:=nil;
  lpDirectory :=nil;

  try
    try
      ShowCmd     :=Ex.PopDS();
      Directory   :=Ex.PopDS();
      Parameters  :=Ex.PopDS();
      FileName    :=Ex.PopDS();
      Operation   :=Ex.PopDS();

      if Length(Operation.Value) > 0 then begin
        GetMem(lpOperation,   Length(Operation.Value)   + 2);
        StrPCopy(lpOperation,   Operation.Value);
      end;
      if Length(FileName.Value)  > 0 then begin
        GetMem(lpFileName,    Length(FileName.Value)    + 2);
        StrPCopy(lpFileName,    FileName.Value);
      end;
      if Length(Parameters.Value)> 0 then begin
        GetMem(lpParameters,  Length(Parameters.Value)  + 2);
        StrPCopy(lpParameters,  Parameters.Value);
      end;
      if Length(Directory.Value) > 0 then begin
        GetMem(lpDirectory,   Length(Directory.Value)   + 2);
        StrPCopy(lpDirectory,   Directory.Value);
      end;

      res:=ShellExecute(Application.MainForm.Handle, lpOperation, lpFileName, lpParameters, lpDirectory, ShowCmd.Value);
      case res of
        0:	                    s:='The operating system is out of memory or resources.';
        ERROR_FILE_NOT_FOUND:   s:='The specified file was not found.';
        ERROR_PATH_NOT_FOUND:   s:='The specified path was not found.';
        ERROR_BAD_FORMAT:	      s:='The .EXE file is invalid (non-Win32 .EXE or error in .EXE image).';
        SE_ERR_ACCESSDENIED:    s:='The operating system denied access to the specified file.';
        SE_ERR_ASSOCINCOMPLETE: s:='The filename association is incomplete or invalid.';
        SE_ERR_DDEBUSY:         s:='The DDE transaction could not be completed because other DDE transactions were being processed.';
        SE_ERR_DDEFAIL:         s:='The DDE transaction failed.';
        SE_ERR_DDETIMEOUT:      s:='The DDE transaction could not be completed because the request timed out.';
        SE_ERR_DLLNOTFOUND:     s:='The specified dynamic-link library was not found.';
        SE_ERR_NOASSOC:         s:='There is no application associated with the given filename extension.';
        SE_ERR_OOM:             s:='There was not enough memory to complete the operation.';
        SE_ERR_SHARE:           s:='A sharing violation occurred.';
        else begin
          s:='';
{
          res:=WaitForSingleObject(res,  INFINITE);
          case res of
            WAIT_FAILED:        s:='WAIT_FAILED';
            WAIT_ABANDONED:     s:='WAIT_ABANDONED';
            WAIT_TIMEOUT:       s:='WAIT_TIMEOUT';
            WAIT_OBJECT_0:      s:='';
          end;
}
        end;
      end;

      Ex.PushDS(TVariant.Create(s));

    except
	  	Error(errIllegalParameters, ['ExecCMD']);
    end;
  finally
    FreeMem(lpOperation);
    FreeMem(lpFileName);
    FreeMem(lpParameters);
    FreeMem(lpDirectory);

    if Assigned(ShowCmd)    then ShowCmd.Free();
    if Assigned(Directory)  then Directory.Free();
    if Assigned(Parameters) then Parameters.Free();
    if Assigned(FileName)   then FileName.Free();
    if Assigned(Operation)  then Operation.Free();
  end;
end;

//////////////////////////////////////////////////////////////////////////////////////////////////////
procedure RegisterWords(Words:TWords);                                       //  func    Imm
begin
  Words.AddWord('SYSTEM', 'UNIT',               BWP(BW_Unit,             0 or atKeyWord));
  Words.AddWord('SYSTEM', 'PROC',               BWP(BW_ProcDefinition,   0 or atKeyWord));
  Words.AddWord('SYSTEM', 'FUNC',               BWP(BW_FuncDefinition,   0 or atKeyWord));
  Words.AddWord('SYSTEM', 'END',                BWP(BW_EndDefinition,    atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'USES',               BWP(BW_Uses,             0 or atKeyWord));
  Words.AddWord('SYSTEM', 'DROPUNIT',           BWP(BW_DropUnit,         atIsFunction));

  Words.AddWord('SYSTEM', '_CreateLocalStack',  BWP(BW_CreateLocalStack, 0));
  Words.AddWord('SYSTEM', '_FreeLocalStack',    BWP(BW_FreeLocalStack,   0));

  Words.AddWord('SYSTEM', '_PushLocalWord',     BWP(BW_PushLocalWord,    0));
  Words.AddWord('SYSTEM', '_AddDSSection',      BWP(BW_AddDSSection,     0));
  Words.AddWord('SYSTEM', '_DropDSSection',     BWP(BW_DropDSSection,    0));
  Words.AddWord('SYSTEM', '_ResumeDSSection',   BWP(BW_ResumeDSSection,  0));

  Words.AddWord('SYSTEM', '_DebugInfo',         BWP(BW_DebugInfo,  0));

  Words.AddWord('SYSTEM', '_Ret',               BWP(BW_Ret,              0));

  Words.AddWord('SYSTEM', 'SWAP',               BWP(BW_Swap,             0 or atKeyWord));
  Words.AddWord('SYSTEM', 'DROP',               BWP(BW_Drop,             0 or atKeyWord));

  Words.AddWord('SYSTEM', 'VAR',                BWP(BW_VAR,              atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', '_CreateVar',         BWP(BW_CreateVar,        0));
  Words.AddWord('SYSTEM', '_ReturnVar',         BWP(BW_ReturnVar,        0));

  Words.AddWord('SYSTEM', '_ApplyProp',         BWP(BW_ApplyProp,        0));
  Words.AddWord('SYSTEM', '_CallMethod',        BWP(BW_CallMethod,       0));

  Words.AddWord('SYSTEM', ',',                  BWP(BW_Comma,            atIsImm));
  Words.AddWord('SYSTEM', '(',                  BWP(BW_OpenRound,        atIsImm));
  Words.AddWord('SYSTEM', ')',                  BWP(BW_CloseRound,       atIsImm));
  Words.AddWord('SYSTEM', '[',                  BWP(BW_OpenSquareBrackets,   atIsImm));
  Words.AddWord('SYSTEM', ']',                  BWP(BW_CloseSquareBrackets,  atIsImm));

  Words.AddWord('SYSTEM', ';',                  BWP(BW_Semicolon,        atIsImm));

  Words.AddWord('SYSTEM', '_GOTO',              BWP(BW_GOTO,             atIsFunction));
  Words.AddWord('SYSTEM', '_IFZEROGOTO',        BWP(BW_IFZEROGOTO,       atIsFunction));
  Words.AddWord('SYSTEM', 'IF',                 BWP(BW_IF,               atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'THEN',               BWP(BW_THEN,             atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'ELSE',               BWP(BW_ELSE,             atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'ENDIF',              BWP(BW_ENDIF,            atIsImm or atKeyWord));

  Words.AddWord('SYSTEM', 'WHILE',              BWP(BW_WHILE,            atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'DO',                 BWP(BW_DO,               atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'ENDWHILE',           BWP(BW_ENDWHILE,         atIsImm or atKeyWord));
  Words.AddWord('SYSTEM', 'ENDW',               BWP(BW_ENDWHILE,         atIsImm or atKeyWord));

  Words.AddWord('SYSTEM', 'Addr',               BWP(BW_GetAddrProc,      atIsImm));

  Words.AddWord('SYSTEM', 'ProcessMessages',    BWP(BW_ProcessMessages,  0));

  Words.AddWord('SYSTEM', 'true',               BWP(BW_True,             0 or atKeyWord));
  Words.AddWord('SYSTEM', 'yes',                BWP(BW_True,             0 or atKeyWord));
  Words.AddWord('SYSTEM', 'false',              BWP(BW_False,            0 or atKeyWord));
  Words.AddWord('SYSTEM', 'no',                 BWP(BW_False,            0 or atKeyWord));
  Words.AddWord('SYSTEM', 'Now',                BWP(BW_Now,              0));
// operators
  Words.AddWord('SYSTEM', '/',                  BWP(BW_Div,              atIsOperator or atPrecedence1));
  Words.AddWord('SYSTEM', '*',                  BWP(BW_Mul,              atIsOperator or atPrecedence1));
  Words.AddWord('SYSTEM', 'mod',                BWP(BW_Mod,              atIsOperator or atPrecedence1 or atKeyWord));
  Words.AddWord('SYSTEM', 'div',                BWP(BW_IntDiv,           atIsOperator or atPrecedence1 or atKeyWord));
  Words.AddWord('SYSTEM', 'and',                BWP(BW_And,              atIsOperator or atPrecedence1 or atKeyWord));

  Words.AddWord('SYSTEM', 'or',                 BWP(BW_Or,               atIsOperator or atPrecedence2 or atKeyWord));
  Words.AddWord('SYSTEM', 'xor',                BWP(BW_Xor,              atIsOperator or atPrecedence2 or atKeyWord));
  Words.AddWord('SYSTEM', '-',                  BWP(BW_Sub,              atIsOperator or atPrecedence2));
  Words.AddWord('SYSTEM', '+',                  BWP(BW_Sum,              atIsOperator or atPrecedence2));

  Words.AddWord('SYSTEM', '>',                  BWP(BW_Greate,           atIsOperator or atPrecedence3));
  Words.AddWord('SYSTEM', '<',                  BWP(BW_Less,             atIsOperator or atPrecedence3));
  Words.AddWord('SYSTEM', '<>',                 BWP(BW_NotEQ,            atIsOperator or atPrecedence3));
  Words.AddWord('SYSTEM', '==',                 BWP(BW_EQ,               atIsOperator or atPrecedence3));
  Words.AddWord('SYSTEM', '>=',                 BWP(BW_GreateEQ,         atIsOperator or atPrecedence3));
  Words.AddWord('SYSTEM', '<=',                 BWP(BW_LessEQ,           atIsOperator or atPrecedence3));

  Words.AddWord('SYSTEM', '=',                  BWP(BW_Assign,           atIsOperator or atPrecedence4));
  Words.AddWord('SYSTEM', '+=',                 BWP(BW_SumAssign,        atIsOperator or atPrecedence4));
  Words.AddWord('SYSTEM', '-=',                 BWP(BW_SubAssign,        atIsOperator or atPrecedence4));
  Words.AddWord('SYSTEM', '*=',                 BWP(BW_MulAssign,        atIsOperator or atPrecedence4));
  Words.AddWord('SYSTEM', '/=',                 BWP(BW_DivAssign,        atIsOperator or atPrecedence4));
//
  Words.AddWord('SYSTEM', 'Print',              BWP(BW_Print,            atIsFunction));
  Words.AddWord('SYSTEM', 'Trunc',              BWP(BW_Trunc,            atIsFunction));
  Words.AddWord('SYSTEM', 'Frac',               BWP(BW_Frac,             atIsFunction));
  Words.AddWord('SYSTEM', 'Round',              BWP(BW_Round,            atIsFunction));

  Words.AddWord('SYSTEM', 'StrCopy',            BWP(BW_StrCopy,          atIsFunction));
  Words.AddWord('SYSTEM', 'StrPos',             BWP(BW_StrPos,           atIsFunction));
  Words.AddWord('SYSTEM', 'StrDelete',          BWP(BW_StrDelete,        atIsFunction));
  Words.AddWord('SYSTEM', 'Length',             BWP(BW_Length,           atIsFunction));
  Words.AddWord('SYSTEM', 'Upper',              BWP(BW_Upper,            atIsFunction));
  Words.AddWord('SYSTEM', 'Trim',               BWP(BW_Trim,             atIsFunction));

  Words.AddWord('SYSTEM', 'ParamStr',           BWP(BW_ParamStr,         atIsFunction));
  Words.AddWord('SYSTEM', 'ParamCount',         BWP(BW_ParamCount,       atIsFunction));
  Words.AddWord('SYSTEM', 'not',                BWP(BW_Not,              atIsFunction or atKeyWord));
  Words.AddWord('SYSTEM', 'MAX',                BWP(BW_Max,              atIsFunction or atKeyWord));
  Words.AddWord('SYSTEM', 'MIN',                BWP(BW_Min,              atIsFunction or atKeyWord));
  Words.AddWord('SYSTEM', 'Str',                BWP(BW_STR,              atIsFunction));
  Words.AddWord('SYSTEM', 'Curr',               BWP(BW_CURR,             atIsFunction));
  Words.AddWord('SYSTEM', 'ShowUnits',          BWP(BW_ShowUnits,        atIsFunction));
  Words.AddWord('SYSTEM', 'ShowUnit',           BWP(BW_ShowUnit,         atIsFunction));
  Words.AddWord('SYSTEM', 'CLS',                BWP(BW_CLS,              atIsFunction));
  Words.AddWord('SYSTEM', 'DateToStr',          BWP(BW_DateToStr,        atIsFunction));
  Words.AddWord('SYSTEM', 'StrToDate',          BWP(BW_StrToDate,        atIsFunction));
  Words.AddWord('SYSTEM', 'GetTickCount',       BWP(BW_GetTickCount,     atIsFunction));

  Words.AddWord('SYSTEM', 'ERROR',              BWP(BW_Error,            atIsFunction));

  Words.AddWord('SYSTEM', 'FilePath',           BWP(BW_FilePath,         atIsFunction));
  Words.AddWord('SYSTEM', 'FileName',           BWP(BW_FileName,         atIsFunction));
  Words.AddWord('SYSTEM', 'FileNameOnly',       BWP(BW_FileNameOnly,     atIsFunction));
  Words.AddWord('SYSTEM', 'FileExists',         BWP(BW_FileExists,       atIsFunction));
  Words.AddWord('SYSTEM', 'DeleteFile',         BWP(BW_DeleteFile,       atIsFunction));
  Words.AddWord('SYSTEM', 'CopyFile',           BWP(BW_CopyFile,         atIsFunction));
  Words.AddWord('SYSTEM', 'MoveFile',           BWP(BW_MoveFile,         atIsFunction));
  Words.AddWord('SYSTEM', 'MakeDir',            BWP(BW_MakeDir,          atIsFunction));

  Words.AddWord('SYSTEM', 'Sleep',              BWP(BW_Sleep,            atIsFunction));
  Words.AddWord('SYSTEM', 'Random',             BWP(BW_Random,           atIsFunction));
  Words.AddWord('SYSTEM', 'MessageDlg',         BWP(BW_MessageDlg,       atIsFunction));

  Words.AddWord('SYSTEM', 'ScreenHeight',       BWP(BW_ScreenHeight,        0));
  Words.AddWord('SYSTEM', 'ScreenWidth',        BWP(BW_ScreenWidth,        0));

{
  Words.AddWord('SYSTEM', 'GetComCtlVersion',   BWP(BW_GetComCtlVersion, atIsFunction));
	Words.AddWord('SYSTEM', 'ComCtlVersionIE3',   BWV($00040046));//The version of ComCtl32.dll installed with Internet Explorer version 3.
	Words.AddWord('SYSTEM', 'ComCtlVersionIE4',   BWV($00040047));//The version of ComCtl32.dll installed with Internet Explorer version 4.
	Words.AddWord('SYSTEM', 'ComCtlVersionIE401', BWV($00040048));//The version of ComCtl32.dll installed with Internet Explorer version 4.01.
	Words.AddWord('SYSTEM', 'ComCtlVersionIE5',   BWV($00050050));//The version of ComCtl32.dll installed with Internet Explorer version 5.
}
  Words.AddWord('SYSTEM', 'LoadLibrary',        BWP(BW_LoadLibrary,   atIsFunction));
  Words.AddWord('SYSTEM', 'FreeLibrary',        BWP(BW_FreeLibrary,   atIsFunction));

//  Words.AddWord('SYSTEM', 'GetInt',           BWP(BW_GetInt,           true,   false));
//  Words.AddWord('SYSTEM', 'GetIntIni',        BWP(BW_GetIntIni,        true,   false));

  Words.AddWord('SYSTEM', 'Terminate',          BWP(BW_Terminate,     atIsFunction));
  Words.AddWord('SYSTEM', 'ExecCMD',            BWP(BW_ExecCMD,       atIsFunction));
  Words.AddWord('SYSTEM', 'ShellExecute',       BWP(BW_ShellExecute,  atIsFunction));
end;

end.

