unit AInterpreter;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

{$INCLUDE DEFINES.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, IniFiles, Menus,

  ARUtils, ErrorMsg, ARVFS;//, SQL;//, Debug;

const
    atIsBase      = $0001;
    atIsImm       = $0002;
    atIsFunction  = $0004;
    atIsOperator  = $0004 or $0008; // +,-,= etc.
    atKeyWord     = $0010;
    atBreakPoint  = $0100;

    atPrecedenceMask  = $7000;     // operators precedence mask
    atPrecedence0     = $0000;     // First (highest), default for all function
    atPrecedence1     = $1000;     // Second
    atPrecedence2     = $2000;     // Third
    atPrecedence3     = $3000;     // Fourth
    atPrecedence4     = $4000;     // Fifth (lowest)
type

  TScriptEventHandler = integer;

  TRefIsMethod = function(APropList:TAIStringList):boolean of object;

  TTermAttr = (taError, taString, taNumber, taWord);

  TTerm = record
    Attr  :TTermAttr;
    Value :string;
  end;

function  ObjAsVar  (o:TObject):variant;
function  VarAsObj  (const v:variant):TObject;
function  IsVariant (const v:variant):boolean;
function  IsObject  (const v:variant):boolean;

type
  TContext = class;
  TBO = class;

  EStepOnRun = class(Exception)
    Ex  :TContext;
    constructor Create(AEx:TContext);
  end;
////////////////////////////////////////////////////////////////////////////////
  TAIProc = procedure(Ex:TContext);
//  PAIProc = ^TAIProc;


  TBaseProc = procedure(Ex:TContext);

  TShowMessageEvent             = procedure(Sender:TObject; Ex:TContext; msg:string) of object;
  TShowErrorEvent               = procedure(Sender:TObject; Ex:TContext; msg:string) of object;

  TShowEditorDialogEvent        = procedure(Sender:TObject) of object;
  TShowProjectManagerDialogEvent= procedure(Sender:TObject) of object;

  TRegisterWordsEvent           = procedure(Sender:TObject) of object;

  TUndeclaredWordEvent          = procedure(Sender:TObject; Ex:TContext; const term:string; var Processed:boolean) of object;

  TBO = class(TObject)
  private
    function    GetText:string;virtual;
    procedure   SetText(s:string);virtual;
  public
    property    Text:string read GetText write SetText;

    constructor Create;
    destructor  Destroy;override;
  end;
////////////////////////////////////////////////////////////////////////////////
  TAddr = class(TBO)
  private
    Attr        :integer;
    function    GetText:string;override;
    procedure   SetText(s:string);override;
    function    GetAddr:pointer;virtual;
    procedure   SetAddr(p:pointer);

    function    GetIsBase:boolean;
    procedure   SetIsBase(AIsBase:boolean);

    function    GetIsImm:boolean;
    procedure   SetIsImm(AIsImm:boolean);

    function    GetIsFunction:boolean;
    procedure   SetIsFunction(AIsFunction:boolean);

    function    GetIsOperator:boolean;

    function    GetBreakPoint:boolean;
    procedure   SetBreakPoint(ABreakPoint:boolean);

    function    GetKeyWord:boolean;
    procedure   SetKeyWord(AKeyWord:boolean);
  public
    Value        :integer;

    property    IsBase      :boolean  read GetIsBase      write SetIsBase;
    property    IsImm       :boolean  read GetIsImm       write SetIsImm;
    property    IsFunction  :boolean  read GetIsFunction  write SetIsFunction;
    property    IsOperator  :boolean  read GetIsOperator;
    property    BreakPoint  :boolean  read GetBreakPoint  write SetBreakPoint;
    property    KeyWord     :boolean  read GetKeyWord      write SetKeyWord;

//    constructor Create(p:pointer; AIsBase:boolean; AIsFunction:boolean; AIsImm:boolean);
    constructor Create(p:pointer; AAttr:integer);
    destructor  Destroy;override;

    procedure   Assign(O:TAddr);
    property    Addr  :pointer  read GetAddr write SetAddr;
  end;
///////////////////////////////////// TData ////////////////////////////////////
  TData = class(TBO)
  private
    PPropList   :TAIStringList;
    PushList    :TAIStringList;

    function    GetPropList:TAIStringList;
    procedure   SetPropList(APropList:TAIStringList);
    function    GetPropListCount:integer;
  protected
    function    GetValue:variant;virtual;
    procedure   SetValue(AValue:variant);virtual;
  public
    ProcessedProp:boolean;
    UseCount    :integer;

    constructor  Create;
    destructor  Destroy;override;
    procedure    Free;

    property     PropListCount:integer read GetPropListCount;
    property     PropList:TAIStringList read GetPropList write SetPropList;
    procedure    PushPropList;
    procedure    PopPropList;

    property     Value:variant read GetValue write SetValue;
    property     Text:string read GetText write SetText;

    procedure    CallMethod(Ex:TContext);virtual;

    procedure    Assign(D:TData);virtual;
  end;

///////////////////////////////////// TVariant /////////////////////////////////
  TVariant = class(TData)
  private
    V            :variant;
  protected
    function    GetValue:variant;override;
    procedure   SetValue(AValue:variant);override;

    function    GetText:string;override;
    procedure   SetText(s:string);override;
  public
    constructor Create(AValue:variant);

    procedure   CallMethod(Ex:TContext);override;
    procedure   Assign(D:TData);override;
  end;

///////////////////////////////////// TDataRef//////////////////////////////////
  TDataRef = class(TData)
  private
  protected
    function    GetValue:variant;override;
    procedure   SetValue(AValue:variant);override;

    function    GetText:string;override;
    procedure   SetText(s:string);override;
  public
    FValue      :TData;

    constructor Create(prop:string;AValue:TData);
    destructor  Destroy;override;

    procedure   Assign(D:TData);override;

    procedure   CallMethod(Ex:TContext);override;
  end;

  TWords = class
  private
  public
    UnitList    :TAIStringList;
    KeyWordList :TAIStringList;

    procedure   AddUnit(AUnitName:string);
    procedure   DropUnit(AUnitName:string);
    function    FindUnit(AUnitName:string):boolean;

    procedure   AddWord(AUnitName:string; AWordName:string; w:TBO);
    procedure   DropWord(AUnitName:string; AWordName:string);
    function    FindWord(AUnitList:TAIStringList; AWordName:string):TBO;

    constructor Create;
    destructor  Destroy;override;
  end;

  TCustomEditorDialog = class(TForm)
    procedure   FormCloseQuery(Sender: TObject; var CanClose: Boolean);virtual;abstract;
    procedure   FormClose(Sender: TObject; var Action: TCloseAction);virtual;abstract;
    procedure   FormShow(Sender: TObject);virtual;abstract;
    procedure   FormDestroy(Sender: TObject);virtual;abstract;
  public
    procedure   ShowErrorPos(AFileName:string;APos:integer);virtual;abstract;
    procedure   ShowStopPos(AFileName:string;APos:integer);virtual;abstract;
    procedure   LoadFile(AFileName:string);virtual;abstract;
    procedure   ClearMsg;virtual;abstract;
    procedure   ShowMessage(msg:string);virtual;abstract;
    procedure   ShowError(msg:string);virtual;abstract;
    procedure   UpdateWatches;virtual;abstract;
  end;

  TCustomProjectManagerDialog = class(TForm)
  end;

  TAInterpreter = class(TComponent)
  private
    FProjectName:TFileName;

    FFileSystem :TVirtualFileSystem;

    FOnShowMessage              :TShowMessageEvent;
    FOnShowError                :TShowErrorEvent;
    FOnShowEditorDialog         :TShowEditorDialogEvent;
    FOnShowProjectManagerDialog :TShowProjectManagerDialogEvent;

    FOnRegisterWords            :TRegisterWordsEvent;

    FUndeclaredWordEvent        :TUndeclaredWordEvent;
  protected
    function    GetFileSystem:TVirtualFileSystem;
    procedure   SetFileSystem(AFileSystem:TVirtualFileSystem);

    procedure   SetProjectName(AProjectName:TFileName);virtual;
    procedure   StepIP(Ex:TContext);
    procedure   ExecString(Ex:TContext;term:string);
    procedure   ExecNumber(Ex:TContext;term:string);
    procedure   ExecWord(Ex:TContext;term:string);
    procedure   Loaded;override;
  public
    CurrContext                 :TContext;

    Words       :TWords;
    ProjectList :TAIStringList;

    EditorDialog        :TCustomEditorDialog;
    ProjectManagerDialog:TCustomProjectManagerDialog;

    CompileWithDebugInfo:boolean;
    StepOnRun   :boolean;

    procedure   ShowMessage(Ex:TContext; msg:string);
    procedure   ClearMsg;
    procedure   ShowError(Ex:TContext; msg:string);

//    procedure   RunScript(scr:PChar);overload;
    procedure   RunScript(Sender:TObject; scr:PChar);
    procedure   RunFile(FileName:string);

    function    GetAddr(AUnitName:string; AWordName:string):TAddr;
    function    GetVarAddr(AUnitName:string; AWordName:string):TData;

    procedure   ExecProc(HOwner:TData; HEventHanler:TScriptEventHandler; Params:array of variant);
    procedure   ExecProcTData(HEventHanler:TScriptEventHandler; Params:array of TData);
    procedure   ExecProcEmpty(HOwner:TData; HEventHanler:TScriptEventHandler);
    function    ExecFuncTData(AddrProc:pointer; Params:array of TData):TData;

    procedure   RunProc(AUnitName:string; AProcName:string; Params:array of variant);
    procedure   RunProcTData(AUnitName:string; AProcName:string; Params:array of TData);
    function    RunFunc(AUnitName:string; AProcName:string; Params:array of variant):TData;
    function    RunFuncTData(AUnitName:string; AProcName:string; Params:array of TData):TData;

    procedure   Interpret(Ex:TContext; NewCS:pointer);
    procedure   Continue(Ex:TContext);
    procedure   Reset;

    procedure   ShowEditorDialog(AFileName:string);virtual;
    procedure   ShowProjectManagerDialog;virtual;
  published
    property    ProjectName :TFileName        read FProjectName  write SetProjectName;
    property    FileSystem:TVirtualFileSystem read GetFileSystem  write SetFileSystem;

    property    OnShowMessage             :TShowMessageEvent
      read FOnShowMessage                  write FOnShowMessage;
    property    OnShowError               :TShowErrorEvent
      read FOnShowError                 write FOnShowError;
    property    OnShowEditorDialog        :TShowEditorDialogEvent
      read FOnShowEditorDialog          write FOnShowEditorDialog;
    property    OnShowProjectManagerDialog:TShowProjectManagerDialogEvent
      read FOnShowProjectManagerDialog  write FOnShowProjectManagerDialog;

    property    OnRegisterWords           :TRegisterWordsEvent
      read FOnRegisterWords             write FOnRegisterWords;

    property    OnUndeclaredWordEvent     :TUndeclaredWordEvent
      read FUndeclaredWordEvent         write FUndeclaredWordEvent;

    constructor Create(AOwner: TComponent);override;
    destructor  Destroy;override;
  end;

  TContext = class
  private
    // list of word local lists
    // on proc call, adds new TAIList
    // (declared local vars)
    // before proc exit, top LocWS item dropped
    BackTermList  :TAIStringList;

    function      GetLocalVar(idx:integer):TData;
    function      GetOp:TAIList;
    function      GetDS:TAIList;
  public
    LocWS         :TAIList;

    Owner         :TAInterpreter;
    Sender        :TObject;

    Buf           :PChar;
    Pos           :integer;

    CurrUnit      :string;
    UnitList      :TAIStringList;
    // local vars list on compile time
    // first words search in LocalWords
    LocalWords    :TAIStringList;
    // to called proc, pointer to current CS (compile stack) placed on proc
    CS            :TAIList;

    DSList        :TAIList;
    // compiled statement stack - PROC, FUNC, WHILE, IF etc.
    StatementList :TAIStringList;

    // compiled func stack list
    OpList        :TAIList;

    // IP (instruction pointer) - cuurent executed addr
    IP            :integer;
    //
    JumpList      :TAIList;
    // on proc call, stack RSIP <- current IP + 1,
    // and RSCS <- current proc CS,
    // CurrCS sets to called proc CS, IP = 0
    RSIP          :TAIList;
    RSCS          :TAIList;

    CompileMode   :boolean;
    CurrExecLine  :integer;

    OldCurrSrcLine:integer;
    CurrSrcLine   :integer;

    // link to TAInterpreter.Words
    Words         :TWords;
    // link to TAInterpreter.ProjectList
    ProjectList   :TAIStringList;

    constructor Create(aBuf:PChar;AOwner:TAInterpreter);
    destructor  Destroy;override;

    procedure   PushRS(NewCS:pointer);
    procedure   PopRS;

    function    GetNextTerm:TTerm;
    procedure   PushBackTerm(term:TTerm);

    function    TopStatement:string;
    procedure   DropTopStatement;

    procedure   PushJamp(JampAddr:integer);
    function    PopJamp:integer;

    function    GetTopCSAsInt:integer;

    procedure   CompileData(d:TData);
    procedure   CompileDataAt(idx:integer;d:TData);
    procedure   CompileAddr(a:TAddr);

    procedure   CreateLocalStack;
    procedure   FreeLocalStack;
//    procedure   AddLocalVar(d:TData);
    procedure   AddLocalVar(VarName:string; d:TData);

    property    LocalVar[idx:integer]:TData read GetLocalVar;

    property    Op:TAIList read GetOp;
    procedure   ResumeOperand(o:TBO);
    procedure   AddOperand(operand:TBO);
    procedure   AddOpSection;
    procedure   DropOpSection;

    property    DS:TAIList read GetDS;
    function    PushDS(p:TData):TData;
    function    PopDS:TData;
    procedure   AddDSSection;
    procedure   DropDSSection;

    procedure   StepIP;
    procedure   Interpret(NewCS:pointer);
    procedure   RunScript(scr:PChar);
    procedure   RunFile(FileName:string);
    procedure   ShowMessage(msg:string);

//    function    GetAddr(AUnitName:string; AWordName:string):TAddr;

    procedure   CompileLineNo;
  end;

//function     BWP(Addr:PAIProc; AAttr:integer):TBO;
function     BWP(Addr:TAIProc; AAttr:integer):TBO;
function     BWV(v:variant):TBO;

var
  BOCount         :integer;
  AddrCount       :integer;
  DataCount       :integer;

  AI:TAInterpreter;

procedure Register;

implementation
{$R AINTERPRETER.RES}

uses
  AIEditor,
  AIProjectManager,
  AIWords,
  AIObject,
  AIList,
  AIRecord,
  AIVarList

{  ,
  AIReportOld,
  AIUserRep,
  AIProgress,
  AIGUI,
  AISQL,
  AIDateEdit,
  AIButton,
  AILabel,
  AIPanel,
  AICheckBox,
  AIEdit,
  AIRadioGroup,
  AIGroupBox,
  AIMenu
}
;


procedure Register;
begin
  RegisterComponents('ARComponents', [TAInterpreter]);
end;

// WARNING - to hold TObject in variant - used varError type!
// other types cause side effects.
function  ObjAsVar(o:TObject):variant;
begin
  VarClear(Result);
  TVarData(Result).VType:=varError;
  TVarData(Result).VUnknown:=o;
end;

function  VarAsObj(const v:variant):TObject;
begin
  if VarType(v) = varError then begin
    Result:=TVarData(v).VUnknown;
  end else begin
    Result:=nil;
  end;
end;

function  IsVariant(const v:variant):boolean;
begin
  Result:=not VarType(v) = varError;
end;

function  IsObject(const v:variant):boolean;
begin
  Result:=VarType(v) = varError;
end;

constructor EStepOnRun.Create(AEx:TContext);
begin
  inherited Create('');

  Ex  :=AEx;
end;

constructor TAInterpreter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Words:=TWords.Create;
  ProjectList:=TAIStringList.Create('ProjectList');
  AI:=Self;
end;

destructor  TAInterpreter.Destroy;
begin
  ProjectList.Free;
  Words.Free;

  inherited;
end;

{
procedure   TAInterpreter.SetDatabase(ADatabase:TDatabase);
begin
  FDatabase:=ADatabase;
  SQL.DefaultDatabaseName:=Database.DatabaseName;
end;
}

procedure   TAInterpreter.SetProjectName(AProjectName:TFileName);
var inf:TIniFile;
begin
  if FProjectName <> AProjectName then begin
    FProjectName:=AProjectName;
    if (not (csDesigning in ComponentState)) and (Length(ProjectName) > 0) then begin
      try
        inf:=TIniFile.Create(ExtractFilePathAndNameOnly(ParamStr(0)) + '.ini');
        try
          inf.WriteString('AIProject', 'ProjectName', ProjectName);
        finally
          inf.Free;
        end;
      except
      end;
    end;
  end;
end;

procedure    TAInterpreter.ExecProcEmpty(HOwner:TData; HEventHanler:TScriptEventHandler);
var Ex:TContext;
begin
  if HEventHanler <> 0 then begin
    Ex:=TContext.Create(nil, Self);
    try
      Ex.PushDS(HOwner);

      Interpret(Ex, pointer(HEventHanler));
    finally
      Ex.Free;
    end;
  end else begin
    HOwner.Free;
  end;
end;

procedure   TAInterpreter.ExecString(Ex:TContext;term:string);
begin
  if Ex.CompileMode then begin
    Ex.CompileData(TVariant.Create(term));
  end else begin
    Ex.PushDS(TVariant.Create(term));
  end;
end;

procedure   TAInterpreter.ExecNumber(Ex:TContext;term:string);
var OldDecSep:char;
    code:integer;
    value:extended;
begin
  OldDecSep:=DecimalSeparator;
  DecimalSeparator:='.';
  Val(term, value, code);
  DecimalSeparator:=OldDecSep;

  if Ex.CompileMode then begin
    Ex.CompileData(TVariant.Create(value));
  end else begin
    Ex.PushDS(TVariant.Create(value));
  end;
end;

procedure   TAInterpreter.ExecWord(Ex:TContext;term:string);
  procedure ExecProp(prop:string);
  var i:integer;
      pl:TAIStringList;
  begin
    if (Length(prop) > 0) then begin
      pl:=TAIStringList.Create('pl');
      try
        StrToPropList(prop, pl);
        for i:=0 to pl.Count - 1 do begin
          if Ex.CompileMode then begin
            Ex.CompileData(TVariant.Create(pl[i]));
            Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
          end else begin
            Ex.PushDS(TVariant.Create(pl[i]));
            BW_ApplyProp(Ex);
          end;
        end;
      finally
        pl.Free;
      end;
    end;
  end;

var idx:integer;
    proc:TAIProc;
    prop:string;
    CurrentWord:TBO;
    NextTerm:TTerm;
    Processed:boolean;
begin
  prop:='';
  prop:=GetProp(term);

  if (Length(term) > 0) then begin
    term:=Upper(term);
    idx:=Ex.LocalWords.IndexOf(term);
    if idx <> -1 then begin

      NextTerm:=Ex.GetNextTerm;
      if NextTerm.Value = '(' then begin
        Ex.AddOpSection;
        Ex.AddOperand(TDataRef.Create(prop, TVariant.Create(idx)));
      end else begin
        Ex.PushBackTerm(NextTerm);
        Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
        Ex.CompileData(TVariant.Create(idx));
        ExecProp(prop);
      end;
    end else begin
      CurrentWord:=Words.FindWord(Ex.UnitList, term);
      if not Assigned(CurrentWord) then begin
        Processed:=false;
        AIObject.ProcessUndeclaredWord(Ex, term, Processed);

        if not Processed then begin
          if Assigned(FUndeclaredWordEvent) then begin
            FUndeclaredWordEvent(Self, Ex, term, Processed);
            if not Processed then begin
              Error(errUndeclaredWord, [term]);
            end;
          end else begin
            Error(errUndeclaredWord, [term]);
          end;
        end;
      end else begin
        if CurrentWord is TAddr then begin
          proc:=TAddr(CurrentWord).Addr;
          if TAddr(CurrentWord).IsImm then begin
            if TAddr(CurrentWord).IsBase then begin
              proc(Ex);
            end else begin
              Interpret(Ex, TAddr(CurrentWord).Addr);
            end;
          end else begin
            if Ex.CompileMode then begin
///
              if TAddr(CurrentWord).IsOperator then begin
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                NextTerm:=Ex.GetNextTerm;
                if NextTerm.Value = '(' then begin
                  Ex.AddOpSection;
                  Ex.AddOperand(TAddr(CurrentWord));
                end else begin
                  Ex.PushBackTerm(NextTerm);
                  Ex.CompileAddr(TAddr(CurrentWord));
                end;
              end;
///
            end else begin
              if TAddr(CurrentWord).IsOperator then begin
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                NextTerm:=Ex.GetNextTerm;
                if NextTerm.Value = '(' then begin
                  Ex.AddOpSection;
                  Ex.AddOperand(TAddr(CurrentWord));
                end else begin
                  Ex.PushBackTerm(NextTerm);
                  if TAddr(CurrentWord).IsBase then begin
                    proc(Ex);
                  end else begin
                    Interpret(Ex, TAddr(CurrentWord).Addr);
                  end;
                end;
              end;
            end;
          end;
        end else begin
          if CurrentWord is TData then begin
            if Ex.CompileMode then begin
              Ex.CompileData(TData(CurrentWord));
              ExecProp(prop);
            end else begin
              Ex.PushDS(TData(CurrentWord));
              ExecProp(prop);
            end
          end else begin
          end;
        end;
      end;
    end;
  end;
end;
{
procedure   TAInterpreter.ExecWord(Ex:TContext;term:string);
var idx:integer;
    proc:TAIProc;
    prop:string;
    CurrentWord:TBO;
    NextTerm:TTerm;
    pl:TAIStringList;
begin
  prop:='';
  prop:=GetProp(term);

  if (Length(term) = 0) then begin
    Ex.CompileData(TVariant.Create(prop));
    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
  end else begin
    term:=Upper(term);
    idx:=Ex.LocalWords.IndexOf(term);
    if idx <> -1 then begin
      if (Length(prop) > 0) then begin
        pl:=TAIStringList.Create('pl');
        StrToPropList(prop, pl);
        try
          NextTerm:=Ex.GetNextTerm;
          if NextTerm.Value = '(' then begin
            Ex.AddOpSection;
            Ex.AddOperand(TDataRef.Create(prop, TVariant.Create(idx)));
          end else begin
            Ex.PushBackTerm(NextTerm);
            Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
            Ex.CompileData(TVariant.Create(idx));
            Ex.CompileData(TVariant.Create(prop));
            Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
          end;
        finally
          pl.Free;
        end;
      end else begin
        Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
        Ex.CompileData(TVariant.Create(idx));
      end;
    end else begin
      CurrentWord:=Words.FindWord(Ex.UnitList, term);
      if not Assigned(CurrentWord) then begin
        Error(errUndeclaredWord, [term]);
      end else begin
        if CurrentWord is TAddr then begin
          proc:=TAddr(CurrentWord).Addr;
          if TAddr(CurrentWord).IsImm then begin
            if TAddr(CurrentWord).IsBase then begin
              proc(Ex);
            end else begin
              Interpret(Ex, TAddr(CurrentWord).Addr);
            end;
          end else begin
            if Ex.CompileMode then begin
///
              if TAddr(CurrentWord).IsOperator then begin
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                NextTerm:=Ex.GetNextTerm;
                if NextTerm.Value = '(' then begin
                  Ex.AddOpSection;
                  Ex.AddOperand(TAddr(CurrentWord));
                end else begin
                  Ex.PushBackTerm(NextTerm);
                  Ex.CompileAddr(TAddr(CurrentWord));
                end;
              end;
///
            end else begin
              if TAddr(CurrentWord).IsOperator then begin
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                NextTerm:=Ex.GetNextTerm;
                if NextTerm.Value = '(' then begin
                  Ex.AddOpSection;
                  Ex.AddOperand(TAddr(CurrentWord));
                end else begin
                  Ex.PushBackTerm(NextTerm);
                  if TAddr(CurrentWord).IsBase then begin
                    proc(Ex);
                  end else begin
                    Interpret(Ex, TAddr(CurrentWord).Addr);
                  end;
                end;
              end;
            end;
          end;
        end else begin
          if CurrentWord is TData then begin
            if Ex.CompileMode then begin
              Ex.CompileData(TData(CurrentWord));
              if Length(prop) > 0 then begin
                Ex.CompileData(TVariant.Create(prop));
                Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
              end;
            end else begin
              Ex.PushDS(TData(CurrentWord));
              Ex.PushDS(TVariant.Create(prop));
              BW_ApplyProp(Ex);
            end
          end else begin
          end;
        end;
      end;
    end;
  end;
end;

}
{
procedure   TAInterpreter.ExecWord(Ex:TContext;term:string);
var idx:integer;
    proc:TAIProc;
    prop:string;
    CurrentWord:TBO;
    NextTerm:TTerm;
    reffunc:TRefIsMethod;
    pl:TAIStringList;
begin
  prop:='';
  prop:=GetProp(term);

//  if (term = '') and (prop <> '') then begin
//    Ex.CompileData(TVariant.Create(prop));
//    Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
//  end else begin
    term:=Upper(term);
    idx:=Ex.LocalWords.IndexOf(term);
    if idx <> -1 then begin
      if (Length(prop) > 0) then begin
        @reffunc:=Ex.LocalWords.Objects[idx];
        pl:=TAIStringList.Create('pl');
        StrToPropList(prop, pl);
        try
          if reffunc(pl) then begin
            NextTerm:=Ex.GetNextTerm;
            if NextTerm.Value = '(' then begin
              Ex.AddOpSection;
            end else begin
              Ex.PushBackTerm(NextTerm);
            end;
            Ex.AddOperand(TDataRef.Create(prop, TVariant.Create(idx)));
          end else begin
            Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
            Ex.CompileData(TVariant.Create(idx));
            Ex.CompileData(TVariant.Create(prop));
            Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
          end;
        finally
          pl.Free;
        end;
      end else begin
        Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_PushLocalWord'));
        Ex.CompileData(TVariant.Create(idx));
      end;
    end else begin
      CurrentWord:=Words.FindWord(Ex.UnitList, term);
      if not Assigned(CurrentWord) then begin
        Error(errUndeclaredWord, [term]);
      end else begin
        if CurrentWord is TAddr then begin
          proc:=TAddr(CurrentWord).Addr;
          if TAddr(CurrentWord).IsImm then begin
            if TAddr(CurrentWord).IsBase then begin
              proc(Ex);
            end else begin
              Interpret(Ex, TAddr(CurrentWord).Addr);
            end;
          end else begin
            if Ex.CompileMode then begin
///
              if TAddr(CurrentWord).IsFunction then begin
                if not TAddr(CurrentWord).IsOperator then begin
                  NextTerm:=Ex.GetNextTerm;
                  if NextTerm.Value = '(' then begin
                    Ex.AddOpSection;
                  end else begin
                    Ex.PushBackTerm(NextTerm);
                  end;
                end;
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                Ex.CompileAddr(TAddr(CurrentWord));
              end;
            end else begin
              if TAddr(CurrentWord).IsFunction then begin
                if not TAddr(CurrentWord).IsOperator then begin
                  NextTerm:=Ex.GetNextTerm;
                  if NextTerm.Value = '(' then begin
                    Ex.AddOpSection;
                  end else begin
                    Ex.PushBackTerm(NextTerm);
                  end;
                end;
                Ex.AddOperand(TAddr(CurrentWord));
              end else begin
                if TAddr(CurrentWord).IsBase then begin
                  proc(Ex);
                end else begin
                  Interpret(Ex, TAddr(CurrentWord).Addr);
                end;
              end;
            end
          end
        end else begin
          if CurrentWord is TData then begin
            if Ex.CompileMode then begin
              Ex.CompileData(TData(CurrentWord));
              if Length(prop) > 0 then begin
                Ex.CompileData(TVariant.Create(prop));
                Ex.CompileAddr(Ex.Owner.GetAddr('SYSTEM', '_ApplyProp'));
              end;
            end else begin
              Ex.PushDS(TData(CurrentWord));
              Ex.PushDS(TVariant.Create(prop));
              BW_ApplyProp(Ex);
            end
          end else begin
          end;
        end;
      end;
    end;
//  end;
end;

}

procedure   TAInterpreter.RunFile(FileName:string);
  procedure   IntRunFile(FileName:string);
  var sl:TAIStringList;
      s:string;
  begin
    sl:=TAIStringList.Create('IE run file');
    try
      try
        sl.LoadFromStream(FileSystem.CreateReadStream(FileName));
      finally
        FileSystem.CloseReadStream;
      end;

      s:=sl.Text;
      RunScript(nil, PChar(s));
    finally
      sl.Free;
    end;
  end;
begin
  if FileSystem.FileExists(FileName) then begin
    IntRunFile(FileName);
  end else begin
    if FileSystem.FileExists(FileSystem.ExtractPath(ProjectName) + FileName) then begin
      IntRunFile(FileSystem.ExtractPath(ProjectName) + FileName);
    end;
  end;
end;

procedure   TAInterpreter.ShowMessage(Ex:TContext; msg:string);
begin
  if Assigned(FOnShowMessage) then begin
    OnShowMessage(Self, Ex, msg);
  end else begin
    if Assigned(EditorDialog) then begin
      EditorDialog.ShowMessage(msg);
    end;
  end;
end;

procedure   TAInterpreter.ClearMsg;
begin
  if Assigned(EditorDialog) then begin
    EditorDialog.ClearMsg;
  end;
end;

procedure   TAInterpreter.ShowError(Ex:TContext; msg:string);
var idx:integer;
begin
  if Assigned(FOnShowError) then begin
    OnShowError(Self, Ex, msg);
  end;

  if Assigned(EditorDialog) then begin
    EditorDialog.ShowMessage(msg);

    idx:=ProjectList.IndexOfName(Ex.CurrUnit);
    if idx <> -1 then begin
//      if Ex.CompileMode then begin
        EditorDialog.ShowErrorPos(ProjectList.Values[Ex.CurrUnit], -Ex.CurrExecLine);
//      end else begin
//        EditorDialog.ShowErrorPos(ProjectList.Values[Ex.CurrUnit], Ex.Pos);
//      end;
    end;
  end;
end;

procedure   TAInterpreter.ShowEditorDialog(AFileName:string);
begin
  if Assigned(FOnShowEditorDialog) then begin
    FOnShowEditorDialog(Self);
  end else begin
    if not Assigned(EditorDialog) then begin
      EditorDialog:=TEditorDialog.Create(Application.MainForm, Self);
    end;
    EditorDialog.Show;
    EditorDialog.LoadFile(AFileName);
  end;
end;

procedure   TAInterpreter.ShowProjectManagerDialog;
begin
  if Assigned(FOnShowProjectManagerDialog) then begin
    FOnShowProjectManagerDialog(Self);
  end else begin
    if not Assigned(ProjectManagerDialog) then begin
      ProjectManagerDialog:=TProjectManagerDialog.Create(Application.MainForm, Self);
    end;
    ProjectManagerDialog.ShowModal;
    ProjectManagerDialog.Free;
    ProjectManagerDialog:=nil;
  end;
end;

///////////////////////////////////////////////////////////////////////////////////
//function    BWP(Addr:PAIProc; AAttr:integer):TBO;
function     BWP(Addr:TAIProc; AAttr:integer):TBO;
begin
  Result:=TAddr.Create(pointer(@Addr), AAttr or atIsBase);
end;

function    BWV(v:variant):TBO;
begin
  Result:=TVariant.Create(v);
  inc(TData(Result).UseCount);
end;

/////////////////////////////////////////// TCONTEXT ////////////////////////////////////

function      TContext.GetOp:TAIList;
begin
  Result:=TAIList(OpList[OpList.Count - 1]);
end;

procedure     TContext.ResumeOperand(o:TBO);
var proc:TAIProc;
begin
  if CompileMode then begin
    if o is TDataRef then begin
      CompileAddr(Owner.GetAddr('SYSTEM', '_PushLocalWord'));
      CompileData(TVariant.Create(TVariant(TDataRef(o).FValue).Value));  // direct access to avoid appling PropList
      CompileData(TVariant.Create(PropListToStr(TDataRef(o).PropList)));
      CS.Add(Owner.GetAddr('SYSTEM', '_CallMethod'));

      dec(TDataRef(o).UseCount);
      TDataRef(o).Free;
    end else begin
      if o is TAddr then begin
        CompileAddr(TAddr(o));
      end else begin
        CompileData(TData(op));
      end;
    end;
  end else begin
    if o is TDataRef then begin
      TData(o).CallMethod(Self);
      dec(TData(o).UseCount);
      TData(o).Free;
    end else begin
      if TBO(o) is TAddr then begin
        if TAddr(o).IsBase then begin
          proc:=TAddr(o).Addr;
          proc(Self);
        end else begin
          Interpret(TAddr(o).Addr);
        end;
      end;
    end;
  end;
end;

procedure      TContext.AddOperand(operand:TBO);
var o:TAddr;
begin
  if Assigned(operand) and (operand is TData) then inc(TData(operand).UseCount);
  if OpList.Count = 0 then begin
    AddOpSection;
  end;

  if operand is TAddr then begin
    while Op.count > 0 do begin
      if TBO(Op[Op.Count - 1]) is TAddr then begin
        o:=TAddr(Op[Op.Count - 1]);
        if ((o.Attr and atPrecedenceMask) <> 0) and
           ((TAddr(operand).Attr and atPrecedenceMask) > (o.Attr and atPrecedenceMask)) then begin
          Op.Delete(Op.Count-1);
          ResumeOperand(o);
        end else begin
          break;
        end;
      end else begin
        break;
      end;
    end;
  end;

  Op.Add(operand);
//  TAIList(OpList[OpList.Count - 1]).Add(operand);
end;

procedure      TContext.AddOpSection;
begin
  OpList.Add(TAIList.Create('New OP Section'));
end;

procedure      TContext.DropOpSection;
var operand:TBO;
begin
  while Op.Count > 0 do begin
    operand:=Op[Op.Count - 1];
    if Assigned(operand) and (operand is TData) then begin
      dec(TData(operand).UseCount);
      TData(operand).Free;
    end;
    Op.Delete(Op.Count - 1);
  end;

  TAIList(OpList[OpList.Count - 1]).Free;;

  OpList.Delete(OpList.Count - 1);
end;

function      TContext.GetDS:TAIList;
begin
  Result:=TAIList(DSList[DSList.Count - 1]);
end;

procedure    TContext.AddDSSection;
begin
  DSList.Add(TAIList.Create('New DS section'));
end;

procedure    TContext.DropDSSection;
//var d:TData;
begin
  while DS.Count > 0 do begin
    PopDS.Free;
//    d:=DS[DS.Count - 1];
//    if Assigned(d) then begin
//      dec(d.UseCount);
//      d.Free;
//    end;
//    DS.Delete(DS.Count - 1);
  end;

  TAIList(DSList[DSList.Count - 1]).Free;

  DSList.Delete(DSList.Count - 1);
end;

function    TContext.PushDS(p:TData):TData;
begin
  DS.Add(p);
  inc(p.UseCount);
  Result:=p;
end;

function    TContext.PopDS:TData;
begin
  if DS.Count = 0 then begin
    Error(errDSEmpty, [0]);
  end;

  Result:=DS.Items[DS.Count-1];
  dec(Result.UseCount);
  DS.Delete(DS.Count-1);
end;

function    TContext.GetLocalVar(idx:integer):TData;
begin
//  Result:=TData(TAIList(LocWS[LocWS.Count - 1])[idx]);
  Result:=TData(TAIStringList(LocWS[LocWS.Count - 1]).Objects[idx]);
end;

procedure    TContext.CreateLocalStack;
begin
//  LocWS.Add(TAIList.Create('proc local stack'));
  LocWS.Add(TAIStringList.Create('proc local stack'));
  AddOpSection;
end;

procedure    TContext.FreeLocalStack;
//var ls:TAIList;
var ls:TAIStringList;
    d:TData;
begin
//  ls:=TAIList(LocWS[LocWS.Count-1]);
  ls:=TAIStringList(LocWS[LocWS.Count-1]);

  while ls.Count > 0 do begin
    d:=TData(ls.Objects[ls.Count-1]);
    dec(d.UseCount);
    d.Free;
    ls.Delete(ls.Count-1);
  end;
  ls.Free;
  LocWS.Delete(LocWS.Count-1);

  DropOpSection;
end;

procedure    TContext.AddLocalVar(VarName:string; d:TData);
begin
  inc(d.UseCount);
  TAIStringList(LocWS[LocWS.Count-1]).AddObject(VarName, d);
  PushDS(d);
end;

procedure    TContext.CompileData(d:TData);
begin
  CS.Add(d);
  inc(d.UseCount);
end;

procedure    TContext.CompileDataAt(idx:integer;d:TData);
begin
  CS[idx]:=d;
  inc(d.UseCount);
end;

procedure    TContext.CompileAddr(a:TAddr);
begin
  CS.Add(a);
end;

function    TContext.GetTopCSAsInt:integer;
begin
  Result:=TVariant(CS[IP]).Value;
end;

procedure   TContext.PushRS(NewCS:pointer);
begin
  RSIP.Add(pointer(IP + 1));
  RSCS.Add(CS);

  IP:=0;
  CS:=NewCS;
end;

procedure   TContext.PopRS;
begin
  if RSIP.Count = 0 then Error(errRSEmpty, [0]);
  IP:=integer(RSIP[RSIP.Count - 1]);
  RSIP.Delete(RSIP.Count - 1);

  CS:=RSCS[RSCS.Count - 1];
  RSCS.Delete(RSCS.Count - 1);
end;

constructor  TContext.Create(aBuf:PChar;AOwner:TAInterpreter);
begin
  Owner       :=AOwner;
  Words       :=Owner.Words;
  ProjectList :=Owner.ProjectList;

  CurrExecLine  :=-1;
  OldCurrSrcLine:=-1;
  CurrSrcLine   :=-1;

  Buf:=aBuf;
  Pos:=0;

  LocalWords    :=TAIStringList.Create('Local words');
  LocWS         :=TAIList.Create('local stack');

  BackTermList  :=TAIStringList.Create('back terms');

  DSList        :=TAIList.Create('DS list');
  DSList.Add(TAIList.Create('DS'));

  OpList        :=TAIList.Create('OP list');
  AddOpSection;

  StatementList :=TAIStringList.Create('statement list');
  RSIP          :=TAIList.Create('RSIP');
  RSCS          :=TAIList.Create('RSCS');
  CompileMode   :=false;
  IP            :=0;
  JumpList      :=TAIList.Create('jump list');

  UnitList      :=TAIStringList.Create('unit list');
  UnitList.Add('SYSTEM');
end;

destructor  TContext.Destroy;
begin
  UnitList.Free;
  LocalWords.Free;

  while LocWS.Count > 0 do FreeLocalStack;
  LocWS.Free;

  while DSList.Count > 0 do begin
    DropDSSection;
  end;
  DSList.Free;

  StatementList.Free;

  while OpList.Count > 0 do begin
    DropOpSection;
  end;
  OpList.Free;

  BackTermList.Free;

  RSIP.Free;
  RSCS.Free;
  JumpList.Free;
  inherited;
end;

procedure   TContext.PushJamp(JampAddr:integer);
begin
  JumpList.Add(pointer(JampAddr));
end;

function    TContext.PopJamp:integer;
begin
  if JumpList.Count = 0 then begin
    Error(errInternalJampListEmpty, [0]);
  end;

  Result:=integer(JumpList.Items[JumpList.Count-1]);
  JumpList.Delete(JumpList.Count-1);
end;

function    TContext.TopStatement:string;
begin
  if Self.StatementList.Count = 0 then begin
    Result:='';
  end else begin
    Result:=Self.StatementList[Self.StatementList.Count - 1];
  end;
end;

procedure   TContext.DropTopStatement;
begin
  if Self.StatementList.Count > 0 then begin
    Self.StatementList.Delete(Self.StatementList.Count - 1);
  end;
end;

procedure   TContext.PushBackTerm(term:TTerm);
begin
  BackTermList.AddObject(term.Value, pointer(Ord(term.Attr)));
end;

function    TContext.GetNextTerm:TTerm;
var Term:TTerm;
    code:integer;
    value:double;
    OldDecSep:char;
    StrTerm:char;
  function Blank(ch:char):boolean;
  begin
    Result:=(ch = ' ') or (ch = #$0d) or (ch = #$0a) or (ch = #$09);
  end;
  function IsAlpha(ch:char):boolean;
  begin
    Result:=IsCharAlpha(ch) or (ch  = '_') or (ch  = '~');
  end;
  function IsDigit(ch:char):boolean;
  begin
    Result:=((ch >= '0') and (ch <='9')) or (ch = '.');
  end;
  function IsAlphaDigit(ch:char):boolean;
  begin
    Result:=IsDigit(ch) or IsAlpha(ch);
  end;
  function IsNotAlphaDigit(ch:char):boolean;
  begin
    Result:=(not IsAlphaDigit(ch)) and (not Blank(ch));
  end;
  procedure incPos;
  begin
    if Buf[Pos] = #$0d then begin
      inc(CurrSrcLine);
    end;
    inc(Pos);
  end;
begin
  if BackTermList.Count > 0 then begin
    term.Value:=BackTermList[BackTermList.Count - 1];
    term.Attr:=TTermAttr(integer(BackTermList.Objects[BackTermList.Count - 1]));
    BackTermList.Delete(BackTermList.Count - 1);
  end else begin
    Term.Attr:=taError;
    Term.Value:='';
    while Buf[Pos] <> #$0 do begin
      if Blank(Buf[Pos]) then begin
        incPos;
      end else begin
        if (Buf[Pos] = '/') and (Buf[Pos+1] <> #$0) and (Buf[Pos+1] = '/') then begin
          while (Buf[Pos] <> #$0) and (Buf[Pos] <> #$0d) and (Buf[Pos] <> #$0a) do incPos;
        end else begin
          if Buf[Pos] = '{' then begin
            while (Buf[Pos] <> #$0) and (Buf[Pos] <> '}') do incPos;
            incPos;
          end else begin
            break;
          end;
        end;
      end;
    end;
{
    if (Buf[Pos] = '''') then begin
      incPos;
      while (Buf[Pos] <> '''') and (Buf[Pos] <> #$0) do begin
        Term.Value:=Term.Value + Buf[Pos]; incPos;
      end;
      if Buf[Pos] = #$0 then begin
        Error(errUnterminatedString, [0]);
      end else begin
        incPos;
        Term.Attr:=taString;
      end;
}
    if (Buf[Pos] = '''') or (Buf[Pos] = '"') then begin
      StrTerm:=Buf[Pos];
      incPos;
      while (Buf[Pos] <> #$0) and ((Buf[Pos] <> StrTerm) or ((Buf[Pos] = StrTerm) and (Buf[Pos + 1] = StrTerm))) do begin
        if (Buf[Pos] = StrTerm) and (Buf[Pos + 1] = StrTerm) then incPos;
        Term.Value:=Term.Value + Buf[Pos]; incPos;
      end;
      if Buf[Pos] = #$0 then begin
        Error(errUnterminatedString, [0]);
      end else begin
        incPos;
        Term.Attr:=taString;
      end;

    end else begin
      if Buf[Pos] in ['(',')','!','#','$','%','&','*','+',',','-','/',':',';','<','=','>','?','@','[','\',']','^','{','|','}'] then begin
        Term.Attr:=taWord;

        Term.Value:=Buf[Pos];
        incPos;

        if (Term.Value[1] in ['<','=','>'])     and (Buf[Pos] in ['<','=','>']) or
           (Term.Value[1] in ['+','-','*','/']) and (Buf[Pos] =  '=')
        then begin
          Term.Value:=Term.Value + Buf[Pos];
          incPos;
        end;

        Result:=Term;
      end else begin
        if IsNotAlphaDigit(Buf[Pos]) then begin
          while IsNotAlphaDigit(Buf[Pos]) and (Buf[Pos] <> #$0) do begin
            Term.Value:=Term.Value + Buf[Pos]; incPos;
          end;
          Term.Attr:=taWord;
          Result:=Term;
        end else begin
          while IsAlphaDigit(Buf[Pos]) do begin
            if Buf[Pos] = '~' then begin
              Term.Value:=Term.Value + ' ';
            end else begin
              Term.Value:=Term.Value + Buf[Pos];
            end;
            incPos;
          end;
          if IsAlpha(Term.Value[1]) then begin
            Term.Attr:=taWord;
          end else begin
            OldDecSep:=DecimalSeparator;
            DecimalSeparator:='.';
            Val(Term.Value, value, code);
            DecimalSeparator:=OldDecSep;
            if code = 0 then begin
              Term.Attr:=taNumber;
            end else begin
              Term.Attr:=taWord;
            end;
          end;
        end;
      end;
    end;
  end;
  Result:=Term;
end;
////////////////////////////////////////////////////////////////////////
procedure   TContext.RunScript(scr:PChar);
begin
  Owner.RunScript(nil, scr);
end;

procedure   TContext.RunFile(FileName:string);
begin
  Owner.RunFile(FileName);
end;

procedure   TContext.ShowMessage(msg:string);
begin
  Owner.ShowMessage(Self, msg);
end;

procedure    TContext.StepIP();
var proc:TAIProc;
    p:pointer;
begin
  p:=CS[IP];
  if TBO(p) is TAddr then begin
    if TAddr(p).IsBase then begin
      proc:=TAddr(p).Addr;
      inc(IP);
      proc(Self);
    end else begin
      PushRS(TAddr(p).Addr);
    end;
  end else begin
    PushDS(TData(p));
    inc(IP);
  end;
end;

procedure   TContext.Interpret(NewCS:pointer);
begin
//  Owner.Interpret(Self, NewCS);
  PushRS(NewCS);

  while true do begin
    if RSIP.Count = 0 then break;
    StepIP();
  end;
end;

////////////////////  BaseObject ////////////////////////////////////////////////////////////////////
constructor TBO.Create;
begin
  inc(BOCount);
end;

destructor  TBO.Destroy;
begin
  dec(BOCount);
  inherited;
end;

function    TBO.GetText:string;
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'CallMethod']);
end;

procedure   TBO.SetText(s:string);
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'CallMethod']);
end;
////////////////////  Addr  ////////////////////////////////////////////////////////////////////

constructor TAddr.Create(p:pointer; AAttr:integer);
begin
  inherited Create;
  inc(AddrCount);
  Value:=integer(p);
  Attr:=AAttr;
//  IsBase:=AIsBase;
//  IsFunction:=AIsFunction;
//  IsImm  :=AIsImm;
end;
destructor  TAddr.Destroy;
begin
  dec(AddrCount);
  inherited;
end;

procedure    TAddr.Assign(O:TAddr);
begin
  Value  :=O.Value;
  IsBase:=O.IsBase;
  IsImm  :=O.IsImm;
end;

function    TAddr.GetText:string;
begin
  Result:=IntToStr(Value);
end;

procedure   TAddr.SetText(s:string);
begin
  Value:=StrToInt(s);
end;

function  TAddr.GetAddr:pointer;
begin
  Result:=pointer(Value);
end;

procedure TAddr.SetAddr(p:pointer);
begin
  Value:=integer(p);
end;

function    TAddr.GetIsBase:boolean;
begin
  Result:=(Attr and atIsBase) <> 0;
end;
procedure   TAddr.SetIsBase(AIsBase:boolean);
begin
  if AIsBase then begin
    Attr:=Attr or atIsBase;
  end else begin
    Attr:=Attr and (not atIsBase);
  end;
end;

function    TAddr.GetIsImm:boolean;
begin
  Result:=(Attr and atIsImm) <> 0;
end;
procedure   TAddr.SetIsImm(AIsImm:boolean);
begin
  if AIsImm then begin
    Attr:=Attr or atIsImm;
  end else begin
    Attr:=Attr and (not atIsImm);
  end;
end;
function    TAddr.GetIsFunction:boolean;
begin
  Result:=(Attr and atIsFunction) <> 0;
end;
function    TAddr.GetIsOperator:boolean;
begin
  Result:=(Attr and atIsOperator) = atIsOperator;
end;
procedure   TAddr.SetIsFunction(AIsFunction:boolean);
begin
  if AIsFunction then begin
    Attr:=Attr or atIsFunction;
  end else begin
    Attr:=Attr and (not atIsFunction);
  end;
end;
function    TAddr.GetBreakPoint:boolean;
begin
  Result:=(Attr and atBreakPoint) <> 0;
end;
procedure   TAddr.SetBreakPoint(ABreakPoint:boolean);
begin
  if ABreakPoint then begin
    Attr:=Attr or atBreakPoint;
  end else begin
    Attr:=Attr and (not atBreakPoint);
  end;
end;
function    TAddr.GetKeyWord:boolean;
begin
  Result:=(Attr and atKeyWord) <> 0;
end;
procedure   TAddr.SetKeyWord(AKeyWord:boolean);
begin
  if AKeyWord then begin
    Attr:=Attr or atKeyWord;
  end else begin
    Attr:=Attr and (not atKeyWord);
  end;
end;
////////////////////  Data ////////////////////////////////////////////////////////////////////
constructor  TData.Create;
begin
  inherited Create;
  inc(DataCount);
  UseCount:=0;
end;

destructor  TData.Destroy;
begin
  dec(DataCount);

  if Assigned(PPropList)  then PPropList.Free;
  if Assigned(PushList)   then PushList.Free;

  inherited;
end;

procedure    TData.Free;
begin
  if UseCount = 0 then Destroy;
end;

function    TData.GetPropList:TAIStringList;
begin
  if not Assigned(PPropList) then PPropList:=TAIStringList.Create('TData get PropList');
  Result:=PPropList;
end;

procedure   TData.SetPropList(APropList:TAIStringList);
begin
  if not Assigned(PPropList) then PPropList:=TAIStringList.Create('TData set PropList');
  PPropList.Assign(APropList);
end;

function    TData.GetPropListCount:integer;
begin
  if not Assigned(PPropList) then begin
    Result:=0;
  end else begin
    Result:=PPropList.Count;
  end;
end;

function    TData.GetValue:variant;
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'GetValue']);
  Result:=0;
end;

procedure   TData.SetValue(AValue:variant);
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'SetValue']);
end;

procedure   TData.CallMethod(Ex:TContext);
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'CallMethod']);
end;

procedure   TData.Assign(D:TData);
begin
  Error(errCallUndeclaredMethod, [Self.ClassName, 'Assign']);
end;

procedure   TData.PushPropList;
begin
  if not Assigned(PushList) then PushList:=TAIStringList.Create('TData PushList');
  PushList.Add(PropList.Text);
end;

procedure   TData.PopPropList;
begin
  PropList.Text:=PushList[PushList.Count - 1];
  PushList.Delete(PushList.Count - 1);
end;
////////////////////  Variant ////////////////////////////////////////////////////////////////////
constructor  TVariant.Create(AValue:variant);
begin
  inherited Create;

  v:=AValue;
end;

function    TVariant.GetText:string;
begin
  ProcessedProp:=false;
  if PropListCount = 0 then begin
    Result:=v;
    ProcessedProp:=true;
  end;
end;

procedure   TVariant.SetText;
begin
  ProcessedProp:=false;
  if PropListCount = 0 then begin
    v:=s;
    ProcessedProp:=true;
  end;
end;

function    TVariant.GetValue:variant;
begin
  ProcessedProp:=false;
  if PropListCount = 0 then begin
    Result:=V;
    ProcessedProp:=true;
  end;
end;

procedure    TVariant.SetValue(AValue:variant);
begin
  ProcessedProp:=false;
  if PropListCount = 0 then begin
    v:=AValue;
    ProcessedProp:=true;
  end;
end;

procedure    TVariant.CallMethod(Ex:TContext);
begin
  ProcessedProp:=false;
end;

procedure    TVariant.Assign(D:TData);
begin
  ProcessedProp:=false;
  if PropListCount = 0 then begin
    Self.Value:=D.Value;
    ProcessedProp:=true;
  end;
end;
////////////////////  DataRef ////////////////////////////////////////////////////////////////////
{
constructor TDataRef.Create(prop:string;AValue:TData);
begin
  inherited Create;

  if AValue is TDataRef then begin
    FValue:=TDataRef(AValue).FValue;

    PropList.Assign(AValue.PropList);
    if Length(prop) > 0 then PropList.Add(prop);
  end else begin
    FValue:=AValue;
    PropList.Clear();
    if Length(prop) > 0 then PropList.Add(prop);
  end;

  inc(FValue.UseCount);
end;
}
constructor TDataRef.Create(prop:string;AValue:TData);
var i:integer;
begin
  inherited Create;

  StrToPropList(prop, PropList);

  if AValue is TDataRef then begin
    FValue:=TDataRef(AValue).FValue;

    for i:=0 to AValue.PropList.Count - 1 do begin
      PropList.Insert(i, AValue.PropList[i]);
    end;
  end else begin
    FValue:=AValue;
  end;

  inc(FValue.UseCount);
end;

destructor  TDataRef.Destroy;
begin
  dec(FValue.UseCount);
  FValue.Free;

//  PropList.Free;
  inherited;
end;

function    TDataRef.GetValue:variant;
begin
  FValue.PushPropList;
    FValue.PropList:=PropList;
    Result:=FValue.Value;
  FValue.PopPropList;
//  FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errPropReadingFromType, [PropListToStr(PropList), FValue.ClassName]);
  end;
end;

procedure    TDataRef.SetValue(AValue:variant);
begin
  FValue.PushPropList;
    FValue.PropList:=PropList;
    FValue.Value:=AValue;
  FValue.PopPropList;
//    FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errPropWritingToType, [PropListToStr(PropList), FValue.ClassName]);
  end;
end;

function    TDataRef.GetText:string;
begin
  FValue.PushPropList;
    FValue.PropList:=PropList;
    Result:=FValue.Text;
  FValue.PopPropList;
//  FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errPropReadingFromType, [PropListToStr(PropList), FValue.ClassName]);
  end;
end;
procedure    TDataRef.SetText(s:string);
begin
  FValue.PushPropList;
    FValue.PropList:=PropList;
    FValue.Text:=s;
  FValue.PopPropList;
//  FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errPropWritingToType, [PropListToStr(PropList), FValue.ClassName]);
  end;
end;

procedure    TDataRef.Assign(D:TData);
begin
  FValue.PushPropList;
    FValue.PropList.Assign(PropList);
    Self.FValue.Assign(D);
  FValue.PopPropList;
//  FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errAssignmentToTypeValue, [FValue.ClassName, D.ClassName, PropListToStr(PropList)]);
  end;
end;

procedure    TDataRef.CallMethod(Ex:TContext);
begin
  FValue.PushPropList;
    FValue.PropList:=PropList;
    Self.FValue.CallMethod(Ex);
  FValue.PopPropList;
//  FValue.PropList.Clear;  //??

  if not FValue.ProcessedProp then begin
    Error(errCallMethod, [FValue.ClassName, PropListToStr(PropList)]);
  end;
end;
/////////////////////////////////////////////////////////////////////////////////////
function  TWords.FindUnit(AUnitName:string):boolean;
begin
  Result:=UnitList.IndexOf(AUnitName) <> -1;
end;

procedure    TWords.AddUnit(AUnitName:string);
var idx:integer;
begin
  idx:=UnitList.IndexOf(AUnitName);
  if idx = -1 then begin
    UnitList.AddObject(AUnitName, TAIStringList.Create('new unit'));
  end else begin
    Error(errUNITAlreadyDefined, [AUnitName]);
  end;
end;

procedure    TWords.DropUnit(AUnitName:string);
var idx:integer;
    lst:TAIStringList;
begin
  idx:=UnitList.IndexOf(AUnitName);
  if idx <> -1 then begin
    lst:=TAIStringList(UnitList.Objects[idx]);
    while lst.Count > 0 do begin
      try
        DropWord(AUnitName, lst[lst.Count - 1]);
      except
      end;
      lst.Delete(lst.Count - 1);
    end;
    TAIStringList(UnitList.Objects[idx]).Free;
    UnitList.Delete(idx);
  end;
end;

procedure    TWords.AddWord(AUnitName:string;AWordName:string; w:TBO);
var idxu,idxw:integer;
begin
  AUnitName:=Upper(AUnitName);
  AWordName:=Upper(AWordName);

  if w is TData then begin
    inc(TData(w).UseCount);
  end else begin
    if TAddr(w).KeyWord then begin
      KeyWordList.Add(AWordName);
    end;
  end;

  idxu:=UnitList.IndexOf(AUnitName);

  if idxu = -1 then begin
    AddUnit(AUnitName);
    TAIStringList(UnitList.Objects[UnitList.Count - 1]).AddObject(AWordName, w);
  end else begin
    idxw:=TAIStringList(UnitList.Objects[idxu]).IndexOf(AWordName);
    if idxw = -1 then begin
      TAIStringList(UnitList.Objects[idxu]).AddObject(AWordName, w);
    end else begin
      Error(errWORDinUNITAlreadyDefined, [AWordName, AUnitName]);
    end;
  end;
end;

function    TWords.FindWord(AUnitList:TAIStringList;AWordName:string):TBO;
var i,idxu,idxw:integer;
begin
  Result:=nil;

  for i:=0 to AUnitList.Count - 1 do begin
    idxu:=UnitList.IndexOf(AUnitList[i]);
    if idxu <> -1 then begin
      idxw:=TAIStringList(UnitList.Objects[idxu]).IndexOf(AWordName);
      if idxw <> -1 then begin
        Result:=TBO(TAIStringList(UnitList.Objects[idxu]).Objects[idxw]);
        break;
      end;
    end;
  end;
end;

procedure    TWords.DropWord(AUnitName:string;AWordName:string);
var idxu,idxw:integer;
    w:TBO;
    lst:TAIList;
begin
  idxu:=UnitList.IndexOf(AUnitName);
  if idxu <> -1 then begin
    idxw:=TAIStringList(UnitList.Objects[idxu]).IndexOf(AWordName);
    if idxw <> -1 then begin
      w:=TBO(TAIStringList(UnitList.Objects[idxu]).Objects[idxw]);
      if w is TData then begin
        try
          dec(TData(w).UseCount);
          TData(w).Free;
        except
          Error(errInternalWhileDropWord, [AWordName, w.Text]);
        end;
      end else begin
        if (w is TAddr) and (not TAddr(w).IsBase) then begin
          lst:=TAIList(TAddr(w).Addr);
          while lst.Count > 0 do begin
            try
              if TBO(lst[lst.Count - 1]) is TData then begin
                dec(TData(lst[lst.Count - 1]).UseCount);
                TData(lst[lst.Count - 1]).Free;
              end;
            except
              Error(errInternalWhileDropWord, [AWordName, TBO(lst[lst.Count - 1]).Text]);
            end;
            lst.Delete(lst.Count - 1);
          end;
          lst.Clear;
          lst.Free;
          w.Free;
        end else begin
        end;
      end;
    end;
  end;
end;

constructor  TWords.Create;
begin
  UnitList    :=TAIStringList.Create('unit list');
  KeyWordList  :=TAIStringList.Create('key word list');
end;

destructor  TWords.Destroy;
begin
  KeyWordList.Free;

  while UnitList.Count > 0 do DropUnit(UnitList[UnitList.Count - 1]);
  UnitList.Free;

  inherited;
end;

function    TAInterpreter.GetAddr(AUnitName:string; AWordName:string):TAddr;
var lst:TAIStringList;
begin
  lst:=TAIStringList.Create('IE GetAddr');
  lst.Add(AUnitName);
  try
    Result:=TAddr(Words.FindWord(lst, AWordName));
    if not Assigned(Result) then begin
      Error(errInternalWordNotFound, [AWordName]);
    end;
  finally
    lst.Free;
  end;
end;

function    TAInterpreter.GetVarAddr(AUnitName:string; AWordName:string):TData;
var lst:TAIStringList;
begin
  lst:=TAIStringList.Create('IE GetVarAddr');
  lst.Add(AUnitName);
  try
    Result:=TData(Words.FindWord(lst, AWordName));
    if not Assigned(Result) then begin
      Error(errInternalWordNotFound, [AWordName]);
    end;
  finally
    lst.Free;
  end;
end;

procedure    TAInterpreter.ExecProc(HOwner:TData; HEventHanler:TScriptEventHandler; Params:array of variant);
var Ex:TContext;
    i:integer;
begin
  if HEventHanler <> 0 then begin
    Ex:=TContext.Create(nil, Self);
    try
      Ex.PushDS(HOwner);

      for i:= 0 to High(Params) do begin
        Ex.PushDS(TVariant.Create(Params[i]));
      end;

      Self.Interpret(Ex, pointer(HEventHanler));
    finally
      Ex.Free;
    end;
  end else begin
    HOwner.Free;
  end;
end;

procedure   TAInterpreter.ExecProcTData(HEventHanler:TScriptEventHandler; Params:array of TData);
var Ex:TContext;
    i:integer;
begin
  if HEventHanler <> 0 then begin
    Ex:=TContext.Create(nil, Self);
    try
      for i:= 0 to High(Params) do begin
        Ex.PushDS(Params[i]);
      end;

      Self.Interpret(Ex, pointer(HEventHanler));
    finally
      Ex.Free;
    end;
  end;
end;

procedure   TAInterpreter.RunProc(AUnitName:string; AProcName:string; Params:array of variant);
var Ex:TContext;
    i:integer;
    p:TAddr;
begin
  RunScript(nil, PChar('uses ' + AUnitName + '; '));

  p:=GetAddr(AUnitName, AProcName);

  if Assigned(p) then begin

    Ex:=TContext.Create(nil, Self);
    try
      for i:= 0 to High(Params) do begin
        Ex.PushDS(TVariant.Create(Params[i]));
      end;

      Self.Interpret(Ex, p.Addr);
    finally
      Ex.Free;
    end;
  end;
end;

procedure   TAInterpreter.RunProcTData(AUnitName:string; AProcName:string; Params:array of TData);
var Ex:TContext;
    i:integer;
    p:TAddr;
begin
  RunScript(nil, PChar('uses ' + AUnitName + '; '));

  p:=GetAddr(AUnitName, AProcName);

  if Assigned(p) then begin

    Ex:=TContext.Create(nil, Self);
    try
      for i:= 0 to High(Params) do begin
        Ex.PushDS(Params[i]);
      end;
      Self.Interpret(Ex, p.Addr);
    finally
      Ex.Free;
    end;
  end;
end;

function    TAInterpreter.RunFuncTData(AUnitName:string; AProcName:string; Params:array of TData):TData;
var Ex:TContext;
    i:integer;
    p:TAddr;
begin
  Result:=nil;
  RunScript(nil, PChar('uses ' + AUnitName + '; '));

  p:=GetAddr(AUnitName, AProcName);

  if Assigned(p) then begin

    Ex:=TContext.Create(nil, Self);
    try
      for i:= 0 to High(Params) do begin
        Ex.PushDS(Params[i]);
      end;

      Self.Interpret(Ex, p.Addr);

      Result:=Ex.PopDS;
    finally
      Ex.Free;
    end;
  end;
end;

function    TAInterpreter.RunFunc(AUnitName:string; AProcName:string; Params:array of variant):TData;
var i:integer;
    arrData:array of TData;
begin
  SetLength(arrData, High(Params) + 1);

  for i:= 0 to High(Params) do begin
    arrData[i]:=TVariant.Create(Params[i]);
  end;

  Result:=RunFuncTData(AUnitName, AProcName, arrData);
end;

function    TAInterpreter.ExecFuncTData(AddrProc:pointer; Params:array of TData):TData;
var Ex:TContext;
    i:integer;
    p:TAddr;
begin
  Result:=nil;

  Ex:=TContext.Create(nil, Self);
  try
    for i:= 0 to High(Params) do begin
      Ex.PushDS(Params[i]);
    end;

    Self.Interpret(Ex, AddrProc);

    Result:=Ex.PopDS;
  finally
    Ex.Free;
  end;
end;

{
function    TAInterpreter.RunFunc(AUnitName:string; AProcName:string; Params:array of variant):TData;
var Ex:TContext;
    i:integer;
    p:TAddr;
begin
  RunScript(PChar('uses ' + AUnitName + '; '));

  p:=GetAddr(AUnitName, AProcName);

  if Assigned(p) then begin

    Ex:=TContext.Create(nil, Self);
    try
      for i:= 0 to High(Params) do begin
        Ex.PushDS(TVariant.Create(Params[i]));
      end;

      Self.Interpret(Ex, p.Addr);

      Result:=Ex.PopDS;
    finally
      Ex.Free;
    end;
  end;
end;
}
function    TAInterpreter.GetFileSystem:TVirtualFileSystem;
begin
  Result:=FFileSystem;
end;

procedure   TAInterpreter.SetFileSystem(AFileSystem:TVirtualFileSystem);
begin
  FFileSystem:=AFileSystem;
end;

procedure   TContext.CompileLineNo;
var p:pointer;
begin
  if Owner.CompileWithDebugInfo and CompileMode then begin
    if OldCurrSrcLine <> CurrSrcLine then begin
      OldCurrSrcLine:=CurrSrcLine;

      p:=Owner.GetAddr('SYSTEM', '_DEBUGINFO');
//      if (CS.Count > 1) and (CS[CS.Count - 2] = p) then begin
//        TVariant(CS[CS.Count - 1]).Value:=CurrSrcLine;
//      end else begin
        CompileAddr(p);
        CompileData(TVariant.Create(CurrSrcLine));
//      end;
    end;
  end;
end;

procedure   TAInterpreter.RunScript(Sender:TObject; scr:PChar);
var Term:TTerm;
    Ex:TContext;
  procedure   CorrectText(p:PChar);
  var i:integer;
  begin
    for i:=0 to StrLen(p) - 1 do begin
      if (Ord(p[i]) < $20) and (Ord(p[i]) <> $0d) and (Ord(p[i]) <> $0a)   then
        p[i]:=' ';
    end;
  end;
begin
  Ex:=nil;
  
  if Assigned(CurrContext) then begin
    try
      Continue(CurrContext);
      Reset;
    except
      on E:EStepOnRun do begin
        CurrContext:=(E as EStepOnRun).Ex;
        EditorDialog.ShowStopPos(ProjectList.Values[CurrContext.CurrUnit], CurrContext.CurrExecLine);
      end;
      on E:Exception do begin
        Reset;
        ShowError(Ex, '<' + Term.Value + '> ' + E.Message);
      end;
    end;
  end else begin
    Ex:=nil;
    CurrContext:=nil;
    try
      try
        CorrectText(scr);

        Ex:=TContext.Create(scr, Self);
        Ex.Sender:=Sender;
        Term:=Ex.GetNextTerm;
        while not ((Term.Value = '') and (Term.Attr <> taString)) do begin
          Ex.CompileLineNo;
          case Term.Attr of
            taString: ExecString(Ex, Term.Value);
            taNumber: ExecNumber(Ex, Term.Value);
            taWord:   ExecWord(Ex, Term.Value);
          end;
          Term:=Ex.GetNextTerm;
        end;
      except
        on E:EStepOnRun do begin
          CurrContext:=(E as EStepOnRun).Ex;
          EditorDialog.ShowStopPos(ProjectList.Values[Ex.CurrUnit], Ex.CurrExecLine);
        end;
        on E:Exception do begin
          Reset;
          ShowError(Ex, '<' + Term.Value + '> ' + E.Message);
        end;
      end;
    finally
//      if Assigned(Ex) and Assigned(CurrContext) and (Ex <> CurrContext) then begin
        Ex.Free;
        CurrContext:=nil;
//      end;
    end;
  end;
end;

procedure    TAInterpreter.StepIP(Ex:TContext);
var proc:TAIProc;
    p:pointer;
begin
  p:=Ex.CS[Ex.IP];
  if TBO(p) is TAddr then begin
    if TAddr(p).IsBase then begin
      proc:=TAddr(p).Addr;
      inc(Ex.IP);
      proc(Ex);
    end else begin
      Ex.PushRS(TAddr(p).Addr);
    end;
  end else begin
    Ex.PushDS(TData(p));
    inc(Ex.IP);
  end;
end;

procedure   TAInterpreter.Interpret(Ex:TContext;NewCS:pointer);
begin
  Ex.PushRS(NewCS);

  while true do begin
    if Ex.RSIP.Count = 0 then break;
    StepIP(Ex);
  end;
end;

procedure   TAInterpreter.Continue(Ex:TContext);
begin
  while true do begin
    if Ex.RSIP.Count = 0 then break;
    StepIP(Ex);
  end;
end;

procedure   TAInterpreter.Reset;
begin
  if Assigned(CurrContext) then begin
    EditorDialog.ShowStopPos(ProjectList.Values[CurrContext.CurrUnit], -1);
    CurrContext.Free;
  end;
  CurrContext:=nil;
end;

procedure   TAInterpreter.Loaded;
var inf:TIniFile;
    idx:integer;
begin
  if csDesigning in ComponentState then exit;

  if not Assigned(FFileSystem) then begin
    FFileSystem:=TStandardFileSystem.Create(Owner);
  end;

  AIWords.RegisterWords(Words);
  AIObject.RegisterWords(Words);
  AIList.RegisterWords(Words);
  AIRecord.RegisterWords(Words);
  AIVarList.RegisterWords(Words);

{
  AIUserRep.RegisterWords(Words);
  AIReportOld.RegisterWords(Words);
  AIProgress.RegisterWords(Words);
  AIEdit.RegisterWords(Words);
  AIRadioGroup.RegisterWords(Words);
  AIGroupBox.RegisterWords(Words);
  AIDateEdit.RegisterWords(Words);
  AIButton.RegisterWords(Words);
  AILabel.RegisterWords(Words);
  AIPanel.RegisterWords(Words);
  AICheckBox.RegisterWords(Words);
  AIMenu.RegisterWords(Words);
  AISQL.RegisterWords(Words);
  AIGUI.RegisterWords(Words);
}
  if Assigned(OnRegisterWords) then OnRegisterWords(Self);


  if not FileExists(ProjectName) then begin
    inf:=TIniFile.Create(ExtractFilePathAndNameOnly(ParamStr(0)) + '.ini');
    try
      ProjectName:=inf.ReadString('AIProject', 'ProjectName', '');
    finally
      inf.Free;
    end;
  end;

  if FileExists(ProjectName) then begin
    try
      ProjectList.LoadFromStream(FileSystem.CreateReadStream(ProjectName));
    finally
      FileSystem.CloseReadStream;
    end;

    idx:=ProjectList.IndexOfName('AutoLoad');
    if idx <> -1 then begin
      RunFile(ProjectList.Values['AutoLoad']);
    end;
  end;
end;

end.


