unit AIMenu;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

{$INCLUDE DEFINES.INC}

interface

uses
  Windows, SysUtils, Graphics, Menus, Forms, Classes, Buttons,
  Controls,

  ErrorMsg, AInterpreter, ARUtils;

procedure RegisterWords(Words:TWords);

var
  EnableDragButton:boolean;

type
  TBMPList = class
  private
    BL          :TAIStringList;
  public
    procedure   Add(BMPName:string; BMPPath:string);
    function    GetBMP(BMPName:string):TBitmap;
    constructor Create;
    destructor  Destroy;override;
  end;

  TAICommand = class
  public
    Owner       :TComponent;
    Interpret   :TAInterpreter;
    Script      :string;
    BitmapName  :string;
    Caption     :string;
    procedure   Assign(acommand:TAICommand);
    procedure   Execute;
    constructor Create(AOwner:TComponent; AInterpret:TAInterpreter);
  end;

  TAISpeedButton = class(TSpeedButton)
  private
    procedure   ButtonClick(Sender: TObject);
    function    GetScript:string;
    procedure   SetScript(AScript:string);
  public
    command     :TAICommand;

    property    Script:string read GetScript write SetScript;
    constructor Create(AOwner:TComponent;AInterpret:TAInterpreter);reintroduce;
    destructor  Destroy;override;

    property    OnStartDrag;
  end;

  TAIMenuItem = class( TMenuItem)
  private
    procedure   MenuItemClick(Sender: TObject);
    function    GetScript:string;
    procedure   SetScript(AScript:string);
  public
    command     :TAICommand;

    property    Script:string read GetScript write SetScript;
    constructor Create(AOwner:TComponent;AInterpret:TAInterpreter);reintroduce;
    destructor  Destroy;override;
  end;

  TDragCommand = class(TDragObject)
  public
    command     :TAICommand;
    constructor Create(acommand:TAICommand);
    destructor  Destroy;override;
  end;

var BMPList:TBMPList;

///////////////////////////////////////////////////////////////////////
implementation
///////////////////////////////////////////////////////////////////////
constructor TDragCommand.Create(acommand:TAICommand);
begin
  command :=TAICommand.Create(nil, acommand.Interpret);
  command.Assign(acommand);
end;

destructor  TDragCommand.Destroy;
begin
  command.Free;
  inherited;
end;
///////////////////////////////////////////////////////////////////////
constructor TBMPList.Create;
begin
  BL:=TAIStringList.Create('BMP list');
end;

destructor  TBMPList.Destroy;
var i:integer;
begin
  for i:=0 to BL.Count - 1 do begin
    BL.Objects[i].Free;
  end;
  BL.Free;
  inherited;
end;

procedure   TBMPList.Add(BMPName:string; BMPPath:string);
var idx:integer;
begin
  idx:=BL.IndexOfName(BMPName);
  if idx = -1 then begin
    BL.AddObject(BMPName + '=' + BMPPath, nil);
  end else begin
    BL[idx]         :=BMPName + '=' + BMPPath;
    BL.Objects[idx] :=nil;
  end;
end;

function    TBMPList.GetBMP(BMPName:string):TBitmap;
var idx:integer;
begin
  Result:=nil;

  idx:=BL.IndexOfName(BMPName);
  if (idx <> -1) then begin
    if Assigned(BL.Objects[idx]) then begin
      Result:=TBitmap(BL.Objects[idx]);
    end else begin
      if FileExists(BL.Values[BL.Names[idx]]) then begin
        Result:=TBitmap.Create;
        Result.LoadFromFile(BL.Values[BL.Names[idx]]);
        BL.Objects[idx]:=Result;
      end;
    end;
  end;
end;

///////////////////////////////////////////////////////////////////////
constructor TAICommand.Create(AOwner:TComponent; AInterpret:TAInterpreter);
begin
  Owner     :=AOwner;
  Interpret :=AInterpret;
end;

procedure   TAICommand.Execute;
begin
  if Assigned(Interpret) and (Script <> '') then begin
    Interpret.RunScript(Owner, PChar(Script));
  end;
end;

procedure   TAICommand.Assign(acommand:TAICommand);
begin
  Owner     :=acommand.Owner;
  Interpret :=acommand.Interpret;
  Script    :=acommand.Script;
  BitmapName:=acommand.BitmapName;
  Caption   :=acommand.Caption;
end;
///////////////////////////////////////////////////////////////////////
constructor TAISpeedButton.Create(AOwner:TComponent;AInterpret:TAInterpreter);
begin
  inherited Create(AOwner);
  command:=TAICommand.Create(Self, AInterpret);
  Self.Flat:=true;
{$IFNDEF D3}
  Self.Transparent:=false;
{$ENDIF}
  OnClick:=ButtonClick;
end;

destructor  TAISpeedButton.Destroy;
begin
  command.Free;
  inherited;
end;

function    TAISpeedButton.GetScript:string;
begin
  Result:=command.Script;
end;

procedure   TAISpeedButton.SetScript(AScript:string);
begin
  command.Script:=AScript;
end;

procedure   TAISpeedButton.ButtonClick(Sender: TObject);
begin
  if not EnableDragButton then command.Execute;
end;

///////////////////////////////////////////////////////////////////////
constructor TAIMenuItem.Create(AOwner:TComponent;AInterpret:TAInterpreter);
begin
  inherited Create(AOwner);
  command:=TAICommand.Create(Self, AInterpret);
  command.Caption:=Caption;
  OnClick:=MenuItemClick;
end;

destructor  TAIMenuItem.Destroy;
begin
  command.Free;
  inherited;
end;

function    TAIMenuItem.GetScript:string;
begin
  Result:=command.Script;
end;

procedure   TAIMenuItem.SetScript(AScript:string);
begin
  command.Script:=AScript;
end;

procedure   TAIMenuItem.MenuItemClick(Sender: TObject);
begin
  command.Execute;
end;
///////////////////////////////////////////////////////////////////////

function  GetMainMenu(Ex:TContext):TMainMenu;
var form:TForm;
begin
  form:=TForm(Ex.Owner.Owner);
  if Assigned(form.Menu) then begin
    Result:=form.Menu;
  end else begin
    Result:=TMainMenu.Create(form);
  end;
end;

procedure BW_DropMainMenu(Ex:TContext);  // DS: MenuName
var MainMenu:TMainMenu;
    MenuName:TData;
    mi:TAIMenuItem;
    i:integer;
begin
  MainMenu:=GetMainMenu(Ex);

  MenuName:=Ex.PopDS;
  try
    for i:=0 to MainMenu.Items.Count - 1 do begin
      if MainMenu.Items[i] is TAIMenuItem then begin
        mi:=MainMenu.Items[i] as TAIMenuItem;
        if CT(mi.Caption, MenuName.Value) then begin
          mi.Free;
//          MainMenu.Items.Delete(i);
          exit;
        end;
      end;
    end;
  finally
    MenuName.Free;
  end;
end;

procedure BW_AddMainMenu(Ex:TContext);  // DS: MenuName
var MainMenu:TMainMenu;
    MenuName:TData;
    mi:TAIMenuItem;
begin
  MainMenu:=GetMainMenu(Ex);

  MenuName:=Ex.PopDS;
  try
    mi:=TAIMenuItem.Create(MainMenu, Ex.Owner);
    mi.Caption:=MenuName.Value;
    MainMenu.Items.Add(mi);
  finally
    MenuName.Free;
  end;
end;

procedure BW_DropSubMenu(Ex:TContext);// DS: MenuName(exsisted), SubMenuName(exsisted)
var MainMenu:TMainMenu;
    MenuName,SubMenuName:TData;
    mi:TAIMenuItem;
    i,j:integer;
begin
  MainMenu:=GetMainMenu(Ex);

  SubMenuName    :=Ex.PopDS;
  MenuName      :=Ex.PopDS;
  try
    for i:=0 to MainMenu.Items.Count - 1 do begin
      if CT(MainMenu.Items[i].Caption, MenuName.Value) then begin

        for j:=0 to MainMenu.Items[i].Count - 1 do begin
          if CT(MainMenu.Items[i].Items[j].Caption, SubMenuName.Value) then begin
            mi:=TAIMenuItem(MainMenu.Items[i].Items[j]);
            MainMenu.Items[i].Delete(j);
            mi.Free;;
            exit;
          end;
        end;
      end;
    end;
  finally
    SubMenuName.Free;
    MenuName.Free;
  end;
end;

procedure BW_AddSubMenu(Ex:TContext);// DS: MenuName(exsisted), SubMenuName(added), ShortCutString, IconName, Script
var MainMenu:TMainMenu;
    MenuName,SubMenuName,ShortCutString,IconName,Script:TData;
    mi:TAIMenuItem;
    mainidx,subidx,i,j:integer;
    bmp:TBitmap;
begin
  MainMenu:=GetMainMenu(Ex);

  Script          :=Ex.PopDS;
  IconName        :=Ex.PopDS;
  ShortCutString  :=Ex.PopDS;
  SubMenuName     :=Ex.PopDS;
  MenuName        :=Ex.PopDS;
  mainidx :=-1;
  subidx  :=-1;
  try
    try
      for i:=0 to MainMenu.Items.Count - 1 do begin
        if CT(MainMenu.Items[i].Caption, MenuName.Value) then begin
          mainidx:=i;
          if not CT('-', SubMenuName.Value) then begin
            for j:=0 to MainMenu.Items[i].Count - 1 do begin
              if CT(MainMenu.Items[i].Items[j].Caption, SubMenuName.Value) then begin
                subidx:=j;
                TMenuItem(MainMenu.Items[mainidx].Items[subidx]).Free;
                break;
              end;
            end;
          end;
          break;
        end;
      end;

      mi:=TAIMenuItem.Create(MainMenu, Ex.Owner);
      mi.Caption  :=SubMenuName.Value;
      mi.Script    :=Script.Value;
      mi.ShortCut  :=TextToShortCut(ShortCutString.Value);

      mi.command.BitmapName:=IconName.Value;
      bmp:=BMPList.GetBMP(IconName.Value);
{$IFNDEF D3}
      if Assigned(bmp) then begin
        mi.Bitmap:=bmp;
      end;
{$ENDIF}

      if mainidx <> -1 then begin
        if subidx = -1 then begin
          MainMenu.Items[mainidx].Add(mi);
{$IFNDEF D3}
          if Assigned(bmp) then begin
            SetMenuItemBitmaps(MainMenu.Items[mainidx].Handle, MainMenu.Items[mainidx].Count - 1, MF_BYPOSITION, bmp.Handle, bmp.Handle);
          end;
{$ENDIF}
        end else begin
          MainMenu.Items[mainidx].Insert(subidx, mi);
{$IFNDEF D3}
          if Assigned(bmp) then begin
            SetMenuItemBitmaps(MainMenu.Items[mainidx].Handle, subidx, MF_BYPOSITION, bmp.Handle, bmp.Handle);
          end;
{$ENDIF}
        end;
      end;
    except
      exit;
    end;
  finally
    Script.Free;
    IconName.Free;
    ShortCutString.Free;
    SubMenuName.Free;
    MenuName.Free;
  end;
end;

procedure BW_AddSubSubMenu(Ex:TContext);// DS: MenuName(exsisted), SubMenuName(exsisted), SubSubMenuName(added), ShortCutString, IconName, Script
var MainMenu:TMainMenu;
    MenuName, SubMenuName, SubSubMenuName, ShortCutString, IconName, Script:TData;
    mi:TAIMenuItem;
    mainidx,subidx,subsubidx,i,j,k:integer;
    bmp:TBitmap;
begin
  MainMenu:=GetMainMenu(Ex);

  Script          :=Ex.PopDS;
  IconName        :=Ex.PopDS;
  ShortCutString  :=Ex.PopDS;
  SubSubMenuName  :=Ex.PopDS;
  SubMenuName     :=Ex.PopDS;
  MenuName        :=Ex.PopDS;
  mainidx   :=-1;
  subidx    :=-1;
  subsubidx :=-1;
  try
    try
      for i:=0 to MainMenu.Items.Count - 1 do begin
        if CT(MainMenu.Items[i].Caption, MenuName.Value) then begin
          mainidx:=i;
          if not CT('-', SubMenuName.Value) then begin
            for j:=0 to MainMenu.Items[i].Count - 1 do begin
              if CT(MainMenu.Items[i].Items[j].Caption, SubMenuName.Value) then begin
                subidx:=j;
                for k:=0 to MainMenu.Items[i].Items[j].Count - 1 do begin
                  if CT(MainMenu.Items[i].Items[j].Items[k].Caption, SubSubMenuName.Value) then begin
                    subsubidx:=k;
                    TMenuItem(MainMenu.Items[mainidx].Items[subidx].Items[subsubidx]).Free;
                    break;
                  end;
                end;
                break;
              end;
            end;
          end;
          break;
        end;
      end;

      mi:=TAIMenuItem.Create(MainMenu, Ex.Owner);
      mi.Caption  :=SubSubMenuName.Value;
      mi.Script   :=Script.Value;
      mi.ShortCut :=TextToShortCut(ShortCutString.Value);

      mi.command.BitmapName:=IconName.Value;
      bmp:=BMPList.GetBMP(IconName.Value);
{$IFNDEF D3}
      if Assigned(bmp) then begin
        mi.Bitmap:=bmp;
      end;
{$ENDIF}

      if (mainidx <> -1) and (subidx <> -1) then begin
        MainMenu.Items[mainidx].Items[subidx].OnClick:=nil;
        if subsubidx = -1 then begin
          MainMenu.Items[mainidx].Items[subidx].Add(mi);
{$IFNDEF D3}
          if Assigned(bmp) then begin
            SetMenuItemBitmaps(MainMenu.Items[mainidx].Items[subidx].Handle, MainMenu.Items[mainidx].Items[subidx].Count - 1, MF_BYPOSITION, bmp.Handle, bmp.Handle);
          end;
{$ENDIF}
        end else begin
          MainMenu.Items[mainidx].Items[subidx].Insert(subsubidx, mi);
{$IFNDEF D3}
          if Assigned(bmp) then begin
            SetMenuItemBitmaps(MainMenu.Items[mainidx].Items[subidx].Handle, subsubidx, MF_BYPOSITION, bmp.Handle, bmp.Handle);
          end;
{$ENDIF}
        end;
      end;
    except
      exit;
    end;
  finally
    Script.Free;
    IconName.Free;
    ShortCutString.Free;
    SubSubMenuName.Free;
    SubMenuName.Free;
    MenuName.Free;
  end;
end;

procedure BW_AddBitmap(Ex:TContext);  // DS: BMPName, BMPPath:str ->
var BMPName, BMPPath:TData;
begin
  BMPPath:=Ex.PopDS;
  BMPName:=Ex.PopDS;
  try
    try
      BMPList.Add(BMPName.Value, BMPPath.Value);
    except
      Error(errIllegalParameters, ['AddBitmap']);
    end;
  finally
    BMPName.Free;
    BMPPath.Free;
  end;
end;
/////////////////////////////////////////////////////////////////////////////
procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('FORMS',  'DropMainMenu',   BWP(BW_DropMainMenu,   atIsFunction));
  Words.AddWord('FORMS',  'AddMainMenu',    BWP(BW_AddMainMenu,    atIsFunction));

  Words.AddWord('FORMS',  'DropSubMenu',    BWP(BW_DropSubMenu,    atIsFunction));
  Words.AddWord('FORMS',  'AddSubMenu',     BWP(BW_AddSubMenu,     atIsFunction));
  Words.AddWord('FORMS',  'AddSubSubMenu',  BWP(BW_AddSubSubMenu,  atIsFunction));

  Words.AddWord('SYSTEM', 'AddBitmap',      BWP(BW_AddBitmap,      atIsFunction));
end;

initialization
  RegisterClass(TAIMenuItem);

  BMPList:=TBMPList.Create;
finalization
  BMPList.Free;
  BMPList:=nil;
end.
