unit AISQL;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes,

  ARUtils, DBTables, AInterpreter, ARSQL;

procedure RegisterWords(Words:TWords);
var
  CurrentAlias:string;

type
  TAISQL = class(TSQL)
  public
    UseCount  :integer;
  end;

  TDataSQL = class(TData)
  protected
    function    GetValue:variant;override;
  public
    Q            :TAISQL;
    constructor  Create;
    destructor  Destroy;override;
    procedure    CallMethod(Ex:TContext);override;
    procedure    Assign(D:TData);override;
  private
  end;

implementation

///////////////////////////////////////////////////////////////////////////////
constructor TDataSQL.Create;
begin
  inherited Create;
  Q:=TAISQL.Create;
  Q.Q.DatabaseName:=CurrentAlias;
  Q.UseCount:=1;
end;

destructor  TDataSQL.Destroy;
begin
  dec(Q.UseCount);
  if Q.UseCount = 0 then begin
    Q.Free;
  end;

  inherited;
end;

procedure    TDataSQL.CallMethod(Ex:TContext);
var v:TData;
begin
  ProcessedProp:=true;
  if PropListCount > 0 then begin
    if CT(PropList[0], 'OPEN')              then Q.Open                         else
    if CT(PropList[0], 'FIRST')             then Q.First                        else
    if CT(PropList[0], 'EXEC')              then Q.Exec                         else
    if CT(PropList[0], 'CLOSE')             then Q.Close                        else
    if CT(PropList[0], 'NEW')               then begin Q.Close;Q.SQL.Clear;end  else
    if CT(PropList[0], 'PREPARE')           then Q.Prepare                      else
    if CT(PropList[0], 'NEXT')              then Q.Next                         else
    if CT(PropList[0], 'ADD')               then begin
      v:=Ex.PopDS;
      try
        Q.SQL.Add(v.Value);
      finally
        v.Free;
      end;
    end else
    ProcessedProp:=false;
  end;
end;

function    TDataSQL.GetValue:variant;
begin
  ProcessedProp:=true;
  if PropListCount > 0 then begin
    if CT(PropList[0], 'SQL')         then Result:=Q.SQL.Text       else
    if CT(PropList[0], 'SQLCOUNT')    then Result:=Q.SQL.Count      else
    if CT(PropList[0], 'S')           then Result:=Q.S[PropList[1]] else
    if CT(PropList[0], 'BLOB')        then Result:=Q.BLOB[PropList[1]] else
    if CT(PropList[0], 'I')           then Result:=Q.I[PropList[1]] else
    if CT(PropList[0], 'D')           then Result:=Q.D[PropList[1]] else
    if CT(PropList[0], 'F')           then Result:=Q.F[PropList[1]] else
    if CT(PropList[0], 'C')           then Result:=Q.C[PropList[1]] else
    if CT(PropList[0], 'G')           then Result:=ObjAsVar(Q.G[PropList[1]]) else
    if CT(PropList[0], 'EOF')         then Result:=integer(Q.EOF)   else
    if CT(PropList[0], 'RECORDCOUNT') then Result:=Q.RecordCount    else
    ProcessedProp:=false;
  end;
end;

procedure    TDataSQL.Assign(D:TData);
begin
  ProcessedProp:=true;
  if PropListCount > 0 then begin
    if CT(PropList[0], 'SQL')         then begin Q.Close; Q.SQL.Clear; Q.SQL.Add(D.Value)  end  else
    if CT(PropList[0], 'S')           then Q.S[PropList[1]]   :=D.Value else
    if CT(PropList[0], 'BLOB')        then Q.BLOB[PropList[1]]:=D.Value else
    if CT(PropList[0], 'I')           then Q.I[PropList[1]]   :=D.Value else
    if CT(PropList[0], 'D')           then Q.D[PropList[1]]   :=D.Value else
    if CT(PropList[0], 'F')           then Q.F[PropList[1]]   :=D.Value else
    if CT(PropList[0], 'C')           then Q.C[PropList[1]]   :=D.Value else
    if CT(PropList[0], 'REQUESTLIVE') then Q.RequestLive      :=D.Value else
//    if CT(PropList[0], 'M')    then Q.ParamByName(PropList[1]).AsFloat   :=D.Value else
    ProcessedProp:=false;
  end else begin
    if D.ClassType = Self.ClassType then begin
      dec(Q.UseCount);
      if Q.UseCount = 0 then begin
        Q.Free;
      end;
      inc(TDataSQL(D).Q.UseCount);
      Q:=TDataSQL(D).Q;
    end else begin
      ProcessedProp:=false;
    end;
  end;
end;
///////////////////////////////////////////////////////////////////////////////

procedure BW_CreateSQL(Ex:TContext);
begin
  Ex.AddLocalVar(TVariant(Ex.CS[Ex.IP]).Value, TDataSQL.Create);
  inc(Ex.IP);
end;

procedure BW_SQL(Ex:TContext);
var VarName:string;
begin
  if Ex.CompileMode then begin
    VarName:=Upper(Ex.GetNextTerm.Value);
    Ex.LocalWords.Add(VarName);
    Ex.CompileAddr(Ex.Owner.GetAddr('SQL', '_CreateSQL'));
    Ex.CompileData(TVariant.Create(VarName));
  end else begin
    Ex.Words.AddWord(Ex.CurrUnit, Ex.GetNextTerm.Value, TDataSQL.Create);
  end;
end;

procedure BW_ReturnSQL(Ex:TContext);
begin
  Ex.LocalWords.Add('RESULT');
  Ex.CompileAddr(Ex.Owner.GetAddr('SQL', '_CreateSQL'));
  Ex.CompileData(TVariant.Create('RESULT'));
end;

procedure  BW_SetCurrentAlias(Ex:TContext);
var a:TData;
begin
  a:=Ex.PopDS;
  try
    CurrentAlias:=a.Text;
  finally
    a.Free;
  end;
end;

procedure RegisterWords(Words:TWords);
begin
  Words.AddWord('SQL', 'SQL',             BWP(BW_SQL,               atIsImm or atKeyWord));
  Words.AddWord('SQL', '_CreateSQL',      BWP(BW_CreateSQL,         0));
  Words.AddWord('SQL', '_ReturnSQL',      BWP(BW_ReturnSQL,         atIsImm));

  Words.AddWord('SQL', 'SetCurrentAlias', BWP(BW_SetCurrentAlias,  atIsFunction));
end;

end.


