unit HMemo;
//************************************************************//
//                     AND Interpreter                       //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, SysUtils, Controls, Classes, Graphics, Forms, Messages,
  Math, ExtCtrls, StdCtrls, Clipbrd,

  ARUtils;

procedure Register;

const
  CR  = #$0D;
  LF  = #$0A;
  TAB = #$09;
  VK_A = $41;
  VK_T = $54;
  VK_W = $57;
  VK_Y = $59;
  VK_Z = $5A;
  VK_I = $49;
  VK_U = $55;

  VK_0 = $30;
  VK_1 = $31;
  VK_2 = $32;
  VK_3 = $33;
  VK_4 = $34;
  VK_5 = $35;
  VK_6 = $36;
  VK_7 = $37;
  VK_8 = $38;
  VK_9 = $39;


  UndoMax = 128;

type
  TStartStopKeyRec = record
    Start         :string;
    Stop          :string;
    Style         :TFontStyles;
    Color         :TColor;
    StopOnLineEnd :boolean;
  end;

  TUndoRec = record
    Text:string;
    CurX:integer;
    CurY:integer;
  end;

const
  StartStopKeys:array[1..4] of TStartStopKeyRec = (
    (Start:'//';  Stop:CR;    Style:[fsBold];   Color:clGreen;  StopOnLineEnd:true),
    (Start:'"';   Stop:'"';   Style:[fsBold];   Color:clBlue;   StopOnLineEnd:false),
    (Start:'''';  Stop:'''';  Style:[fsBold];   Color:clBlue;   StopOnLineEnd:false),
    (Start:'{';   Stop:'}';   Style:[fsBold];   Color:clGreen;  StopOnLineEnd:false)
  );

type

  TKeyWords = class(TStringList)
    public
      constructor Create;
      destructor  Destroy;override;

      function    Add(const S: string): Integer; override;
  end;

  TLines = class(TStringList)
    private
      function    GetSSIdx(Index: integer):integer;
      procedure   SetSSIdx(Index: integer; ASSIdx:integer);

      function    CulcSSIdxAfter(Line:integer):integer;
    public
      constructor Create;
      destructor  Destroy;override;

      function    Add(const S: string): Integer; override;
      procedure   Delete(Index: Integer); override;
      procedure   Insert(Index: Integer; const S: string); override;
      procedure   LoadFromFile(const FileName: string); override;
      procedure   Assign(Source: TPersistent); override;

      property    SSIdx[Index: integer]:integer read GetSSIdx write SetSSIdx;
  end;

  THMemo  = class(TCustomControl)
  private
    FTopY       :integer;
    FLeftX      :integer;
    FontDX      :integer;
    FontDY      :integer;
    FScrollUpdate:boolean;

    EOF         :boolean;
    BOF         :boolean;
    EOL         :boolean;

    FCurX       :integer;
    FCurY       :integer;

    FInMouseDrag:boolean;

    FTabStops   :integer;

    UndoList    :array[0..UndoMax - 1] of TUndoRec;
    UndoPos     :integer;

    BookMarkList:array[0..9] of integer;

    FLines      :TLines;
    FOnChange   :TNotifyEvent;
    FOnChangeStatus:TNotifyEvent;

    FGutter     :boolean;
    FGutterWidth:integer;

    FModified   :boolean;

    FColorLine  :integer;
    FLineColor  :TColor;

    FScrollBars :TScrollStyle;
    procedure   SetScrollBars(AScrollBars:TScrollStyle);

    procedure   SetSelStart(p:integer);
    procedure   SetModified(AModified:boolean);

    function    GetLines:TLines;
    procedure   SetLines(ALines: TLines);

    procedure   SetGutter(AGutter:boolean);
    procedure   SetGutterWidth(AGutterWidth:integer);

    function    TabSpaces:string;
    procedure   SaveUndo;

    procedure   SetCurX(ACurX:integer);
    procedure   SetCurY(ACurY:integer);
    procedure   UpdateCaret;
    procedure   ShowCaret;
    procedure   HideCaret;

    procedure   FontChange(Sender: TObject);
    procedure   LinesChange(Sender: TObject);

    function    GetS(posx:integer; posy:integer):char;
    procedure   SetS(posx:integer; posy:integer; c:char);

    function    GetVSTrackPos:integer;
    function    GetVSPos:integer;
    procedure   SetVSPos(AVSPos:integer);
    function    GetVSMax:integer;
    procedure   SetVSMax(AVSMax:integer);
    function    GetVSPage:cardinal;
    procedure   SetVSPage(AVSPage:cardinal);

    function    GetHSTrackPos:integer;
    function    GetHSPos:integer;
    procedure   SetHSPos(AHSPos:integer);
    function    GetHSMax:integer;
    procedure   SetHSMax(AHSMax:integer);
    function    GetHSPage:cardinal;
    procedure   SetHSPage(AHSPage:cardinal);

    procedure   SetTopY(ATopY:integer);
    procedure   SetLeftX(ALeftX:integer);

    function    ViewY:integer;
    function    ViewX:integer;

    procedure   BeginNextWord;
    procedure   BeginPriorWord;
    function    GetNextChar(var X:integer; var Y:integer):char;
    function    GetPriorChar(var X:integer; var Y:integer):char;

    function    GetSelText:string;
    procedure   SetSelText(s:string);

{
    function    GetLineColor(LineIdx:integer):TColor;
    procedure   SetLineColor(LineIdx:integer; AColor:TColor);
}
    function    GetText:string;
    procedure   SetText(AText:string);

  protected
    procedure   CreateParams(var p: TCreateParams);override;
    procedure   CreateWnd;override;
    procedure   KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure   KeyPress(var Key: Char);override;
    procedure   KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   DblClick; override;
    procedure   Loaded; override;

    procedure   WMVSCROLL(var m: TWMScroll ); message WM_VSCROLL;
    procedure   WMHSCROLL(var m: TWMScroll ); message WM_HSCROLL;
    procedure   WMSetFocus(var Message: TMessage); message WM_SETFOCUS;
    procedure   WMKillFocus(var Message: TMessage); message WM_KILLFOCUS;
    procedure   WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure   WMSize(var Message: TWMSize); message WM_SIZE;
    procedure   WMEraseBkgnd(var Message: TWMErasebkgnd); message WM_ERASEBKGND;

    procedure   ChangeStatus;virtual;

    function    MouseToCursor(X:integer; Y:integer):TPoint;
    function    MouseInSelection(mx:integer; my:integer):boolean;

    procedure   ScrollY(FromLine:integer;ToLine:integer;dy:integer);
    procedure   RepaintOn(yb:integer; ye:integer);
    function    InSelection:boolean;
  public
    KeyWords    :TStringList;

    KeyWordFontStyle:TFontStyles;
    KeyWordFontColor:TColor;

    SelBegin    :TPoint;
    SelEnd      :TPoint;

    Msg         :procedure(s:string);

    procedure   Paint;override;
    procedure   MyPaint;
    procedure   PaintLines(BeginLine:integer; EndLine:integer);

    constructor Create(AOwner:TComponent);override;
    destructor  Destroy;override;
    property    S[posx:integer; posy:integer]:char read GetS write SetS;

    property    VSPos :integer  read GetVSPos   write SetVSPos;
    property    VSMax :integer  read GetVSMax   write SetVSMax;
    property    VSPage:cardinal read GetVSPage  write SetVSPage;
    property    HSPos :integer  read GetHSPos   write SetHSPos;
    property    HSMax :integer  read GetHSMax   write SetHSMax;
    property    HSPage:cardinal read GetHSPage  write SetHSPage;

    property    TopY :integer   read FTopY      write SetTopY;
    property    LeftX:integer   read FLeftX     write SetLeftX;

    property    CurX:integer    read  FCurX     write SetCurX;
    property    CurY:integer    read  FCurY     write SetCurY;

    procedure   CopyToClipboard;
    procedure   CutToClipboard;
    procedure   PasteFromClipboard;
    procedure   Undo;
    procedure   Redo;
    procedure   ClearSelection;

    procedure   SetColorLine(ALine:integer; AColor:TColor);

    function    GetWordAt(x:integer; y:integer):string;
    procedure   SelectWordAt(x:integer; y:integer);

    property    SelText:string  read GetSelText write SetSelText;

//    property    LineColor[LineIdx:integer]:TColor read GetLineColor write SetLineColor;

    procedure   ClearUndo;

    function    FindText(txt:string; MatchCase: boolean; WholeWord:boolean): boolean;

    function    GetSSIdx(Before:integer):integer;
  published
    {standart VCL property}
    property    Align;
    property    Color default clWindow;
    property    Ctl3D;
    property    DragCursor;
    property    DragMode;
    property    Enabled;
    property    Font;
    property    ParentColor default False;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    OnClick;
    property    OnDblClick;
    property    OnDragDrop;
    property    OnDragOver;
    property    OnEndDrag;
    property    OnEnter;
    property    OnExit;
    property    OnKeyDown;
    property    OnKeyPress;
    property    OnKeyUp;
    property    OnMouseDown;
    property    OnMouseMove;
    property    OnMouseUp;
    property    SelStart:integer write SetSelStart;

    property    OnChange:TNotifyEvent       read FOnChange        write FOnChange;
    property    OnChangeStatus:TNotifyEvent read FOnChangeStatus  write FOnChangeStatus;
    property    TabStops:integer            read FTabStops        write FTabStops default 2;
    property    Lines:TLines                read GetLines         write SetLines;
    property    Modified:boolean            read FModified        write SetModified;

    property    Text:string                  read GetText          write SetText;

    property    Gutter:boolean              read FGutter          write SetGutter;
    property    GutterWidth:integer         read FGutterWidth     write SetGutterWidth;
    property    ScrollBars:TScrollStyle     read FScrollBars      write SetScrollBars default ssNone;
  end;

implementation
{$R *.RES}

procedure Register;
begin
  RegisterComponents('ARComponents', [THMemo]);
end;

function    IsWordBreak(ch:char):boolean;
begin
  Result:=ch in [#0,#$0d,#$0a,#$09,' ','(',')','!','#','$','%','&','*','+',',','-','/',':',';','<','=','>','?','@','[','\',']','^','{','|','}','~','"','''','.'];
end;

function    Max(a:integer; b:integer):integer;
begin
  if a > b then Result:=a else Result:=b;
end;

function    Min(a:integer; b:integer):integer;
begin
  if a < b then Result:=a else Result:=b;
end;
/////////////////////////////////////////////////
constructor TKeyWords.Create;
begin
  inherited;

  Sorted    :=true;
  Duplicates:=dupIgnore;
end;

destructor  TKeyWords.Destroy;
begin
  inherited;
end;

function    TKeyWords.Add(const S: string): integer;
begin
  Result:=-1;

  if IndexOf(S) = -1 then begin
    inherited Add(ANSIUpperCase(S));
  end;
end;

/////////////////////////////////////////////////
constructor TLines.Create;
begin
  inherited;
end;

destructor  TLines.Destroy;
begin
  inherited;
end;

procedure   TLines.LoadFromFile(const FileName: string);
begin
  inherited LoadFromFile(FileName);

  GetSSIdx(Count - 1);//force reculc ssidx
end;

procedure   TLines.Assign(Source: TPersistent);
begin
  inherited Assign(Source);

//  ClearSSIdx;
  GetSSIdx(Count - 1);//force reculc ssidx
end;

function    TLines.GetSSIdx(Index: integer):integer;
begin
  if (Index < 0) or (Index >= Count) then begin
    Result:=-1;
  end else begin
    Result:=integer(Objects[Index]);
    if Result = 0 then begin
      Result:=CulcSSIdxAfter(Index - 1);
      Objects[Index]:=pointer(Result);
    end;
  end;
end;

procedure   TLines.SetSSIdx(Index: integer; ASSIdx:integer);
begin
  if (Index >= 0) and (Index < Count) and (Objects[Index] <> pointer(ASSIdx)) then begin
    Objects[Index]:=pointer(ASSIdx);
  end;
end;

function    TLines.CulcSSIdxAfter(Line:integer):integer;
  function CulcGetSSIdx(s:string; ASSIdx:integer):integer;
  var pos, i, j:integer;
  begin
    Result:=ASSIdx;
    pos:=1;

    try
      while pos <= Length(s) do begin
        if IsWordBreak(s[pos]) then begin
          if Result = -1 then begin
            for i:=1 to High(StartStopKeys) do begin
              if s[pos] = StartStopKeys[i].Start[1] then begin
                if Length(StartStopKeys[i].Start) = 1 then begin
                  Result:=i;
                  break;
                end else begin
                  inc(pos); if pos = Length(s) then exit;
                  for j:=2 to Length(StartStopKeys[i].Start) do begin
                    if s[pos] = StartStopKeys[i].Start[j] then begin
                      Result:=i;
                      break;
                    end;
                    inc(pos); if pos = Length(s) then exit;
                  end;
                end;
              end;
              if Result <> -1 then break;
            end;
          end else begin
            if s[pos] = StartStopKeys[Result].Stop[1] then begin
              if Length(StartStopKeys[Result].Stop) = 1 then begin
                Result:=-1;
              end else begin
                inc(pos); if pos = Length(s) then exit;
                for j:=2 to Length(StartStopKeys[Result].Stop) do begin
                  if s[pos] = StartStopKeys[Result].Stop[j] then begin
                    Result:=-1;
                    break;
                  end;
                  inc(pos); if pos = Length(s) then exit;
                end;
              end;
            end;
          end;
        end;
        inc(pos);
      end;
    finally
      if (Result <> -1) and StartStopKeys[Result].StopOnLineEnd then Result:=-1;
    end;
  end;
var ssi:integer;
begin
  if (Line < 0) or (Line >= Count) then begin
    Result:=-1;
  end else begin
    if Line = 0 then begin
      ssi:=-1;
    end else begin
      ssi:=SSIdx[Line - 1];
    end;
    Result:=CulcGetSSIdx(Self[Line], ssi);
  end;
end;

function    TLines.Add(const S: string): Integer;
begin
  inherited Add(S);

  SSIdx[Count - 1]:=CulcSSIdxAfter(Count - 2);

  Result:=Count - 1;
end;

procedure   TLines.Delete(Index: Integer);
begin
  if Index < (Count - 1) then begin
    SSIdx[Index + 1]:=SSIdx[Index];

    inherited Delete(Index);
  end;
end;

procedure   TLines.Insert(Index: Integer; const S: string);
begin
  if Index < (Count - 1) then begin
    SSIdx[Index + 1]:=SSIdx[Index];
  end;

  inherited Insert(Index, S);
end;

/////////////////////////////////////////////////

constructor THMemo.Create(AOwner:TComponent);
var i:integer;
begin
  inherited Create(AOwner);
{$IFDEF D4}
  DoubleBuffered:=true;
{$ENDIF}
  ControlStyle:=[csClickEvents, csCaptureMouse, csDoubleClicks];

  FontDX:=8;
  FontDY:=13;

  FColorLine:=-1;
  FLineColor:=-1;

  FLines:=TLines.Create;
  FLines.OnChange:=LinesChange;

  KeyWords:=TKeyWords.Create;

{
  // to test keywords
  KeyWords.Add('UNIT');
  KeyWords.Add('USES');
  KeyWords.Add('PROC');
  KeyWords.Add('FUNC');
  KeyWords.Add('VAR');
  KeyWords.Add('Q');
  KeyWords.Add('FUNC');
  KeyWords.Add('COPYFILES');
  KeyWords.Add('FILESFROM');
  KeyWords.Add('FILESTO');
  KeyWords.Add('LIST');
  KeyWords.Add('L');
  KeyWords.Add('FILEMASK');
  KeyWords.Add('PATHFROM');
  KeyWords.Add('PATHTO');
  KeyWords.Add('FILEPATH');
  KeyWords.Add('RESULT');
  KeyWords.Add('TRUE');
  KeyWords.Add('FIRST');
  KeyWords.Add('WHILE');
  KeyWords.Add('NOT');
  KeyWords.Add('EOF');
  KeyWords.Add('DO');
  KeyWords.Add('COPYFILE');
  KeyWords.Add('STR');
  KeyWords.Add('FALSE');
  KeyWords.Add('PRINT');
  KeyWords.Add('CAN');
  KeyWords.Add('COPY');
  KeyWords.Add('ENDIF');
  KeyWords.Add('NEXT');
  KeyWords.Add('ENDWHILE');
  KeyWords.Add('END');
  KeyWords.Add('DELETEFILES');
  KeyWords.Add('PATH');
  KeyWords.Add('DELETEFILE');
  KeyWords.Add('DELETE');
  KeyWords.Add('FUNCTION');
  KeyWords.Add('BEGIN');
  KeyWords.Add('PROCEDURE');
  KeyWords.Add('CLASS');
  KeyWords.Add('TYPE');
  KeyWords.Add('CONSTRUCTOR');
  KeyWords.Add('IF');
  KeyWords.Add('THEN');
  KeyWords.Add('ELSE');
  KeyWords.Add('INHERITED');
  KeyWords.Add('INTERFACE');
  KeyWords.Add('IMPLEMENTATION');
}

//  KeyWordFont:=TFont.Create;

  Cursor      :=crIBeam;
  ParentColor :=False;
  Color       :=clWindow;
  TabStop     :=true;

  Gutter      :=true;
  GutterWidth :=30;

  for i:=0 to 9 do BookMarkList[i]:=-1;
end;

destructor  THMemo.Destroy;
begin
  if Focused then DestroyCaret;
  FLines.OnChange:=nil;

  KeyWords.Free;
  Lines.Free;

  inherited;
end;

procedure   THMemo.CreateParams(var p: TCreateParams);
begin
  inherited CreateParams(p);
  with p do begin
    Style   :=Style or WS_TABSTOP;

    if ScrollBars in [ssBoth, ssVertical] then begin
      Style:=Style or WS_VSCROLL;
    end;
    if ScrollBars in [ssBoth, ssHorizontal] then begin
      Style:=Style or WS_HSCROLL;
    end;

    ExStyle :=ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

procedure   THMemo.CreateWnd;
begin
//  Align:=alClient;

  KeyWordFontColor:=Font.Color;
  KeyWordFontStyle:=[fsBold];
  Font.OnChange:=FontChange;
//  Font.Name:='COURIER';
//  Font.Size:=10;

  inherited;

  FontChange(Self);
  HSMax:=1000;

  if TabStops = 0 then TabStops:=2;
end;

procedure   THMemo.Loaded;
begin
  inherited;

end;

function    THMemo.InSelection:boolean;
begin
  Result:=(SelBegin.x <> SelEnd.x) or (SelBegin.y <> SelEnd.y);
end;

procedure   THMemo.FontChange(Sender: TObject);
begin
  Canvas.Font:=Font;

  FontDX:=Canvas.TextWidth('O') + 1;
  FontDY:=Canvas.TextHeight('O');
end;

function    THMemo.ViewY:integer;
begin
  Result:=ClientHeight div FontDY + 1;
end;

function    THMemo.ViewX:integer;
begin
  if Gutter then begin
    Result:=Max((ClientWidth - GutterWidth) div FontDX, 0);
  end else begin
    Result:=ClientWidth div FontDX;
  end;
end;

function    THMemo.GetS(posx:integer; posy:integer):char;
begin
  inc(posx);
  if posy < Lines.Count then begin
    if posx <= Length(Lines[posy]) then begin
      Result:=Lines[posy][posx];
      EOL:=false;
      EOF:=false;
    end else begin
      EOL:=true;
      Result:=CR;
    end;
  end else begin
    EOF:=true;
    EOL:=true;
    Result:=CR;
  end;
end;

procedure   THMemo.SetS(posx:integer; posy:integer; c:char);
begin
  if posy < Lines.Count then begin
    if posx < Length(Lines[posy]) then begin
      PChar(Lines[posy])[posx]:=c;
    end else begin
      while Length(Lines[posy]) <= posx do Lines[posy]:=Lines[posy] + ' ';
      PChar(Lines[posy])[posx]:=c;
    end;
  end else begin
    while Lines.Count <= posy do Lines.AddObject('', nil);
    if posx <= Length(Lines[posy]) then begin
      PChar(Lines[posy])[posx]:=c;
    end else begin
      while Length(Lines[posy]) < posx do Lines[posy]:=Lines[posy] + ' ';
      PChar(Lines[posy])[posx]:=c;
    end;
  end;
end;

///////////////////////////////////////////////////////////////////
function    THMemo.GetVSTrackPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nTrackPos;
end;
function    THMemo.GetVSPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nPos;
end;
procedure   THMemo.SetVSPos(AVSPos:integer);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssVertical] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_POS;
    si.nPos:=AVSPos;
    SetScrollInfo(Handle, SB_VERT, si, true);
  end;
end;

function    THMemo.GetVSMax:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nMax;
end;
procedure   THMemo.SetVSMax(AVSMax:integer);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssVertical] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_RANGE;
    si.nMin:=0;
    si.nMax:=AVSMax;
    SetScrollInfo(Handle, SB_VERT, si, true);
  end;
end;

function    THMemo.GetVSPage:cardinal;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nPage;
end;
procedure   THMemo.SetVSPage(AVSPage:cardinal);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssVertical] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_PAGE;
    si.nPage:=AVSPage;
    SetScrollInfo(Handle, SB_VERT, si, true);
  end;
end;

function    THMemo.GetHSTrackPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nTrackPos;
end;
function    THMemo.GetHSPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nPos;
end;
procedure   THMemo.SetHSPos(AHSPos:integer);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssHorizontal] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_POS;
    si.nPos:=AHSPos;
    SetScrollInfo(Handle, SB_HORZ, si, true);
  end;
end;

function    THMemo.GetHSMax:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nMax;
end;
procedure   THMemo.SetHSMax(AHSMax:integer);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssHorizontal] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_RANGE;
    si.nMin:=0;
    si.nMax:=AHSMax;

    SetScrollInfo(Handle, SB_HORZ, si, true);
  end;
end;

function    THMemo.GetHSPage:cardinal;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nPage;
end;
procedure   THMemo.SetHSPage(AHSPage:cardinal);
var si: TScrollInfo;
begin
  if ScrollBars in [ssBoth, ssHorizontal] then begin
    si.cbSize:=SizeOf(si);
    si.fMask:=SIF_PAGE;
    si.nPage:=AHSPage;
    SetScrollInfo(Handle, SB_HORZ, si, true);
  end;
end;
///////////////////////////////////////////////////////////////////

procedure   THMemo.WMVSCROLL(var m: TWMScroll );
begin
  if FScrollUpdate then exit;
  FScrollUpdate:=true;

  case m.ScrollCode of
    SB_LINEDOWN:begin
      TopY:=TopY + 1;
    end;
    SB_LINEUP  :begin
      TopY:=TopY - 1;
    end;
    SB_PAGEUP  :begin
      TopY:=TopY - ViewY;
    end;
    SB_PAGEDOWN:begin
      TopY:=TopY + ViewY;
    end;
    SB_THUMBPOSITION:begin
      TopY:=VSPos;
    end;
    SB_THUMBTRACK:begin
      TopY:=GetVSTrackPos;
    end;
  end;

  if CurY = (TopY - 1)      then CurY:=TopY;
  if CurY = (TopY + ViewY)  then CurY:=TopY + ViewY - 1;
  while CurY < (TopY - 1)     do CurY:=CurY + ViewY;
  while CurY > (TopY + ViewY) do CurY:=CurY - ViewY;

  FScrollUpdate:=false;
end;

procedure   THMemo.WMHSCROLL(var m: TWMScroll );
begin
  if FScrollUpdate then exit;
  FScrollUpdate:=true;

  case m.ScrollCode of
    SB_LINEDOWN:begin
      LeftX:=LeftX + 1;
    end;
    SB_LINEUP  :begin
      LeftX:=LeftX - 1;
    end;
    SB_PAGEUP  :begin
      LeftX:=LeftX - ViewX;
    end;
    SB_PAGEDOWN :begin
      LeftX:=LeftX + ViewX;
    end;
    SB_THUMBPOSITION:begin
      LeftX:=HSPos;
    end;
    SB_THUMBTRACK:begin
      LeftX:=GetHSTrackPos;
    end;
  end;

  if CurX = (LeftX - 1)      then CurX:=LeftX;
  if CurX = (LeftX + ViewX)  then CurX:=LeftX + ViewX - 1;
  while CurX < (LeftX - 1)     do CurX:=CurX + ViewX;
  while CurX > (LeftX + ViewX) do CurX:=CurX - ViewX;

  FScrollUpdate:=false;
end;

procedure   THMemo.ScrollY(FromLine:integer;ToLine:integer;dy:integer);
var r, cr:TRect;
begin
  if dy = 0 then exit;

  r :=Rect(0, FromLine * FontDY,    ClientWidth, ToLine * FontDY);
  if dy < 0 then begin
    cr:=Rect(0, r.Top,  ClientWidth, r.Bottom);
  end else begin
    cr:=Rect(0, r.Top,  ClientWidth, Min(ClientHeight, r.Bottom + dy * FontDY));
  end;

  ScrollWindow(Self.Handle, 0, dy * FontDY, @r, @cr);
  Update;
end;

procedure   THMemo.SetTopY(ATopY:integer);
var old:integer;
begin
  old:=FTopY;
  FTopY:=Max(ATopY, 0);
  VSPos:=FTopY;
  if Abs(old - FTopY) >= ViewY then begin
    MyPaint;
  end else begin
    if old < FTopY then begin
      ScrollY(0, ViewY, old - FTopY);
    end else begin
      ScrollY(0, ViewY, old - FTopY);
    end;
  end;
end;

procedure   THMemo.SetLeftX(ALeftX:integer);
begin
  if FLeftX <> ALeftX then begin
    FLeftX:=Min(Max(ALeftX, 0), 1000);
    HSPos:=FLeftX;
    MyPaint;
  end;
end;

procedure   THMemo.WMSetFocus(var Message: TMessage);
begin
  Self.SetFocus;
  DestroyCaret;
  CreateCaret(Handle, 0, 2, FontDY);
  Windows.ShowCaret(Handle);
end;

procedure   THMemo.WMKillFocus(var Message: TMessage);
begin
  DestroyCaret;
end;

procedure   THMemo.SetCurX(ACurX:integer);
begin
  HideCaret;
  FCurX:=Max(ACurX, 0);
  UpdateCaret;
end;

procedure   THMemo.SetCurY(ACurY:integer);
begin
  HideCaret;
  FCurY:=Min(Max(ACurY, 0), Lines.Count - 1);
  UpdateCaret;
end;

procedure   THMemo.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := DLGC_WANTARROWS or DLGC_WANTCHARS or {DLGC_WANTALLKEYS or }DLGC_WANTTAB;
end;

procedure   THMemo.WMSize(var Message: TWMSize);
begin
  inherited;
  Realign;

  VSPage:=ViewY;
  HSPage:=ViewX;
end;

procedure   THMemo.ShowCaret;
begin
  if Focused then begin
    Windows.ShowCaret(Handle);
  end;
end;

procedure   THMemo.HideCaret;
begin
  if Focused then begin
    Windows.HideCaret(Handle);
  end;
end;

procedure   THMemo.UpdateCaret;
var tTopY, tLeftX:integer;
begin
  tTopY :=TopY;
  tLeftX:=LeftX;

  if CurY = (tTopY - 1)      then tTopY:=CurY;
  if CurY = (tTopY + ViewY)  then tTopY:=CurY - ViewY + 1;
  while CurY < (tTopY - 1)     do tTopY:=tTopY - ViewY;
  while CurY > (tTopY + ViewY) do tTopY:=tTopY + ViewY;

  if CurX = (tLeftX - 1)      then tLeftX:=CurX;
  if CurX = (tLeftX + ViewX)  then tLeftX:=CurX - ViewX + 1;
  while CurX < (tLeftX - 1)     do tLeftX:=tLeftX - ViewX;
  while CurX > (tLeftX + ViewX) do tLeftX:=tLeftX + ViewX;

  TopY :=tTopY;
  LeftX:=tLeftX;

  if Focused then begin
    if Gutter then begin
      SetCaretPos((CurX - LeftX)* FontDX + GutterWidth, (CurY - TopY) * FontDY);
    end else begin
      SetCaretPos((CurX - LeftX)* FontDX, (CurY - TopY) * FontDY);
    end;
  end;

  ShowCaret;
  ChangeStatus;
end;

function    THMemo.GetNextChar(var X:integer; var Y:integer):char;
begin
  EOL:=false;
  EOF:=false;
  BOF:=false;
  Result:=CR;

  X:=X + 1;
  if Y < Lines.Count then begin
    if X < Length(Lines[Y]) then begin
      Result:=S[X, Y];
    end else
    if X = Length(Lines[Y]) then begin
      Result:=CR;
      EOL:=true;
    end else
    if X > Length(Lines[Y]) then begin
      Y:=Y + 1;
      X:=0;
      if Y < Lines.Count then begin
        Result:=S[X, Y];
        EOL:=false;
      end else begin
        EOF:=true;
        EOL:=true;
        Result:=CR;
      end;
    end;
  end else begin
    EOF:=true;
    EOL:=true;
    Result:=CR;
  end;
end;

function    THMemo.GetPriorChar(var X:integer; var Y:integer):char;
begin
  EOL:=false;
  EOF:=false;
  BOF:=false;
  Result:=CR;

  if Y >= 0 then begin
    if (X - 1) >= 0 then begin
      X:=X - 1;
      Result:=S[X, Y];
    end else
    if (X - 1) < 0 then begin
      if Y > 0 then begin
        Y:=Y - 1;
        X:=Length(Lines[Y]);
        Result:=S[X, Y];
        EOL:=true;
      end else begin
        Y:=0;
        X:=0;
        Result:=S[X, Y];
        BOF:=true;
      end;
    end;
  end else begin
    EOF:=true;
    EOL:=true;
    Result:=CR;
  end;
end;

procedure   THMemo.BeginNextWord;
var ch:char;
    X,Y:integer;
begin
  X:=CurX;
  Y:=CurY;

  ch:=S[X, Y];
  if ch = CR then begin
    inc(Y); X:=0;
  end;

  while (not BOF) and (not IsWordBreak(ch)) do ch:=GetNextChar(X, Y);

  while (not EOF) and (IsWordBreak(ch) and (ch <> CR)) do ch:=GetNextChar(X, Y);

  CurX:=X;
  CurY:=Y;
end;

procedure   THMemo.BeginPriorWord;
var ch:char;
    X,Y:integer;
begin
  X:=CurX;
  Y:=CurY;

  ch:=GetPriorChar(X, Y);
  while (not BOF) and (IsWordBreak(ch)) do ch:=GetPriorChar(X, Y);
  while (not BOF) and (not IsWordBreak(ch)) do ch:=GetPriorChar(X, Y);

  GetNextChar(X, Y);

  CurX:=X;
  CurY:=Y;
end;

function    THMemo.MouseInSelection(MX:integer; MY:integer):boolean;
begin
  Result:=((MX + MY * 1000) >= Min((SelBegin.x + SelBegin.y * 1000), (SelEnd.x + SelEnd.y * 1000)))
          and
          ((MX + MY * 1000) <= Max((SelBegin.x + SelBegin.y * 1000), (SelEnd.x + SelEnd.y * 1000)))
end;

function    THMemo.MouseToCursor(X:integer; Y:integer):TPoint;
begin
  if Gutter then begin
    Result.x:=(X - GutterWidth + FontDX div 2) div FontDX + LeftX;
  end else begin
    Result.x:=(X + FontDX div 2) div FontDX;
  end;
  Result.y  :=(Y - FontDY * 1 div 4) div FontDY + TopY;
end;

procedure   THMemo.MouseMove(Shift: TShiftState; X, Y: Integer);
var p, OldSelEnd:TPoint;
begin
  inherited MouseMove(Shift, X, Y);

  p:=MouseToCursor(X, Y);

  if ssLeft in Shift then begin
    OldSelEnd:=SelEnd;
    CurX:=p.x;
    CurY:=p.y;
    UpdateCaret;
    if not FInMouseDrag then begin
      SelEnd:=Point(CurX, CurY);
      Self.Cursor:=crDrag;
      RepaintOn(OldSelEnd.y,  SelEnd.y);
    end;
  end else begin
    if MouseInSelection(p.x, p.y) then begin
      Self.Cursor:=crArrow;
    end else begin
      Self.Cursor:=crIBeam;
    end;
  end;
end;

procedure   THMemo.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var p, OldSelBegin, OldSelEnd:TPoint;
begin
  SetFocus;
  
  inherited MouseDown(Button, Shift, X, Y);

  p:=MouseToCursor(X, Y);

  CurX:=p.x;
  CurY:=p.y;

  if MouseInSelection(p.x, p.y) then begin
    FInMouseDrag:=true;
    Self.Cursor:=crDrag;
  end else begin
    UpdateCaret;

    OldSelBegin:=SelBegin;
    OldSelEnd  :=SelEnd;

    SelBegin:=Point(CurX, CurY);
    SelEnd  :=Point(CurX, CurY);

    RepaintOn(Min(OldSelBegin.y, OldSelEnd.y), Max(OldSelBegin.y, OldSelEnd.y));
  end;
end;

procedure   THMemo.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var p:TPoint;
    s:string;
    OldSelBeg, OldSelEnd:TPoint;
    NewSelBeg, NewSelEnd:TPoint;
    shft, size:integer;
begin
  inherited MouseUp(Button, Shift, X, Y);

  if FInMouseDrag then begin
    p:=MouseToCursor(X, Y);
    CurX:=p.x;
    CurY:=p.y;

    OldSelBeg:=SelBegin;
    OldSelEnd:=SelEnd;
    NewSelBeg:=p;
    NewSelEnd:=p;
    if not MouseInSelection(p.x, p.y) then begin
      s:=SelText;

      if ssCtrl in Shift then begin
        shft:=0;
      end else begin
        shft:=(NewSelBeg.x - OldSelBeg.x) + 1000 * (NewSelBeg.y - OldSelBeg.y);
      end;
      size:=(SelEnd.x  - SelBegin.x)    + 1000 * (SelEnd.y  - SelBegin.y);
      if shft >= 0 then begin
        SelBegin:=NewSelBeg;
        SelEnd  :=NewSelEnd;
        SelText:=s;

        if not (ssCtrl in Shift) then begin
          SelBegin:=OldSelBeg;
          SelEnd  :=OldSelEnd;
          ClearSelection;
        end;

        SelBegin:=NewSelBeg;
        SelEnd  :=Point((SelBegin.x  + 1000 * SelBegin.y  + size) mod 1000, (SelBegin.x  + 1000 * SelBegin.y  + size) div 1000);
      end else begin
        if not (ssCtrl in Shift) then begin
          ClearSelection;
          MyPaint;
        end;

        SelBegin:=NewSelBeg;
        SelEnd  :=NewSelEnd;
        CurX:=SelBegin.x;
        CurY:=SelBegin.y;
        SelText:=s;
        SelBegin:=NewSelBeg;
        SelEnd  :=Point((SelBegin.x  + 1000 * SelBegin.y  + size) mod 1000, (SelBegin.x  + 1000 * SelBegin.y  + size) div 1000);
      end;
      CurX:=SelBegin.x;
      CurY:=SelBegin.y;
      MyPaint;
    end;
  end;
  FInMouseDrag:=false;
  Self.Cursor:=crIBeam;
end;

procedure   THMemo.KeyUp(var Key: Word; Shift: TShiftState);
begin
  inherited KeyUp(Key, Shift)
end;

function    THMemo.GetSelText:string;
var B,E:integer;
    pb,pe:TPoint;
    x,y:integer;
    ch:char;
begin
  Result:='';

  if InSelection then begin

    B:=SelBegin.x + SelBegin.y * 1000;
    E:=SelEnd.x + SelEnd.y * 1000;

    if B <= E then begin
      pb.x:=SelBegin.x; pe.x:=SelEnd.x;
      pb.y:=SelBegin.y; pe.y:=SelEnd.y;
    end else begin
      pb.x:=SelEnd.x; pe.x:=SelBegin.x;
      pb.y:=SelEnd.y; pe.y:=SelBegin.y;
    end;

    x:=pb.x; y:=pb.y;

    while (x + y * 1000) < (pe.x + pe.y * 1000) do begin
      ch:=S[x, y];
      Result:=Result + ch;
      if ch = CR then begin
        Result:=Result + LF;
      end;
      GetNextChar(x, y);
    end;
  end;
end;

procedure   THMemo.ClearSelection;
var B,E:integer;
    pb,pe:TPoint;
    i:integer;
    t,t2:string;
begin
  if InSelection then begin

    B:=SelBegin.x + SelBegin.y * 1000;
    E:=SelEnd.x + SelEnd.y * 1000;

    if B <= E then begin
      pb.x:=SelBegin.x; pe.x:=SelEnd.x;
      pb.y:=SelBegin.y; pe.y:=SelEnd.y;
    end else begin
      pb.x:=SelEnd.x; pe.x:=SelBegin.x;
      pb.y:=SelEnd.y; pe.y:=SelBegin.y;
    end;

//    x:=pb.x; y:=pb.y;

    CurX:=pb.x;
    CurY:=pb.y;
    SelBegin:=Point(0, CurY);
    SelEnd  :=Point(0, CurY);

    case (pe.y - pb.y) of
      0:    begin
              t:=Lines[pb.y];
              System.Delete(t, pb.x + 1, pe.x - pb.x);
              Lines[pb.y]:=t;
            end;
      1:    begin
              t:=Lines[pb.y];
              System.Delete(t, pb.x + 1, Length(t) - pb.x);
              t2:=Lines[pb.y + 1];
              System.Delete(t2, 1, pe.x);
              Lines[pb.y]:=t + t2;
              Lines.Delete(pb.y + 1);

              ScrollY(pb.y - TopY + 1, ViewY, -1);
            end;
      else  begin
              for i:=pb.y + 1 to pe.y - 1 do begin
                Lines.Delete(pb.y + 1);
              end;

              t:=Lines[pb.y];
              System.Delete(t, pb.x + 1, Length(t) - pb.x);
              t2:=Lines[pb.y + 1];
              System.Delete(t2, 1, pe.x);
              Lines[pb.y]:=t + t2;
              Lines.Delete(pb.y + 1);

              ScrollY(pb.y - TopY + 1, ViewY, -(pe.y - pb.y));
            end;
    end;
  end;
end;

procedure   THMemo.SetSelText(s:string);
var i, x, y, cnt:integer;
    t:string;
begin
  SaveUndo;
  ClearSelection;

  x:=CurX; y:=CurY;

  while y >= Lines.Count do begin
    Lines.AddObject('', nil);
  end;

  t:=Lines[y];
  cnt:=1;
  for i:=1 to Length(s) do begin
    case s[i] of
      CR:   begin
              Lines[y]:=System.Copy(t, 1, x);
              y:=y + 1;
              Lines.Insert(y, System.Copy(t, x + 1, Length(t)));
              t:=Lines[y];
              x:=0;
              inc(cnt);
            end;
      LF: ;
      else  begin
              while x > Length(t) do t:=t + ' ';
              inc(x);
              System.Insert(s[i], t, x);
            end;
    end;

    Lines[y]:=t;
  end;

  ScrollY(CurY - TopY + 1, ViewY, cnt - 1);
  PaintLines(CurY, CurY + cnt);
  CurX:=x; CurY:=y;
  SelBegin:=Point(CurX, CurY);
  SelEnd  :=Point(CurX, CurY);
end;

procedure   THMemo.KeyDown(var Key: Word; Shift: TShiftState);
var ProcessKey:boolean;
    x, y, n, OldCurX, OldCurY:integer;
    t:string;
    OldSelBeg, OldSelEnd:TPoint;
begin
  inherited KeyDown(Key, Shift);

  OldCurX:=CurX;
  OldCurY:=CurY;
  OldSelBeg:=SelBegin;
  OldSelEnd:=SelEnd;

  ProcessKey:=false;

  case key of
    VK_HOME:  begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  CurX:=0;
                  CurY:=0;
                  TopY:=0;
                  LeftX:=0;
                end else begin
                  CurX:=0;
                end;
              end;
    VK_END :  begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  CurY:=Lines.Count - 1;
                  CurX:=Length(Lines[CurY]);
                  if CurY >= (TopY + ViewY - 1) then TopY:=CurY - ViewY div 2;
                end else begin
                  if CurY >= Lines.Count then begin
                    CurX:=0;
                  end else begin
                    CurX:=Length(Lines[CurY]);
                  end;
                end;
              end;
    VK_PRIOR: begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  CurY:=TopY;
                end else begin
                  CurY:=CurY - ViewY;
                end;
              end;
    VK_NEXT:  begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  CurY:=TopY + ViewY - 1;
                end else begin
                  CurY:=CurY + ViewY;
                end;
              end;
    VK_UP:    begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  TopY:=TopY - 1;
                  CurY:=Min(CurY, TopY + ViewY - 1);
                end else begin
                  CurY:=CurY - 1;
                end;
              end;
    VK_DOWN:  begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  TopY:=TopY + 1;
                  CurY:=Max(CurY, TopY);
                end else begin
                  CurY:=CurY + 1;
                end;
              end;
    VK_LEFT:  begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  BeginPriorWord;
                end else begin
                  CurX:=CurX - 1;
                end;
              end;
    VK_RIGHT: begin
                ProcessKey:=true;
                if ssCtrl in Shift then begin
                  BeginNextWord;
                end else
                if ssAlt in Shift then begin
                  x:=CurX; y:=CurY;
                  GetNextChar(x, y);
                  CurX:=x; CurY:=y;
                end else
                begin
                  CurX:=CurX + 1;
                end;
              end;
    VK_BACK:  begin
                ProcessKey:=true;
                if (not InSelection) or ((SelBegin.x = SelEnd.x) and (SelBegin.y = SelEnd.y)) then begin
                  if ssCtrl in Shift then begin
                    SelEnd  :=Point(CurX, CurY);
                    BeginPriorWord;
                    SelBegin:=Point(CurX, CurY);
                    SelText:='';
                  end else begin
                    if CurX > 0 then begin
                      x:=CurX; y:=CurY;
                      GetPriorChar(x, y);
                      CurX:=x; CurY:=y;

                      if S[CurX, CurY] = CR then begin
                        CurX:=Length(Lines[CurY]);
                      end;
                      SelBegin:=Point(CurX, CurY);

                      x:=CurX; y:=CurY;
                      GetNextChar(x, y);

                      SelEnd  :=Point(x, y);
                      SelText:='';
                    end;
                  end;
                end else begin
                  SelText:='';
                end;
              end;
    VK_DELETE:begin
                ProcessKey:=true;

                if ssShift in Shift then begin
                  CutToClipboard;
                end else
                if not InSelection then begin
                  if ssCtrl in Shift then begin
                    SelBegin:=Point(CurX, CurY);
                    BeginNextWord;
                    SelEnd  :=Point(CurX, CurY);
                    SelText:='';
                  end else begin
                    if S[CurX, CurY] = CR then begin
                      if CurY >= Lines.Count then begin
                        CurX:=0;
                      end else begin
                        CurX:=Length(Lines[CurY]);
                      end;
                    end;
                    SelBegin:=Point(CurX, CurY);

                    x:=CurX; y:=CurY;
                    GetNextChar(x, y);

                    SelEnd  :=Point(x, y);
                    SelText:='';
                  end;
                end else begin
                  SelText:='';
                end;
              end;
    VK_INSERT:begin
                ProcessKey:=true;

                if ssCtrl in Shift then begin
                  CopyToClipboard;
                end else
                if ssShift in Shift then begin
                  PasteFromClipboard;
                end else
                ;
              end;
    VK_W:     begin
                if ssCtrl in Shift then begin
                  SelectWordAt(CurX, CurY);
                end;
              end;
    VK_0..VK_9:
              begin
                if ssCtrl in Shift then begin
                  if ssShift in Shift then begin
                    if BookMarkList[Key - VK_0] <> CurY then begin;
                      BookMarkList[Key - VK_0]:=CurY;
                    end else begin
                      BookMarkList[Key - VK_0]:=-1;
                    end;
                  end else begin
                    if BookMarkList[Key - VK_0] <> -1 then begin
                      CurY:=BookMarkList[Key - VK_0];
                    end;
                  end;
                  MyPaint;
                end;
              end;
    VK_TAB:   begin
                ProcessKey:=true;

                if TabStops = 0 then TabStops:=2;
                n:=TabStops - (CurX - (CurX div TabStops) * TabStops);//- (CurX mod 2);//);
                t:='';
                while n > 0 do begin
                  t:=t + ' ';
                  dec(n);
                end;
                SelText:=t;
              end;
    VK_Z:     begin
                if ssCtrl in Shift then begin
                  if ssShift in Shift then begin
                    Redo;
                  end else begin
                    Undo;
                  end;
                end;
              end;
    VK_T:     begin
                if ssCtrl in Shift then begin
                  SelBegin:=Point(CurX, CurY);
                  BeginNextWord;
                  SelEnd  :=Point(CurX, CurY);
                  SelText:='';
                end;
              end;
    VK_Y:     begin
                if ssCtrl in Shift then begin
                  Lines.Delete(CurY);
                  ScrollY(CurY - TopY, ViewY, -1);
                end;
              end;
    VK_I:     begin
                if (ssCtrl in Shift) and (ssShift in Shift) and InSelection then begin
                  SaveUndo;
                  for y:=Min(SelBegin.y, SelEnd.y) to Max(SelBegin.y, SelEnd.y) - 1 do begin
                    if y < Lines.Count then begin
                      t:=Lines[y];
                      System.Insert(TabSpaces, t, 1);
                      Lines[y]:=t;
                    end;
                  end;
                  RepaintOn(Min(SelBegin.y, SelEnd.y), Max(SelBegin.y, SelEnd.y));
                end;
              end;
    VK_U:     begin
                if (ssCtrl in Shift) and (ssShift in Shift) and InSelection then begin
                  SaveUndo;
                  for y:=Min(SelBegin.y, SelEnd.y) to Max(SelBegin.y, SelEnd.y) - 1 do begin
                    if y < Lines.Count then begin
                      t:=Lines[y];
                      if TabStops = 0 then TabStops:=2;
                      System.Delete(t, 1, TabStops);
                      Lines[y]:=t;
                    end;
                  end;
                  RepaintOn(Min(SelBegin.y, SelEnd.y), Max(SelBegin.y, SelEnd.y));
                end;
              end;
  end;

  UpdateCaret;

  if ProcessKey then begin
    if ssShift in Shift then begin
      if (SelEnd.x <> CurX) or (SelEnd.y <> CurY) then begin
        SelEnd:=Point(CurX, CurY);
        RepaintOn(OldSelEnd.y,  SelEnd.y);
      end;
    end else begin
      if (OldCurX <> CurX) or (OldCurY <> CurY) then begin
        SelBegin:=Point(CurX, CurY);
        SelEnd  :=Point(CurX, CurY);
        RepaintOn(Min(OldSelBeg.y, OldSelEnd.y), Max(OldSelBeg.y, OldSelEnd.y));
      end;
    end;
  end;
end;

procedure   THMemo.RepaintOn(yb:integer; ye:integer);
begin
  PaintLines(Min(yb, ye), Max(yb, ye));
end;

procedure   THMemo.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);

  if (Key > #$1B) and (Key <> #$7f) then begin
    SelText:= Key
  end else begin
    case Key of
      CR:   begin
              SelText:= #$0D#$0A;
            end;
      #$03: begin
              CopyToClipboard;
            end;
      #$16: begin
              PasteFromClipboard;
            end;
      #$18: begin
              CutToClipboard;
            end;
      else begin
        Key:=#0;
      end;
    end
  end;
end;

procedure   THMemo.CopyToClipboard;
begin
  if InSelection then begin
    Clipboard.SetTextBuf(PChar(SelText));
  end;
end;

procedure   THMemo.CutToClipboard;
begin
  CopyToClipboard;

  SelText:='';
end;

procedure   THMemo.PasteFromClipboard;
begin
  if Clipboard.HasFormat(CF_TEXT) then begin
    SelText:=Clipboard.AsText;
  end;
end;

procedure   THMemo.ClearUndo;
var i:integer;
begin
  for i:=0 to UndoMax - 1 do begin
    UndoList[i].Text:='';
    UndoList[i].CurX:=0;
    UndoList[i].CurY:=0;
    UndoPos:=0;
  end;
end;

procedure   THMemo.SaveUndo;
begin
  UndoList[UndoPos].Text:=Lines.Text;
  UndoList[UndoPos].CurX:=CurX;
  UndoList[UndoPos].CurY:=CurY;
  UndoPos:=(UndoPos + UndoMax + 1) mod UndoMax;
  UndoList[UndoPos].Text:='';
  UndoList[UndoPos].CurX:=0;
  UndoList[UndoPos].CurY:=0;
end;

procedure   THMemo.Undo;
begin
  UndoPos:=(UndoPos + UndoMax - 1) mod UndoMax;
  if Length(UndoList[UndoPos].Text) <> 0 then begin
    Lines.Text:=UndoList[UndoPos].Text;
    CurX:=UndoList[UndoPos].CurX;
    CurY:=UndoList[UndoPos].CurY;
  end;
  MyPaint;
end;

procedure   THMemo.Redo;
begin
  if Length(UndoList[(UndoPos + UndoMax + 1) mod UndoMax].Text) <> 0 then begin
    UndoPos:=(UndoPos + UndoMax + 1) mod UndoMax;

    Lines.Text:=UndoList[UndoPos].Text;
    CurX:=UndoList[UndoPos].CurX;
    CurY:=UndoList[UndoPos].CurY;

    MyPaint;
  end;
end;

function    THMemo.GetWordAt(x:integer; y:integer):string;
begin
  Result:='';
  if not IsWordBreak(S[x, y]) then begin
    while (x >= 0) and (not IsWordBreak(S[x, y])) do dec(x);
    inc(x);
    while not IsWordBreak(S[x, y]) do begin
      Result:=Result + S[x, y];
      inc(x);
    end;
  end;
end;

procedure   THMemo.SelectWordAt(x:integer; y:integer);
begin
  if not IsWordBreak(S[x, y]) then begin
    while (x >= 0) and (not IsWordBreak(S[x, y])) do dec(x);
    inc(x);
    SelBegin:=Point(x, y);
    while not IsWordBreak(S[x, y]) do inc(x);
    SelEnd:=Point(x, y);
    MyPaint;
  end;
end;

procedure   THMemo.DblClick;
begin
  inherited DblClick;
  SelectWordAt(CurX, CurY);
end;

{
function    THMemo.GetLineColor(LineIdx:integer):TColor;
begin
  if (LineIdx < Lines.Count) and Assigned(Lines.Objects[LineIdx]) then begin
    Result:=integer(Lines.Objects[LineIdx]);
  end else begin
    Result:=0;
  end;
end;

procedure   THMemo.SetLineColor(LineIdx:integer; AColor:TColor);
begin
  if LineIdx < Lines.Count then begin
    Lines.Objects[LineIdx]:=pointer(AColor);
  end;
  MyPaint;
end;
}

function    THMemo.TabSpaces:string;
var n:integer;
begin
  if TabStops = 0 then TabStops:=2;
  n:=TabStops;
  Result:='';
  while n > 0 do begin
    Result:=Result + ' ';
    dec(n);
  end;
end;

procedure   THMemo.SetLines(ALines: TLines);
begin
  FLines.Assign(ALines);
  CurX:=0; CurY:=0;
end;

function    THMemo.GetLines:TLines;
begin
  Result:=FLines;
end;

procedure   THMemo.SetGutter(AGutter:boolean);
begin
  FGutter:=AGutter;
//  MyPaint;
end;

procedure   THMemo.SetGutterWidth(AGutterWidth:integer);
begin
  FGutterWidth:=AGutterWidth;
//  MyPaint;
end;

procedure   THMemo.ChangeStatus;
begin
  if not ((csLoading in ComponentState) or (csReading in ComponentState)) then begin
    if Assigned(OnChangeStatus) then OnChangeStatus(Self);
  end;
end;

procedure   THMemo.LinesChange(Sender: TObject);
begin
  if VSMax <> Lines.Count then begin
    VSMax:=Lines.Count;
  end;
  if Assigned(OnChange) then OnChange(Self);
  Modified:=true;
  ChangeStatus;
  Invalidate;
end;

procedure   THMemo.SetModified(AModified:boolean);
begin
  if HandleAllocated then FModified:=AModified;
end;

procedure   THMemo.SetSelStart(p:integer);
var x,y,i:integer;
begin
  x:=0; y:=0;
  for i:=0 to Lines.Count - 1 do begin
    if p <= (Length(Lines[i]) + 2) then begin
      x:=p;
      break;
    end else begin
      x:=0;
      inc(y);
      dec(p, Length(Lines[i]) + 2);
    end;
  end;
  CurX:=x; CurY:=Y;
end;

function     THMemo.GetText:string;
begin
  Result:=Lines.Text;
end;

procedure   THMemo.SetText(AText:string);
begin
  Lines.Text:=AText;
end;

function    THMemo.FindText(txt:string; MatchCase: boolean; WholeWord:boolean): boolean;
var x, y, p:integer;
    ss:string;
begin
  Result:=false;
  if Length(txt) = 0 then exit;

  if not MatchCase then txt:=ANSIUpperCase(txt);

  x:=CurX;
  y:=CurY;
  while y < Lines.Count - 1 do begin
    ss:=System.Copy(Lines[y], x + 1, Length(Lines[y]));
    if not MatchCase then ss:=ANSIUpperCase(ss);
    p:=System.Pos(txt, ss);
    if p > 0 then begin
      if WholeWord then begin
        if not (IsWordBreak(S[x + p - 2, y]) and IsWordBreak(S[x + p + Length(txt) - 1, y])) then begin
          x:=x + p - 1 + Length(txt);
          continue;
        end;
      end;

      SelBegin:=Point(x + p - 1, y);
      SelEnd  :=Point(x + p - 1 + Length(txt), y);
      CurX    :=SelEnd.X;
      CurY    :=Y;
      Result  :=true;
      MyPaint;
      break;
    end;
    x:=0;
    inc(y);
  end;
end;


procedure   THMemo.WMEraseBkgnd(var Message: TWMErasebkgnd);
begin
  Message.Result:=1;
end;

function    THMemo.GetSSIdx(Before:integer):integer;
var ss:string;
    pos, i, j:integer;
begin
  ss:=Lines.Text;
  Result:=-1;
  pos:=1;

  while Before >= 0 do begin
    if ss[pos] = CR then begin
      dec(Before);
      inc(pos);
      if ss[pos] = LF then inc(pos);
      if (Result <> -1) and StartStopKeys[Result].StopOnLineEnd then begin
        Result:=-1;
      end;
    end else begin
      if IsWordBreak(ss[pos]) then begin
        if Result = -1 then begin
          for i:=1 to High(StartStopKeys) do begin
            if ss[pos] = StartStopKeys[i].Start[1] then begin
              if Length(StartStopKeys[i].Start) = 1 then begin
                Result:=i;
                break;
              end else begin
                inc(pos);
                for j:=2 to Length(StartStopKeys[i].Start) do begin
                  if ss[pos] = StartStopKeys[i].Start[j] then begin
                    Result:=i;
                    break;
                  end;
                  inc(pos);
                end;
              end;
            end;
            if Result <> -1 then break;
          end;
        end else begin
          if ss[pos] = StartStopKeys[Result].Stop[1] then begin
            if Length(StartStopKeys[Result].Stop) = 1 then begin
              Result:=-1;
            end else begin
              inc(pos);
              for j:=2 to Length(StartStopKeys[Result].Stop) do begin
                if ss[pos] = StartStopKeys[Result].Stop[j] then begin
                  Result:=-1;
                  break;
                end;
                inc(pos);
              end;
            end;
          end;
        end;
      end;
      inc(pos);
    end;
  end;
end;

procedure   THMemo.MyPaint;
begin
  Paint;
end;

procedure   THMemo.Paint;
var r:TRect;
    yt, yb:integer;
begin
  r:=Canvas.ClipRect;
  if not EqualRect(r, Rect(0, 0, 0, 0)) then begin
    yt:=TopY + (r.Top div FontDY);
    yb:=TopY + (r.Bottom div FontDY);
    PaintLines(yt, Max(yt,yb));
  end else begin
//    PaintLines(TopY, ClientHeight div FontDY);
  end;
end;

procedure   THMemo.PaintLines(BeginLine:integer; EndLine:integer);
var SSIdx:integer;
    procedure PaintGutter(Line:integer);
    var YPos:integer;
    begin
      if Gutter then begin
        YPos:=(Line - TopY) * FontDY;
        Canvas.Brush.Color:=clBtnFace;
        Canvas.FillRect(Rect(0, YPos, GutterWidth, YPos + FontDY));
        Canvas.Pen.Color:=clWhite;
        Canvas.MoveTo(GutterWidth - 2, YPos);
        Canvas.LineTo(GutterWidth - 2, YPos + FontDY);
      end;
    end;

    procedure PaintLinesInt;
    var x,xx,y,yy,i,j:integer;
        p,B,E:integer;
        ch:char;
        w:string;
        XPos,YPos:integer;
        BrushColor,FontColor:TColor;
        FontStyle:TFontStyles;
        EOL:boolean;
        KeyWordCnt:integer;
    begin
      Canvas.Font.Assign(Font);

      KeyWordCnt:=0;

        if SSIdx <> -1 then begin
          FontColor:=StartStopKeys[SSIdx].Color;
          FontStyle:=StartStopKeys[SSIdx].Style;
        end else begin
          FontColor:=Font.Color;
          FontStyle:=Font.Style;
        end;

        YPos:=(BeginLine - TopY) * FontDY;

        for y:=BeginLine to EndLine  do begin
          Lines.SSIdx[y]:=SSIdx;

          PaintGutter(y);

          for i:=0 to 9 do begin
            if BookMarkList[i] = y then begin
              Canvas.Brush.Color:=clLime;
              Canvas.Font.Color:=clBlack;
              Canvas.TextOut(0, (y - TopY) * FontDY, IntToStr(i));

              break;
            end;
          end;

          XPos:=0; if Gutter then XPos:=XPos + GutterWidth;

          if y > Lines.Count - 1 then begin
            if Canvas.Brush.Color <> Color then Canvas.Brush.Color:=Color;
            Canvas.FillRect(Rect(XPos, YPos, ClientWidth, ClientHeight));

            for yy:=y to TopY + ViewY do begin
              PaintGutter(yy);
            end;
            break;
          end;

          for x:=0 to Length(Lines[y]) do begin

            ch :=S[x, y];
            EOL:= ch = CR;

            if (y = FColorLine) then begin
              BrushColor:=FLineColor;
            end else begin
              BrushColor:=0;
              if InSelection then begin
                p:=x + 1000 * y;
                B:=SelBegin.x + SelBegin.y * 1000;
                E:=SelEnd.x + SelEnd.y * 1000;

                if (p >= Min(B, E)) and (p < Max(B, E)) then begin
                  BrushColor:=clHighlight;
                end;
              end;
            end;

            if KeyWordCnt > 0 then begin
              dec(KeyWordCnt);
            end else begin
              if SSIdx = -1 then begin
                FontColor:=Font.Color;
                FontStyle:=Font.Style;
                if not IsWordBreak(ch) then begin
                  w:=''; xx:=x;
                  while not IsWordBreak(S[xx, y]) do begin
                    w:=w + S[xx, y]; inc(xx);
                  end;
                  if (Length(w) > 0) and (KeyWords.IndexOf(w) <> -1) then begin
                    FontColor:=KeyWordFontColor;
                    FontStyle:=KeyWordFontStyle;
                  end;
                  KeyWordCnt:=Length(w) - 1;
                end else begin
                  for i:=1 to High(StartStopKeys) do begin
                    if ch = StartStopKeys[i].Start[1] then begin
                      if Length(StartStopKeys[i].Start) = 1 then begin
                        SSIdx:=i;
                        FontColor:=StartStopKeys[SSIdx].Color;
                        FontStyle:=StartStopKeys[SSIdx].Style;
                        break;
                      end else begin
                        xx:=x + 1;
                        for j:=2 to Length(StartStopKeys[i].Start) do begin
                          if S[xx, y] = StartStopKeys[i].Start[j] then begin
                            SSIdx:=i;
                            FontColor:=StartStopKeys[SSIdx].Color;
                            FontStyle:=StartStopKeys[SSIdx].Style;
                            break;
                          end;
                          inc(xx);
                        end;
                      end;
                    end;
                    if SSIdx <> -1 then break;
                  end;
                  if SSIdx = -1 then begin
                    FontColor:=Font.Color;
                    FontStyle:=Font.Style;
                  end;
                end;
              end else begin
                if IsWordBreak(ch) then begin
                  FontStyle:=StartStopKeys[SSIdx].Style;
                  FontColor:=StartStopKeys[SSIdx].Color;

                  if ch = StartStopKeys[SSIdx].Stop[1] then begin
                    if Length(StartStopKeys[SSIdx].Stop) = 1 then begin
                      SSIdx:=-1;
                    end else begin
                      xx:=x + 1;
                      for j:=2 to Length(StartStopKeys[SSIdx].Stop) do begin
                        if S[xx, y] = StartStopKeys[SSIdx].Stop[j] then begin
                          SSIdx:=-1;
                          break;
                        end;
                        inc(xx);
                      end;
                    end;
                  end;
                end;
              end;
            end;

            if BrushColor = 0 then BrushColor:=Color;
            if Canvas.Brush.Color <> BrushColor then Canvas.Brush.Color:=BrushColor;

            if (x >= LeftX) and (x <= LeftX + ViewX) then begin
              Canvas.FillRect(Rect(XPos, YPos, XPos + FontDX, YPos + FontDY));

              if (not EOL) and (ch <> ' ') and (ch <> CR) then begin
                if Canvas.Font.Style <> FontStyle then Canvas.Font.Style:=FontStyle;
                if Canvas.Font.Color <> FontColor then Canvas.Font.Color:=FontColor;
                Canvas.TextOut(XPos, YPos, ch);
              end;
              inc(XPos, FontDX);
            end;
          end;

          if Length(Lines[y]) < (LeftX + ViewX) then begin
            Canvas.FillRect(Rect(XPos, YPos, ClientWidth, YPos + FontDY));
          end;

          inc(YPos, FontDY);
        end;
//      RestoreDC(Canvas.Handle, SavedDC);
    end;
begin
  if not Assigned(Lines) then exit;
  
  BeginLine :=Max(BeginLine, TopY);
  EndLine   :=Min(EndLine, TopY + ViewY);

  if Assigned(Msg) then Msg(Format('%d: %d - %d', [ViewY, BeginLine - TopY, EndLine - TopY]));

  HideCaret;

  if (BeginLine > 0) and (BeginLine < Lines.Count) then begin
    SSIdx:=Lines.SSIdx[BeginLine];
  end else begin
    SSIdx:=-1;
  end;

  PaintLinesInt;
  inc(EndLine);
  BeginLine:=EndLine;

  while (EndLine < Lines.Count) and (Lines.SSIdx[EndLine] <> SSIdx) do begin
    if EndLine <= (TopY + ViewY - 1) then begin
      PaintLinesInt;
    end else begin
      Lines.SSIdx[EndLine]:=SSIdx;
      SSIdx:=Lines.CulcSSIdxAfter(EndLine);
    end;
    inc(EndLine);
    BeginLine:=EndLine;
  end;

  if Focused then begin
    if Gutter then begin
      SetCaretPos((CurX - LeftX)* FontDX + GutterWidth, (CurY - TopY) * FontDY);
    end else begin
      SetCaretPos((CurX - LeftX)* FontDX, (CurY - TopY) * FontDY);
    end;
  end;
  
  ShowCaret;
end;

procedure   THMemo.SetColorLine(ALine:integer; AColor:TColor);
var oldline:integer;
begin
  try
  if FColorLine <> ALine then begin
    oldline:=FColorLine;
    FColorLine:=-1;
    FLineColor:=-1;
    PaintLines(oldline, oldline);
    FColorLine:=ALine;
    FLineColor:=AColor;
    PaintLines(FColorLine, FColorLine);
  end else begin
    if FLineColor <> AColor then begin
      FLineColor:=AColor;
      PaintLines(FColorLine, FColorLine);
    end;
  end;
  except
  end;
end;

procedure   THMemo.SetScrollBars(AScrollBars:TScrollStyle);
begin
  if FScrollBars <> AScrollBars then begin
    FScrollBars:=AScrollBars;

    ShowScrollBar(Handle,	SB_BOTH, false);
    case FScrollBars of
      ssNone:       ShowScrollBar(Handle,	SB_BOTH, false);
      ssBoth:       ShowScrollBar(Handle,	SB_BOTH, true);
      ssVertical:   ShowScrollBar(Handle,	SB_VERT, true);
      ssHorizontal: ShowScrollBar(Handle,	SB_HORZ, true);
    end;
  end;
end;

end.






