unit ARButtons;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, ComCtrls, Graphics, Classes, Menus, Dialogs, Forms, ExtCtrls,
  Controls, Buttons, StdCtrls, SysUtils, Math,

  ARUtils, ARRepCommon;

const
  DropDownWidth = 14;

type
  TARGlyphButton = class(TSpeedButton)
  private
    FCaption    :string;

    procedure   SetCaption(ACaption:string);
  protected
    procedure   Paint; override;
    procedure   UpdateWidth;virtual;
  public
    Image       :TBitmap;
    constructor Create(AOwner: TComponent); override;
  published
    property    Caption:string  read  FCaption  write SetCaption;
  end;

//  TARButton
  TARButton = class(TARGlyphButton)
  private
		FPopupForm      :TForm;
		FOnChange       :TNotifyEvent;
		FChecked        :boolean;
    FLoaded         :boolean;
		FPopupFormHiding:boolean;

		procedure   PopupDeactivate(Sender: TObject);
  	procedure   PopupKeyPress(Sender:TObject; var Key: Char);
		procedure   SetChecked(AChecked:boolean);
  protected
    procedure   CloseUp(Action:boolean);
    procedure   RecreatePopupForm;virtual;
    procedure   DoClick;virtual;
    procedure   DoChange;virtual;
    procedure   Loaded;override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure   Paint;override;
    procedure   UpdateWidth;override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   Popup(Sender:TObject);
  published
		property    OnChange:TNotifyEvent read FOnChange    write FOnChange;
		property    Checked:boolean       read FChecked     write SetChecked;
  end;

// TARColorButton
  TARColorButton = class(TARButton)
  private
    FColor      :TColor;
    FDocColors  :TAIList;
    FDocBtns    :TAIList;

    FStdBtns    :TAIList;
    FStdPanel   :TPanel;
    FDocPanel   :TPanel;
    FMorePanel  :TPanel;
		FOnGetDocColors:TNotifyEvent;

    procedure   SetColor(AColor:TColor);
  protected
    procedure   Paint; override;
    procedure   RecreatePopupForm;override;
		procedure   ColorBtnClick(Sender: TObject);
		procedure   MoreBtnClick(Sender: TObject);
    procedure   DoClick;override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    procedure   ClearDocColor;
    procedure   AddDocColor(AColor:TColor);
  published
    property    Color:TColor                read FColor           write SetColor;
		property    OnGetDocColors:TNotifyEvent read FOnGetDocColors  write FOnGetDocColors;
  end;

// TARGridButton
  TARGridButton = class(TARColorButton)
  private
    FGridPanel  :TPanel;
    FGridKind   :TFrameStyle;
    FGridWidth  :integer;
    FShowColors :boolean;

    FGridBtns   :TAIList;

    procedure   SetGridWidth(AGridWidth:integer);
    procedure   SetGridKind(AGridKind:TFrameStyle);
    procedure   SetShowColors(AShowColors:boolean);
  protected
    procedure   Paint; override;
    procedure   RecreatePopupForm;override;
		procedure   GridBtnClick(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property    GridWidth:integer     read  FGridWidth  write SetGridWidth;
    property    GridKind:TFrameStyle  read  FGridKind   write SetGridKind;
    property    ShowColors:boolean    read  FShowColors write SetShowColors;
  end;

// TARComboButton
  TARComboButton = class(TARButton)
  private
    FListBox    :TListBox;
    FMainPanel  :TPanel;

    function    GetItems:TStrings;
    procedure   SetItems(AItems:TStrings);
    function    GetValue:string;virtual;
    procedure   SetValue(AValue:string);virtual;
    function    GetItemIndex:integer;
    procedure   SetItemIndex(AItemIndex:integer);
    function    GetPopupWidth:integer;
    procedure   SetPopupWidth(APopupWidth:integer);
  protected
    procedure   Paint; override;
    procedure   RecreatePopupForm;override;
    procedure   LBClick(Sender:TObject);virtual;
    procedure   SelfClick(Sender:TObject);virtual;
    procedure   UpdateWidth;override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

  published
    property    Items:TStrings      read  GetItems      write SetItems;
    property    Value:string        read  GetValue      write SetValue;
    property    ItemIndex:integer   read  GetItemIndex  write SetItemIndex;
    property    PopupWidth:integer  read  GetPopupWidth write SetPopupWidth default 100;
  end;

// TARFontButton
  TARFontButton = class(TARComboButton)
  private
  protected
    procedure   Paint; override;
    procedure   LBDrawItem(Control:TWinControl; Index:Integer; Rect:TRect; State:TOwnerDrawState);
    procedure   UpdateWidth;override;
  public
    constructor Create(AOwner: TComponent); override;
  published
  end;

// TARComboMoreButton
  TARComboMoreButton = class(TARComboButton)
  private
    FOnMoreBtnClick:TNotifyEvent;
    FMoreBtn    :TSpeedButton;
  protected
		procedure   MoreBtnClick(Sender: TObject);
    procedure   RecreatePopupForm;override;
    procedure   UpdateWidth;override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property    OnMoreBtnClick:TNotifyEvent  read FOnMoreBtnClick write FOnMoreBtnClick;
  end;

// TARComboEditButton
  TARComboEditButton = class(TARComboButton)
  private
		FEditForm   :TForm;
    FEdit       :TEdit;
    FValue      :string;

		FEditFormHiding:boolean;
    function    GetValue:string;override;
    procedure   SetValue(AValue:string);override;
  protected
    procedure   CloseUpEdit(Action:boolean);
    procedure   RecreateEditForm;
    procedure   SelfClick(Sender:TObject);override;
  	procedure   EditFormDeactivate(Sender:TObject);
  	procedure   EditFormKeyPress(Sender:TObject; var Key: Char);
    procedure   PopupEditForm(Sender:TObject);
    procedure   UpdateWidth;override;
  public
    constructor Create(AOwner: TComponent); override;
  published
  end;

procedure   Register;

implementation

procedure   Register;
begin
{
  RegisterComponents('ARComponents', [TARGlyphButton]);
  RegisterComponents('ARComponents', [TARColorButton]);
  RegisterComponents('ARComponents', [TARGridButton]);
  RegisterComponents('ARComponents', [TARComboButton]);
  RegisterComponents('ARComponents', [TARFontButton]);
  RegisterComponents('ARComponents', [TARComboMoreButton]);
  RegisterComponents('ARComponents', [TARComboEditButton]);
}
end;

var
  BtnGlyph:TBitmap;

constructor TARGlyphButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  Width:=23;
  Height:=22;
  Constraints.MinHeight :=22;
  Constraints.MinWidth  :=23;
  Flat:=true;
  Transparent:=false;
end;

procedure   TARGlyphButton.UpdateWidth;
var w:integer;
begin
  if Assigned(Image) and (not Image.Empty) then begin
    w:=Max(23, Image.Width + 0);
  end else begin
    w:=Max(23, Canvas.TextWidth(FCaption) + 7);
  end;
  if w <> Width then Width:=w;
end;

procedure   TARGlyphButton.SetCaption(ACaption:string);
begin
  FCaption:=ACaption;
  UpdateWidth;
end;

procedure   TARGlyphButton.Paint;
var w, h:integer;
begin
  inherited;

  if Assigned(Image) and (not Image.Empty) then begin
    w:=(ClientHeight - Image.Height) div 2 + 1;
    Canvas.Draw(w, w, Image);
  end else begin
    if Length(Caption) > 0 then begin
      h:=Canvas.TextHeight(Caption);
      Canvas.TextOut((ClientHeight - h) div 2 - 1, (ClientHeight - h) div 2 - 1, Caption);
    end;
  end;
end;


//////////////////////////////////////////////////////////////////////
//  TARButton
constructor TARButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

	FPopupForm             	  :=TForm.Create(Self);
	FPopupForm.FormStyle   	  :=fsStayOnTop;
	FPopupForm.BorderIcons 	  :=[];
	FPopupForm.BorderStyle 	  :=bsNone;
	FPopupForm.Color       	  :=clRed;
	FPopupForm.OnDeactivate	  :=PopupDeactivate;
	FPopupForm.KeyPreview  	  :=true;
	FPopupForm.OnKeyPress     :=PopupKeyPress;
end;

destructor  TARButton.Destroy;
begin
  inherited Destroy;
end;

procedure   TARButton.UpdateWidth;
begin
  inherited;
  Width:=Width + DropDownWidth;
end;

procedure   TARButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if X < (ClientWidth - DropDownWidth) then begin
    inherited;
  end else begin
    Popup(Self);
  end;
end;

procedure   TARButton.SetChecked(AChecked:boolean);
begin
  FChecked:=AChecked;
  Invalidate;
end;

procedure   TARButton.Loaded;
begin
  inherited;
  FLoaded:=true;
end;

procedure   TARButton.DoClick;
begin
  if Assigned(OnClick) then OnClick(Self);
end;

procedure   TARButton.DoChange;
begin
  if ComponentState * [csLoading, csUpdating] = [] then begin
    if Assigned(OnChange) then OnChange(Self);
  end;
end;

procedure   TARButton.PopupDeactivate(Sender: TObject);
begin
	if not FPopupFormHiding then begin
		CloseUp(false);
	end;
end;

procedure   TARButton.CloseUp(Action:boolean);
begin
	FPopupFormHiding:=true;
	FPopupForm.Hide;
	FPopupFormHiding:=false;

  Down:=false;

	if Action then begin
		DoClick;
	end;
end;

procedure   TARButton.PopupKeyPress(Sender:TObject; var Key: Char);
begin
  if Key = #$1b then begin
    CloseUp(false);
    Key:=#$0;
  end;
  inherited;
end;

procedure   TARButton.RecreatePopupForm;
begin
end;

procedure   TARButton.Popup(Sender:Tobject);
var P:TPoint;
		Y:integer;
begin
	RecreatePopupForm;

	P := Parent.ClientToScreen(Point(0, 0));
	P.X:=P.X + Left;
	Y := P.Y + Top + Height;
	if Y + FPopupForm.Height > Screen.Height then Y := P.Y - FPopupForm.Height;
	FPopupForm.SetBounds(P.X, Y, FPopupForm.Width, FPopupForm.Height);

	FPopupForm.Show;
	FPopupForm.SetFocus;
end;

procedure   TARButton.Paint;
var r:TRect;
begin
  inherited;
  if not BtnGlyph.Empty then begin
    Canvas.Draw(ClientWidth - DropDownWidth + 1, (ClientHeight - BtnGlyph.Height) div 2 - 1, BtnGlyph);
  end;

  if MouseInControl then begin
    r:=Rect(0, 0, ClientWidth - DropDownWidth - 0, ClientHeight);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RIGHT);

    r:=Rect(ClientWidth - DropDownWidth, 0, ClientWidth, ClientHeight);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
  end;
end;
//////////////////////////////////////////////////////////////////////
type
  TColorBtn = class(TSpeedButton)
  protected
    procedure Paint;override;
  public
    Color       :TColor;
  end;

procedure     TColorBtn.Paint;
begin
  inherited;
  Canvas.Pen.Color   :=Color and TColor($ff7f7f7f);
  Canvas.Brush.Color :=Color;
  Canvas.Rectangle(Rect((Width - 13) div 2, (Height - 13) div 2, (Width - 13) div 2 + 13, (Height - 13) div 2 + 13));
end;

//////////////////////////////////////////////////////////////////////
var
  StdColors:array[0..23] of TColor = (
    clBlack,    clWhite,    clMaroon,   clGreen,    clOlive,    clNavy,     clPurple,   clTeal,
    clGray,     clSilver,   clRed,      clLime,     clYellow,   clBlue,     clFuchsia,  clAqua,
    $00404040,  $00E0E0E0,  $00C0C0FF,  $00C0FFC0,  $00C0FFFF,  $00FFC0C0,  $00FFC0FF,  $00FFFFC0
  );

// TARColorButton
constructor TARColorButton.Create(AOwner: TComponent);
var i:integer;
    sb:TColorBtn;
    sb2:TSpeedButton;
    lb:TLabel;
begin
  inherited Create(AOwner);

  FDocColors  :=TAIList.Create('TARColorButton.FDocColors');
  FDocBtns    :=TAIList.Create('TARColorButton.FDocBtns');
  FStdBtns    :=TAIList.Create('TARColorButton.FStdBtns');

  FStdPanel       :=TPanel.Create(Self);
  FStdPanel.Parent:=FPopupForm;
  FStdPanel.Align :=alTop;
  FStdPanel.Height:=89;

  lb:=TLabel.Create(Self);
  lb.Parent :=FStdPanel;
  lb.Caption:='Standard Colors';
  lb.Left   :=8;
  lb.Top    :=8;

  for i:=0 to High(StdColors) do begin
    sb:=TColorBtn.Create(Self);
    sb.Parent:=FStdPanel;
    sb.SetBounds(8 + (i mod 8) * 19, 24 + (i div 8) * 19, 19, 19);
    sb.Color:=StdColors[i];
    sb.Tag:=StdColors[i];

    sb.OnClick:=ColorBtnClick;
    sb.Flat:=true;
    sb.GroupIndex:=3;
    sb.AllowAllUp:=true;
    FStdBtns.Add(sb);
  end;

  FDocPanel       :=TPanel.Create(Self);
  FDocPanel.Parent:=FPopupForm;
  FDocPanel.Align :=alTop;

  lb:=TLabel.Create(Self);
  lb.Parent :=FDocPanel;
  lb.Caption:='Document''s Colors';
  lb.Left   :=8;
  lb.Top    :=8;

  FDocPanel.Height:=23;

  FMorePanel       :=TPanel.Create(Self);
  FMorePanel.Parent:=FPopupForm;
  FMorePanel.Align :=alClient;

  sb2:=TSpeedButton.Create(Self);
  sb2.Parent:=FMorePanel;
  sb2.SetBounds(8, 6, 153, 22);
  sb2.Caption:='More Colors...';
  sb2.Transparent:=false;
  sb2.OnClick:=MoreBtnClick;
  sb2.Flat:=true;

  FMorePanel.Height:=36;

  Width:=37;
end;

procedure   TARColorButton.AddDocColor(AColor:TColor);
var sb:TColorBtn;
    i:integer;
begin
  if (AColor <> clBlack) and (AColor <> clWhite) then begin
    if FDocColors.IndexOf(pointer(AColor)) = -1 then begin
      FDocColors.Add(pointer(AColor));
      sb:=TColorBtn.Create(Self);
      FDocBtns.Add(sb);
      i:=FDocBtns.Count - 1;
      sb.Parent:=FDocPanel;
      sb.SetBounds(8 + (i mod 8) * 19, 24 + (i div 8) * 19, 19, 19);
      sb.Tag:=AColor;
      sb.Color:=AColor;
      sb.OnClick:=ColorBtnClick;
      sb.Flat:=true;
    end;
  end;
end;

destructor  TARColorButton.Destroy;
begin
  FDocColors.Free;
  FDocBtns.Free;

  inherited Destroy;
end;

procedure   TARColorButton.ColorBtnClick(Sender: TObject);
begin
  Color:=TColorBtn(Sender).Color;
  CloseUp(false);
  DoChange;
end;

procedure   TARColorButton.SetColor(AColor:TColor);
begin
  FColor:=AColor;
  Invalidate;
end;

procedure   TARColorButton.Paint;
begin
  inherited;

  if Color <> clNone then begin
    Canvas.Brush.Color:=Color;
    Canvas.Pen.Color  :=Color;
  end else begin
    Canvas.Brush.Color:=clBtnFace;
    Canvas.Pen.Color  :=clBlack;
  end;
  Canvas.Rectangle(2, Height - 6, Height - 2, Height - 2);
end;

procedure   TARColorButton.ClearDocColor;
var i:integer;
begin
  FDocColors.Clear;

  for i:=0 to FDocBtns.Count - 1 do begin
    TColorBtn(FDocBtns[i]).Free;
  end;
  FDocBtns.Clear();
end;

procedure   TARColorButton.RecreatePopupForm;
var i:integer;
begin
  inherited;
	if Assigned(OnGetDocColors) then OnGetDocColors(Self);

  FPopupForm.Width:=170;

  FDocPanel.Visible :=FDocColors.Count > 0;
  FDocPanel.Height  :=(FDocColors.Count div 8 + 1) * 19 + 33;

  FPopupForm.Height:= FStdPanel.Height +
                      FDocPanel.Height * integer(FDocPanel.Visible) +
                      FMorePanel.Height;

  for i:=0 to FStdBtns.Count - 1 do begin
    if TColorBtn(FStdBtns[i]).Color = Color then begin
      TColorBtn(FStdBtns[i]).Down:=true;
      break;
    end;
  end;

  for i:=0 to FDocBtns.Count - 1 do begin
    if TColorBtn(FDocBtns[i]).Color = Color then begin
      TColorBtn(FDocBtns[i]).Down:=true;
      break;
    end;
  end;
end;

procedure   TARColorButton.MoreBtnClick(Sender: TObject);
var cd:TColorDialog;
begin
  CloseUp(false);

  cd:=TColorDialog.Create(Self);
  cd.Options:=[cdFullOpen];
  cd.Color:=Color;
  if cd.Execute then begin
    Color:=cd.Color;
    DoChange;
  end;
end;

procedure   TARColorButton.DoClick;
begin
  if Color <> clNone then begin
    inherited;
  end;
end;
////////////////////////////////////////////////////////////////////////////////////////////////
// TARGridButton
procedure   TARGridButton.SetGridWidth(AGridWidth:integer);
begin
  FGridWidth:=AGridWidth;
  Invalidate;
end;

procedure   TARGridButton.SetGridKind(AGridKind:TFrameStyle);
begin
  FGridKind:=AGridKind;
  Invalidate;
end;

type
  TGridBtn = class(TSpeedButton)
  protected
    procedure Paint;override;
  public
    GridWidth :integer;
  end;

procedure     TGridBtn.Paint;
var x, y:integer;
begin
  inherited;
  x:=(Width - 70) div 2;
  y:=(Height - 11) div 2;
  Canvas.TextOut(x + 2, y + (11 - Canvas.TextHeight('0')) div 2, IntToStr(GridWidth));
  if GridWidth > 0 then begin
    Canvas.Brush.Color:=clBlack;
    Canvas.FillRect(Rect(x + Canvas.TextWidth('0') + 5, (Height - GridWidth) div 2, Width - 5, (Height - GridWidth) div 2 + GridWidth));
  end;
end;

constructor TARGridButton.Create(AOwner: TComponent);
var i:integer;
    sb:TGridBtn;
    lb:TLabel;
begin
  inherited;

  FGridBtns:=TAIList.Create('TARGridButton.FGridBtns:');

  FGridPanel       :=TPanel.Create(Self);
  FGridPanel.Parent:=FPopupForm;
  FGridPanel.Align :=alTop;
  FGridPanel.Height:=109;

  lb:=TLabel.Create(Self);
  lb.Parent :=FGridPanel;
  lb.Caption:='Width (px)';
  lb.Left   :=8;
  lb.Top    :=8;

  for i:=0 to 7 do begin
    sb:=TGridBtn.Create(Self);
    sb.Parent:=FGridPanel;
    sb.SetBounds(8 + (i mod 2) * 76, 24 + (i div 2) * 19, 76, 19);
    sb.GridWidth:=i;
    sb.Tag:=i;

    sb.OnClick:=GridBtnClick;
    sb.Flat:=true;
    sb.GroupIndex:=3;
    sb.AllowAllUp:=true;

    FGridBtns.Add(sb);
  end;
end;

destructor  TARGridButton.Destroy;
begin
  FGridBtns.Free;

  inherited;
end;

procedure   TARGridButton.GridBtnClick(Sender: TObject);
begin
  GridWidth:=TGridBtn(Sender).GridWidth;
  CloseUp(false);
  DoChange;
end;

procedure   TARGridButton.RecreatePopupForm;
begin
  inherited;
  FPopupForm.Height:= FStdPanel.Height  * integer(FStdPanel.Visible) +
                      FDocPanel.Height  * integer(FDocPanel.Visible) +
                      FMorePanel.Height * integer(FMorePanel.Visible) +
                      FGridPanel.Height;

  if (GridWidth >= 0) and (GridWidth < FGridBtns.Count) then begin
    TGridBtn(FGridBtns[GridWidth]).Down:=true;
  end;
end;

procedure   TARGridButton.SetShowColors(AShowColors:boolean);
begin
  FShowColors:=AShowColors;

  FStdPanel.Visible :=FShowColors;
  FDocPanel.Visible :=FShowColors;
  FMorePanel.Visible:=FShowColors;
  RecreatePopupForm;
end;

procedure   TARGridButton.Paint;
var R: TRect;
begin
  inherited;

  if not Checked then begin
    Canvas.Brush.Color:=clBtnFace;
    Canvas.FillRect(Rect(1, 1, Width - DropDownWidth - 2, Height - 1));
  end else begin
    Canvas.Brush.Color:=$00F0F0F0;
    Canvas.FillRect(Rect(1, 1, Width - DropDownWidth - 2, Height - 1));

    R := Rect(-1, -1, Width - DropDownWidth + 0, Height);
    DrawEdge(Canvas.Handle, R, BDR_SUNKEN, BF_TOP or BF_RIGHT or BF_BOTTOM);
    DrawEdge(Canvas.Handle, R, EDGE_ETCHED, BF_LEFT);
  end;

  if Assigned(Image) then begin
    Canvas.Draw((Height - Image.Height) div 2 - 1, (Height - Image.Height) div 2 - 1, Image);
  end;

  if GridWidth > 0 then begin
    Canvas.Brush.Color:=Color;
    if frLeft   in GridKind then Canvas.FillRect(Rect(4,                 4,                  4+GridWidth, Height-5));
    if frRight  in GridKind then Canvas.FillRect(Rect(Height-5-GridWidth,4,                  Height-5,    Height-5));
    if frTop    in GridKind then Canvas.FillRect(Rect(4,                 4,                  Height-5,    4+GridWidth));
    if frBottom in GridKind then Canvas.FillRect(Rect(4,                 Height-5-GridWidth, Height-5,    Height-5));
  end else begin
    Canvas.Brush.Color:=clBtnFace;
    Canvas.Pen.Color  :=clBlack;
    if [frLeft, frRight, frTop, frBottom] = GridKind then begin
      Canvas.Rectangle(Rect(4,    4,    Height-4,   Height-4));
      Canvas.Rectangle(Rect(4+3,  4+3,  Height-4-3, Height-4-3));
    end else begin
      if frLeft   in GridKind then Canvas.Rectangle(Rect(4,          4,          4+4,      Height-5));
      if frRight  in GridKind then Canvas.Rectangle(Rect(Height-5-4, 4,          Height-5, Height-5));
      if frTop    in GridKind then Canvas.Rectangle(Rect(4,          4,          Height-5, 4+4));
      if frBottom in GridKind then Canvas.Rectangle(Rect(4,          Height-5-4, Height-5, Height-5));
    end;
  end;
end;

// TARComboButton
constructor TARComboButton.Create(AOwner: TComponent);
begin
  inherited;

  Self.OnClick:=SelfClick;

  FMainPanel:=TPanel.Create(Self);
  FMainPanel.Parent :=FPopupForm;
  FMainPanel.Align  :=alClient;
  FMainPanel.BevelInner:=bvLowered;//bvRaised;
  FMainPanel.BevelOuter:=bvRaised;

  FListBox:=TListBox.Create(Self);
  FListBox.Parent     :=FMainPanel;
  FListBox.BorderStyle:=bsNone;
  FListBox.Align      :=alClient;
  FListBox.OnClick    :=LBClick;

  Constraints.MinWidth:=50;
end;

procedure   TARComboButton.UpdateWidth;
begin
  inherited;
  Width:=50;
end;

destructor  TARComboButton.Destroy;
begin
  FListBox.Free;
  FMainPanel.Free;

  inherited;
end;

procedure   TARComboButton.SelfClick(Sender:TObject);
begin
  Popup(Self);
end;

function    TARComboButton.GetItems:TStrings;
begin
  Result:=FListBox.Items;
end;

procedure   TARComboButton.SetItems(AItems:TStrings);
begin
  FListBox.Items.Assign(AItems);
end;

function    TARComboButton.GetValue:string;
begin
  if FListBox.ItemIndex >= 0 then begin
    Result:=FListBox.Items[FListBox.ItemIndex];
  end else begin
    Result:='';
  end;
end;

procedure   TARComboButton.SetValue(AValue:string);
begin
  FListBox.ItemIndex:=FListBox.Items.IndexOf(AValue);
  Invalidate;
end;

function    TARComboButton.GetItemIndex:integer;
begin
  Result:=FListBox.ItemIndex;
end;

procedure   TARComboButton.SetItemIndex(AItemIndex:integer);
begin
  FListBox.ItemIndex:=AItemIndex;
end;

function    TARComboButton.GetPopupWidth:integer;
begin
  Result:=FPopupForm.Width;
end;

procedure   TARComboButton.SetPopupWidth(APopupWidth:integer);
begin
  FPopupForm.Width:=APopupWidth;
end;

procedure   TARComboButton.RecreatePopupForm;
begin
  inherited;

  FPopupForm.Width:=Width;
end;

procedure   TARComboButton.LBClick(Sender:TObject);
begin
  CloseUp(false);
  Value:=FListBox.Items[FListBox.ItemIndex];
  DoChange;
  Invalidate;
end;

procedure   TARComboButton.Paint;
begin
  inherited;

  Canvas.Brush.Color:=clWhite;
  Canvas.FillRect(Rect(2, 2, Width - DropDownWidth - 4, Height - 3));

  Canvas.TextRect(Rect(2, 2, Width - DropDownWidth - 4, Height - 3),
                  4,
                  (Height - 3 - Canvas.TextHeight(Value)) div 2,
                  Value);
end;

// TARFontButton
constructor TARFontButton.Create(AOwner: TComponent);
begin
  inherited;

  FListBox.Style      :=lbOwnerDrawFixed;
  FListBox.BorderStyle:=bsNone;
  FListBox.ItemIndex  :=0;
  FListBox.OnDrawItem :=LBDrawItem;
  FListBox.Items.Assign(Screen.Fonts);
  Value:='Arial';

  Constraints.MinWidth:=180;
end;

procedure   TARFontButton.UpdateWidth;
begin
  inherited;
  Width:=180;
end;

procedure   TARFontButton.LBDrawItem(Control:TWinControl; Index:Integer; Rect:TRect; State:TOwnerDrawState);
begin
  if Index <> -1 then begin
    TListBox(Control).Canvas.Font.Size:=10;
    TListBox(Control).Canvas.Font.Name:=FListBox.Items[Index];
    TListBox(Control).Canvas.TextRect(Rect, Rect.Left + 5, Rect.Top, FListBox.Items[Index]);
  end;
end;

procedure   TARFontButton.Paint;
begin
  inherited;

  Canvas.Font.Name:=Value;
  Canvas.Font.Size:=10;
  Canvas.TextRect(Rect(2, 2, Width - DropDownWidth - 4, Height - 3),
                  4,
                  (Height - 3 - Canvas.TextHeight(Value)) div 2,
                  Value);
end;

// TARComboMoreButton
constructor TARComboMoreButton.Create(AOwner: TComponent);
begin
  inherited;

  FMoreBtn       :=TSpeedButton.Create(Self);
  FMoreBtn.Parent:=FMainPanel;
  FMoreBtn.Caption:='Define...';
  FMoreBtn.Transparent:=false;
  FMoreBtn.OnClick:=MoreBtnClick;
  FMoreBtn.Flat:=true;

  Constraints.MinWidth:=180;
end;

procedure   TARComboMoreButton.UpdateWidth;
begin
  inherited;
  Width:=180;
end;

procedure   TARComboMoreButton.RecreatePopupForm;
begin
  inherited;

  FListBox.Align :=alTop;
  FListBox.Height:=FMainPanel.Height - 30;

  FMoreBtn.SetBounds(3, FMainPanel.Height - 26, FMainPanel.Width - 6, 22);
end;

procedure   TARComboMoreButton.MoreBtnClick(Sender: TObject);
begin
  CloseUp(false);
  if Assigned(FOnMoreBtnClick) then FOnMoreBtnClick(Self);
end;

// TARComboEditButton
constructor TARComboEditButton.Create(AOwner: TComponent);
begin
  inherited;

	FEditForm              :=TForm.Create(Self);
	FEditForm.FormStyle    :=fsStayOnTop;
	FEditForm.BorderIcons  :=[];
	FEditForm.BorderStyle  :=bsNone;
	FEditForm.Color        :=clGreen;
	FEditForm.OnDeactivate :=EditFormDeactivate;
	FEditForm.KeyPreview   :=true;
	FEditForm.OnKeyPress   :=EditFormKeyPress;

  FEdit:=TEdit.Create(Self);
  FEdit.Parent      :=FEditForm;
  FEdit.BorderStyle :=bsNone;

  Constraints.MinWidth:=50;
end;

procedure   TARComboEditButton.UpdateWidth;
begin
  inherited;
  Width:=50;
end;

procedure   TARComboEditButton.EditFormDeactivate(Sender: TObject);
begin
	if not FEditFormHiding then begin
		CloseUpEdit(true);
	end;
end;

procedure   TARComboEditButton.SelfClick(Sender:TObject);
begin
  PopupEditForm(Self);
end;

procedure   TARComboEditButton.CloseUpEdit(Action:boolean);
begin
	FEditFormHiding:=true;
	FEditForm.Hide;
	FEditFormHiding:=false;

  Down:=false;

	if Action then begin
    Value:=FEdit.Text;
		DoChange;
	end;
end;

procedure   TARComboEditButton.EditFormKeyPress(Sender:TObject; var Key: Char);
begin
  if Key = #$1b then begin
    CloseUpEdit(false);
    Key:=#$0;
  end else
  if Key = #$0d then begin
    CloseUpEdit(true);
    Key:=#$0;
  end;

  inherited;
end;

procedure   TARComboEditButton.RecreateEditForm;
begin
  inherited;
end;

procedure   TARComboEditButton.PopupEditForm(Sender:TObject);
var P:TPoint;
begin
	P:=ClientToScreen(Point(0, 0));
	FEditForm.SetBounds(P.X + 2, P.Y + 2, Width - DropDownWidth - 4, Height - 4);

  FEdit.SetBounds(0, 0, FEditForm.Width, FEditForm.Height);
  FEdit.Text:=Value;
  
	FEditForm.Show;
	FEditForm.SetFocus;
end;

function    TARComboEditButton.GetValue:string;
begin
  Result:=FValue;
end;

procedure   TARComboEditButton.SetValue(AValue:string);
begin
  inherited SetValue(AValue);
  FValue:=AValue;
end;

initialization
  BtnGlyph                 :=TBitmap.Create;
  BtnGlyph.Handle          :=LoadBitmap(0, PChar(OBM_COMBO));
  BtnGlyph.Transparent     :=true;
  BtnGlyph.TransparentMode :=tmAuto;
finalization
  BtnGlyph.Free;
end.
