unit ARCellProp;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ExtCtrls, StdCtrls, Buttons,

  AREditRep, ExtDlgs;

type
  TARCellPropDlg = class(TForm)
    Image: TGroupBox;
    Path: TEdit;
    LinkToFile: TCheckBox;
    Browse: TBitBtn;
    OKBtn: TBitBtn;
    CancelBtn: TBitBtn;
    ApplyBtn: TBitBtn;
    TextOrientation: TRadioGroup;
    OpenPictureDialog: TOpenPictureDialog;
    CellTextRG: TGroupBox;
    CellText: TMemo;
    procedure FormShow(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure ApplyBtnClick(Sender: TObject);
    procedure BrowseClick(Sender: TObject);
    procedure SomeChange(Sender: TObject);
    procedure TextOrientationChanging(Sender: TObject; var AllowChange: Boolean);
    procedure TextOrientationRGClick(Sender: TObject);
    procedure PathChange(Sender: TObject);
    procedure LinkToFileClick(Sender: TObject);
    procedure CellTextChange(Sender: TObject);
  private
    { Private declarations }
  public
    rep       :TAREditRep;
    Modified  :boolean;
  end;

var
  ARCellPropDlg: TARCellPropDlg;

implementation

{$R *.DFM}

procedure TARCellPropDlg.CancelBtnClick(Sender: TObject);
begin
  ModalResult:=mrCancel;
end;

procedure TARCellPropDlg.OKBtnClick(Sender: TObject);
begin
  ApplyBtnClick(Sender);
  ModalResult:=mrOK;
end;

procedure TARCellPropDlg.BrowseClick(Sender: TObject);
begin
  OpenPictureDialog.FileName:=Path.Text;
  if OpenPictureDialog.Execute() then begin
    Path.Text:=OpenPictureDialog.FileName;
  end;
end;

procedure TARCellPropDlg.SomeChange(Sender: TObject);
begin
  Modified:=true;
  ApplyBtn.Enabled:=true;
end;

procedure TARCellPropDlg.TextOrientationChanging(Sender: TObject; var AllowChange: Boolean);
begin
  AllowChange:=true;

  SomeChange(Sender);
end;

procedure TARCellPropDlg.TextOrientationRGClick(Sender: TObject);
begin
  SomeChange(Sender);
end;

procedure TARCellPropDlg.PathChange(Sender: TObject);
begin
  Path.Color:=clWhite;
  SomeChange(Sender);
end;

procedure TARCellPropDlg.LinkToFileClick(Sender: TObject);
begin
  LinkToFile.AllowGrayed:=false;
  SomeChange(Sender);
end;

procedure TARCellPropDlg.CellTextChange(Sender: TObject);
begin
  CellText.Color:=clWhite;
  SomeChange(Sender);
end;

procedure TARCellPropDlg.ApplyBtnClick(Sender: TObject);
var i:integer;
begin
  for i:=0 to rep.SelCount - 1 do begin
    if Path.Color <> clBtnFace then begin
      rep.Sel[i].PictureFileName:=Path.Text;
    end;
    if LinkToFile.State <> cbGrayed then begin
      rep.Sel[i].LinkPictureToFile:=LinkToFile.Checked;
    end;
    if CellText.Color <> clBtnFace then begin
      rep.Sel[i].Value:=CellText.Text;
    end;
    if TextOrientation.ItemIndex <> -1 then begin
      rep.Sel[i].TextAngle:=TextOrientation.ItemIndex * 90;
    end;
  end;
  ApplyBtn.Enabled:=false;
end;

procedure TARCellPropDlg.FormShow(Sender: TObject);
var i:integer;
begin
  if rep.SelCount > 0 then begin
    Path.Text                 :=rep.CurrCell.PictureFileName;
    LinkToFile.Checked        :=rep.CurrCell.LinkPictureToFile;
    CellText.Text             :=rep.CurrCell.Value;
    TextOrientation.ItemIndex :=rep.CurrCell.TextAngle div 90;

    for i:=0 to rep.SelCount - 1 do begin
      if rep.Sel[i].PictureFileName <> Path.Text then begin
        Path.Color:=clBtnFace;
      end;
      if rep.Sel[i].LinkPictureToFile <> LinkToFile.Checked then begin
        LinkToFile.State:=cbGrayed;
      end;
      if rep.Sel[i].Value <> CellText.Text then begin
        CellText.Color:=clBtnFace;
      end;
      if rep.Sel[i].TextAngle <> (TextOrientation.ItemIndex * 90) then begin
        TextOrientation.ItemIndex:=-1;
      end;
    end;

  end else begin
    Close();
  end;
end;

end.
