unit AREditorPageSetup;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons,

  ARResStrings, ARRepCommon, Printers, ARPrinter;

type
  TARPageSetupDlg = class(TForm)
    MarginsGB: TGroupBox;
    LeftEdit: TEdit;
    TopEdit: TEdit;
    RightEdit: TEdit;
    BottomEdit: TEdit;
    LeftLabel: TLabel;
    TopLabel: TLabel;
    RightLabel: TLabel;
    BottomLabel: TLabel;
    CancelBtn: TBitBtn;
    OKBtn: TBitBtn;
    OrientationRG: TRadioGroup;
    PaperSizeCB: TComboBox;
    PaperSizeLabel: TLabel;
    procedure OKBtnClick(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
  public
    Units         :TARUnits;

    LeftMargin    :extended;
    TopMargin     :extended;
    RightMargin   :extended;
    BottomMargin  :extended;

    Orientation   :TPrinterOrientation;
    PaperSize     :TARPaperSize;

    function  Execute:boolean;
  end;

implementation

{$R *.DFM}

function  TARPageSetupDlg.Execute:boolean;
  function SizeToText(bound:extended):string;
  begin
    Result:=Format('%.2f',[ConvertTo(bound, Units)]);
  end;
begin
  MarginsGB.Caption:=Format(rsAREditorPrintSetup_Margins + ' (%s)', [UnitsShortName(Units)]);

  LeftEdit.Text   :=SizeToText(LeftMargin);
  TopEdit.Text    :=SizeToText(TopMargin);
  RightEdit.Text  :=SizeToText(RightMargin);
  BottomEdit.Text :=SizeToText(BottomMargin);

  PaperSizeCB.ItemIndex    :=integer(PaperSize);
  OrientationRG.ItemIndex  :=integer(Orientation);

  Result:=ShowModal = mrOK;
end;

procedure TARPageSetupDlg.OKBtnClick(Sender: TObject);
  function TextToSize(bound:string):extended;
  begin
    Result:=ConvertFrom(StrToFloat(bound), Units);
  end;
var Edit:TEdit;
begin
  Edit:=LeftEdit;   
  try
                      LeftMargin    :=TextToSize(Edit.Text);
    Edit:=TopEdit;    TopMargin     :=TextToSize(Edit.Text);
    Edit:=RightEdit;  RightMargin   :=TextToSize(Edit.Text);
    Edit:=BottomEdit; BottomMargin  :=TextToSize(Edit.Text);
  except
    Edit.SelectAll;
    Edit.SetFocus;
    exit;
  end;

  PaperSize    :=TARPaperSize(PaperSizeCB.ItemIndex);
  Orientation  :=TPrinterOrientation(OrientationRG.ItemIndex);

  ModalResult:=mrOK;
end;

procedure TARPageSetupDlg.CancelBtnClick(Sender: TObject);
begin
  ModalResult:=mrCancel;
end;

procedure TARPageSetupDlg.FormCreate(Sender: TObject);
begin
  Caption               :=rsAREditorPrintSetup_PageSetup;
  MarginsGB.Caption     :=rsAREditorPrintSetup_Margins;
  LeftLabel.Caption     :=rsAREditorPrintSetup_LeftLabel;
  TopLabel.Caption      :=rsAREditorPrintSetup_TopLabel;
  RightLabel.Caption    :=rsAREditorPrintSetup_RightLabel;
  BottomLabel.Caption   :=rsAREditorPrintSetup_BottomLabel;
  CancelBtn.Caption     :=rsAREditorPrintSetup_CancelBtn;
  OKBtn.Caption         :=rsAREditorPrintSetup_OKBtn;
  OrientationRG.Caption :=rsAREditorPrintSetup_Orientation;
  PaperSizeLabel.Caption:=rsAREditorPrintSetup_PaperSize;

  OrientationRG.Items.Clear;
  OrientationRG.Items.Add(rsAREditorPrintSetup_Portrait);
  OrientationRG.Items.Add(rsAREditorPrintSetup_Landscape);
end;

end.
