unit ARFrame;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes, SyncObjs, Windows, Controls, Graphics, ImgList,

  ARUtils, ARGridRep, ARRepCommon;

type
  TARFrameThread = class(TThread)
  private
    ARGridRepList :TAIList;
    CS            :TCriticalSection;
  protected
    procedure     Execute; override;
    procedure     ShowFrame;
  public
    constructor   Create(CreateSuspended: Boolean);
    destructor    Destroy;override;
    procedure     Add(Rep:TARGridRep);
    procedure     Delete(Rep:TARGridRep);
  end;


implementation
{$R ARGridRepFR8.res}

var
    FrameBMPList:TImageList;
    FrameBMP:TBitmap;
    FrameBmpIdx :integer;

constructor   TARFrameThread.Create(CreateSuspended: Boolean);
begin
  ARGridRepList:=TAIList.Create('TARFrameThread ARGridList');
  CS           :=TCriticalSection.Create;
  inherited;
  FreeOnTerminate:=false;
end;

destructor    TARFrameThread.Destroy;
begin
  CS.Enter;
    Terminate;
    WaitFor;
  CS.Leave;
  CS.Free;
  ARGridRepList.Free;
  inherited;
end;

procedure   TARFrameThread.ShowFrame;
var OldClipRgn, ClipRgn:HRGN;
    r:TRect;
    i:integer;
    Rep:TARGridRep;
    FocusedCellStyle:TFocusedCellStyle;
const d = 0;
begin
  FocusedCellStyle:=fcsBlackAnt;
  CS.Enter;
  try
    if ARGridRepList.Count > 0 then begin
      for i:=0 to ARGridRepList.Count - 1 do begin
        Rep:=TARGridRep(ARGridRepList[i]);

        if  (not Assigned(Rep.Parent)) or
            (not (Rep.Parent is TWinControl)) or
            (csDestroying	in Rep.ComponentState) or
            (not Rep.Focused) or
            (not Rep.Visible) or
            (not (Rep.FocusedCellStyle in [fcsBlackAnt, fcsRedAnt]))
        then continue;

        if (Rep.FOldCurrBandIdx <> Rep.CurrBandIdx) or (Rep.FOldCurrCellIdx <> Rep.CurrCellIdx) then begin
          if  (Rep.FOldCurrBandIdx >= 0) and (Rep.FOldCurrBandIdx <= Rep.SrcRep.BandCount - 1) and
              (Rep.FOldCurrCellIdx >= 0) and (Rep.FOldCurrCellIdx <= Rep.SrcRep.Bands[Rep.FOldCurrBandIdx].CellCount - 1) then begin

            //force repaint last cell
            with Rep.SrcRep.Bands[Rep.FOldCurrBandIdx].Cells[Rep.FOldCurrCellIdx] do begin
              BeginUpdate;
                Changed:=true;
              EndUpdate;
            end;

          end;
          Rep.FOldCurrBandIdx:=Rep.CurrBandIdx;
          Rep.FOldCurrCellIdx:=Rep.CurrCellIdx;
        end;

        r:=Rep.FrameCellClipRect(Rep.CurrBandIdx, Rep.CurrCellIdx);
        if not IsRectEmpty(r) then begin
          ClipRgn:=CreateRectRgn(r.Left, r.Top,  r.Right, r.Bottom);
          OldClipRgn:=SelectObject(Rep.Canvas.Handle, ClipRgn);
          try
            Rep.Canvas.Brush.Color  :=clRed;
            FrameBMPList.GetBitmap(FrameBmpIdx, FrameBMP);
            Rep.Canvas.Brush.Bitmap:=FrameBMP;

            r:=Rep.FrameCellRect(Rep.CurrBandIdx, Rep.CurrCellIdx);
            if not IsRectEmpty(r) then begin
              r.Left  :=r.Left + d;
              r.Top   :=r.Top + d;
              r.Right :=r.Right - d;
              r.Bottom:=r.Bottom - d;
              Rep.Canvas.FrameRect(r);
            end;
            Rep.Canvas.Brush.Color:=clBlack;
          finally
            SelectObject(Rep.Canvas.Handle, OldClipRgn);
            DeleteObject(ClipRgn);
          end;
        end;
        FocusedCellStyle:=Rep.FocusedCellStyle;
      end;

      FrameBmpIdx:=(FrameBmpIdx + 1) mod 4;
      if FocusedCellStyle = fcsRedAnt then begin
        FrameBmpIdx:=FrameBmpIdx + 8;
      end;
    end;
  finally
    CS.Leave;
  end;
end;

procedure   TARFrameThread.Execute;
begin
  while not Terminated do begin
    Synchronize(ShowFrame);
    Sleep(100);
  end;
end;

procedure   TARFrameThread.Add(Rep:TARGridRep);
begin
  CS.Enter;
    if ARGridRepList.IndexOf(Rep) = -1 then begin
      ARGridRepList.Add(Rep);
    end;
  CS.Leave;
end;

procedure   TARFrameThread.Delete(Rep:TARGridRep);
var idx:integer;
begin
  CS.Enter;
    idx:=ARGridRepList.IndexOf(Rep);
    if idx <> -1 then begin
      ARGridRepList.Delete(idx);
    end;
  CS.Leave;
end;

initialization
  IsMultiThread:=true;

  FrameBMPList:=TImageList.CreateSize(8, 8);
  FrameBMPList.GetResource(rtBitmap, 'FRAMES', 8, [], clBlue);

  FrameBMP    :=TBitmap.Create();

finalization
  FrameBMP.Free();
  FrameBMPList.Free();
end.


