unit ARInplaceEditor;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes, Menus, Controls, StdCtrls, Forms, Messages, SysUtils, Windows,

  ARGridRep;

type
  TARInplaceEditorDlg = class(TForm)
    Editor: TMemo;
    PopupMenu: TPopupMenu;
    OK1: TMenuItem;
    Cancel1: TMenuItem;
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure OK1Click(Sender: TObject);
    procedure Cancel1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    Rep         :TARGridRep;
    CancelOnExit:boolean;

    procedure   WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    procedure   CreateParams(var p: TCreateParams); override;
		procedure   KeyDown(var Key: Word; Shift: TShiftState);override;
    procedure   KeyPress(var Key: Char);override;
    procedure   DoExit;override;

    procedure   CancelEdit;
    procedure   EndEdit;
  public
    constructor Create(AOwner:TComponent);override;
    destructor  Destroy;override;

    procedure   RunEdit(ch:char);
  end;

implementation

{$R *.DFM}

constructor TARInplaceEditorDlg.Create(AOwner:TComponent);
begin
  if AOwner is TARGridRep then begin
    inherited Create(AOwner);
    Rep     :=TARGridRep(AOwner);
    Parent  :=TARGridRep(AOwner);
    Visible :=false;
  end else begin
    raise Exception.Create('Owner of TARInplaceEditorDlg must be TARGridRep');
  end;
end;

destructor  TARInplaceEditorDlg.Destroy;
begin
  inherited;
end;

procedure   TARInplaceEditorDlg.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := DLGC_WANTMESSAGE or DLGC_WANTARROWS or DLGC_WANTCHARS or DLGC_WANTALLKEYS;// or DLGC_WANTTAB;
end;

procedure   TARInplaceEditorDlg.CreateParams(var p: TCreateParams);
begin
  inherited CreateParams(p);
  with p do begin
//    ExStyle :=ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

procedure   TARInplaceEditorDlg.CancelEdit;
begin
  CancelOnExit:=true;

  Rep.SetFocus;
end;

procedure   TARInplaceEditorDlg.EndEdit;
begin
  CancelOnExit:=false;

  Rep.SetFocus;
end;

procedure   TARInplaceEditorDlg.DoExit;
begin
  inherited;

  try
    if (not CancelOnExit) and Assigned(rep.CurrCell) then begin
      rep.CurrCell.Value:=Editor.Text;

      if Editor.Modified then begin
        if Assigned(Rep.OnCellChange) then Rep.OnCellChange(Rep.CurrBand, Rep.CurrCell);
        if Assigned(Rep.OnAfterEdit)  then Rep.OnAfterEdit(rep);
      end;
    end;
  finally
    Close;
    rep.InEditing:=false;
  end;
end;

procedure   TARInplaceEditorDlg.RunEdit(ch:char);
var r:TRect;
    border:integer;
begin
  Parent:=rep;
  
  rep.PopupMenu:=nil;
  rep.InEditing:=true;
  r:=Rep.CellRect(Rep.CurrBandIdx, Rep.CurrCellIdx);

  border:=Rep.CurrCell.FrameWidth;

  SetBounds(r.Left + border,
            r.Top + border,
            r.Right - r.Left - 2 * border - 1,
            r.Bottom - r.Top - 2 * border - 1);

  Editor.Color        :=Rep.CurrCell.Color;
  BorderStyle         :=bsNone;
  Editor.WordWrap     :=Rep.CurrCell.WordWrap;
  Editor.Alignment    :=TAlignment(rep.CurrCell.HAlign);
  Editor.Font.Size    :=rep.CurrCell.FontSize;
  Editor.Font.Style   :=rep.CurrCell.FontStyle;
  Editor.Font.Name    :=rep.CurrCell.FontName;
  Editor.Font.Color   :=rep.CurrCell.FontColor;
  Editor.Font.Charset :=rep.SrcRep.Font.Charset;

  Editor.Text:=rep.CurrCell.Value;

  Show;

  if ch <> #0 then begin
    Editor.Perform(WM_CHAR, Ord(ch), 0);
  end;
end;

procedure TARInplaceEditorDlg.FormShow(Sender: TObject);
begin
  rep.CurrCell.BeginUpdate;
  Editor.SelectAll;
  Editor.SetFocus;
end;

procedure TARInplaceEditorDlg.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose:=true;
end;

procedure TARInplaceEditorDlg.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  rep.CurrCell.EndUpdate;
  Action:=caFree;
end;

procedure   TARInplaceEditorDlg.KeyDown(var Key: Word; Shift: TShiftState);
begin
	case Key of
    VK_ESCAPE:
      begin
     		CancelEdit;
       	Key := 0;
     	end;
    VK_RETURN:
      begin
        if not (ssCtrl in Shift) then begin
          EndEdit;
        end else begin
          Editor.Perform(WM_CHAR, VK_RETURN, 0);
        end;

       	Key := 0;
      end;
		else  begin
          end;
	end;

	inherited KeyDown(Key, Shift);
end;

procedure TARInplaceEditorDlg.OK1Click(Sender: TObject);
begin
  EndEdit;
end;

procedure TARInplaceEditorDlg.Cancel1Click(Sender: TObject);
begin
	CancelEdit;
end;

procedure TARInplaceEditorDlg.KeyPress(var Key: Char);
begin
  if Key in [#32..#255] then begin
    Editor.Perform(WM_CHAR, Ord(Key), 0);
  end else begin
    inherited KeyPress(Key);
  end;
end;

end.

