unit ARPreviewForm;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes, ComCtrls, StdCtrls, Mask, Controls, ExtCtrls, Buttons, ToolWin,
  Forms, SysUtils,

  ARPrinter, ARResStrings;

type

  TARPreviewDialog = class(TForm)
    ToolBar: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    StatusBar: TStatusBar;
    CloseButton: TButton;
    PrinterSetupBtn: TSpeedButton;
    FirstPageBtn: TSpeedButton;
    ZoomTo100Btn: TSpeedButton;
    ZoomToWidthBtn: TSpeedButton;
    ZoomToFitBtn: TSpeedButton;
    PrintBtn: TSpeedButton;
    PriorPageBtn: TSpeedButton;
    NextPageBtn: TSpeedButton;
    LastPageBtn: TSpeedButton;
    CloseBtn: TSpeedButton;
    ShowNPABtn: TSpeedButton;
    ToolButton5: TToolButton;
    Panel1: TPanel;
    ZoomEdit: TMaskEdit;
    ZoomUpDown: TUpDown;
    procedure PrinterSetupBtnClick(Sender: TObject);
    procedure PrintBtnClick(Sender: TObject);
    procedure ZoomToFitBtnClick(Sender: TObject);
    procedure ZoomToWidthBtnClick(Sender: TObject);
    procedure ZoomTo100BtnClick(Sender: TObject);
    procedure FirstPageBtnClick(Sender: TObject);
    procedure PriorPageBtnClick(Sender: TObject);
    procedure NextPageBtnClick(Sender: TObject);
    procedure LastPageBtnClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure ZoomUpDownClick(Sender: TObject; Button: TUDBtnType);
    procedure CloseButtonClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ZoomEditChange(Sender: TObject);
    procedure ShowNPABtnClick(Sender: TObject);
  private
    function    GetARPrinter:TARPrinter;
    procedure   SetARPrinter(AARPrinter:TARPrinter);
    procedure   UpdateStatusBar;
    procedure   ARPreviewChangeView(Sender: TObject);
  public
    ARPreview   :TARPreview;

    property    ARPrinter:TARPrinter  read  GetARPrinter  write SetARPrinter;
  end;

implementation

{$R *.DFM}

procedure TARPreviewDialog.FormResize(Sender: TObject);
begin
  ZoomUpDown.Position:=ARPreview.Zoom;
end;

procedure TARPreviewDialog.PrinterSetupBtnClick(Sender: TObject);
begin
  ARPreview.ARPrinter.PrinterSetupDialog;
  ARPreview.Invalidate;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.PrintBtnClick(Sender: TObject);
begin
  ARPreview.ARPrinter.PrintDialog;
  ARPreview.Invalidate;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.ZoomToFitBtnClick(Sender: TObject);
begin
  ARPreview.ZoomToFit;
  ZoomUpDown.Position:=ARPreview.Zoom;
end;

procedure TARPreviewDialog.ZoomToWidthBtnClick(Sender: TObject);
begin
  ARPreview.ZoomToWidth;
  ZoomUpDown.Position:=ARPreview.Zoom;
end;

procedure TARPreviewDialog.ZoomTo100BtnClick(Sender: TObject);
begin
  ARPreview.Zoom:=100;
  ZoomUpDown.Position:=ARPreview.Zoom;
end;

procedure TARPreviewDialog.FirstPageBtnClick(Sender: TObject);
begin
  ARPreview.CurrPage:=1;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.PriorPageBtnClick(Sender: TObject);
begin
  ARPreview.CurrPage:=ARPreview.CurrPage - 1;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.NextPageBtnClick(Sender: TObject);
begin
  ARPreview.CurrPage:=ARPreview.CurrPage + 1;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.LastPageBtnClick(Sender: TObject);
begin
  ARPreview.CurrPage:=100000;
  UpdateStatusBar;
end;

function    TARPreviewDialog.GetARPrinter:TARPrinter;
begin
  Result:=ARPreview.ARPrinter
end;

procedure    TARPreviewDialog.SetARPrinter(AARPrinter:TARPrinter);
begin
  ARPreview.ARPrinter:=AARPrinter;
end;

procedure   TARPreviewDialog.CloseBtnClick(Sender: TObject);
begin
  ModalResult:=mrOK;
end;

procedure    TARPreviewDialog.UpdateStatusBar;
begin
  StatusBar.Panels[0].Text:=Format(rsARPreviewForm_Pages, [ARPreview.CurrPage, ARPreview.PageCount]);
end;

procedure TARPreviewDialog.CloseButtonClick(Sender: TObject);
begin
  ModalResult:=mrOK;
end;

procedure TARPreviewDialog.ARPreviewChangeView(Sender: TObject);
begin
  ZoomUpDown.Position:=ARPreview.Zoom;
  UpdateStatusBar;
end;

procedure TARPreviewDialog.FormCreate(Sender: TObject);
begin
  ARPreview:=TARPreview.Create(Self);
  ARPreview.Parent:=Self;
  ARPreview.Align:=alClient;
  ARPreview.OnChangeView:=ARPreviewChangeView;

  ZoomUpDown.Min:=ARPreviewZoomMin;
  ZoomUpDown.Max:=ARPreviewZoomMax;

  Caption               :=rsARPreviewForm_Caption;
  PrinterSetupBtn.Hint  :=rsARPreviewForm_PrinterSetupBtn;
  FirstPageBtn.Hint     :=rsARPreviewForm_FirstPageBtn;
  ZoomTo100Btn.Hint     :=rsARPreviewForm_ZoomTo100Btn;
  ZoomToWidthBtn.Hint   :=rsARPreviewForm_ZoomToWidthBtn;
  ZoomToFitBtn.Hint     :=rsARPreviewForm_ZoomToFitBtn;
  PrintBtn.Hint         :=rsARPreviewForm_PrintBtn;
  PriorPageBtn.Hint     :=rsARPreviewForm_PriorPageBtn;
  NextPageBtn.Hint      :=rsARPreviewForm_NextPageBtn;
  LastPageBtn.Hint      :=rsARPreviewForm_LastPageBtn;
  CloseBtn.Hint         :=rsARPreviewForm_CloseBtn;
  ZoomEdit.Hint         :=rsARPreviewForm_Zoom;
  ZoomUpDown.Hint       :=rsARPreviewForm_Zoom;
  ShowNPABtn.Hint       :=rsARPreviewForm_ShowNPA;
end;

procedure TARPreviewDialog.FormShow(Sender: TObject);
begin
  ZoomEdit.OnChange:=ZoomEditChange;
  ARPreview.CurrPage:=1;
  UpdateStatusBar;
  ARPreview.ZoomToFit;
end;

procedure TARPreviewDialog.ZoomUpDownClick(Sender: TObject; Button: TUDBtnType);
begin
  ARPreview.Zoom:=ZoomUpDown.Position;
end;

procedure TARPreviewDialog.ZoomEditChange(Sender: TObject);
begin
  try
    ARPreview.Zoom:=StrToInt(Trim(ZoomEdit.Text));
  except
  end;
end;

procedure TARPreviewDialog.ShowNPABtnClick(Sender: TObject);
begin
  ARPreview.ShowNonPrintingArea:=ShowNPABtn.Down;
end;

end.


