unit ARPrinter;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

{$INCLUDE DEFINES.INC}

interface

uses
  Classes, Graphics, Printers, ExtCtrls, Windows, Controls, Messages, Forms,
  Math, Dialogs,

  ARUtils;

type
  TARPaperSize = (
    Default,    Letter, LetterSmall,  Tabloid,  Ledger, Legal,  Statement,
    Executive,  A3,     A4,           A4Small,  A5,     B4,     B5,
    Folio,      Quarto, qr10X14,      qr11X17,  Note,   Env9,   Env10,
    Env11,      Env12,  Env14,        CSheet,   DSheet, ESheet, Custom
  );

  TZoomMode = (zmFixed, zmToWidth, zmToFit);

const
   { Actual paper sizes for all the known paper types }
   cARPaperSizeMetrics : array[Letter..ESheet, 0..1] of extended =
      ((215.9, 279.4), { Letter }
       (215.9, 279.4), { Letter small }
       (279.4, 431.8), { Tabloid }
       (431.8, 279.4), { Ledger }
       (215.9, 355.6), { Legal }
       (139.7, 215.9), { Statement }
       (190.5, 254.0), { Executive }
       (297.0, 420.0), { A3 }
       (210.0, 297.0), { A4 }
       (210.0, 297.0), { A4 small }
       (148.0, 210.0), { A5 }
       (250.0, 354.0), { B4 }
       (182.0, 257.0), { B5 }
       (215.9, 330.2), { Folio }
       (215.0, 275.0), { Quarto }
       (254.0, 355.6), { 10X14 }
       (279.4, 431.8), { 11X17 }
       (215.9, 279.0), { Note }
       (98.43, 225.4), { Envelope #9 }
       (104.8, 241.3), { Envelope #10 }
       (114.3, 263.5), { Envelope #11 }
       (101.6, 279.4), { Envelope #12 - might be wrong !! }
       (127.0, 292.1), { Envelope #14 }
       (100.0, 100.0),
       (100.0, 100.0),
       (100.0, 100.0));

  ARPreviewZoomMin = 10;
  ARPreviewZoomMax = 500;

type
  TARPrinter = class(TComponent)
  private
    PageList    :TAIList;

    FCurPage    :integer;
    FCanvas     :TCanvas;
    FMetafile   :TMetafile;

    FFromPage   :integer;
    FToPage     :integer;
    FPaperSize  :TARPaperSize;
    FCopies     :integer;
    FTitle      :string;
    FOrientation:TPrinterOrientation;

    FPPIX       :integer;
    FPPIY       :integer;
    FLeftMargin :integer;
    FTopMargin  :integer;
    FPageWidth  :integer;
    FPageHeight :integer;
    FPrintingWidth  :integer;
    FPrintingHeight :integer;

    FOnPrinterChange:TNotifyEvent;

    procedure   SetPaperSize(APaperSize:TARPaperSize);

    function    GetCopies:integer;
    procedure   SetCopies(ACopies:integer);

    function    GetCollation:boolean;

    function    GetOrientation:TPrinterOrientation;
    procedure   SetOrientation(AOrientation:TPrinterOrientation);

    function    GetTitle:string;
    procedure   SetTitle(ATitle:string);
    procedure   SetCurPage(ACurPage:integer);
    function    GetPageCount:integer;

    function    GetPaperWidth:integer;
    function    GetPaperHeight:integer;
    function    GetPrintWidth:integer;
    function    GetPrintHeight:integer;

    function    GetLeftMargin:integer;
    function    GetTopMargin :integer;
    function    GetRightMargin:integer;
    function    GetBottomMargin :integer;

    property    Metafile:TMetafile              read  FMetafile;
  public
    function    RefreshPrinter:boolean;

    property    PaperSize:TARPaperSize          read  FPaperSize      write SetPaperSize;

    property    PrinterPPIX           :integer  read  FPPIX;
    property    PrinterPPIY           :integer  read  FPPIY;
    property    PrinterLeftMargin     :integer  read  FLeftMargin;
    property    PrinterTopMargin      :integer  read  FTopMargin;
    property    PrinterPrintingWidth  :integer  read  FPrintingWidth;
    property    PrinterPrintingHeight :integer  read  FPrintingHeight;

    property    PaperHeight:integer             read  GetPaperHeight;
    property    PaperWidth:integer              read  GetPaperWidth;
    property    PrintHeight:integer             read  GetPrintHeight;
    property    PrintWidth:integer              read  GetPrintWidth;
    property    LeftMargin:integer              read  GetLeftMargin;
    property    TopMargin:integer               read  GetTopMargin;
    property    RightMargin:integer             read  GetRightMargin;
    property    BottomMargin:integer            read  GetBottomMargin;

    property    CurrPage:integer                read  FCurPage        write SetCurPage;
    property    PageCount:integer               read  GetPageCount;
    property    Canvas:TCanvas                  read  FCanvas         write FCanvas;
    property    Title:string                    read  GetTitle        write SetTitle;

    property    FromPage:integer                read  FFromPage       write FFromPage;
    property    ToPage:integer                  read  FToPage         write FToPage;
    property    Copies:integer                  read  GetCopies       write SetCopies;
    property    Orientation:TPrinterOrientation read  GetOrientation  write SetOrientation;
    property    Collation:boolean               read  GetCollation;

    procedure   BeginDoc;
    procedure   NewPage;
    procedure   EndDoc;
    procedure   Abort;

    procedure   Preview;
    procedure   Print;
    procedure   PrintDialog;
    procedure   PrinterSetupDialog;

    property    OnPrinterChange:TNotifyEvent  read  FOnPrinterChange  write FOnPrinterChange;
    constructor Create(AOwner:TComponent);override;
    destructor  Destroy;override;
  end;

  TARPreview  = class(TPanel)
  private
    FTopY       :integer;
    FLeftX      :integer;

    FPageDX     :integer;
    FPageDY     :integer;

    FScrollUpdate :boolean;

    FZoom       :integer;

    FARPrinter  :TARPrinter;

    FZoomMode   :TZoomMode;

    FOnChangeView:TNotifyEvent;
    FShowNonPrintingArea:boolean;

    procedure   SetShowNonPrintingArea(AShowNonPrintingArea:boolean);

    procedure   SetARPrinter(AARPrinter:TARPrinter);

    function    GetCurPage:integer;
    procedure   SetCurPage(ACurPage:integer);
    function    GetPageCount:integer;

    function    GetViewPageDX:integer;
    function    GetViewPageDY:integer;

    procedure   SetZoom(AZoom:integer);

    procedure   SetTopY(ATopY:integer);
    procedure   SetLeftX(ALeftX:integer);

    function    GetVSTrackPos:integer;
    function    GetVSPos:integer;
    procedure   SetVSPos(AVSPos:integer);
    function    GetVSMax:integer;
    procedure   SetVSMax(AVSMax:integer);
    function    GetVSPage:integer;
    procedure   SetVSPage(AVSPage:integer);

    function    GetHSTrackPos:integer;
    function    GetHSPos:integer;
    procedure   SetHSPos(AHSPos:integer);
    function    GetHSMax:integer;
    procedure   SetHSMax(AHSMax:integer);
    function    GetHSPage:integer;
    procedure   SetHSPage(AHSPage:integer);

    procedure   SetPageDX(APageDX:integer);
    procedure   SetPageDY(APageDY:integer);

    procedure   MouseWheelUp(Sender:TObject; Shift:TShiftState; MousePos:TPoint; var Handled:boolean);
    procedure   MouseWheelDown(Sender:TObject; Shift:TShiftState; MousePos:TPoint; var Handled:boolean);
  protected
    procedure   Paint;override;
    procedure   CreateParams(var p: TCreateParams);override;
    procedure   KeyDown(var Key: Word; Shift: TShiftState); override;

    procedure   WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure   WMVScroll   (var Message: TWMScroll);     message WM_VSCROLL;
    procedure   WMHScroll   (var Message: TWMScroll);     message WM_HSCROLL;
    procedure   WMSize      (var Message: TWMSize);       message WM_SIZE;
    procedure   WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;

    property    VSPos :integer        read GetVSPos     write SetVSPos;
    property    VSMax :integer        read GetVSMax     write SetVSMax;
    property    VSPage:integer        read GetVSPage    write SetVSPage;
    property    HSPos :integer        read GetHSPos     write SetHSPos;
    property    HSMax :integer        read GetHSMax     write SetHSMax;
    property    HSPage:integer        read GetHSPage    write SetHSPage;

    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer);override;

    procedure   UpdateScrolls;

    property    ViewPageDX:integer          read GetViewPageDX;
    property    ViewPageDY:integer          read GetViewPageDY;
  public
    property    ARPrinter:TARPrinter        read FARPrinter           write SetARPrinter;

    property    TopY :integer               read FTopY                write SetTopY;
    property    LeftX:integer               read FLeftX               write SetLeftX;

    property    PageDX:integer              read FPageDX              write SetPageDX;
    property    PageDY:integer              read FPageDY              write SetPageDY;

    property    PageCount:integer           read GetPageCount;

    procedure   ZoomToFit;
    procedure   ZoomToWidth;

    constructor Create(AOwner:TComponent);override;

//    property    Canvas;
  published
    property    Zoom:integer                read FZoom                write SetZoom default 100;
    property    CurrPage:integer            read GetCurPage           write SetCurPage;
    property    ShowNonPrintingArea:boolean read FShowNonprintingArea write SetShowNonprintingArea;
    property    OnChangeView:TNotifyEvent   read  FOnChangeView       write FOnChangeView;
  end;

procedure Register;

implementation
{$R *.RES}

uses ARPreviewForm;

function    PrinterInstalled:boolean;
begin
  Result:=Printer.Printers.Count > 0;
end;
///////////////////////////////////////////////////////////////////////
constructor  TARPrinter.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  PageList:=TAIList.Create('TARPrinter PageList');
  FCurPage:=0;
end;

destructor  TARPrinter.Destroy;
begin
  Abort;

  PageList.Free;
  inherited;
end;

type
  TPrinterDevice = class
    Driver      :string;
    Device      :string;
    Port        :string;
//    constructor Create(ADriver, ADevice, APort: PChar);
//    function    IsEqual(ADriver, ADevice, APort: PChar): Boolean;
  end;

function      TARPrinter.RefreshPrinter:boolean;
var NewPPIX       :integer;
    NewPPIY       :integer;
    NewLeftMargin :integer;
    NewTopMargin  :integer;
    NewPageWidth  :integer;
    NewPageHeight :integer;
begin
  Result:=false;

  if Printer.Printers.Count = 0 then begin
    FPPIX       :=Screen.PixelsPerInch;
    FPPIY       :=Screen.PixelsPerInch;
    FLeftMargin :=0;
    FTopMargin  :=0;
    FPageWidth  :=Round(cARPaperSizeMetrics[PaperSize,     Ord(Orientation)] / 25.4 * FPPIX);
    FPageHeight :=Round(cARPaperSizeMetrics[PaperSize, 1 - Ord(Orientation)] / 25.4 * FPPIY);
  end else begin
    FPPIX:=printer.PageHeight; // FORCE set IC, ie. Printer.Handle is IC now

    NewPPIX           :=GetDeviceCaps(Printer.Handle, LOGPIXELSX);
    if FPPIX <> NewPPIX then begin
      FPPIX := NewPPIX;
      Result:=true;
    end;
    NewPPIY           :=GetDeviceCaps(Printer.Handle, LOGPIXELSY);
    if FPPIY <> NewPPIY then begin
      FPPIY := NewPPIY;
      Result:=true;
    end;
    NewLeftMargin     :=GetDeviceCaps(Printer.Handle, PHYSICALOFFSETX);
    if FLeftMargin <> NewLeftMargin then begin
      FLeftMargin := NewLeftMargin;
      Result:=true;
    end;
    NewTopMargin      :=GetDeviceCaps(Printer.Handle, PHYSICALOFFSETY);
    if FTopMargin <> NewTopMargin then begin
      FTopMargin := NewTopMargin;
      Result:=true;
    end;
    NewPageWidth  :=GetDeviceCaps(Printer.Handle, PHYSICALWIDTH);
    if FPageWidth <> NewPageWidth then begin
      FPageWidth := NewPageWidth;
      Result:=true;
    end;
    NewPageHeight :=GetDeviceCaps(Printer.Handle, PHYSICALHEIGHT);
    if FPageHeight <> NewPageHeight then begin
      FPageHeight := NewPageHeight;
      Result:=true;
    end;
  end;
end;

procedure    TARPrinter.SetPaperSize(APaperSize:TARPaperSize);
var Driver, Device, Port:PChar;
    DeviceMode:THandle;
    DevMode: PDeviceMode;
begin
  FPaperSize:=APaperSize;
  if Printer.Printers.Count > 0 then begin
    GetMem(Driver,  256);
    GetMem(Device,  256);
    GetMem(Port,    256);
    try
      Printer.GetPrinter(Driver, Device, Port, DeviceMode);
        DevMode:=GlobalLock(DeviceMode);
          DevMode^.dmPaperSize:=Ord(APaperSize);
        GlobalUnlock(DeviceMode);
//    Printer.SetPrinter(Driver, Device, Port, DeviceMode);
    finally
      FreeMem(Port);
      FreeMem(Device);
      FreeMem(Driver);
    end;
  end;
end;

function      TARPrinter.GetPaperWidth:integer;
begin
  Result:=Round(FPageWidth * Screen.PixelsPerInch / PrinterPPIX);
end;

function      TARPrinter.GetPaperHeight:integer;
begin
  Result:=Round(FPageHeight * Screen.PixelsPerInch / PrinterPPIY);
end;

function      TARPrinter.GetPrintWidth:integer;
begin
  Result:=Round(Printer.PageWidth * Screen.PixelsPerInch / PrinterPPIX);
end;

function      TARPrinter.GetPrintHeight:integer;
begin
  Result:=Round(Printer.PageHeight * Screen.PixelsPerInch / PrinterPPIY);
end;

function      TARPrinter.GetLeftMargin:integer;
begin
  Result:=Round(PrinterLeftMargin * Screen.PixelsPerInch / PrinterPPIX);
end;

function      TARPrinter.GetTopMargin :integer;
begin
  Result:=Round(PrinterTopMargin  * Screen.PixelsPerInch / PrinterPPIY);
end;

function      TARPrinter.GetRightMargin:integer;
begin
  Result:=PaperWidth - LeftMargin;
end;

function      TARPrinter.GetBottomMargin :integer;
begin
  Result:=PaperHeight - TopMargin;
end;

procedure     TARPrinter.BeginDoc;
begin
  Abort;

  RefreshPrinter;

    FMetafile   :=TMetafile.Create;
    FMetafile.Width:=PaperWidth;
    FMetafile.Height:=PaperHeight;
    FCanvas   :=TMetafileCanvas.Create(FMetafile, 0);

  PageList.Add(FMetafile);
  FCurPage:=0;
end;

procedure    TARPrinter.NewPage;
begin
  FCanvas.Free;

    FMetafile   :=TMetafile.Create;
    FMetafile.Width:=PaperWidth;
    FMetafile.Height:=PaperHeight;

    FCanvas   :=TMetafileCanvas.Create(FMetafile, 0);

  PageList.Add(FMetafile);
  FCurPage:=PageCount;
end;

procedure    TARPrinter.EndDoc;
begin
  FCanvas.Free;
  FCanvas:=nil;

  FFromPage:=Min(PageCount, 1);
  FToPage  :=PageCount;
end;

procedure    TARPrinter.Abort;
var i:integer;
begin
  if (PageList.Count > 0) and Assigned(FCanvas) then begin
    FCanvas.Free;
    FCanvas:=nil;
  end;

  FMetafile:=nil;
  for i:=0 to PageList.Count - 1 do begin
    TMetafile(PageList[i]).Free;
  end;
  PageList.Clear;
  FCurPage  :=0;
  FFromPage :=0;
  FToPage   :=0;
end;

procedure    TARPrinter.SetCurPage(ACurPage:integer);
begin
  ACurPage:=Min(Max(ACurPage, 1), PageCount);

  if FCurPage <> ACurPage then begin
     FCurPage:=ACurPage;

    FMetafile:=PageList[FCurPage - 1];
  end;
end;

function    TARPrinter.GetPageCount:integer;
begin
  Result:=PageList.Count;
end;

function    TARPrinter.GetTitle:string;
begin
  if PrinterInstalled then begin
    Result:=Printer.Title;
    FTitle:=Result;
  end else begin
    Result:=FTitle;
  end;
end;

procedure    TARPrinter.SetTitle(ATitle:string);
begin
  FTitle:=ATitle;
  
  if PrinterInstalled then begin
    Printer.Title:=ATitle;
  end;
end;

function    TARPrinter.GetCopies:integer;
begin
  Result:=FCopies;
end;

procedure    TARPrinter.SetCopies(ACopies:integer);
begin
  FCopies:=ACopies;
  if PrinterInstalled then begin
    Printer.Copies:=FCopies;
  end;
end;

function    TARPrinter.GetCollation:boolean;
begin
  if PrinterInstalled then begin
    Result:=pcCollation in Printer.Capabilities;
  end else begin
    Result:=false;
  end;
end;

function    TARPrinter.GetOrientation:TPrinterOrientation;
begin
  if PrinterInstalled then begin
    Result:=Printer.Orientation;
    FOrientation:=Result;
  end else begin
    Result:=FOrientation;
  end;
end;

procedure    TARPrinter.SetOrientation(AOrientation:TPrinterOrientation);
begin
  FOrientation:=AOrientation;
  if PrinterInstalled then begin
    Printer.Orientation:=AOrientation;
  end;
end;

procedure    TARPrinter.PrinterSetupDialog;
var dlg:TPrinterSetupDialog;
begin
  dlg:=TPrinterSetupDialog.Create(Application.MainForm);
  try
    if dlg.Execute then begin
      if RefreshPrinter and Assigned(OnPrinterChange) then OnPrinterChange(Self);
    end;
  finally
    dlg.Free;
  end;
end;

procedure    TARPrinter.PrintDialog;
var dlg: TPrintDialog;
begin
  dlg:=TPrintDialog.Create(Application.MainForm);
  try
    dlg.MinPage :=1;
    dlg.MaxPage :=PageCount;
    dlg.FromPage:=FromPage;
    dlg.ToPage  :=ToPage;
    dlg.Copies  :=Copies;
    dlg.Collate :=true;
    dlg.Options :=[poPageNums,poWarning,poHelp,poDisablePrintToFile];

    if dlg.Execute then begin
      if RefreshPrinter and Assigned(OnPrinterChange) then OnPrinterChange(Self);

      FromPage:=dlg.FromPage;
      ToPage  :=dlg.ToPage;
      Copies  :=dlg.Copies;
      Print;
    end;
  finally
    dlg.Free;
  end;
end;

procedure    TARPrinter.Print;
var i:integer;
    SavedDC:integer;
begin
  if PageCount > 0 then begin

    Printer.BeginDoc;
    Printer.Title:=Title;

    try
      for i:=FromPage to ToPage do begin
        CurrPage:=i;

        SavedDC := SaveDC(Printer.Canvas.Handle);
        try
          SetMapMode      (Printer.Canvas.Handle, MM_ANISOTROPIC);
          SetWindowExtEx  (Printer.Canvas.Handle, Screen.PixelsPerInch,  Screen.PixelsPerInch,  nil);
          SetViewportExtEx(Printer.Canvas.Handle, PrinterPPIX , PrinterPPIY, nil);
          Printer.Canvas.Draw(0, 0, Metafile);
        finally
          RestoreDC(Printer.Canvas.Handle, SavedDC);
        end;

        if i < ToPage then Printer.NewPage;
      end;

      Printer.EndDoc;
    except
      Printer.Abort;
    end;
  end;
end;

procedure   TARPrinter.Preview;
var ARPreviewDialog:TARPreviewDialog;
begin
//  ARPreviewDialog:=TARPreviewDialog.Create(Application.MainForm);
//  ARPreviewDialog.ARPrinter:=Self;
//  ARPreviewDialog.ShowModal;
//  fixed by Ben Hay (ben_hay@bigfoot.com) 08.15.2000

  ARPreviewDialog:=TARPreviewDialog.Create(Application.MainForm);
  try
    ARPreviewDialog.ARPrinter:=Self;
    ARPreviewDialog.ShowModal;
  finally
    ARPreviewDialog.Free;
  end;
end;

//////////////////////////////////////////////////////////////////////////////////////
function    TARPreview.GetVSTrackPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nTrackPos;
end;

function    TARPreview.GetVSPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nPos;
end;

procedure   TARPreview.SetVSPos(AVSPos:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_POS;
  si.nPos:=AVSPos;
  SetScrollInfo(Handle, SB_VERT, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetVSMax:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nMax;
end;

procedure   TARPreview.SetVSMax(AVSMax:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_RANGE;
  si.nMin:=0;
  si.nMax:=AVSMax;
  SetScrollInfo(Handle, SB_VERT, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetVSPage:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_VERT, si);
  Result:=si.nPage;
end;

procedure   TARPreview.SetVSPage(AVSPage:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_PAGE;
  si.nPage:=AVSPage;
  SetScrollInfo(Handle, SB_VERT, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetHSTrackPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nTrackPos;
end;

function    TARPreview.GetHSPos:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nPos;
end;

procedure   TARPreview.SetHSPos(AHSPos:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_POS;
  si.nPos:=AHSPos;
  SetScrollInfo(Handle, SB_HORZ, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetHSMax:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nMax;
end;

procedure   TARPreview.SetHSMax(AHSMax:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_RANGE;
  si.nMin:=0;
  si.nMax:=AHSMax;
  SetScrollInfo(Handle, SB_HORZ, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetHSPage:integer;
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_ALL;
  GetScrollInfo(Handle, SB_HORZ, si);
  Result:=si.nPage;
end;

procedure   TARPreview.SetHSPage(AHSPage:integer);
var si: TScrollInfo;
begin
  si.cbSize:=SizeOf(si);
  si.fMask:=SIF_PAGE;
  si.nPage:=AHSPage;
  SetScrollInfo(Handle, SB_HORZ, si, true);

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

constructor TARPreview.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

{$IFNDEF D3}
  OnMouseWheelUp:=MouseWheelUp;
  OnMouseWheelDown:=MouseWheelDown;
{$ENDIF}

  ControlStyle:=[csClickEvents, csCaptureMouse, csDoubleClicks];

  ParentColor :=False;
  Color       :=clBtnFace;
  TabStop     :=true;

  FZoom       :=100;
  FZoomMode   :=zmFixed;
  DoubleBuffered:=false;

  ARPrinter   :=TARPrinter.Create(Self);
end;

procedure   TARPreview.CreateParams(var p: TCreateParams);
begin
  inherited CreateParams(p);
  with p do begin
    Style   :=(Style or WS_VSCROLL or WS_HSCROLL) or WS_TABSTOP  ;

    ExStyle :=ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

procedure   TARPreview.WMVSCROLL(var Message: TWMScroll );
begin
  if FScrollUpdate then exit;
  FScrollUpdate:=true;

  case Message.ScrollCode of
    SB_LINEDOWN:begin
                  TopY:=TopY + 10;
                end;
    SB_LINEUP  :begin
                  TopY:=TopY - 10;
                end;
    SB_PAGEUP  :begin
                  TopY:=TopY - ClientHeight;
                end;
    SB_PAGEDOWN:begin
                  TopY:=TopY + ClientHeight;
                end;
    SB_THUMBPOSITION:begin
                  TopY:=VSPos;
                end;
    SB_THUMBTRACK:begin
                  TopY:=GetVSTrackPos;
                end;
  end;

  FScrollUpdate:=false;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure   TARPreview.WMHSCROLL(var Message: TWMScroll );
begin
  if FScrollUpdate then exit;
  FScrollUpdate:=true;

  case Message.ScrollCode of
    SB_LINEDOWN:begin
                  LeftX:=LeftX + 10;
                end;
    SB_LINEUP  :begin
                  LeftX:=LeftX - 10;
                end;
    SB_PAGEUP  :begin
                  LeftX:=LeftX - ClientWidth;
                end;
    SB_PAGEDOWN :begin
                  LeftX:=LeftX + ClientWidth;
                end;
    SB_THUMBPOSITION:begin
                  LeftX:=HSPos;
                end;
    SB_THUMBTRACK:begin
                  LeftX:=GetHSTrackPos;
                end;
  end;

  FScrollUpdate:=false;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure   TARPreview.SetTopY(ATopY:integer);
begin
  FTopY:=Min(Max(ATopY, 0), ViewPageDY - ClientHeight);
  VSPos:=FTopY;
  Paint;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure   TARPreview.SetLeftX(ALeftX:integer);
begin
  FLeftX:=Min(Max(ALeftX, 0), ViewPageDX - ClientWidth);
  HSPos:=FLeftX;
  Paint;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure    TARPreview.SetPageDX(APageDX:integer);
begin
  FPageDX :=APageDX;
  HSMax   :=ViewPageDX;
  HSPage  :=ClientWidth;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure    TARPreview.SetPageDY(APageDY:integer);
begin
  FPageDY :=APageDY;
  VSMax   :=ViewPageDY;
  VSPage  :=ClientHeight;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetViewPageDX:integer;
begin
  Result:=Round(PageDX * Zoom / 100.0);
end;

function    TARPreview.GetViewPageDY:integer;
begin
  Result:=Round(PageDY * Zoom / 100.0);
end;

function    TARPreview.GetCurPage:integer;
begin
  Result:=ARPrinter.CurrPage;
end;

procedure    TARPreview.SetCurPage(ACurPage:integer);
begin
  ARPrinter.CurrPage:=ACurPage;
  Paint;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

function    TARPreview.GetPageCount:integer;
begin
  Result:=ARPrinter.PageCount;
end;

procedure   TARPreview.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);

  case key of
    VK_HOME:  begin
                if ssCtrl in Shift then begin
                  CurrPage:=0;
                end else begin
                  TopY:=0;
                end;
              end;
    VK_END :  begin
                if ssCtrl in Shift then begin
                  CurrPage:=PageCount;
                end else begin
                TopY:=100000;
                end;
              end;
    VK_PRIOR: begin
                TopY:=TopY - ClientHeight;
              end;
    VK_NEXT:  begin
                TopY:=TopY + ClientHeight;
              end;
    VK_UP:    begin
                TopY:=TopY - 20;
              end;
    VK_DOWN:  begin
                TopY:=TopY + 20;
              end;
    VK_LEFT:  begin
                CurrPage:=CurrPage - 1;
              end;
    VK_RIGHT: begin
                CurrPage:=CurrPage + 1;
              end;
    VK_ADD:   begin
                Zoom:=Zoom + 10;
              end;
    VK_SUBTRACT:begin
                Zoom:=Zoom - 10;
              end;
  end;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure   TARPreview.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := DLGC_WANTARROWS or DLGC_WANTCHARS;
end;

procedure    TARPreview.Paint;
var OrgX, OrgY:integer;
    MemDC:HDC;
    MemBitmap, OldBitmap: HBITMAP;
    MemCanvas:TCanvas;
begin
  UpdateScrolls;

  PageDX:=ARPrinter.PaperWidth;
  PageDY:=ARPrinter.PaperHeight;

  if ClientWidth >  ViewPageDX then begin
    OrgX:=(ClientWidth -  ViewPageDX) div 2
  end else begin
    OrgX:=Max(Min(-LeftX, 0),  ClientWidth - ViewPageDX);
  end;

  if ClientHeight > ViewPageDY then begin
    OrgY:=(ClientHeight - ViewPageDY) div 2
  end else begin
    OrgY:=Max(Min(-TopY, 0),  ClientHeight - ViewPageDY);
  end;

  MemBitmap       :=CreateCompatibleBitmap(Canvas.Handle, ClientRect.Right, ClientRect.Bottom);
  MemDC           :=CreateCompatibleDC(Canvas.Handle);
  MemCanvas       :=TCanvas.Create;
  MemCanvas.Handle:=MemDC;

  OldBitmap := SelectObject(MemDC, MemBitmap);
  try
    MemCanvas.Brush.Color:=clBtnFace;
    MemCanvas.FillRect(ClientRect);

    //paint white page
    MemCanvas.Brush.Color:=clGray;
    MemCanvas.FillRect (Rect(OrgX + ViewPageDX,          OrgY + 5, OrgX + ViewPageDX + 5, OrgY + ViewPageDY + 5));
    MemCanvas.FillRect (Rect(OrgX + 5,          OrgY + ViewPageDY, OrgX + ViewPageDX + 5, OrgY + ViewPageDY + 5));

    MemCanvas.Brush.Color:=clWIndow;
    MemCanvas.Rectangle(Rect(OrgX, OrgY, OrgX + ViewPageDX, OrgY + ViewPageDY));


    SetMapMode(MemDC, MM_ISOTROPIC);

    SetViewportOrgEx(MemDC, OrgX, OrgY, nil);
//    SetWindowExtEx  (MemDC,        100,        100,  nil);
//    SetViewportExtEx(MemDC, Zoom div 1, Zoom div 1,  nil);

    SetWindowExtEx  (MemDC,      1000,      1000, nil);
    SetViewportExtEx(MemDC, Zoom * 10, Zoom * 10, nil);

    if PrinterInstalled then begin
      if ShowNonPrintingArea then begin
        MemCanvas.Brush.Color:=clLtGray;
        MemCanvas.Rectangle(                   0,                   0,  ARPrinter.PaperWidth,  ARPrinter.PaperHeight);
        MemCanvas.Brush.Color:=clWhite;
        MemCanvas.Rectangle(ARPrinter.LeftMargin, ARPrinter.TopMargin,  ARPrinter.RightMargin, ARPrinter.BottomMargin);
      end;

      MemCanvas.Draw(ARPrinter.LeftMargin, ARPrinter.TopMargin, ARPrinter.Metafile);

    end else begin
      MemCanvas.Draw(0, 0, ARPrinter.Metafile);
    end;

    SetViewportOrgEx(MemDC, 0, 0, nil);
    SetWindowExtEx  (MemDC, 1, 1, nil);
    SetViewportExtEx(MemDC, 1, 1, nil);

    BitBlt(Canvas.Handle, 0, 0, ClientRect.Right, ClientRect.Bottom, MemDC, 0, 0, SRCCOPY);
  finally
    SelectObject(MemDC, OldBitmap);
    MemCanvas.Free;
    DeleteDC(MemDC);
    DeleteObject(MemBitmap);
  end;
end;

procedure   TARPreview.UpdateScrolls;
begin
  if FScrollUpdate then exit;
  FScrollUpdate:=true;

    HSPos:=LeftX;
    VSPos:=TopY;

    HSMax   :=ViewPageDX;
    HSPage  :=ClientWidth;

    VSMax   :=ViewPageDY;
    VSPage  :=ClientHeight;

  FScrollUpdate:=false;
end;

procedure    TARPreview.SetARPrinter(AARPrinter:TARPrinter);
begin
  if Assigned(FARPrinter) then begin
    FARPrinter.Free;
  end;

  FARPrinter:=AARPrinter;
end;

procedure   TARPreview.SetShowNonPrintingArea(AShowNonPrintingArea:boolean);
begin
  if FShowNonPrintingArea <> AShowNonPrintingArea then begin
    FShowNonPrintingArea:=AShowNonPrintingArea;
    Paint;

    if Assigned(FOnChangeView) then FOnChangeView(Self);
  end;
end;

procedure   TARPreview.SetZoom(AZoom:integer);
begin
  AZoom:=Min(ARPreviewZoomMax, Max(ARPreviewZoomMin, AZoom));
  if FZoom <> AZoom then begin
    FZoom:=AZoom;
    FZoomMode:=zmFixed;
    Paint;

    if Assigned(FOnChangeView) then FOnChangeView(Self);
  end;
end;

procedure   TARPreview.ZoomToWidth;
begin
  Zoom:=Round(100 * ClientWidth / ARPrinter.PaperWidth) - 3;
  FZoomMode:=zmToWidth;
end;

procedure   TARPreview.ZoomToFit;
begin
  Zoom:=Min(Round(100 * ClientWidth / ARPrinter.PaperWidth), Round(100 * ClientHeight / ARPrinter.PaperHeight)) - 2;
  FZoomMode:=zmToFit;
end;

procedure   TARPreview.WMSize(var Message: TWMSize);
begin
  inherited;
  Realign;

  VSPage:=ClientHeight;
  HSPage:=ClientWidth;

  case FZoomMode of
    zmToFit:  ZoomToFit;
    zmToWidth:ZoomToWidth;
  end;

  if Assigned(FOnChangeView) then FOnChangeView(Self);
end;

procedure   TARPreview.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result:=1;
end;

procedure   TARPreview.MouseWheelUp(Sender:TObject; Shift:TShiftState; MousePos:TPoint; var Handled:boolean);
begin
  TopY:=TopY - 20;
end;

procedure   TARPreview.MouseWheelDown(Sender:TObject; Shift:TShiftState; MousePos:TPoint; var Handled:boolean);
begin
  TopY:=TopY + 20;
end;

const OldY:integer = -1000;
      OldX:integer = -1000;

procedure   TARPreview.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  OldX:=X;
  OldY:=Y;
end;

procedure   TARPreview.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  if ssLeft in Shift then begin
    if (OldX <> X) or (OldY <> Y) then begin
      LeftX :=LeftX + (OldX - X);
      TopY  :=TopY  + (OldY - Y);

      OldX:=X;
      OldY:=Y;
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('ARComponents', [TARPreview]);
end;

end.


