unit ARRep;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

{$INCLUDE DEFINES.INC}

interface

uses
  Classes, Windows, SysUtils, Math, Graphics, 

  ARRepCommon, ARUtils, ARPrinter, ARVFS,

  Printers;

const
  CellCnt :integer = 0;
  BandCnt :integer = 0;
  RepCnt  :integer = 0;

type
  TColorArray = array[Low(TFrameStyles) .. High(TFrameStyles)] of TColor;
  TWidthsArray= array[Low(TFrameStyles) .. High(TFrameStyles)] of integer;

  TARTemplate = class(TAIStringList)
  end;

  TARBandList = class;
  TARBand = class;
////////////////// TARCell //////////////////

  TARCellOption = (
    arcoWordWrap,
    arcoAutoStretch,
    arcoAutoSize,
    arcoTile,
    arcoSelected,
    arcoFocused,
    arcoChanged,
    arcoIsMoneyValue,
    arcoIsRTF,
    arcoShared,
    arcoInUse
  );

  TARCellOptions = set of TARCellOption;

  TARCell    = class(TPersistent)
  private
    FPicture      :TPicture;

    FFrameColors  :TColorArray;
    FFrameWidths  :TWidthsArray;

    FFontName     :TFontName;
    FFontSize     :integer;
    FFontColor    :TColor;
    FFontStyle    :TFontStyles;
    FTextAngle    :integer;

    FWidth        :integer;
    FHAlign       :THAlign;
    FVAlign       :TVAlign;
    FColor        :TColor;
    FValue        :string;

    Options       :TARCellOptions;

    FUserVar      :string;
    FUserVarValue :integer;

    FOnChange     :TNotifyEvent;

    InUpdateCnt   :integer;

    FARTemplate   :TARTemplate;

    PrintPos      :integer;
    TextPos       :integer;

    FLinkPictureToFile:boolean;
    FPictureFileName:string;

    function      GetLinkPictureToFile:boolean;
    procedure     SetLinkPictureToFile(ALinkPictureToFile:boolean);
    function      GetPictureFileName:string;
    procedure     SetPictureFileName(const APictureFileName:string);

    function      GetPicture:TPicture;
    procedure     SetPicture(APicture:TPicture);

    function      GetS:string;
    procedure     SetS(const AValue:string);

    function      GetF:extended;
    procedure     SetF(AValue:extended);

    function      GetI:integer;
    procedure     SetI(AValue:integer);

    function      GetD:TDateTime;
    procedure     SetD(AValue:TDateTime);

    function      GetT:TDateTime;
    procedure     SetT(AValue:TDateTime);

    function      GetC:currency;
    procedure     SetC(AValue:currency);

    function      GetM:currency;
    procedure     SetM(AValue:currency);

    function      GetG:TGraphic;
    procedure     SetG(AG:TGraphic);

    function      GetFrameStyle:TFrameStyle;
    function      GetFrameColor:TColor;
    function      GetFrameWidth:integer;

    procedure     SetFrameStyle(AFrameStyle:TFrameStyle);
    procedure     SetFrameColor(AFrameColor:TColor);
    procedure     SetFrameWidth(AFrameWidth:integer);

    procedure     SetFrameColors(index:TFrameStyles; AFrameColor:TColor);
    procedure     SetFrameWidths(index:TFrameStyles; AFrameWidth:integer);

    function      GetFrameColors(index:TFrameStyles):TColor;
    function      GetFrameWidths(index:TFrameStyles):integer;

    procedure     SetFontName(const AFontName:TFontName);
    procedure     SetFontSize(AFontSize:integer);
    procedure     SetFontColor(AFontColor:TColor);
    procedure     SetFontStyle(AFontStyle:TFontStyles);
    procedure     SetTextAngle(ATextAngle:integer);

    procedure     SetWidth(AWidth:integer);
    procedure     SetHAlign(AHAlign:THAlign);
    procedure     SetVAlign(AVAlign:TVAlign);
    procedure     SetColor(AColor:TColor);
    function      GetWordWrap:boolean;
    procedure     SetWordWrap(AWordWrap:boolean);
    function      GetAutoStretch:boolean;
    procedure     SetAutoStretch(AAutoStretch:boolean);
    function      GetAutoSize:boolean;
    procedure     SetAutoSize(AAutoSize:boolean);
    function      GetIsRTF:boolean;
    procedure     SetIsRTF(AIsRTF:boolean);
    function      GetTile:boolean;
    procedure     SetTile(ATile:boolean);
    procedure     SetValue(const AValue:string);
    function      GetSelected:boolean;
    procedure     SetSelected(ASelected:boolean);
    function      GetFocused:boolean;
    procedure     SetFocused(AFocused:boolean);
    function      GetShared:boolean;
    procedure     SetShared(AShared:boolean);
    function      GetInUse:boolean;
  protected
    function      InUpdate:boolean;
    procedure     Invalidate;
    function      GetARTemplate:TARTemplate;
    procedure     SetARTemplate(const AARTemplate:TARTemplate);
    procedure     ARTemplateChange(Sender:TObject);
    procedure     PaintGraphic(g:TGraphic; Canvas:TCanvas; r:TRect);
    procedure     PaintFrames(Canvas:TCanvas; r:TRect);

    function      GetTopCell:TARCell;
    function      GetBottomCell:TARCell;
  public
    Owner         :TARBand;
    Changed       :boolean;

    VarNameList   :TAIStringList;
    IsMoneyValue  :boolean;
    property      FrameColors[index:TFrameStyles]:TColor  read GetFrameColors write SetFrameColors;
    property      FrameWidths[index:TFrameStyles]:integer read GetFrameWidths write SetFrameWidths;
    procedure     CreateVarNameList;

    function      Height(Canvas:TCanvas):integer;

    constructor   Create;
    destructor    Destroy;override;
    procedure     Assign(ACell:TARCell);reintroduce;

    procedure     BeginUpdate;
    procedure     EndUpdate;

    property      Focused:boolean           read  GetFocused     write  SetFocused;

    function      GetTopBand:TARBand;
    function      GetBottomBand:TARBand;
    procedure     PaintCell(Canvas:TCanvas; r:TRect; ClipRect:TRect; SelectedCellStyle:TSelectedCellStyle; FocusedCellStyle:TFocusedCellStyle);
    procedure     ProcessValue;
    property      InUse             :boolean      read  GetInUse;
    function      ContinuePrint(Canvas:TCanvas; r:TRect; ClipRect:TRect):integer;
    procedure     PreparePrint;
  published
    property      ARTemplate:TARTemplate          read  GetARTemplate         write SetARTemplate;

    property      FrameStyle        :TFrameStyle  read  GetFrameStyle         write SetFrameStyle;
    property      FrameColor        :TColor       read  GetFrameColor         write SetFrameColor;
    property      FrameWidth        :integer      read  GetFrameWidth         write SetFrameWidth;

    property      FontName          :TFontName    read  FFontName             write  SetFontName;
    property      FontSize          :integer      read  FFontSize             write  SetFontSize;
    property      FontColor         :TColor       read  FFontColor            write  SetFontColor;
    property      FontStyle         :TFontStyles  read  FFontStyle            write  SetFontStyle;
    property      TextAngle         :integer      read  FTextAngle            write  SetTextAngle;

    property      Width             :integer      read  FWidth                write  SetWidth;
    property      HAlign            :THAlign      read  FHAlign               write  SetHAlign;
    property      VAlign            :TVAlign      read  FVAlign               write  SetVAlign;
    property      Color             :TColor       read  FColor                write  SetColor;
    property      WordWrap          :boolean      read  GetWordWrap           write  SetWordWrap;
    property      AutoStretch       :boolean      read  GetAutoStretch        write  SetAutoStretch;
    property      AutoSize          :boolean      read  GetAutoSize           write  SetAutoSize;
    property      Tile              :boolean      read  GetTile               write  SetTile;
    property      Value             :string       read  FValue                write  SetValue;
    property      Selected          :boolean      read  GetSelected           write  SetSelected;
    property      Picture           :TPicture     read  GetPicture            write  SetPicture;
    property      LinkPictureToFile :boolean      read  GetLinkPictureToFile  write  SetLinkPictureToFile;
    property      PictureFileName   :string       read  GetPictureFileName    write  SetPictureFileName;

    property      IsRTF             :boolean      read  GetIsRTF              write  SetIsRTF;
    property      Shared            :boolean      read  GetShared             write  SetShared;

    property      S:string                        read  GetS                  write  SetS;
    property      F:extended                      read  GetF                  write  SetF;
    property      I:integer                       read  GetI                  write  SetI;
    property      D:TDateTime                     read  GetD                  write  SetD;
    property      T:TDateTime                     read  GetT                  write  SetT;
    property      C:currency                      read  GetC                  write  SetC;
    property      M:currency                      read  GetM                  write  SetM;
    property      G:TGraphic                      read  GetG                  write  SetG;


    property      UserVar           :string       read  FUserVar              write  FUserVar;
    property      UserVarValue      :integer      read  FUserVarValue         write  FUserVarValue;

    property      OnChange          :TNotifyEvent read  FOnChange             write FOnChange;
  end;
////////////////////////////////////////////////////////// TARBandList
  TARBandList = class(TPersistent)
  private
    FOwner        :TARBandList;
    FTop          :integer;
    FHeight       :integer;
    FBands        :TAIList;
    FExpanded     :boolean;

    FName         :string;
    FMark         :TBitmap;
    FID           :integer;

    FOnChange     :TNotifyEvent;
    FOnBandChange :TARBandChangeEvent;
    FOnCellChange :TARCellChangeEvent;

    InUpdateCnt   :integer;
    Changed       :boolean;

    FARTemplate   :TARTemplate;

    procedure     SetOwner(AOwner:TARBandList);
    function      GetMarked:boolean;
    procedure     SetMarked(AMarked:boolean);
    function      GetMark:TBitmap;
    procedure     SetMark(AMark:TBitmap);
    procedure     SetName(const AName:string);
    function      GetExpanded:boolean;virtual;
    procedure     SetExpanded(AExpanded:boolean);virtual;
    function      GetExpandable:boolean;
    function      GetItems(index:integer):TARBandList;
    function      GetBandCount:integer;
    procedure     Invalidate;virtual;
    procedure     SetHeight(AHeight:integer);
    procedure     ReculcTops(idxband:integer = 0);
  protected
    procedure     Expand;
    procedure     Collapse;
    function      InUpdate:boolean;
    function      GetARTemplate:TARTemplate;virtual;
    procedure     SetARTemplate(const AARTemplate:TARTemplate);
    procedure     ARTemplateChange(Sender:TObject);virtual;
  public
    property      Owner:TARBandList               read FOwner         write SetOwner;

    constructor   Create;
    destructor    Destroy;override;

    procedure     CellChange(band:TObject; cell:TObject);
    procedure     BandChange(band:TObject);

    function      IndexOf(ABand:TARBandList):integer;

    procedure     Add(ABand:TARBandList);
    procedure     Delete(index: Integer);
    procedure     DeleteAll;
    function      Move(CurIndex, NewIndex: Integer):boolean;
    procedure     Insert(index: Integer; ABand:TARBandList);

    procedure     BeginUpdate;virtual;
    procedure     EndUpdate;virtual;
    property      Items[idx:integer]:TARBandList  read  GetItems;default;
  published
    property      ARTemplate:TARTemplate          read  GetARTemplate write SetARTemplate;

    property      Name:string                     read  FName         write SetName;
    property      BandCount:integer               read  GetBandCount;
    property      Expanded:boolean                read  GetExpanded   write SetExpanded;
    property      Expandable:boolean              read  GetExpandable;
    property      Marked:boolean                  read  GetMarked     write SetMarked;
    property      Mark:TBitmap                    read  GetMark       write SetMark;
    property      ID:integer                      read  FID           write FID;
    property      Height:integer                  read  FHeight       write SetHeight;

    property      OnChange:TNotifyEvent           read  FOnChange     write FOnChange;
    property      OnBandChange:TARBandChangeEvent read  FOnBandChange write FOnBandChange;
    property      OnCellChange:TARCellChangeEvent read  FOnCellChange write FOnCellChange;
  end;
////////////////////////////////////////////////////////// TARBand
  TARBand = class (TARBandList)
  private
    CellList      :TAIList;
    FBandState    :TARBandState;

    function      GetSelected:boolean;
    procedure     SetSelected(ASelected:boolean);

    function      GetWidth:integer;
    function      GetCell(idxcell:integer):TARCell;
    procedure     SetCell(idxcell:integer; ACell:TARCell);

    function      GetCellCount:integer;
    function      GetLefts(index:integer):integer;
    function      GetRights(index:integer):integer;
    procedure     CellChange(Sender:TObject);
  protected
    function      GetARTemplate:TARTemplate;override;
    procedure     ARTemplateChange(Sender:TObject);override;
  public
    constructor   Create;
    destructor    Destroy;override;

    procedure     Assign(ABand:TARBand);reintroduce;
    function      NewCell:TARCell;
    procedure     DeleteCells;
    procedure     DeleteCell(idxcell:integer);
    procedure     Split(ACell:TARCell);
    procedure     Move(idxcellfrom:integer;idxcellto:integer);
    function      IndexOf(ACell:TARCell):integer;
    property      Cells[index:integer]:TARCell    read  GetCell     write SetCell;
    property      Lefts[index:integer]:integer    read  GetLefts;
    property      Rights[index:integer]:integer   read  GetRights;
  published
    property      Width:integer                   read  GetWidth;
    property      CellCount:integer               read  GetCellCount;

    property      BandState:TARBandState          read  FBandState  write FBandState;

    property      Selected:boolean                read  GetSelected write SetSelected;
  end;
////////////////////////////////////////////////////////// TARRep
  TARRep  = class (TARBandList)
  private
    FUserVarList:TAIStringList;
    FFont       :TFont;

    procedure   SetFont(AFont:TFont);
    procedure   SetUserVarList(AUserVarList:TAIStringList);

    function    GetBand(idxband:integer):TARBand;
    function    GetCell(idxband:integer;idxcell:integer):TARCell;
    function    GetTops(idxband:integer):integer;

  protected
    function      GetARTemplate:TARTemplate;override;
    procedure     ARTemplateChange(Sender:TObject);override;
  public
    Orientation :TPrinterOrientation;
    PaperSize   :TARPaperSize;
    LeftMargin  :extended;
    TopMargin   :extended;
    RightMargin :extended;
    BottomMargin:extended;

    FileSystem  :TVirtualFileSystem;

{$WARNINGS OFF}
    property    Bands[idxband:integer]:TARBand                  read  GetBand;
{$WARNINGS ON}
    property    Cells[idxband:integer;idxcell:integer]:TARCell  read  GetCell;
    property    Tops[idxband:integer]:integer                   read  GetTops;

    property    UserVarList:TAIStringList                       read  FUserVarList  write SetUserVarList;

    procedure   Save(AFileName:string);
    procedure   Load(AFileName:string);

    constructor Create;
    destructor  Destroy;override;

    function    NewBand:TARBand;
    procedure   DeleteBand(idxband:integer);
    function    MaxHeight(ABand:TARBand):integer;

  published
    property    Font:TFont                      read FFont         write SetFont;
    property    OnBandChange:TARBandChangeEvent read FOnBandChange write FOnBandChange;
    property    OnCellChange:TARCellChangeEvent read FOnCellChange write FOnCellChange;
  end;

var
  ShowHiddenItems:boolean;

implementation
////////////////// TARCell ////////////////////////////////////
constructor  TARCell.Create;
begin
  inc(CellCnt);

  BeginUpdate;
    FrameStyle  :=DefFrameStyle;
    FrameColor  :=DefFrameColor;
    FrameWidth  :=DefFrameWidth;

    FontName    :=DefFontName;
    FontSize    :=DefFontSize;
    FontColor   :=DefFontColor;
    FontStyle   :=DefFontStyle;
    TextAngle   :=0;
    
    Color       :=DefBkColor;
    HAlign      :=DefHAlign;
    VAlign      :=DefVAlign;
    Width       :=DefWidth;
    WordWrap    :=DefWordWrap;
    AutoStretch :=DefAutoStretch;
    Tile        :=DefTile;
    IsRTF       :=false;
    Shared      :=false;

    PictureFileName   :='';
    LinkPictureToFile :=true;

    Value       :=DefCellValue;
  EndUpdate;
end;

destructor  TARCell.Destroy;
begin
  if Assigned(FARTemplate)  then  FARTemplate.Free;
  if Assigned(FPicture)     then  FPicture.Free;
  if Assigned(VarNameList)  then  VarNameList.Free;
  dec(CellCnt);

  inherited
end;

procedure   TARCell.Assign(ACell:TARCell);
var i:TFrameStyles;
begin
  BeginUpdate;
    FrameStyle  :=ACell.FrameStyle;
    for i:=Low(TFrameStyles) to High(TFrameStyles) do begin
      FrameColors[i]:=ACell.FrameColors[i];
      FrameWidths[i]:=ACell.FrameWidths[i];
    end;

    FontName    :=ACell.FontName;
    FontSize    :=ACell.FontSize;
    FontColor   :=ACell.FontColor;
    FontStyle   :=ACell.FontStyle;
    TextAngle   :=ACell.TextAngle;
    
    Color       :=ACell.Color;
    HAlign      :=ACell.HAlign;
    VAlign      :=ACell.VAlign;
    Width       :=ACell.Width;
    WordWrap    :=ACell.WordWrap;
    AutoStretch :=ACell.AutoStretch;
    Tile        :=ACell.Tile;

    UserVar     :=ACell.UserVar;
    UserVarValue:=ACell.UserVarValue;
    AutoSize    :=ACell.AutoSize;

    Value       :=ACell.Value;

    IsRTF       :=ACell.IsRTF;
    Shared      :=ACell.Shared;

    PictureFileName   :=ACell.PictureFileName;
    LinkPictureToFile :=ACell.LinkPictureToFile;

    if Assigned(ACell.FPicture) then begin
      Picture.Assign(ACell.FPicture);
    end;
  EndUpdate;
end;

procedure   TARCell.CreateVarNameList;
var p:integer;
    v:string;
begin
  if not Assigned(VarNameList) then VarNameList:=TAIStringList.Create('TARCell.VarNameList');
  VarNameList.Clear;

  v:=Value;

  while Length(v) > 0 do begin
    p:=System.Pos(BeginVarToken, v);
    if p = 0 then begin
      VarNameList.AddObject(v, pointer(0));//simply text
      break;
    end else begin
      if p > 1 then begin
        VarNameList.AddObject(System.Copy(v, 1, p - 1), pointer(0));//simply text
      end;

      System.Delete(v, 1, p);                     // and delete
                                                  // start var def
      p:=System.Pos(EndVarToken, v);
      if p = 0 then begin                         // unfinished var def
        break;
      end else begin
        VarNameList.AddObject(System.Copy(v, 1, p-1), pointer(1));//var name
        System.Delete(v, 1, p);
      end;
    end;
  end;
end;

function    TARCell.GetPicture:TPicture;
begin
  if not Assigned(FPicture) then FPicture:=TPicture.Create;
  Result:=FPicture;
end;

function    TARCell.GetLinkPictureToFile:boolean;
begin
  Result:=FLinkPictureToFile;
end;

function    TARCell.GetPictureFileName:string;
begin
  Result:=FPictureFileName;
end;

procedure   TARCell.SetPicture(APicture:TPicture);
begin
  BeginUpdate;
    if not Assigned(FPicture) then FPicture:=TPicture.Create;
    FPicture.Assign(APicture);
    Changed:=true;
  EndUpdate;
end;

procedure   TARCell.SetLinkPictureToFile(ALinkPictureToFile:boolean);
begin
  BeginUpdate;
    FLinkPictureToFile:=ALinkPictureToFile;
    Changed:=true;
  EndUpdate;
end;

procedure   TARCell.SetPictureFileName(const APictureFileName:string);
begin
  if FPictureFileName <> Trim(APictureFileName) then begin
    BeginUpdate;

      FPictureFileName:=Trim(APictureFileName);

      if FileExists(FPictureFileName) then begin
        Picture.LoadFromFile(FPictureFileName);
      end else begin
        if Assigned(FPicture) then begin
          FPicture.Free;
          FPicture:=nil;
        end;
      end;
      Changed:=true;

    EndUpdate;
  end;
end;

procedure   TARCell.ProcessValue;
var  fname:string;
begin
  if (Length(Value) > 0) and (Value[1] = '{') and (Value[Length(Value)] = '}') then begin
    fname:=Value;
    fname:=System.Copy(Value, 2, Length(Value) - 2);

    if FileExists(fname) then begin
      Picture.LoadFromFile(fname);
    end else begin
      if Assigned(FPicture) then begin
        FPicture.Free;
        FPicture:=nil;
      end;
    end;
  end else begin
  end;
end;

procedure TARCell.PaintGraphic(g:TGraphic; Canvas:TCanvas; r:TRect);
var xx, yy:integer;
    x,y:integer;
begin
  if AutoSize then begin
    Canvas.StretchDraw(r, g);
  end else begin
    if Tile then begin
      for xx:=0 to ((r.Right - r.Left) div (g.Width - 1)) + 1 do begin
        for yy:=0 to ((r.Bottom - r.Top) div (g.Height - 1)) + 1 do begin
          Canvas.Draw(r.Left + xx * (g.Width - 1), r.Top + yy * (g.Height - 1), g);
        end;
      end;
    end else begin
      case VAlign of
        avTop   : y:=r.Top;
        avCenter: y:=r.Top + (r.Bottom - r.Top - g.Height) div 2;
        avBottom: y:=r.Bottom - g.Height;
        else      y:=r.Top;
      end;

      case HAlign of
        ahLeft  : x:=r.Left;
        ahCenter: x:=r.Left + (r.Right - r.Left - g.Width) div 2;
        ahRight : x:=r.Right - g.Width;
        else      x:=r.Left;
      end;
      Canvas.Draw(x, y, g);
    end;
  end;
end;

procedure    TARCell.PaintFrames(Canvas:TCanvas; r:TRect);
begin
  if frLeft   in FrameStyle then begin
    Canvas.Brush.Color:=FrameColors[frLeft];    Canvas.FillRect(Rect(r.Left, r.Top, r.Left + FrameWidths[frLeft], r.Bottom));
  end;
  if frRight  in FrameStyle then begin
    Canvas.Brush.Color:=FrameColors[frRight];   Canvas.FillRect(Rect(r.Right - FrameWidths[frRight], r.Top, r.Right, r.Bottom));
  end;
  if frTop    in FrameStyle then begin
    Canvas.Brush.Color:=FrameColors[frTop];     Canvas.FillRect(Rect(r.Left, r.Top, r.Right, r.Top + FrameWidths[frTop]));
  end;
  if frBottom in FrameStyle then begin
    Canvas.Brush.Color:=FrameColors[frBottom];  Canvas.FillRect(Rect(r.Left, r.Bottom - FrameWidths[frBottom], r.Right, r.Bottom));
  end;
end;

procedure    TARCell.PaintCell(Canvas:TCanvas; r:TRect; ClipRect:TRect; SelectedCellStyle:TSelectedCellStyle; FocusedCellStyle:TFocusedCellStyle);
var rf:TRect;
    cell:TARCell;
begin
  if r.Right <= 0 then exit;

  try
    if InUse then begin
      cell:=GetTopCell;
      if cell <> Self then begin
        GetTopCell.PaintCell(Canvas, r, ClipRect, SelectedCellStyle, FocusedCellStyle);
      end;
    end else begin
      if Selected and (SelectedCellStyle = scsSelected) then begin
        Canvas.Brush.Color:=clHighlight;
      end else begin
        Canvas.Brush.Color:=Color;
      end;

      if Shared then begin
        Canvas.Pen.Color:=clBlack;
        Canvas.Rectangle(r.Left, r.Top, r.Left + Width, r.Bottom);
        Canvas.MoveTo(r.Left,         r.Top); Canvas.LineTo(r.Left + Width, r.Bottom);
        Canvas.MoveTo(r.Left + Width, r.Top); Canvas.LineTo(r.Left,         r.Bottom);
      end else begin
        if Selected and (SelectedCellStyle = scsSelected) then begin
          Canvas.Brush.Color:=clHighlight;
        end else begin
          Canvas.Brush.Color:=Color;
        end;
        Canvas.FillRect(Rect(r.Left, r.Top, r.Left + Width, r.Bottom));
        rf.Left  :=r.Left   + FrameWidth;
        rf.Top   :=r.Top    + FrameWidth;
        rf.Right :=r.Right  - FrameWidth;
        rf.Bottom:=r.Bottom - FrameWidth;

        // draw picture
        if Assigned(FPicture) and Assigned(FPicture.Graphic) then begin
          PaintGraphic(FPicture.Graphic, Canvas, rf);
        end;

        // draw text
        if Canvas.Font.Name  <> FontName  then Canvas.Font.Name :=FontName;
        if Canvas.Font.Size  <> FontSize  then Canvas.Font.Size :=FontSize;

        if Selected and (SelectedCellStyle = scsSelected) then begin
          if Canvas.Font.Color <> clHighlightText then Canvas.Font.Color:=clHighlightText;
        end else begin
          if Canvas.Font.Color <> FontColor then Canvas.Font.Color:=FontColor;
        end;

        if Canvas.Font.Style <> FontStyle then Canvas.Font.Style:=FontStyle;

        DrawRectText(Canvas, Value, rf,  ClipRect, HAlign, VAlign, WordWrap, IsRTF, TextAngle);

        // draw frames
        if ShowHiddenItems then begin
          Canvas.Brush.Color:=clLtGray;
          Canvas.Brush.Style:=bsDiagCross;
          Canvas.FrameRect(r);
          Canvas.Brush.Style:=bsSolid;
        end;

        PaintFrames(Canvas, r);

        if Selected then begin
          case SelectedCellStyle of
            scsInvert:  begin // inverted rect
                          Canvas.DrawFocusRect(Rect(r.Left, r.Top, r.Right - 1, r.Bottom - 1));
                          Canvas.DrawFocusRect(Rect(r.Left + 1, r.Top + 1, r.Right - 2, r.Bottom - 2));
                          Canvas.DrawFocusRect(Rect(r.Left + 2, r.Top + 2, r.Right - 3, r.Bottom - 3));
                          Canvas.CopyMode:=cmDstInvert;
                          Canvas.CopyRect(r, Canvas, r);
                          Canvas.CopyMode:=cmSrcCopy;;
                        end;
            scsGray:    begin // gray rect
                          Canvas.Brush.Color:=$02181818;//$02444444;
                          Canvas.CopyMode:=cmPatInvert;
                          r.Left:=r.Left + 1;
                          r.Top :=r.Top  + 1;
                          Canvas.CopyRect(r, Canvas, r);
                          Canvas.CopyMode:=cmSrcCopy;;
                        end;
          end;
        end;

        if Focused then begin
          case FocusedCellStyle of
            fcsRect:      begin
                            Canvas.Pen.Color:=clGray;
                            Canvas.Pen.Width:=3;
                            Canvas.Brush.Style:=bsClear;
                            Canvas.RoundRect(r.Left + 0, r.Top + 0, r.Right - 1, r.Bottom - 1, 0, 0);
                            Canvas.Brush.Style:=bsSolid;
                            Canvas.Pen.Width:=1;
                          end;
            fcsRoundRect: begin
                            Canvas.Pen.Color:=clGray;
                            Canvas.Pen.Width:=3;
                            Canvas.Brush.Style:=bsClear;
                            Canvas.RoundRect(r.Left + 0, r.Top + 0, r.Right - 1, r.Bottom - 1, 10, 10);
                            Canvas.Brush.Style:=bsSolid;
                            Canvas.Pen.Width:=1;
                          end;
          end;
        end;
      end;
    end;
  finally
  end;
end;

function  TARCell.Height(Canvas:TCanvas):integer;
begin
  Result:=TextHeight(Canvas, Value, Width - FrameWidth * 2, HAlign, VAlign, WordWrap, IsRTF, TextPos) + FrameWidth * 2 + 2;
  if AutoStretch and Assigned(FPicture) then begin
    Result:=Max(Result, Picture.Height - PrintPos);
  end;
end;

procedure   TARCell.PreparePrint;
begin
  PrintPos:=0;
  TextPos :=0;
end;

function    TARCell.ContinuePrint(Canvas:TCanvas; r:TRect; ClipRect:TRect):integer;
var rf:TRect;
begin
  r.Top   :=r.Top    - PrintPos;

  Result:=Min(r.Bottom, ClipRect.Bottom);

  if r.Right <= 0 then exit;

  if not InUse then begin
    Canvas.Brush.Color:=Color;
    if Shared then begin
      Canvas.Pen.Color:=clBlack;
      Canvas.Rectangle(r.Left, r.Top, r.Left + Width, r.Bottom);
      Canvas.MoveTo(r.Left,         r.Top); Canvas.LineTo(r.Left + Width, r.Bottom);
      Canvas.MoveTo(r.Left + Width, r.Top); Canvas.LineTo(r.Left,         r.Bottom);
    end else begin
      Canvas.Brush.Color:=Color;
      Canvas.FillRect(Rect(r.Left, r.Top, r.Left + Width, r.Bottom));

      rf.Left:=r.Left + FrameWidth; rf.Right :=r.Right  - FrameWidth;

      rf.Top :=r.Top  + FrameWidth; rf.Bottom:=r.Bottom - FrameWidth;

      // draw picture
      if Assigned(FPicture) and Assigned(FPicture.Graphic) then begin
        PaintGraphic(FPicture.Graphic, Canvas, rf);
      end;

      // draw text
      if Canvas.Font.Name   <> FontName   then Canvas.Font.Name  :=FontName;
      if Canvas.Font.Size   <> FontSize   then Canvas.Font.Size  :=FontSize;
      if Canvas.Font.Color  <> FontColor  then Canvas.Font.Color :=FontColor;
      if Canvas.Font.Style  <> FontStyle  then Canvas.Font.Style :=FontStyle;

      PrintRectText(Canvas, Value, rf,  ClipRect, HAlign, VAlign, WordWrap, IsRTF, TextPos);

      PaintFrames(Canvas, r);
    end;
    PrintPos:=Max(0, ClipRect.Bottom - r.Top);
  end;
end;

function    TARCell.GetS:string;
begin
  Result:=Value;
end;
procedure   TARCell.SetS(const AValue:string);
begin
  Value:=AValue;
end;
function    TARCell.GetF:extended;
begin
  Result:=StrToFloat(Value);
end;
procedure   TARCell.SetF(AValue:extended);
begin
  Value:=Format('%f', [AValue]);
end;
function    TARCell.GetI:integer;
begin
  Result:=StrToInt(Value);
end;
procedure   TARCell.SetI(AValue:integer);
begin
  Value:=IntToStr(AValue);
end;
function    TARCell.GetD:TDateTime;
begin
  Result:=StrToDate(Value);
end;
procedure   TARCell.SetD(AValue:TDateTime);
begin
  Value:=DateToStr(AValue);
end;
function    TARCell.GetT:TDateTime;
begin
  Result:=StrToTime(Value);
end;
procedure   TARCell.SetT(AValue:TDateTime);
begin
  Value:=TimeToStr(AValue);
end;
function    TARCell.GetC:currency;
begin
  Result:=StrToCurr(Value);
end;
procedure   TARCell.SetC(AValue:currency);
begin
  Value:=CurrToStr(AValue);
end;
function    TARCell.GetM:currency;
var p:integer;
    s:string;
begin
  s:=Value;
  p:=System.Pos(ARMoneySeparator, s);
  if p > 0 then begin
    s[p]:=DecimalSeparator;
  end;
  Result:=StrToCurr(s);
end;
procedure   TARCell.SetM(AValue:currency);
var OldDecSep:char;
begin
  OldDecSep:=DecimalSeparator;
  DecimalSeparator:=ARMoneySeparator;
  Value:=Format('%.2f', [AValue]);
  DecimalSeparator:=OldDecSep;
end;

function      TARCell.GetG:TGraphic;
begin
  Result:=Picture.Graphic;
end;

procedure     TARCell.SetG(AG:TGraphic);
begin
  Picture.Assign(AG);
end;

procedure     TARCell.SetFrameStyle(AFrameStyle:TFrameStyle);
var index:TFrameStyles;
begin
  BeginUpdate;

    for index:=Low(TFrameStyles) to High(TFrameStyles) do begin
      if index in AFrameStyle then begin
        if FFrameWidths[index] = 0 then FFrameWidths[index]:=1;
      end else begin
        if FFrameWidths[index] > 0 then FFrameWidths[index]:=0;
      end;
    end;
    Changed:=true;

  EndUpdate;
end;

function      TARCell.GetFrameStyle:TFrameStyle;
var index:TFrameStyles;
begin
  Result:=[];
  for index:=Low(TFrameStyles) to High(TFrameStyles) do begin
    if FFrameWidths[index] > 0 then Result:=Result + [index];
  end;
end;

function      TARCell.GetFrameColor:TColor;
begin
  if  (FFrameColors[frLeft]  = FFrameColors[frTop]) and
      (FFrameColors[frTop]   = FFrameColors[frRight]) and
      (FFrameColors[frRight] = FFrameColors[frBottom]) then begin
    Result:=FFrameColors[frLeft];
  end else begin
    Result:=clNone;
  end;
end;

procedure     TARCell.SetFrameColor(AFrameColor:TColor);
var index:TFrameStyles;
begin
  if FrameColor <> AFrameColor then begin
    BeginUpdate;
      for index:=Low(TFrameStyles) to High(TFrameStyles) do begin
        FFrameColors[index]:=AFrameColor;
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetFrameWidth:integer;
begin
  if  (FFrameWidths[frLeft]   = FFrameWidths[frTop]) and
      (FFrameWidths[frTop]    = FFrameWidths[frRight]) and
      (FFrameWidths[frRight]  = FFrameWidths[frBottom]) then begin
    Result:=FFrameWidths[frLeft];
  end else begin
    Result:=1;
  end;
end;

procedure     TARCell.SetFrameWidth(AFrameWidth:integer);
var index:TFrameStyles;
begin
  BeginUpdate;
    for index:=Low(TFrameStyles) to High(TFrameStyles) do begin
      if FFrameWidths[index] > 0 then begin
        FFrameWidths[index]:=AFrameWidth;
      end;
    end;
    Changed:=true;
  EndUpdate;
end;

procedure     TARCell.SetFrameColors(index:TFrameStyles; AFrameColor:TColor);
begin
  if FFrameColors[index] <> AFrameColor then begin
    BeginUpdate;
      FFrameColors[index]:=AFrameColor;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetFrameWidths(index:TFrameStyles; AFrameWidth:integer);
begin
  if FFrameWidths[index] <> AFrameWidth then begin
    BeginUpdate;
      FFrameWidths[index]:=AFrameWidth;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetFrameColors(index:TFrameStyles):TColor;
begin
  Result:=FFrameColors[index];
end;

function      TARCell.GetFrameWidths(index:TFrameStyles):integer;
begin
  Result:=FFrameWidths[index];
end;

procedure     TARCell.SetFontName(const AFontName:TFontName);
begin
  if FFontName <> AFontName then begin
    BeginUpdate;
      FFontName:=AFontName;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetFontSize(AFontSize:integer);
begin
  if FFontSize <> AFontSize then begin
    BeginUpdate;
      FFontSize:=AFontSize;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetFontColor(AFontColor:TColor);
begin
  if FFontColor <> AFontColor then begin
    BeginUpdate;
      FFontColor:=AFontColor;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetFontStyle(AFontStyle:TFontStyles);
begin
  if FFontStyle <> AFontStyle then begin
    BeginUpdate;
      FFontStyle:=AFontStyle;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetTextAngle(ATextAngle:integer);
begin
  if FTextAngle <> ATextAngle then begin
    BeginUpdate;
      FTextAngle:=ATextAngle;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetWidth(AWidth:integer);
begin
  if FWidth <> AWidth then begin
    BeginUpdate;
      FWidth:=AWidth;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetHAlign(AHAlign:THAlign);
begin
  if FHAlign <> AHAlign then begin
    BeginUpdate;
      FHAlign:=AHAlign;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetVAlign(AVAlign:TVAlign);
begin
  if FVAlign <> AVAlign then begin
    BeginUpdate;
      FVAlign:=AVAlign;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetColor(AColor:TColor);
begin
  if FColor <> AColor then begin
    BeginUpdate;
      FColor:=AColor;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetWordWrap(AWordWrap:boolean);
begin
  if AWordWrap <> (arcoWordWrap in Options) then begin
    BeginUpdate;
      if AWordWrap then begin
        Options:=Options + [arcoWordWrap];
      end else begin
        Options:=Options - [arcoWordWrap];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetWordWrap:boolean;
begin
  Result:=arcoWordWrap in Options;
end;

procedure     TARCell.SetAutoStretch(AAutoStretch:boolean);
begin
  if AAutoStretch <> (arcoAutoStretch in Options) then begin
    BeginUpdate;
      if AAutoStretch then begin
        Options:=Options + [arcoAutoStretch];
      end else begin
        Options:=Options - [arcoAutoStretch];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetAutoStretch:boolean;
begin
  Result:=arcoAutoStretch in Options;
end;

procedure     TARCell.SetAutoSize(AAutoSize:boolean);
begin
  if AAutoSize <>(arcoAutoSize in Options) then begin
    BeginUpdate;
      if AAutoSize then begin
        Options:=Options + [arcoAutoSize];
      end else begin
        Options:=Options - [arcoAutoSize];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetAutoSize:boolean;
begin
  Result:=arcoAutoSize in Options;
end;

procedure     TARCell.SetIsRTF(AIsRTF:boolean);
begin
  if AIsRTF <>(arcoIsRTF in Options) then begin
    BeginUpdate;
      if AIsRTF then begin
        Options:=Options + [arcoIsRTF];
      end else begin
        Options:=Options - [arcoIsRTF];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetIsRTF:boolean;
begin
  Result:=arcoIsRTF in Options;
end;

procedure     TARCell.SetTile(ATile:boolean);
begin
  if ATile <> (arcoTile in Options) then begin
    BeginUpdate;
      if ATile then begin
        Options:=Options + [arcoTile];
      end else begin
        Options:=Options - [arcoTile];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetTile:boolean;
begin
  Result:=arcoTile in Options;
end;

procedure     TARCell.SetValue(const AValue:string);
begin
  if FValue <> AValue then begin
    BeginUpdate;
      FValue:=AValue;
      Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARCell.SetSelected(ASelected:boolean);
begin
  if ASelected <> (arcoSelected in Options) then begin
    BeginUpdate;
      if ASelected then begin
        Options:=Options + [arcoSelected];
      end else begin
        Options:=Options - [arcoSelected];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetSelected:boolean;
begin
  Result:=arcoSelected in Options;
end;

procedure     TARCell.SetFocused(AFocused:boolean);
begin
  if AFocused <> (arcoFocused in Options) then begin
    BeginUpdate;
      if AFocused then begin
        Options:=Options + [arcoFocused];
      end else begin
        Options:=Options - [arcoFocused];
      end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetFocused:boolean;
begin
  Result:=arcoFocused in Options;
end;

procedure     TARCell.SetShared(AShared:boolean);
begin
  if AShared <> (arcoShared in Options) then begin
    BeginUpdate;
    if AShared then begin
      Options:=Options + [arcoShared];
    end else begin
      Options:=Options - [arcoShared];
    end;
      Changed:=true;
    EndUpdate;
  end;
end;

function      TARCell.GetShared:boolean;
begin
  Result:=arcoShared in Options;
end;

function      TARCell.GetInUse:boolean;
begin
  Result:=Assigned(Owner) and (Owner <> GetTopBand);
end;

procedure     TARCell.Invalidate;
begin
  if Changed and (not InUpdate) then begin
    if Assigned(OnChange) then OnChange(Self);
    Changed:=false;
  end;
end;

function      TARCell.GetTopCell:TARCell;
var ib, ic, L:integer;
    band:TARBand;
    bandlist:TARBandList;
    found:boolean;
begin
  Result:=Self;
  if Shared then begin
    band:=Owner;
    if Assigned(band) then begin
      bandlist:=band.Owner;
      if Assigned(bandlist) then begin
        L:=band.Lefts[band.IndexOf(Self)];
        for ib:=bandlist.IndexOf(band) - 1 downto 0 do begin
          found:=false;
          for ic:=0 to TARBand(bandlist.Items[ib]).CellCount - 1 do begin
            if (L = TARBand(bandlist.Items[ib]).Lefts[ic]) and
               (Width = TARBand(bandlist.Items[ib]).Cells[ic].Width)
            then begin
              Result:=TARBand(bandlist.Items[ib]).Cells[ic];
              found:=TARBand(bandlist.Items[ib]).Cells[ic].Shared;
              break;
            end;
          end;
          if not found then break;
        end;
      end;
    end;
  end;
end;

function      TARCell.GetTopBand:TARBand;
var ib, ic, L:integer;
    bandlist:TARBandList;
    found:boolean;
begin
  Result:=Owner;
  if Shared then begin
    if Assigned(Result) then begin
      bandlist:=Result.Owner;
      if Assigned(bandlist) then begin
        L:=Result.Lefts[Result.IndexOf(Self)];
        for ib:=bandlist.IndexOf(Result) - 1 downto 0 do begin
          found:=false;
          for ic:=0 to TARBand(bandlist.Items[ib]).CellCount - 1 do begin
            if (L = TARBand(bandlist.Items[ib]).Lefts[ic]) and
               (Width = TARBand(bandlist.Items[ib]).Cells[ic].Width)
            then begin
              Result:=TARBand(bandlist.Items[ib]);
              found:=TARBand(bandlist.Items[ib]).Cells[ic].Shared;
              break;
            end;
          end;
          if not found then break;
        end;
      end;
    end;
  end;
end;

function      TARCell.GetBottomBand:TARBand;
var ib, ic, L:integer;
    bandlist:TARBandList;
    found:boolean;
begin
  Result:=TARBand(Owner);

  if Assigned(Result) then begin
    bandlist:=Result.Owner;
    if Assigned(bandlist) then begin
      L:=Result.Lefts[Result.IndexOf(Self)];
      for ib:=bandlist.IndexOf(Result) + 1 to bandlist.BandCount - 1 do begin
        found:=false;
        for ic:=0 to TARBand(bandlist.Items[ib]).CellCount - 1 do begin
          if (TARBand(bandlist.Items[ib]).Cells[ic].Shared) and
             (L = TARBand(bandlist.Items[ib]).Lefts[ic]) and
             (Width = TARBand(bandlist.Items[ib]).Cells[ic].Width)
          then begin
            Result:=TARBand(bandlist.Items[ib]);
            found:=true;
            break;
          end;
        end;
        if not found then break;
      end;
    end;
  end;
end;

function      TARCell.GetBottomCell:TARCell;
var ib, ic, L:integer;
    band:TARBand;
    bandlist:TARBandList;
    found:boolean; 
begin
  Result:=Self;

  band:=TARBand(Owner);
  if Assigned(band) then begin
    bandlist:=band.Owner;
    if Assigned(bandlist) then begin
      L:=band.Lefts[band.IndexOf(Self)];
      for ib:=bandlist.IndexOf(band) + 1 to bandlist.BandCount - 1 do begin
        found:=false;
        for ic:=0 to TARBand(bandlist.Items[ib]).CellCount - 1 do begin
          if (TARBand(bandlist.Items[ib]).Cells[ic].Shared) and
             (L = TARBand(bandlist.Items[ib]).Lefts[ic]) and
             (Width = TARBand(bandlist.Items[ib]).Cells[ic].Width)
          then begin
            Result:=TARBand(bandlist.Items[ib]).Cells[ic];
            found:=true;
            break;
          end;
        end;
        if not found then break;
      end;
    end;
  end;
end;

procedure     TARCell.BeginUpdate;
begin
  inc(InUpdateCnt);
end;

procedure     TARCell.EndUpdate;
begin
  dec(InUpdateCnt);
  Invalidate;
end;

function      TARCell.InUpdate:boolean;
begin
  Result:=InUpdateCnt <> 0;
end;

procedure     TARCell.ARTemplateChange(Sender:TObject);
var p, i:integer;
    sname, sparam, stmp:string;
    iparam, iparam2, code:integer;
    L:TARTemplate;
    sl:TAIStringList;
    gr:TGraphic;
    newversion:boolean;
begin
  sl:=TAIStringList.Create('TARCell.ARTemplateChange sl');
  try
    L:=FARTemplate;

    newversion:=false;

    sname :=TrimLeft(L[0]);
    if sname = 'CELL' then begin
      BeginUpdate;

      FontStyle:=[];
      FrameWidth:=0;
      FrameColor:=0;
      TextAngle :=0;
      FrameStyle:=[];
      gr:=nil;

      for i:=1 to L.Count - 1 do begin
        p:=System.Pos('=', L[i]);
        if p <> 0 then begin
          sname :=TrimLeft(System.Copy(L[i], 1, p-1));
          sparam:=System.Copy(L[i], p+1, 2000000000);
          iparam2:=-1;
          System.Val(sparam, iparam, code);
          if code > 0 then begin
            stmp:=System.Copy(sparam, code + 1, 2000000000);
            System.Val(stmp, iparam2, code);
          end;
        end else begin
          sname :=TrimLeft(L[i]);
          sparam:='';
          iparam:=-1;
          iparam2:=-1;
        end;

        if sname = 'SHARED'     then Shared   :=true                        else
        if sname = 'ISRTF'      then IsRTF    :=true                        else
        if sname = 'FONTNAME'   then FontName :=sparam                      else
        if sname = 'FONTSIZE'   then FontSize :=iparam                      else
        if sname = 'FONTCOLOR'  then FontColor:=iparam                      else
        if sname = 'TEXTANGLE'  then TextAngle:=iparam                      else
        if sname = 'BOLD'       then FontStyle:=FontStyle  + [fsBold]       else
        if sname = 'ITALIC'     then FontStyle:=FontStyle  + [fsItalic]     else
        if sname = 'UNDERLINE'  then FontStyle:=FontStyle  + [fsUnderline]  else
        if sname = 'STRIKEOUT'  then FontStyle:=FontStyle  + [fsStrikeOut]  else

        if sname = 'FRAMECOLOR' then begin
          if not newversion then FrameColor  :=iparam
        end else
        if sname = 'FRAMEWIDTH' then begin
          if not newversion then FrameWidth  :=iparam
        end else

        if sname = 'FRAMELEFT'    then begin
          newversion:=true;
          if iparam = -1 then iparam:=1;
          FrameWidths[frLeft]:=iparam;   FrameColors[frLeft]:=iparam2;
        end else
        if sname = 'FRAMERIGHT'   then begin
          newversion:=true;
          if iparam = -1 then iparam:=1;
          FrameWidths[frRight]:=iparam;  FrameColors[frRight]:=iparam2;
        end else
        if sname = 'FRAMETOP'     then begin
          newversion:=true;
          if iparam = -1 then iparam:=1;
          FrameWidths[frTop]:=iparam;    FrameColors[frTop]:=iparam2;
        end else
        if sname = 'FRAMEBOTTOM'  then begin
          newversion:=true;
          if iparam = -1 then iparam:=1;
          FrameWidths[frBottom]:=iparam; FrameColors[frBottom]:=iparam2;
        end else

        if sname = 'COLOR'       then Color       :=iparam else

        if sname = 'VALUE'       then Value       :=sparam else
        if sname = 'COMMAVALUE'  then begin sl.CommaText:=sparam; Value:=sl.Text end else

        if sname = 'HALIGN'       then HAlign      :=THAlign(iparam) else
        if sname = 'VALIGN'       then VAlign      :=TVAlign(iparam) else
        if sname = 'WIDTH'        then Width       :=iparam else
        if sname = 'WORDWRAP'     then WordWrap    :=boolean(iparam) else
        if sname = 'OBJECTTYPE'   then UserVar     :=sparam else
        if sname = 'AUTOSIZE'     then AutoSize    :=boolean(iparam) else
        if sname = 'AUTOSTRETCH'  then AutoStretch :=boolean(iparam) else
        if sname = 'TILE'         then Tile        :=boolean(iparam) else

        if sname = 'PICTUREFILENAME' then begin
          PictureFileName:=sparam;
        end else

        if sname = 'LINKPICTURETOFILE' then begin
          LinkPictureToFile:=boolean(iparam);
        end else

        if sname = 'PICTURETYPE'  then begin
          if Assigned(gr) then gr.Free;
          if sparam = 'BITMAP'    then begin
            gr:=TBitmap.Create();
          end else
          if sparam = 'ICON'      then begin
            gr:=TIcon.Create();
          end else
          if sparam = 'METAFILE'  then begin
            gr:=TMetafile.Create();
          end;
        end else

        if sname = 'PICTURE' then begin
          if Assigned(gr) then begin
            StringToGraphic(sparam, gr);
            Picture.Assign(gr);
          end;
        end else

        if sname = 'ENDCELL'     then break;
      end;
      Changed:=true;
      EndUpdate;
    end;
  finally
    sl.Free;
  end;
end;

function      TARCell.GetARTemplate:TARTemplate;
var sl:TAIStringList;
begin
  sl:=TAIStringList.Create('TARCell.GetARTemplate sl');
  try
    if not Assigned(FARTemplate) then begin
      FARTemplate:=TARTemplate.Create('TARCell.FARTemplate');
      FARTemplate.OnChange:=ARTemplateChange;
    end;
    FARTemplate.OnChange:=nil;
      FARTemplate.Clear;
      Result:=FARTemplate;

      FARTemplate.Add('  CELL');

      if Shared then  FARTemplate.Add('    SHARED');
      if IsRTF  then  FARTemplate.Add('    ISRTF');

      FARTemplate.Add(Format('    FONTNAME=%s',   [FontName]));
      FARTemplate.Add(Format('    FONTSIZE=%d',   [FontSize]));
      FARTemplate.Add(Format('    FONTCOLOR=%d',  [FontColor]));
      FARTemplate.Add(Format('    TEXTANGLE=%d',  [TextAngle]));

      if fsBold       in FontStyle then  FARTemplate.Add('    BOLD');
      if fsItalic     in FontStyle then  FARTemplate.Add('    ITALIC');
      if fsUnderline  in FontStyle then  FARTemplate.Add('    UNDERLINE');
      if fsStrikeOut  in FontStyle then  FARTemplate.Add('    STRIKEOUT');

      if FrameWidths[frLeft]  > 0 then
        FARTemplate.Add(Format('    FRAMELEFT=%d,%d',   [FrameWidths[frLeft],   FrameColors[frLeft]]));
      if FrameWidths[frRight] > 0 then
        FARTemplate.Add(Format('    FRAMERIGHT=%d,%d',  [FrameWidths[frRight],  FrameColors[frRight]]));
      if FrameWidths[frTop]   > 0 then
        FARTemplate.Add(Format('    FRAMETOP=%d,%d',    [FrameWidths[frTop],    FrameColors[frTop]]));
      if FrameWidths[frBottom]> 0 then
        FARTemplate.Add(Format('    FRAMEBOTTOM=%d,%d', [FrameWidths[frBottom], FrameColors[frBottom]]));

      FARTemplate.Add(Format('    FRAMECOLOR=%d', [FrameColor]));
      FARTemplate.Add(Format('    FRAMEWIDTH=%d', [FrameWidth]));
      FARTemplate.Add(Format('    COLOR=%d',      [Color]));

      sl.Text:=Value;
      if sl.Count <= 1 then begin
        FARTemplate.Add(Format('    VALUE=%s',        [Value]));
      end else begin
        FARTemplate.Add(Format('    COMMAVALUE=%s',   [sl.CommaText]));
      end;

      FARTemplate.Add(Format('    HALIGN=%d',       [Ord(HAlign)]));
      FARTemplate.Add(Format('    VALIGN=%d',       [Ord(VAlign)]));
      FARTemplate.Add(Format('    WIDTH=%d',        [Width]));

      FARTemplate.Add(Format('    WORDWRAP=%d',     [integer(WordWrap)]));
      FARTemplate.Add(Format('    OBJECTTYPE=%s',   [UserVar]));
      FARTemplate.Add(Format('    AUTOSTRETCH=%d',  [integer(AutoStretch)]));

      if Length(PictureFileName) > 0 then begin
        FARTemplate.Add(Format('    PICTUREFILENAME=%s',    [PictureFileName]));
      end;

      if Assigned(FPicture) or (Length(PictureFileName) > 0) then begin
        FARTemplate.Add(Format('    LINKPICTURETOFILE=%d',  [integer(LinkPictureToFile)]));
        FARTemplate.Add(Format('    FITTOCELL=%d',  [integer(AutoSize)]));
        FARTemplate.Add(Format('    AUTOSIZE=%d',   [integer(AutoSize)]));
        FARTemplate.Add(Format('    TILE=%d',       [integer(Tile)]));
      end;

      if (not LinkPictureToFile) and Assigned(FPicture) then begin
        if FPicture.Graphic is TBitmap then begin
          FARTemplate.Add('    PICTURETYPE=BITMAP')
        end else
        if FPicture.Graphic is TIcon then begin
          FARTemplate.Add('    PICTURETYPE=ICON')
        end else
        if FPicture.Graphic is TMetafile then begin
          FARTemplate.Add('    PICTURETYPE=METAFILE');
        end;

        FARTemplate.Add(Format('    PICTURE=%s',  [GraphicToString(FPicture.Graphic)]));
      end;
      FARTemplate.Add('  ENDCELL');
    FARTemplate.OnChange:=ARTemplateChange;
  finally
    sl.Free;
  end;
end;

procedure     TARCell.SetARTemplate(const AARTemplate:TARTemplate);
begin
  if not Assigned(FARTemplate) then begin
    FARTemplate:=TARTemplate.Create('TARCell.FARTemplate');
    FARTemplate.OnChange:=ARTemplateChange;
  end;

  BeginUpdate;
    FARTemplate.Assign(AARTemplate);
    Changed:=true;
  EndUpdate;
end;

////////////////////////////////////////////////////////// TARBandList
constructor TARBandList.Create;
begin
  FBands  :=TAIList.Create('TARBandList.bands');
  Expanded:=false;
  Owner   :=nil;
end;

destructor  TARBandList.Destroy;
begin
  Collapse;

  DeleteAll;
  FBands.Free;

  if Assigned(FARTemplate) then begin
    FARTemplate.Free;
    FARTemplate:=nil;
  end;

  inherited;
end;

function    TARBandList.IndexOf(ABand:TARBandList):integer;
begin
  Result:=FBands.IndexOf(ABand);
end;

function    TARBandList.GetBandCount:integer;
begin
  Result:=FBands.Count;
end;

function    TARBandList.GetItems(index:integer):TARBandList;
begin
  Result:=FBands[index];
end;

procedure   TARBandList.Add(ABand:TARBandList);
var exp:boolean;
begin
  BeginUpdate;
    exp:=Expanded;
    Expanded:=false;

    ABand.Owner:=Self;
    FBands.Add(ABand);

    if FBands.Count = 1 then begin
      ABand.FTop:=0;
    end else begin
      ABand.FTop:=TARBandList(FBands[FBands.Count - 2]).FTop + TARBandList(FBands[FBands.Count - 2]).Height;
    end;

    ABand.OnBandChange:=BandChange;
    ABand.OnCellChange:=CellChange;

    Expanded:=exp;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.Delete(index: Integer);
var exp:boolean;
begin
  BeginUpdate;
    exp:=Expanded;
    Expanded:=false;

    TARBandList(FBands[index]).Owner:=nil;
    TARBandList(FBands[index]).Free;
    FBands.Delete(index);
    ReculcTops(index);

    Expanded:=exp;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.DeleteAll;
var exp:boolean;
begin
  BeginUpdate;
    exp:=Expanded;
    Expanded:=false;

    while FBands.Count > 0 do begin
      TARBandList(FBands[0]).Free();
      FBands.Delete(0);
    end;

  Expanded:=exp;
  Changed:=true;
  EndUpdate;
end;

function  TARBandList.Move(CurIndex, NewIndex: Integer):boolean;
var exp:boolean;
begin
  Result:=false;
  if (CurIndex <> NewIndex) and  (NewIndex >=0) and (NewIndex < BandCount) then begin

    BeginUpdate;
      exp:=Expanded;
      Expanded:=false;

{
      h:=TARBandList(FBands[CurIndex]).Height;
      if CurIndex < NewIndex then begin
        h:=-h;
      end;
      for i:=CurIndex + 1 to NewIndex - 1 do begin
        TARBandList(FBands[i]).FTop:=TARBandList(FBands[i]).FTop + h;
      end;
}
      FBands.Move(CurIndex, NewIndex);

      ReculcTops();

      Expanded:=exp;
    Changed:=true;
    EndUpdate;

    Result:=true;
  end;
end;

procedure   TARBandList.Insert(index: Integer; ABand:TARBandList);
var exp:boolean;
    i, h:integer;
begin
  BeginUpdate;
    exp:=Expanded;
    Expanded:=false;

    ABand.Owner:=Self;

    h:=ABand.Height;
    for i:=index to FBands.Count - 1 do begin
      TARBandList(FBands[index]).FTop:=TARBandList(FBands[index]).FTop + h;
    end;
    FBands.Insert(index, ABand);

    Expanded:=exp;
  Changed:=true;
  EndUpdate;
end;

function    TARBandList.GetExpanded:boolean;
begin
  Result:=FExpanded;
end;

function    TARBandList.GetExpandable:boolean;
begin
  Result:=BandCount > 0;
end;

procedure   TARBandList.SetExpanded(AExpanded:boolean);
begin
  BeginUpdate;
    if AExpanded then begin
      Expand;
    end else begin
      Collapse;
    end;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.Collapse;
var i, idx, minidx:integer;
    exp:boolean;
begin
  BeginUpdate;
    if FExpanded then begin
      FExpanded:=false;
      if Assigned(Owner) then begin
        exp:=Owner.Expanded;
        Owner.Expanded:=false;
        minidx:=Owner.BandCount - 1;
        for i:=0 to BandCount - 1 do begin
          idx:=Owner.IndexOf(FBands[i]);
          if idx >= 0 then begin
            Owner.FBands.Delete(idx);
            minidx:=Min(minidx, idx);
          end;
        end;
        Owner.ReculcTops(minidx);
        Owner.Expanded:=exp;
      end;
      Changed:=true;
    end;
  EndUpdate;
end;

procedure   TARBandList.Expand;
var i, idx:integer;
    exp:boolean;
begin
  BeginUpdate;
    if not FExpanded then begin
      Collapse;
      FExpanded:=true;
      if Assigned(Owner) then begin
        idx:=Owner.IndexOf(Self);
        if idx >= 0 then begin
          exp:=Owner.Expanded;
          Owner.Expanded:=false;
          for i:=0 to BandCount - 1 do begin
            Owner.FBands.Insert(idx + i + 1, FBands[i]);
          end;
          Owner.ReculcTops(idx);
          Owner.Expanded:=exp;
        end;
      end;
    end;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.ReculcTops(idxband:integer);
var i, oldtop:integer;
begin
  if (idxband >= 0) and (idxband < FBands.Count) then begin
    if idxband = 0 then begin
      oldtop   :=0;
    end else begin
      oldtop   :=TARBandList(FBands[idxband - 1]).FTop + TARBandList(FBands[idxband - 1]).Height;
    end;
    for i:=idxband to FBands.Count - 1 do begin
      TARBandList(FBands[i]).FTop:=oldtop;
      oldtop   :=oldtop + TARBandList(FBands[i]).Height;
    end;
  end;
end;

function    TARBandList.GetMarked:boolean;
begin
  Result:=not FMark.Empty;
end;

procedure   TARBandList.SetMarked(AMarked:boolean);
begin
  BeginUpdate;
    if AMarked <> Marked then begin
      if AMarked then begin
        FMark.Handle := LoadBitmap(0, PChar(OBM_COMBO));
      end else begin
        FMark.ReleaseHandle;
      end;
    end;
  Changed:=true;
  EndUpdate;
end;

function    TARBandList.GetMark:TBitmap;
begin
  Result:=FMark;
end;

procedure   TARBandList.SetMark(AMark:TBitmap);
begin
  BeginUpdate;
    FMark.Assign(AMark);
  Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.SetName(const AName:string);
begin
  if FName <> AName then begin
    BeginUpdate;
      FName:=AName;
    Changed:=true;
    EndUpdate;
  end;
end;

procedure   TARBandList.Invalidate;
begin
  if Changed and (not InUpdate) then begin
    if Assigned(OnChange) then OnChange(Self);
    Changed:=false;
  end;
end;

procedure   TARBandList.BeginUpdate;
begin
  inc(InUpdateCnt);
end;

procedure   TARBandList.EndUpdate;
begin
  dec(InUpdateCnt);
  Invalidate;
end;

function    TARBandList.InUpdate:boolean;
begin
  Result:=InUpdateCnt <> 0;
end;

procedure   TARBandList.BandChange(band:TObject);
begin
  if not InUpdate then begin
    if Assigned(OnBandChange) then OnBandChange(band);
  end;
end;

procedure   TARBandList.CellChange(band:TObject; cell:TObject);
begin
  if not InUpdate then begin
    if Assigned(OnCellChange) then OnCellChange(band, cell);
  end;
end;

function    TARBandList.GetARTemplate:TARTemplate;
begin
  if not Assigned(FARTemplate) then begin
    FARTemplate:=TARTemplate.Create('TARBandList.FARTemplate');
    FARTemplate.OnChange:=ARTemplateChange;
  end;

  Result:=FARTemplate;
end;

procedure   TARBandList.SetARTemplate(const AARTemplate:TARTemplate);
begin
  if not Assigned(FARTemplate) then begin
    FARTemplate:=TARTemplate.Create('TARBandList.FARTemplate');
    FARTemplate.OnChange:=ARTemplateChange;
  end;

  BeginUpdate;
    FARTemplate.Assign(AARTemplate);
    Changed:=true;
  EndUpdate;
end;

procedure   TARBandList.ARTemplateChange(Sender:TObject);
begin
end;

procedure   TARBandList.SetHeight(AHeight:integer);
begin
  if FHeight <> AHeight then begin
    BeginUpdate;
      FHeight:=AHeight;

      if Assigned(Owner) then begin
        Owner.ReculcTops(Owner.FBands.IndexOf(Self) + 1);
      end;

    Changed:=true;
    EndUpdate;
  end;
end;

procedure     TARBandList.SetOwner(AOwner:TARBandList);
var i, idx:integer;
begin
  if FOwner <> AOwner then begin
    BeginUpdate;

      FOwner:=AOwner;
      if Assigned(Owner) then begin
        idx:=Owner.FBands.IndexOf(Self);
        if idx <> -1 then begin
          for i:=Owner.FBands.IndexOf(Self) to Owner.FBands.Count - 1 do begin
            TARBandList(Owner.FBands[i]).FTop:=TARBandList(Owner.FBands[i]).FTop + FHeight;
          end;
        end;
      end else begin
        FTop:=0;
      end;

    Changed:=true;
    EndUpdate;
  end;
end;

////////////////////// TARBand ////////////////////////////////////
constructor  TARBand.Create;
begin
  BeginUpdate;
    inherited;

    inc(BandCnt);

    CellList:=TAIList.Create('TARBand.CellList');

    FMark:=TBitmap.Create;

    Name    :=DefBandName;
    Height  :=DefHeight;

    NewCell;
    NewCell;
    NewCell;
    NewCell;
  Changed:=true;
  EndUpdate;
end;

destructor  TARBand.Destroy;
begin
  DeleteCells;
  CellList.Free;

  FMark.Free;

  if Assigned(FARTemplate) then begin
    FARTemplate.Free;
    FARTemplate:=nil;
  end;

  dec(BandCnt);
  inherited;
end;

function    TARBand.GetCellCount:integer;
begin
  try
    if Assigned(CellList) then begin
      Result:=CellList.Count;
    end else begin
      Result:=0;
    end;
  except
    Result:=0;
  end;
end;

function    TARBand.GetCell(idxcell:integer):TARCell;
begin
  if idxcell < CellCount then begin
    Result:=CellList[idxcell];
  end else begin
    Result:=nil;
  end;
end;

procedure   TARBand.SetCell(idxcell:integer;ACell:TARCell);
begin
  CellList[idxcell]:=ACell;
end;

function    TARBand.NewCell:TARCell;
begin
  BeginUpdate;
    Result:=TARCell.Create;
    Result.Owner:=Self;
    CellList.Add(Result);
    Result.OnChange:=CellChange;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBand.CellChange(Sender:TObject);
begin
  if not InUpdate then begin
    if Assigned(OnCellChange) then OnCellChange(Self, Sender);
  end;
end;

procedure   TARBand.DeleteCells;
begin
  BeginUpdate;
    while CellList.Count > 0 do begin
      TARCell(CellList[0]).Free();
      CellList.Delete(0);
    end;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBand.DeleteCell(idxcell:integer);
begin
  BeginUpdate;
    if idxcell < CellCount then begin
      Cells[idxcell].Free;
      CellList.Delete(idxcell);
    end;

    if CellCount = 0 then begin
      NewCell;
    end;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBand.Split(ACell:TARCell);
var idxcell:integer;
    w:integer;
    ncell:TARCell;
begin
  BeginUpdate;
    idxcell:=IndexOf(ACell);
    if idxcell >= 0 then begin
      w:=ACell.Width;
      ncell:=NewCell;
      ncell.Assign(ACell);
      ncell.Width :=(w div 2);
      ACell.Width  :=(w div 2) + (w mod 2);
      CellList.Move(CellCount - 1, idxcell + 1);
    end;
  Changed:=true;
  EndUpdate;
end;

procedure   TARBand.Assign(ABand:TARBand);
var i:integer;
begin
  BeginUpdate;
    DeleteCells;
    for i:=0 to ABand.CellList.Count - 1 do begin
      NewCell.Assign(ABand.CellList[i]);
    end;

    Name    :=ABand.Name;
    Height  :=ABand.Height;
  Changed:=true;
  EndUpdate;
end;

function    TARBand.GetLefts(index:integer):integer;
var left,i:integer;
begin
  Result:=0;
  if (CellCount > 1) and (index < CellCount) then begin
    left:=0;
    for i:=0 to CellCount - 1 do begin
      if i=index then begin
        Result:=left;
        break;
      end else begin
        left:=left + Cells[i].Width;// + 1;
      end;
    end;
  end;
end;

function    TARBand.GetRights(index:integer):integer;
begin
  Result:=Lefts[index] + Cells[index].Width;
end;

function    TARBand.IndexOf(ACell:TARCell):integer;
begin
  Result:=CellList.IndexOf(ACell);
end;

function    TARBand.GetSelected:boolean;
var idxcell:integer;
begin
  Result:=false;
  for idxcell:=0 to CellCount - 1 do begin
    Result:=Result or Cells[idxcell].Selected;
  end;
end;

procedure   TARBand.SetSelected(ASelected:boolean);
var idxcell:integer;
begin
  for idxcell:=0 to CellCount - 1 do begin
    Cells[idxcell].Selected:=ASelected;
  end;
end;

function    TARBand.GetWidth:integer;
var idxcell:integer;
begin
  Result:=0;
  for idxcell:=0 to CellCount - 1 do begin
    Result:=Result + Cells[idxcell].Width + 1;
  end;
end;

procedure   TARBand.Move(idxcellfrom:integer;idxcellto:integer);
begin
  BeginUpdate;
    CellList.Move(idxcellfrom, idxcellto);
  Changed:=true;
  EndUpdate;
end;

function    TARBand.GetARTemplate:TARTemplate;
var idxcell:integer;
begin
  if not Assigned(FARTemplate) then begin
    FARTemplate:=TARTemplate.Create('TARBand.FARTemplate');
    FARTemplate.OnChange:=ARTemplateChange;
  end;

  FARTemplate.OnChange:=nil;
  FARTemplate.Clear;

  Result:=FARTemplate;

  FARTemplate.Add('BAND');
  FARTemplate.Add(Format('  NAME=%s',   [Name]));
  FARTemplate.Add(Format('  HEIGHT=%d', [Height]));

  for idxcell:=0 to CellCount - 1 do begin
    FARTemplate.AddStrings(Cells[idxcell].ARTemplate);
  end;

  FARTemplate.Add('ENDBAND');

  FARTemplate.OnChange:=ARTemplateChange;
end;

procedure   TARBand.ARTemplateChange(Sender:TObject);
var p, i:integer;
    sname, sparam:string;
    iparam, code:integer;
    L:TARTemplate;
    sl:TARTemplate;
    cell:TARCell;
begin
  L:=FARTemplate;

  sname :=TrimLeft(L[0]);
  if sname = 'BAND' then begin
    sl:=TARTemplate.Create('TARBand.ARTemplateChange sl');
    try
      BeginUpdate;

      DeleteCells;
      i:=1;
      while i < L.Count do begin
        p:=System.Pos('=', L[i]);
        if p <> 0 then begin
          sname :=TrimLeft(System.Copy(L[i], 1, p-1));
          sparam:=System.Copy(L[i], p+1, 2000000000);
          System.Val(sparam, iparam, code);
        end else begin
          sname :=TrimLeft(L[i]);
          sparam:='';
          iparam:=0;
        end;
        sname:=AnsiUpperCase(sname);
        inc(i);
        if sname = 'ENDBAND'  then break else
        if sname = 'NAME'     then Name  :=sparam else
        if sname = 'HEIGHT'   then Height:=iparam else
        if sname = 'CELL'     then begin
          sl.Clear;
          sl.Add('    CELL');
          while i < L.Count do begin
            if CT('ENDCELL', TrimLeft(L[i])) then begin
              cell:=NewCell;
              cell.ARTemplate:=sl;
              inc(i);
              break;
            end;
            sl.Add(L[i]);
            inc(i);
          end;
        end;
      end;
      EndUpdate;
    finally
      sl.Free;
    end;
  end;
end;

/////////////////////// TARRep ///////////////////////////////////
constructor  TARRep.Create;
begin
  BeginUpdate;
    inherited;

    inc(RepCnt);

    FUserVarList  :=TAIStringList.Create('TARRep.FUserVarList');

    FFont         :=TFont.Create;
    FFont.Name    :=DefFontName;
    FFont.Size    :=DefFontSize;
    FFont.Color   :=DefFontColor;
    FFont.Style   :=DefFontStyle;
    FFont.Charset :=DefFontCharset;

    Orientation   :=DefOrientation;
    PaperSize     :=DefPaperSize;
    LeftMargin    :=DefLeftMargin;
    TopMargin     :=DefTopMargin;
    RightMargin   :=DefRightMargin;
    BottomMargin  :=DefBottomMargin;

    NewBand;
    NewBand;
    NewBand;
    NewBand;
  Changed:=true;
  EndUpdate;
end;

destructor  TARRep.Destroy;
begin
  FFont.Free;
  FUserVarList.Free;

  if Assigned(FARTemplate) then begin
    FARTemplate.Free;
    FARTemplate:=nil;
  end;

  dec(RepCnt);
  inherited
end;

function    TARRep.NewBand:TARBand;
var i:integer;
begin
  BeginUpdate;
    Result:=TARBand.Create;
    Result.OnBandChange:=BandChange;
    Result.OnCellChange:=CellChange;
    Result.Name:=Format('%s %d', [Result.Name, BandCount]);
    for i:=0 to Result.CellCount - 1 do begin
      Result.Cells[i].Value:=Format('%s(%d,%d)', [Result.Cells[i].Value, BandCount, i]);
    end;
    Add(Result);
  Changed:=true;
  EndUpdate;
end;

function    TARRep.GetBand(idxband:integer):TARBand;
begin
  if BandCount > 0 then begin
    Result:=TARBand(Items[Min(Max(idxband, 0), BandCount - 1)]);
  end else begin
    Result:=nil;
  end;
end;

function    TARRep.GetTops(idxband:integer):integer;
//var i:integer;
begin
  Result:=0;

{
  IdxBand:=Min(IdxBand, BandCount - 1);
  for i:=0 to IdxBand - 1 do begin
    Result:=Result + Bands[i].Height;
  end;
}

  if (idxband >= 0) and (idxband < BandCount) then begin
    Result:=Bands[idxband].FTop;
  end;
end;

function    TARRep.GetCell(idxband:integer;idxcell:integer):TARCell;
begin
  if Assigned(Bands[idxband]) then begin
    Result:=Bands[idxband].Cells[idxcell];
  end else begin
    Result:=nil;
  end;
end;

procedure    TARRep.DeleteBand(idxband:integer);
begin
  BeginUpdate;
    Delete(idxband);
  Changed:=true;
  EndUpdate;
end;

procedure    TARRep.SetUserVarList(AUserVarList:TAIStringList);
begin
  FUserVarList.Assign(AUserVarList);
end;

procedure    TARRep.SetFont(AFont:TFont);
begin
  BeginUpdate;
    FFont.Assign(AFont);
  Changed:=true;
  EndUpdate;
end;

procedure    TARRep.ARTemplateChange(Sender:TObject);
var p:integer;
    band:TARBand;
    sname, sparam:string;
    iparam, code:integer;
    L:TARTemplate;
    sl:TARTemplate;
begin
  sl:=TARTemplate.Create('TARRep.ARTemplateChange sl');
  L:=TARTemplate.Create('TARRep.ARTemplateChange L');
  BeginUpdate;
  try
    L.Assign(FARTemplate);
    DeleteAll;

    while L.Count > 0 do begin

      p:=System.Pos('=', L[0]);
      if p <> 0 then begin
        sname :=TrimLeft(System.Copy(L[0], 1, p-1));
        sparam:=System.Copy(L[0], p+1, 2000000000);
        System.Val(sparam, iparam, code);
      end else begin
        sname :=TrimLeft(L[0]);
        sparam:='';
        iparam:=0;
      end;
      L.Delete(0);
      Font.Style   :=[];
      sname :=ANSIUpperCase(sname);
      sparam:=ANSIUpperCase(sparam);

      if sname = 'ORIENTATION'  then if sparam = 'LANDSCAPE' then Orientation:=poLandscape else Orientation:=poPortrait else
      if sname = 'LEFTMARGIN'   then LeftMargin     :=iparam else
      if sname = 'TOPMARGIN'    then TopMargin      :=iparam else
      if sname = 'RIGHTMARGIN'  then RightMargin    :=iparam else
      if sname = 'BOTTOMMARGIN' then BottomMargin   :=iparam else
      if sname = 'PAPERSIZE'    then PaperSize      :=TARPaperSize(iparam) else

      if sname = 'FONTNAME'     then Font.Name    :=sparam else
      if sname = 'FONTSIZE'     then Font.Size    :=iparam else
      if sname = 'FONTCOLOR'    then Font.Color   :=iparam else
      if sname = 'FONTCHARSET'  then Font.Charset :=iparam else
      if sname = 'BOLD'         then Font.Style   :=Font.Style  + [fsBold] else
      if sname = 'ITALIC'       then Font.Style   :=Font.Style  + [fsItalic] else
      if sname = 'UNDERLINE'    then Font.Style   :=Font.Style  + [fsUnderline] else
      if sname = 'STRIKEOUT'    then Font.Style   :=Font.Style  + [fsStrikeOut] else

      if sname = 'OBJECTLIST' then begin
        FUserVarList.Clear;
        while L.Count > 0 do begin
          p:=System.Pos('=', L[0]);
          if p <> 0 then begin
            sname :=ANSIUpperCase(TrimLeft(System.Copy(L[0], 1, p-1)));
            sparam:=System.Copy(L[0], p+1, 2000000000);
            System.Val(sparam, iparam, code);
          end else begin
            sname :=TrimLeft(L[0]);
            sparam:='';
            iparam:=0;
          end;
          L.Delete(0);
          if sname = 'ENDOBJECTLIST' then begin
            break;
          end else begin
            FUserVarList.AddObject(sname, pointer(iparam));
          end
        end;
      end else

      if sname = 'BAND' then begin
        sl.Clear;
        sl.Add('  BAND');
        while L.Count > 0 do begin
          if CT('ENDBAND', TrimLeft(L[0])) then begin
            L.Delete(0);
            band:=NewBand;
            band.ARTemplate:=sl;
            break;
          end;
          sl.Add(L[0]);
          L.Delete(0);
        end;
      end;
    end;
  finally
    L.Free;
    sl.Free;
    EndUpdate;
  end;
end;

function    TARRep.GetARTemplate:TARTemplate;
var i, idxband:integer;
    sl:TAIStringList;
begin
  if not Assigned(FARTemplate) then begin
    FARTemplate:=TARTemplate.Create('TARRep.FARTemplate');
    FARTemplate.OnChange:=ARTemplateChange;
  end;

  sl:=TAIStringList.Create('TARRep.GetARTemplate sl');
  try
    FARTemplate.OnChange:=nil;
    FARTemplate.Clear;

    if Orientation = poLandscape then FARTemplate.Add('ORIENTATION=LANDSCAPE') else FARTemplate.Add('ORIENTATION=PORTRAIT');
    FARTemplate.Add(Format('LEFTMARGIN=%d',   [Round(LeftMargin)]));
    FARTemplate.Add(Format('TOPMARGIN=%d',    [Round(TopMargin)]));
    FARTemplate.Add(Format('RIGHTMARGIN=%d',  [Round(RightMargin)]));
    FARTemplate.Add(Format('BOTTOMMARGIN=%d', [Round(BottomMargin)]));
    FARTemplate.Add(Format('PAPERSIZE=%d',    [integer(PaperSize)]));

    FARTemplate.Add(Format('FONTNAME=%s',     [Font.Name]));
    FARTemplate.Add(Format('FONTSIZE=%d',     [Font.Size]));
    FARTemplate.Add(Format('FONTCOLOR=%d',    [Font.Color]));
    FARTemplate.Add(Format('FONTCHARSET=%d',  [Font.Charset]));

    if fsBold       in Font.Style then  FARTemplate.Add('BOLD');
    if fsItalic     in Font.Style then  FARTemplate.Add('ITALIC');
    if fsUnderline  in Font.Style then  FARTemplate.Add('UNDERLINE');
    if fsStrikeOut  in Font.Style then  FARTemplate.Add('STRIKEOUT');

    if FUserVarList.Count > 0 then begin
      FARTemplate.Add('OBJECTLIST');
      for i:=0 to FUserVarList.Count - 1 do begin
        FARTemplate.Add(Format('  %s=%d',  [FUserVarList[i], integer(FUserVarList.Objects[i])]));
      end;
      FARTemplate.Add('ENDOBJECTLIST');
    end;

    for idxband:=0 to BandCount - 1 do begin
      FARTemplate.AddStrings(Bands[idxband]. ARTemplate);
    end;

    Result:=FARTemplate;
    FARTemplate.OnChange:=ARTemplateChange;
  finally
    sl.Free;
  end;
end;

procedure   TARRep.Load(AFileName:string);
var L:TARTemplate;
begin
  L:=TARTemplate.Create('TARRep.Load L');

  try
    if not Assigned(FileSystem) then begin
      L.LoadFromFile(AFileName);
    end else begin
      try
        L.LoadFromStream(FileSystem.CreateReadStream(AFileName));
      finally
        FileSystem.CloseReadStream;
      end;
    end;

    ARTemplate:=L;
  finally
    L.Free;
  end;

end;

procedure   TARRep.Save(AFileName:string);
var L:TARTemplate;
begin
  L:=ARTemplate;

  if not Assigned(FileSystem) then begin
    L.SaveToFile(AFileName);
  end else begin
    try
      L.SaveToStream(FileSystem.CreateWriteStream(AFileName));
    finally
      FileSystem.CloseWriteStream;
    end;
  end;
end;

function    TARRep.MaxHeight(ABand:TARBand):integer;
var ib, ic, top:integer;
    bottomband:TARBand;
begin
  Result:=ABand.Height;
  ib:=IndexOf(ABand);
  top:=Tops[ib];
  for ic:=0 to ABand.CellCount - 1 do begin
    bottomband:=ABand.Cells[ic].GetBottomBand;
    if bottomband <> ABand then begin
      Result:=Max(Result, (Tops[IndexOf(bottomband)] + bottomband.Height - top));
    end;
  end;
end;

end.


