unit ARRulerBar;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, ExtCtrls, Classes, Controls, Graphics, SysUtils, Math,
  Forms,

  ARRepCommon;

type

  TARRulerBarOrientation = (rbHorizontal, rbVertical);

//  TARRulerBar = class(TCustomControl)
  TARRulerBar = class(TCustomPanel)
  private
    FOrientation:TARRulerBarOrientation;

    FGutter     :integer;

    FPageWidth  :extended;
    FLeftMargin :extended;
    FRightMargin:extended;
    FUnits      :TARUnits;
    FSelBegin   :extended;
    FSelEnd     :extended;
    FCur        :integer;

    FOnChange   :TNotifyEvent;
    FInUpdate   :integer;
    FModified   :boolean;

    FSelColor   :TColor;
    FRulerColor :TColor;

    procedure   SetSelColor(ASelColor:TColor);
    procedure   SetRulerColor(ARulerColor:TColor);

    procedure   SetOrientation(AOrientation:TARRulerBarOrientation);

    procedure   SetCur(ACur:integer);
    procedure   SetGutter(AGutter:integer);

    procedure   SetPageWidth(APageWidth:extended);
    procedure   SetLeftMargin(ALeftMargin:extended);
    procedure   SetRightMargin(ARightMargin:extended);
    procedure   SetUnits(AUnits:TARUnits);
    procedure   SetSelBegin(ASelBegin:extended);
    procedure   SetSelEnd(ASelEnd:extended);
  protected
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer);override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure   Loaded;override;
    procedure   Paint;override;
    procedure   DoUpdate;
    function    InUpdate:boolean;

    function    UnitsToText(Val:extended):string;
  public
    constructor Create(AOwner:TComponent);override;
    destructor  Destroy;override;

    procedure   BeginUpdate;
    procedure   EndUpdate;
  published
    property    Orientation:TARRulerBarOrientation  read  FOrientation  write SetOrientation;

    property    Gutter:integer                      read  FGutter       write SetGutter;

    property    PageWidth:extended                  read  FPageWidth    write SetPageWidth;
    property    LeftMargin:extended                 read  FLeftMargin   write SetLeftMargin;
    property    RightMargin:extended                read  FRightMargin  write SetRightMargin;
    property    Units:TARUnits                      read  FUnits        write SetUnits;
    property    SelBegin:extended                   read  FSelBegin     write SetSelBegin;
    property    SelEnd:extended                     read  FSelEnd       write SetSelEnd;
    property    Cur:integer                         read  FCur          write SetCur;

    property    SelColor:TColor                     read  FSelColor     write SetSelColor;
    property    RulerColor:TColor                   read  FRulerColor   write SetRulerColor;
    property    Color;
    property    Font;

    property    OnChange:TNotifyEvent               read  FOnChange     write FOnChange;
  end;

implementation
{$R *.RES}

constructor TARRulerBar.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

  Align:=alTop;
  DoubleBuffered:=true;
//  Color:=$00C0FFFF;
end;

destructor  TARRulerBar.Destroy;
begin
  inherited;
end;

procedure   TARRulerBar.Loaded;
begin
  inherited;

  Caption:='';
end;

procedure   TARRulerBar.SetOrientation(AOrientation:TARRulerBarOrientation);
begin
  if FOrientation <> AOrientation then begin
    FOrientation:=AOrientation;
    case FOrientation of
      rbHorizontal: begin
                      Align :=alTop;
                      Height:=35;
                    end;
      rbVertical:   begin
                      Align:=alLeft;
                      Width:=35;
                    end;
    end;
    Invalidate;
  end;
end;

procedure   TARRulerBar.SetGutter(AGutter:integer);
begin
  if FGutter <> AGutter then begin
    FGutter:=AGutter;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetCur(ACur:integer);
begin
  if FCur <> ACur then begin
    FCur:=ACur;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetSelColor(ASelColor:TColor);
begin
  if FSelColor <> ASelColor then begin
    FSelColor:=ASelColor;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetRulerColor(ARulerColor:TColor);
begin
  if FRulerColor <> ARulerColor then begin
    FRulerColor:=ARulerColor;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetPageWidth(APageWidth:extended);
begin
  if FPageWidth <> APageWidth then begin
    FPageWidth:=APageWidth;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetLeftMargin(ALeftMargin:extended);
begin
  if FLeftMargin <> ALeftMargin then begin
    FLeftMargin:=ALeftMargin;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetRightMargin(ARightMargin:extended);
begin
  if FRightMargin <> ARightMargin then begin
    FRightMargin:=ARightMargin;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetUnits(AUnits:TARUnits);
begin
  if FUnits <> AUnits then begin
    FUnits:=AUnits;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetSelBegin(ASelBegin:extended);
begin
  if FSelBegin <> ASelBegin then begin
    FSelBegin:=ASelBegin;
    FModified:=true;
    DoUpdate;
  end;
end;

procedure   TARRulerBar.SetSelEnd(ASelEnd:extended);
begin
  if FSelEnd <> ASelEnd then begin
    FSelEnd:=ASelEnd;
    FModified:=true;
    DoUpdate;
  end;
end;

function    TARRulerBar.UnitsToText(Val:extended):string;
begin
  if (Units = uPix) or (Round(Val) = Val) then begin
    Result:=IntToStr(Round(Val));
  end else begin
    Result:=Format('%.2f', [Val]);
  end;
end;

type TMovedMargin = (mmNone, mmLeft, mmRight);

var DownX:integer;
    DownMargin:extended;
    MovedMargin:TMovedMargin;

procedure   TARRulerBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Orientation = rbHorizontal then begin
    if ssLeft in Shift then begin
      DownX:=X;
    end;
  end;
end;

procedure   TARRulerBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var d:integer;
begin
  if Orientation = rbHorizontal then begin
    BeginUpdate;

      if not (ssLeft in Shift) then begin
        if Abs(X - (Gutter - ToPix(LeftMargin))) < 5 then begin
          Cursor      :=crHSplit;
          MovedMargin :=mmLeft;
          DownMargin  :=LeftMargin;
        end else begin
          if Abs(X - (Gutter - ToPix(LeftMargin) + ToPix(PageWidth) - ToPix(RightMargin))) < 5 then begin
            Cursor      :=crHSplit;
            MovedMargin :=mmRight;
            DownMargin  :=RightMargin;
          end else begin
            Cursor:= crDefault;
            MovedMargin :=mmNone;
          end;
        end;
      end else begin
        if ssLeft in Shift then begin
          d:= X - DownX;
          case MovedMargin of
            mmNone  : ;
            mmLeft  : begin
                        LeftMargin  :=Max(DownMargin - FromPix(d), 0);
                        if Assigned(FOnChange) then FOnChange(Self);
                      end;
            mmRight : begin
                        RightMargin :=Max(DownMargin - FromPix(d), 0);
                        if Assigned(FOnChange) then FOnChange(Self);
                      end;
          end;
        end;
      end;
      Cur:=X;

    EndUpdate;
  end else begin
    Cur:=Y;
  end;
end;

procedure   TARRulerBar.DoUpdate;
begin
  if (not InUpdate) and FModified then begin
    FModified:=false;
    Invalidate;
  end;
end;

function    TARRulerBar.InUpdate:boolean;
begin
  Result:=FInUpdate <> 0;
end;

procedure   TARRulerBar.BeginUpdate;
begin
  inc(FInUpdate);
end;

procedure   TARRulerBar.EndUpdate;
begin
  dec(FInUpdate);
  DoUpdate;
end;

procedure ShowText(Canvas:TCanvas; r:TRect; angle:integer; text:string);
var hfnt, hfntPrev:HFONT;
    plf:TLogFont;
    th:integer;
begin
  th:=Canvas.TextWidth(text);

  with plf do begin
    lfHeight        :=Canvas.Font.Height;
    lfWidth         :=0;
    lfEscapement    :=angle * 10;
    lfOrientation   :=angle * 10;
    lfWeight        :=FW_NORMAL;

    lfItalic        :=byte(fsItalic in Canvas.Font.Style);
    lfUnderline     :=byte(fsUnderline in Canvas.Font.Style);
    lfStrikeOut     :=byte(fsStrikeOut in Canvas.Font.Style);
    lfCharSet       :=byte(Canvas.Font.Charset);
    lfQuality       :=DEFAULT_QUALITY;
    lfOutPrecision  :=OUT_DEFAULT_PRECIS;
    lfClipPrecision :=CLIP_DEFAULT_PRECIS;
    case Canvas.Font.Pitch of
      fpVariable: lfPitchAndFamily := VARIABLE_PITCH;
      fpFixed:    lfPitchAndFamily := FIXED_PITCH;
      else        lfPitchAndFamily := DEFAULT_PITCH;
    end;
    StrPCopy(lfFaceName, Canvas.Font.Name)
  end;

  SetBkMode(Canvas.Handle, TRANSPARENT);

  hfnt    :=CreateFontIndirect(plf);
  hfntPrev:=SelectObject(Canvas.Handle, hfnt);
  try
    TextOut(Canvas.Handle,
            r.Left,
            r.Bottom - (r.Bottom - r.Top - th) div 2,
            PChar(text),
            Length(text)
           );
  finally
    SelectObject(Canvas.Handle, hfntPrev);
    DeleteObject(hfnt);
    SetBkMode(Canvas.Handle, OPAQUE);
  end;
end;

procedure   TARRulerBar.Paint;
var r:TRect;
    u:extended;
    s:string;
    i:integer;
  procedure HShowPixGrid;
  var k:integer;
  begin
    for k:=0 to ToPix(PageWidth) - ToPix(LeftMargin) - ToPix(RightMargin) do begin
      if (k mod 100) = 0 then begin
        Canvas.MoveTo(Gutter + k, 8); Canvas.LineTo(Gutter + k, 22);
        s:=UnitsToText(k);
        Canvas.TextOut(Gutter + k + 2, 3, s);
      end else begin
        if (k mod 50) = 0 then begin
          Canvas.MoveTo(Gutter + k, 15); Canvas.LineTo(Gutter + k, 22);
        end else begin
          if (k mod 10) = 0 then begin
            Canvas.MoveTo(Gutter + k, 19); Canvas.LineTo(Gutter + k, 22);
          end;
        end;
      end;
    end;
  end;
  procedure HShowMMGrid;
  begin
    u:=0;
    repeat
      i:=ToPix(u);
      Canvas.MoveTo(Gutter + i, 8);
      Canvas.LineTo(Gutter + i, 22);
      s:=UnitsToText(u / UnitRatio(uMM));
      Canvas.TextOut(Gutter + i +2, 3, s);

      u:=u + 50; i:=ToPix(u);
      if u > (PageWidth - LeftMargin - RightMargin) then break;
      Canvas.MoveTo(Gutter + i, 15);
      Canvas.LineTo(Gutter + i, 22);

      u:=u + 50;
    until u > (PageWidth - LeftMargin - RightMargin);
  end;
  procedure HShowCMGrid;
  begin
    u:=0;
    repeat
      i:=ToPix(u);
      Canvas.MoveTo(Gutter + i, 8); Canvas.LineTo(Gutter + i, 22);
      s:=UnitsToText(u / UnitRatio(uCM));
      Canvas.TextOut(Gutter + i + 2, 3, s);

      u:=u + 50; i:=ToPix(u);
      if u > (PageWidth - LeftMargin - RightMargin) then break;
      Canvas.MoveTo(Gutter + i, 15); Canvas.LineTo(Gutter + i, 22);

      u:=u + 50;
    until u > (PageWidth - LeftMargin - RightMargin);
  end;
  procedure HShowInGrid;
  var k:integer;
  begin
    u:=0;
    repeat
      i:=ToPix(u);
      Canvas.MoveTo(Gutter + i, 8); Canvas.LineTo(Gutter + i, 22);
      s:=UnitsToText(u / UnitRatio(uIn));
      Canvas.TextOut(Gutter + i + 2, 3, s);

      for k:=1 to 4 do begin
        u:=u + 254 / 10; i:=ToPix(u);
        if u > (PageWidth - LeftMargin - RightMargin) then break;
        Canvas.MoveTo(Gutter + i, 19); Canvas.LineTo(Gutter + i, 22);
      end;

      u:=u + 254 / 10; i:=ToPix(u);
      if u > (PageWidth - LeftMargin - RightMargin) then break;
      Canvas.MoveTo(Gutter + i, 15); Canvas.LineTo(Gutter + i, 22);

      for k:=1 to 4 do begin
        u:=u + 254 / 10;
        if u > (PageWidth - LeftMargin - RightMargin) then break;
        i:=ToPix(u);
        Canvas.MoveTo(Gutter + i, 19); Canvas.LineTo(Gutter + i, 22);
      end;

      u:=u + 254 / 10;
    until u > (PageWidth - LeftMargin - RightMargin);
  end;
  ///////////////////////
  procedure VShowPixGrid;
  var k:integer;
  begin
    for k:=0 to ClientHeight - Gutter do begin
      if (k mod 100) = 0 then begin
        Canvas.MoveTo(8, Gutter + k); Canvas.LineTo(22, Gutter + k);
        s:=UnitsToText(k);
        ShowText(Canvas, Rect(3, Gutter + k + 2, 34, Gutter + k + Canvas.TextWidth(s)), 90, s);
      end else begin
        if (k mod 50) = 0 then begin
          Canvas.MoveTo(15, Gutter + k); Canvas.LineTo(22, Gutter + k);
        end else begin
          if (k mod 10) = 0 then begin
            Canvas.MoveTo(19, Gutter + k); Canvas.LineTo(22, Gutter + k);
          end;
        end;
      end;
    end;
  end;
  procedure VShowMMGrid;
  begin
    u:=0; i:=0;
    repeat
      Canvas.MoveTo(8, Gutter + i); Canvas.LineTo(22, Gutter + i);
      s:=UnitsToText(u / UnitRatio(uMM));
      ShowText(Canvas, Rect(3, Gutter + i + 2, 34, Gutter + i + Canvas.TextWidth(s)), 90, s);

      u:=u + 50; i:=ToPix(u);
      if i > (ClientHeight - Gutter) then break;
      Canvas.MoveTo(15, Gutter + i); Canvas.LineTo(22, Gutter + i);

      u:=u + 50; i:=ToPix(u);
    until i > (ClientHeight - Gutter);
  end;
  procedure VShowCMGrid;
  begin
    u:=0; i:=0;
    repeat
      Canvas.MoveTo(8, Gutter + i); Canvas.LineTo(22, Gutter + i);
      s:=UnitsToText(u / UnitRatio(uCM));
      ShowText(Canvas, Rect(3, Gutter + i + 2, 34, Gutter + i + Canvas.TextWidth(s)), 90, s);

      u:=u + 50; i:=ToPix(u);
      if i > (ClientHeight - Gutter) then break;
      Canvas.MoveTo(15, Gutter + i); Canvas.LineTo(22, Gutter + i);

      u:=u + 50; i:=ToPix(u);
    until i > (ClientHeight - Gutter);
  end;
  procedure VShowInGrid;
  var k:integer;
  begin
    u:=0; i:=0;
    repeat
      Canvas.MoveTo(8, Gutter + i); Canvas.LineTo(22, Gutter + i);
      s:=UnitsToText(u / UnitRatio(uIn));
      ShowText(Canvas, Rect(3, Gutter + i + 2, 34, Gutter + i + Canvas.TextWidth(s)), 90, s);

      for k:=1 to 4 do begin
        u:=u + 254 / 10; i:=ToPix(u);
        if i > (ClientHeight - Gutter) then break;
        Canvas.MoveTo(19, Gutter + i); Canvas.LineTo(22, Gutter + i);
      end;

      u:=u + 254 / 10; i:=ToPix(u);
      if i > (ClientHeight - Gutter) then break;
      Canvas.MoveTo(15, Gutter + i); Canvas.LineTo(22, Gutter + i);

      for k:=1 to 4 do begin
        u:=u + 254 / 10; i:=ToPix(u);
        if i > (ClientHeight - Gutter) then break;
        Canvas.MoveTo(19, Gutter + i); Canvas.LineTo(22, Gutter + i);
      end;

      u:=u + 254 / 10; i:=ToPix(u);
    until i > (ClientHeight - Gutter);
  end;
  ///////////////////////
  procedure ShowGrid;
  begin
    SetBkMode(Canvas.Handle, TRANSPARENT);
    case Orientation of
      rbHorizontal: begin
                      case Units of
                        uPix: HShowPixGrid;
                        uMM:  HShowMMGrid;
                        uCm:  HShowCMGrid;
                        uIn:  HShowInGrid;
                      end;
                    end;
      rbVertical:   begin
                      case Units of
                        uPix: VShowPixGrid;
                        uMM:  VShowMMGrid;
                        uCm:  VShowCMGrid;
                        uIn:  VShowInGrid;
                      end;
                    end;
    end;
    SetBkMode(Canvas.Handle, OPAQUE);
  end;
var a, b, c, d:integer;
begin
  inherited;

  Canvas.Font.Assign(Font);

  case Orientation of
    rbHorizontal:begin
      a:=Gutter - ToPix(LeftMargin);
      b:=Gutter;
      c:=Gutter - ToPix(LeftMargin) + ToPix(PageWidth) - ToPix(RightMargin);
      d:=Gutter - ToPix(LeftMargin) + ToPix(PageWidth);

      Canvas.Pen.Color  :=Color;
      Canvas.Brush.Color:=RulerColor;//$00C0FFFF;
      r:=Rect(a, 2, d, 22);
      Canvas.Rectangle(r);

      Canvas.Brush.Color:=clWhite;
      Canvas.Pen.Color:=clBlack;
      ShowGrid;

      Canvas.Pen.Color  :=Color;
      Canvas.Brush.Color:=Color;
      Canvas.Rectangle(d, 2, ClientWidth, 34);

      Canvas.Pen.Color:=clRed;
      Canvas.MoveTo(a, 0);
      Canvas.LineTo(a, ClientHeight);

      Canvas.Pen.Color:=clBlack;
      Canvas.MoveTo(b, 0);
      Canvas.LineTo(b, ClientHeight);

      Canvas.Pen.Color:=clRed;
      Canvas.MoveTo(c, 0);
      Canvas.LineTo(c, ClientHeight);

      Canvas.Pen.Color:=clBlack;
      Canvas.MoveTo(d, 0);
      Canvas.LineTo(d, ClientHeight);

      Canvas.Pen.Color  :=clBlack;
      Canvas.Brush.Color:=SelColor;//$00FFFFC0;
      Canvas.Rectangle(a + 1, 3, b, 21);
      Canvas.Rectangle(c + 1, 3, d, 21);
      Canvas.Rectangle(Gutter + ToPix(SelBegin), 22, Gutter + ToPix(SelEnd), 34);

      s:=Format('%f', [LeftMargin  / UnitRatio(Units)]);
      Canvas.TextRect(Rect(a + 2, 4, b - 1, 20), (a + b - Canvas.TextWidth(s)) div 2, 5, s);

      s:=Format('%f', [RightMargin  / UnitRatio(Units)]);
      Canvas.TextRect(Rect(c + 2, 4, d - 1, 20), (c + d - Canvas.TextWidth(s)) div 2, 5, s);

      s:=UnitsToText((SelEnd - SelBegin)  / UnitRatio(Units));
      Canvas.TextRect(Rect(Gutter + ToPix(SelBegin) + 1, 23, Gutter + ToPix(SelEnd) -1 , 33),
                      Gutter + (ToPix(SelBegin) + ToPix(SelEnd) - Canvas.TextWidth(s)) div 2,
                      21, s);

      SetBkMode(Canvas.Handle, TRANSPARENT);
      s:=UnitsToText(SelBegin / UnitRatio(Units));
      Canvas.TextOut(Gutter + ToPix(SelBegin) - Canvas.TextWidth(s) - 2, 21, s);
      SetBkMode(Canvas.Handle, OPAQUE);

      Canvas.Pen.Color:=clGreen;
      Canvas.MoveTo(FCur, 0);
      Canvas.LineTo(FCur, ClientHeight);
    end;
    rbVertical:begin
      Canvas.Pen.Color  :=Color;
      Canvas.Brush.Color:=RulerColor;//$00C0FFFF;
      r:=Rect(2, 1, 22, ClientHeight);
      Canvas.Rectangle(r);

      Canvas.Pen.Color  :=clBlack;
      Canvas.Brush.Color:=SelColor;//$00FFFFC0;
      Canvas.Rectangle(22, Gutter + ToPix(SelBegin), 34, Gutter + ToPix(SelEnd) + 1);

      Canvas.Brush.Color:=clWhite;
      Canvas.Pen.Color:=clBlack;
      ShowGrid;

      s:=UnitsToText((SelEnd - SelBegin)  / UnitRatio(Units));
      ShowText(Canvas, Rect(21, Gutter + ToPix(SelBegin), 34, Gutter + ToPix(SelEnd)), 90, s);

      SetBkMode(Canvas.Handle, TRANSPARENT);
      s:=UnitsToText(SelBegin / UnitRatio(Units));
      ShowText(Canvas, Rect(21, Gutter + ToPix(SelBegin) - Canvas.TextWidth(s) - 2, 34, Gutter + ToPix(SelBegin)), 90, s);
      SetBkMode(Canvas.Handle, OPAQUE);

      Canvas.Pen.Color:=clGreen;
      Canvas.MoveTo(0, FCur);
      Canvas.LineTo(ClientWidth, FCur);
    end;
  end;
end;

end.


