unit ARToolsCustomize;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ToolWin, ComCtrls, Buttons,

  ARActions, ARButtons, CheckLst, ExtCtrls;

type
  TARToolsCustomizeDlg = class(TForm)
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    CategoryLB: TListBox;
    CloseBtn: TBitBtn;
    Memo: TMemo;
    TBL: TCheckListBox;
    Button1: TButton;
    TBD3: TScrollBox;
    TBD: TToolBar;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure CloseBtnClick(Sender: TObject);
    procedure CategoryLBClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure TBLClickCheck(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    OldOnHint :TNotifyEvent;

    procedure ShowHint(Sender: TObject);
    procedure ControlEndDoc(Sender, Target: TObject; X, Y: Integer);
  public
    { Public declarations }
  end;

implementation

uses AREditorMain;

{$R *.DFM}

procedure TARToolsCustomizeDlg.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TARToolsCustomizeDlg.FormCreate(Sender: TObject);
begin
  CategoryLB.Items.Assign(AREditorDlg.ActList.CategoryList);
end;

procedure TARToolsCustomizeDlg.CategoryLBClick(Sender: TObject);
var cat:string;
    i:integer;
    ctrl:TDummyControl;
begin
  while TBD.ControlCount > 0 do begin
    TBD.Controls[0].Free;
  end;

  cat:=CategoryLB.Items[CategoryLB.ItemIndex];

  for i:=0 to AREditorDlg.ActList.Count - 1 do begin
    if ANSICompareText(cat, Flat(AREditorDlg.ActList.Item[i].Category)) = 0 then begin
      ctrl:=TDummyControl(AREditorDlg.ActList.Item[i].CreateControl(Self, TBD));
      if Assigned(ctrl) then begin
        ctrl.DragMode :=dmAutomatic;
        ctrl.DragKind :=dkDock;
        ctrl.OnEndDock:=ControlEndDoc;
        ctrl.Enabled  :=true;
      end;
    end;
  end;
end;

procedure TARToolsCustomizeDlg.ShowHint(Sender: TObject);
begin
  Memo.Text:=Application.Hint;
end;

procedure TARToolsCustomizeDlg.FormShow(Sender: TObject);
var i, j:integer;
    tbr:TToolBar;
begin
  OldOnHint:=Application.OnHint;
  Application.OnHint:=ShowHint;

  for i:=0 to AREditorDlg.ComponentCount - 1 do begin
    if AREditorDlg.Components[i] is TToolBar then begin
      tbr:=TToolBar(AREditorDlg.Components[i]);
      tbr.DockSite:=true;
      tbr.DragKind:=dkDock;
      tbr.DragMode:=dmAutomatic;
      for j:=0 to tbr.ControlCount - 1 do begin
        with TDummyControl(tbr.Controls[j]) do begin
          DragMode  :=dmAutomatic;
          DragKind  :=dkDock;
          OnEndDock :=ControlEndDoc;
        end;
      end;
    end;
  end;

  TBL.Items.Clear;

  for i:=0 to AREditorDlg.ActList.CategoryList.Count - 1 do begin
    tbr:=TToolBar(AREditorDlg.FindComponent(AREditorDlg.ActList.CategoryList[i] + 'ToolBar'));
    if Assigned(tbr) then begin
      TBL.Items.Add(AREditorDlg.ActList.CategoryList[i]);
      TBL.Checked[i]:=tbr.Visible;
      TBL.Items.Objects[i]:=tbr;
    end;
  end;
end;

procedure TARToolsCustomizeDlg.TBLClickCheck(Sender: TObject);
var tbr:TToolBar;
begin
  tbr:=TToolBar(TBL.Items.Objects[TBL.ItemIndex]);

  if Assigned(tbr) then begin
    if tbr.Visible then begin
      tbr.Visible:=false;
      TBL.Checked[TBL.ItemIndex]:=false;
    end else begin
      tbr.Visible:=true;
      TBL.Checked[TBL.ItemIndex]:=true;
    end;
  end;
end;

procedure TARToolsCustomizeDlg.FormClose(Sender: TObject; var Action: TCloseAction);
var i, j:integer;
    tbr:TToolBar;
begin
  Application.OnHint:=OldOnHint;

  for i:=0 to ComponentCount - 1 do begin
    if (Components[i] is TControl) and (TControl(Components[i]).Parent = TBD) then begin
      TControl(Components[i]).Free;
    end;
  end;

  for i:=0 to AREditorDlg.ComponentCount - 1 do begin
    if AREditorDlg.Components[i] is TToolBar then begin
      tbr:=TToolBar(AREditorDlg.Components[i]);
      tbr.DockSite:=false;
      for j:=0 to tbr.ControlCount - 1 do begin
        with TDummyControl(tbr.Controls[j]) do begin
          DragMode:=dmManual;
          DragKind:=dkDrag;
          OnEndDock:=nil;
        end;
      end;
    end;
  end;

  Action:=caFree;
end;

procedure TARToolsCustomizeDlg.ControlEndDoc(Sender, Target: TObject; X, Y: Integer);
begin
  if (Sender is TControl) and (not (TControl(Target) is TBD.ClassType{TToolBar})) then begin
    TControl(Sender).ManualDock(TBD);
  end else begin
    AREditorMain.ResizeToolBar(TToolBar(TControl(Sender).Parent));
  end;
end;

procedure TARToolsCustomizeDlg.Button1Click(Sender: TObject);
begin
  AREditorDlg.RestoreDefaultTools;
end;

end.
