unit ARUndo;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes,

  ARUtils;

const
  UndoLimitDefault = 128;//32767;

type
  TUndo = class
  private
    Buf         :TAIStringList;
    Pos         :integer;
    Capacity    :integer;
    function    GetPos(p:integer):integer;
  public
    constructor Create(UndoLimit:integer);
    destructor  Destroy;override;

    procedure   Clear;
    function    CanUndo:boolean;
    function    CanRedo:boolean;
    function    GetUndo:string;
    function    GetRedo:string;
    procedure   Add(s:string);
  end;

implementation

function    TUndo.GetPos(p:integer):integer;
begin
  while p < 0  do inc(p, Capacity);
  while p >= Capacity do dec(p, Capacity);
  Result:=p;
end;

constructor TUndo.Create(UndoLimit:integer);
var i:integer;
begin
  Buf     :=TAIStringList.Create('TUndo.Buf');
  Pos     :=0;
  Capacity:=(UndoLimit);

  for i:=0 to Capacity - 1 do Buf.AddObject('', nil);
end;

destructor  TUndo.Destroy;
begin
  Buf.Free;
end;

procedure   TUndo.Clear;
var i:integer;
begin
  for i:=0 to Capacity - 1 do Buf.AddObject('', nil);
  Pos:=0;
end;

function    TUndo.CanUndo:boolean;
begin
  Result:=Buf.Objects[GetPos(pos - 1)] <> nil;
end;

function    TUndo.CanRedo:boolean;
begin
  Result:=Buf.Objects[GetPos(pos + 1)] <> nil;
end;

function    TUndo.GetUndo:string;
begin
  Result:='';
  if CanUndo then begin
    pos:=GetPos(pos - 1);
    Result:=Buf[pos];
  end;
end;

function    TUndo.GetRedo:string;
begin
  Result:='';
  if CanRedo then begin
    pos:=GetPos(pos + 1);
    Result:=Buf[pos];
  end;
end;

procedure   TUndo.Add(s:string);
begin
  pos:=GetPos(pos + 1);

  Buf[pos]:=s;
  Buf.Objects[pos]:=pointer(-1);

  Buf[GetPos(pos + 1)]:='';
  Buf.Objects[GetPos(pos + 1)]:=nil;
end;

end.
