unit ARUserRep;
//************************************************************//
//                         AND Report                        //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Classes, Controls, Graphics, DB, SysUtils, Math, 

  ARRepCommon, ARUtils, ARRep, ARGridRep;

type

  TARUserRep = class(TARGridRep)
  private
    PictureList :TAIStringList;
    GroupBands  :TAIStringList;

    oldSrcRep   :TARRep;

    NextTopBandState:TARBandState;

    FPath       :string;

//    function    GetCell(BandName:string; IdxBand:integer; ValName:string):TARCell;
    function    GetBandCount(BandName:string):integer;

  protected
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer);override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);override;
    procedure   Loaded;override;
    procedure   InternalAddBands(BandName:string; bandlist:TAIList);

    function    GetTemplate:TARTemplate;override;
    procedure   SetTemplate(ATemplate:TARTemplate);override;
    procedure   ApplyValues(src:TARCell; dst:TARCell);override;

    procedure   SetPath(APath:string);

    function    GetS(VarName:string):string;
    procedure   SetS(VarName:string; VarValue:string);

    function    GetRTF(VarName:string):string;
    procedure   SetRTF(VarName:string; VarValue:string);

    function    GetF(VarName:string):extended;
    procedure   SetF(VarName:string; VarValue:extended);

    function    GetI(VarName:string):integer;
    procedure   SetI(VarName:string; VarValue:integer);override;

    function    GetD(VarName:string):TDateTime;
    procedure   SetD(VarName:string; VarValue:TDateTime);

    function    GetT(VarName:string):TDateTime;
    procedure   SetT(VarName:string; VarValue:TDateTime);

    function    GetC(VarName:string):currency;
    procedure   SetC(VarName:string; VarValue:currency);

    function    GetM(VarName:string):currency;
    procedure   SetM(VarName:string; VarValue:currency);

    function    GetG(VarName:string):TGraphic;
    procedure   SetG(VarName:string; VarValue:TGraphic);

    function    GetU(VarName:string):integer;
    procedure   SetU(VarName:string; UserVarValue:integer);

    function    GetTitle:string;
    procedure   SetTitle(ATitle:string);

//    function    CreateCanvasImage(JobName:string):boolean;
  public
    Vars        :TAIStringList;

    procedure   SetDSValues(DS:TDataSet);

    procedure   Load(FileName:string);

    function    AddBands(BandName:string; ToBand:TARBand = nil; Expanded:boolean = true):TARBand;
    function    AddSubBands(BandName:string; Expanded:boolean):TARBand;
    function    AddSubBandsToBand(BandName:string; ToBand:TARBand; Expanded:boolean):TARBand;

    function    AddGroupBands(const BandName:string; Value:variant; ToBand:TARBand = nil; Expanded:boolean = true):TARBand;
    function    NewGroup(const BandName:string; Value:variant):boolean;

    function    InsertBands(InsertIdxBand:integer; BandName:string):integer;
    procedure   DeleteBand(IdxBand:integer);
    procedure   DeleteBands;
    procedure   ClearValues;

    procedure   MoveToPriorBand;override;
    procedure   MoveToNextBand;override;

    constructor Create(AOwner:TComponent);override;
    destructor  Destroy;override;

//    property    Cells[BandName:string; IdxBand:integer; ValName:string]:TARCell read GetCell;
    function    Cells(ValName:string):TARCell;overload;
    function    Cells(IdxBand:integer; IdxCell:integer):TARCell;overload;
    procedure   SetFocus;override;

    procedure   SaveToFile(FFile :string);
    procedure   LoadFromFile(FFile :string);

    procedure   PageBreak;
    procedure   BeginPara;
    procedure   EndPara;

    property    S[VarName:string]:string            read  GetS      write SetS;
    property    RTF[VarName:string]:string          read  GetRTF    write SetRTF;
    property    F[VarName:string]:extended          read  GetF      write SetF;
    property    I[VarName:string]:integer           read  GetI      write SetI;
    property    D[VarName:string]:TDateTime         read  GetD      write SetD;
    property    T[VarName:string]:TDateTime         read  GetT      write SetT;
    property    G[VarName:string]:TGraphic          read  GetG      write SetG;
    property    C[VarName:string]:currency          read  GetC      write SetC;
    property    M[VarName:string]:currency          read  GetM      write SetM;

    property    U[VarName:string]:integer           read  GetU      write SetU;
    property    BandCount[BandName:string]:integer  read GetBandCount;
  published
    property    Path:string                         read  FPath     write SetPath;
    property    Title:string                        read  GetTitle  write SetTitle;
  end;

procedure Register;

implementation
{$R *.RES}

procedure Register;
begin
  RegisterComponents('ARComponents', [TARUserRep]);
end;

type  TValueFlag = (vfNone, vfMoney, vfRTF);

constructor TARUserRep.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

  NextTopBandState:=[];

  Gutter        :=0;

//  ShowBandTitle :=false;
  ShowBandTitle :=true;

  Vars          :=TAIStringList.Create('TARUserRep.Create Vars');
  PictureList   :=TAIStringList.Create('TARUserRep.Create PictureList');
  PictureList.Sorted:=true;
  GroupBands    :=TAIStringList.Create('TARUserRep.Create GroupBands');

  OrgRep        :=TARRep.Create;
  oldSrcRep     :=nil;

  DeleteBands;
end;

destructor  TARUserRep.Destroy;
var i:integer;
begin
  ClearValues;

  GroupBands.Free;

  for i:=0 to PictureList.Count - 1 do begin
    TPicture(PictureList.Objects[i]).Free;
  end;
  PictureList.Free;

  Vars.Free;

  if Assigned(oldSrcRep) then begin
    SrcRep:=oldSrcRep;
  end;

  OrgRep.Free;

  inherited;
end;

procedure   TARUserRep.MouseMove(Shift: TShiftState; X, Y: Integer);
var cell:TARCell;
    idxband,idxcell:integer;
begin
  if ssLeft in Shift then begin
    cell:=MouseToCell(X, Y, idxband, idxcell);
    if Assigned(cell) and (CurrCell <> cell) then begin
      CurrBandIdx:=idxband;
      CurrCellIdx:=idxcell;
//      SelAdd(CurrCell);
      SelAdd(CurrBandIdx, CurrCellIdx);
      if Assigned(OnSelectCell) then OnSelectCell(Self, CurrBandIdx, CurrCellIdx);
    end;
  end;
end;

procedure   TARUserRep.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var idxband,idxcell:integer;
begin
  inherited;

  if Assigned(MouseToCell(X, Y, idxband, idxcell)) then begin
    CurrBandIdx :=idxband;
    CurrCellIdx :=idxcell;
    if not (ssCtrl in Shift) then ClearSel;
    SelAdd(CurrBandIdx, CurrCellIdx);
    if Assigned(OnSelectCell) then OnSelectCell(Self, CurrBandIdx, CurrCellIdx);
  end else begin
    if (idxband <> -1) and (X < Gutter) then begin

      if SrcRep[idxband].Expandable then begin
        SrcRep[idxband].Expanded:= not SrcRep[idxband].Expanded;
        Invalidate;
        if SrcRep[idxband].Expanded then begin
          if Assigned(OnExpandBand) then OnExpandBand(Self, idxband);
        end else begin
          if Assigned(OnCollapseBand) then OnCollapseBand(Self, idxband);
        end;
      end else begin
        if Assigned(OnSelectBand) then OnSelectBand(Self, idxband);
      end;
    end;
  end;
end;

procedure   TARUserRep.DeleteBands;
begin
  inherited DeleteBands;
  Invalidate;
end;

procedure   TARUserRep.ClearValues;
var i:integer;
begin
  Vars.Clear;

  for i:=0 to PictureList.Count - 1 do begin
    TPicture(PictureList.Objects[i]).Free;
  end;
  PictureList.Clear;
end;
//////////////////////////////////////////
function    TARUserRep.GetS(VarName:string):string;
begin
  Result:=Vars.Values[ANSIUpperCase(VarName)];
end;

function    TARUserRep.GetRTF(VarName:string):string;
begin
  Result:=Vars.Values[ANSIUpperCase(VarName)];
end;

function    TARUserRep.GetF(VarName:string):extended;
begin
  try
    Result:=StrToFloat(Vars.Values[ANSIUpperCase(VarName)]);
  except
    Result:=0.0;
  end;
end;

function    TARUserRep.GetC(VarName:string):currency;
begin
  try
    Result:=StrToCurr(Vars.Values[ANSIUpperCase(VarName)]);
  except
    Result:=0.0;
  end;
end;

function    TARUserRep.GetM(VarName:string):currency;
var p:integer;
    s:string;
begin
  try
    s:=Vars.Values[ANSIUpperCase(VarName)];
    p:=System.Pos(ARMoneySeparator, s);
    if p > 0 then begin
      s[p]:=DecimalSeparator;
    end;
    Result:=StrToCurr(s);
  except
    Result:=0.0;
  end;
end;

function    TARUserRep.GetI(VarName:string):integer;
begin
  try
    Result:=StrToInt(Vars.Values[ANSIUpperCase(VarName)]);
  except
    Result:=0;
  end;
end;

function    TARUserRep.GetD(VarName:string):TDateTime;
begin
  try
    Result:=StrToDate(Vars.Values[ANSIUpperCase(VarName)]);
  except
    Result:=Now;
  end;
end;

function    TARUserRep.GetT(VarName:string):TDateTime;
begin
  try
    Result:=StrToDateTime(Vars.Values[ANSIUpperCase(VarName)]);
  except
    Result:=Now;
  end;
end;

procedure   TARUserRep.SetS(VarName:string; VarValue:string);
var idx:integer;
begin
  Vars.Values[ANSIUpperCase(VarName)]:=VarValue;

  idx:=Vars.IndexOfName(ANSIUpperCase(VarName));
  if idx >= 0 then begin
    Vars.Objects[idx]:=pointer(Ord(vfNone));
  end;
end;

procedure   TARUserRep.SetRTF(VarName:string; VarValue:string);
var idx:integer;
begin
  Vars.Values[ANSIUpperCase(VarName)]:=VarValue;

  idx:=Vars.IndexOfName(ANSIUpperCase(VarName));
  if idx >= 0 then begin
    Vars.Objects[idx]:=pointer(Ord(vfRTF));
  end;
end;

procedure   TARUserRep.SetF(VarName:string; VarValue:extended);
begin
  Vars.Values[ANSIUpperCase(VarName)]:=Format('%f', [VarValue]);
  Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfNone));
end;

procedure   TARUserRep.SetI(VarName:string; VarValue:integer);
begin
  Vars.Values[ANSIUpperCase(VarName)]:=IntToStr(VarValue);
  Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfNone));
end;

procedure   TARUserRep.SetD(VarName:string; VarValue:TDateTime);
begin
  Vars.Values[ANSIUpperCase(VarName)]:=DateToStr(VarValue);
  Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfNone));
end;

procedure   TARUserRep.SetT(VarName:string; VarValue:TDateTime);
begin
  Vars.Values[ANSIUpperCase(VarName)]:=DateTimeToStr(VarValue);
  Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfNone));
end;

procedure   TARUserRep.SetC(VarName:string; VarValue:currency);
begin
  Vars.Values[ANSIUpperCase(VarName)]:=Format('%m', [VarValue]);
  Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfNone));
end;

procedure   TARUserRep.SetM(VarName:string; VarValue:currency);
var OldDecSep:char;
begin
  OldDecSep:=DecimalSeparator;
  DecimalSeparator:=ARMoneySeparator;
  try
    Vars.Values[ANSIUpperCase(VarName)]:=Format('%.2f', [VarValue]);
    Vars.Objects[Vars.IndexOfName(ANSIUpperCase(VarName))]:=pointer(Ord(vfMoney));
  finally
    DecimalSeparator:=OldDecSep;
  end;
end;

function    TARUserRep.GetU(VarName:string):integer;
var idx:integer;
begin
  VarName:=ANSIUpperCase(VarName);
  idx:=SrcRep.UserVarList.IndexOf(VarName);
  if idx = -1 then begin
    Result:=0;
  end else begin
    Result:=integer(SrcRep.UserVarList.Objects[idx]);
  end;
end;

procedure   TARUserRep.SetU(VarName:string; UserVarValue:integer);
var idx:integer;
begin
  VarName:=ANSIUpperCase(VarName);
  idx:=SrcRep.UserVarList.IndexOf(VarName);
  if idx = -1 then begin
    SrcRep.UserVarList.AddObject(VarName, pointer(UserVarValue));
  end else begin
    SrcRep.UserVarList.Objects[idx]:=pointer(UserVarValue);
  end;
end;

function    TARUserRep.GetG(VarName:string):TGraphic;
var idx:integer;
begin
  Result:=nil;
  idx:=PictureList.IndexOf(ANSIUpperCase(VarName));
  if idx <> -1 then begin
    Result:=TPicture(PictureList.Objects[idx]).Graphic;
  end;
end;

procedure   TARUserRep.SetG(VarName:string; VarValue:TGraphic);
var idx:integer;
begin
  VarName:=ANSIUpperCase(VarName);
  idx:=PictureList.IndexOf(VarName);

  if idx = -1 then begin
    PictureList.AddObject(VarName, TPicture.Create);
    idx:=PictureList.IndexOf(VarName);
  end;
  TPicture(PictureList.Objects[idx]).Assign(VarValue);
end;

function    TARUserRep.GetTitle:string;
begin
  Result:=prn.Title;
end;

procedure   TARUserRep.SetTitle(ATitle:string);
begin
  prn.Title:=ATitle;
end;

procedure   TARUserRep.DeleteBand(IdxBand:integer);
begin
  SrcRep.DeleteBand(IdxBand);
  Invalidate;
end;

////////////////
procedure TARUserRep.InternalAddBands(BandName:string; bandlist:TAIList);
var idxband,idxcell:integer;
    b:TARBand;
    H:integer;
begin
  SetARRTF;
  SrcRep.Font.Assign(OrgRep.Font);
  for idxband:=0 to OrgRep.BandCount - 1 do begin
    if CT(OrgRep.Bands[idxband].Name, BandName) then begin
      b:=TARBand.Create;
      bandlist.Add(b);
      b.Assign(OrgRep.Bands[idxband]);
      for idxcell:=0 to b.CellCount - 1 do begin
        ApplyValues(OrgRep.Bands[idxband].Cells[idxcell], b.Cells[idxcell]);
        b.Cells[idxcell].ProcessValue;

        if b.Cells[idxcell].AutoStretch then begin
          Canvas.Font.Size    :=b.Cells[idxcell].FontSize;
          Canvas.Font.Style   :=b.Cells[idxcell].FontStyle;
          Canvas.Font.Name    :=b.Cells[idxcell].FontName;
          Canvas.Font.Charset :=Self.SrcRep.Font.Charset;

          H:=b.Cells[idxcell].Height(Canvas);
          if H > b.Height then begin
            b.Height:=H;
          end;
        end;
      end;
    end;
  end;
end;

function    TARUserRep.AddSubBands(BandName:string; Expanded:boolean):TARBand;
var b:TARBand;
begin
  Result:=nil;
  if SrcRep.BandCount > 0 then begin
    b:=SrcRep.Bands[SrcRep.BandCount - 1];

    Result:=AddSubBandsToBand(BandName, b, Expanded);
  end;
end;

function  TARUserRep.AddBands(BandName:string; ToBand:TARBand = nil; Expanded:boolean = true):TARBand;
var i:integer;
    bl:TAIList;
begin
  Result:=nil;
  bl:=TAIList.Create('TARUserRep.AddBands bl');
  try
    InternalAddBands(BandName, bl);

    if Assigned(ToBand) then begin
      ToBand.Expanded:=Expanded;
      for i:=0 to bl.Count - 1 do ToBand.Add(bl[i]);
    end else begin
      for i:=0 to bl.Count - 1 do SrcRep.Add(bl[i]);
    end;
    if bl.Count > 0 then begin
      Result:=bl[bl.Count - 1];

      if not Assigned(ToBand) then begin
        // set band state to first added band
        TARBand(bl[0]).BandState:=NextTopBandState;
        NextTopBandState:=[];
      end;
    end;
  finally
    bl.Free;
  end;
  Invalidate;
end;
function    TARUserRep.AddSubBandsToBand(BandName:string; ToBand:TARBand; Expanded:boolean):TARBand;
var i:integer;
    bl:TAIList;
begin
  Result:=nil;
  bl:=TAIList.Create('TARUserRep.AddSubBands bl');
  try
    InternalAddBands(BandName, bl);
    ToBand.Expanded:=Expanded;
    for i:=0 to bl.Count - 1 do ToBand.Add(bl[i]);

    if bl.Count > 0 then begin
      Result:=bl[bl.Count - 1];
    end;
  finally
    bl.Free;
  end;
  Invalidate;
end;

function    TARUserRep.InsertBands(InsertIdxBand:integer; BandName:string):integer;
var  idxband,idxcell:integer;
    b:TARBand;
    H:integer;
begin
  Result:=0;
  SrcRep.Font.Assign(OrgRep.Font);
  for idxband:=0 to OrgRep.BandCount - 1 do begin
    if CT(OrgRep.Bands[idxband].Name, BandName) then begin
			inc(Result);
      b:=NewBand;
      if InsertIdxBand < SrcRep.BandCount - 1 then begin
        SrcRep.Move(SrcRep.BandCount - 1, InsertIdxBand + 1);
      end;
      b.Assign(OrgRep.Bands[idxband]);
      for idxcell:=0 to b.CellCount - 1 do begin
        ApplyValues(OrgRep.Bands[idxband].Cells[idxcell], b.Cells[idxcell]);
        b.Cells[idxcell].ProcessValue;

        if b.Cells[idxcell].AutoStretch then begin
          Canvas.Font.Size    :=b.Cells[idxcell].FontSize;
          Canvas.Font.Style   :=b.Cells[idxcell].FontStyle;
          Canvas.Font.Name    :=b.Cells[idxcell].FontName;
          Canvas.Font.Charset :=Self.SrcRep.Font.Charset;

          H:=b.Cells[idxcell].Height(Canvas);
          if H > b.Height then begin
            b.Height:=H;
          end;
        end;
      end;
    end;
  end;
  Invalidate;
end;

procedure TARUserRep.ApplyValues(src:TARCell; dst:TARCell);
var p:integer;
    i,idx:integer;
begin
  inherited;
  
  dst.Value:='';
  dst.IsMoneyValue:=false;

  if Assigned(src.VarNameList) then begin
    for i:=0 to src.VarNameList.Count - 1 do begin
      if integer(src.VarNameList.Objects[i]) = 0 then begin
        dst.Value:=dst.Value + src.VarNameList[i];
      end else begin
        dst.Value:=dst.Value + S[src.VarNameList[i]];
        idx:=Vars.IndexOfName(src.VarNameList[i]);
        if idx >= 0 then begin
          dst.IsMoneyValue:=Vars.Objects[idx] = pointer(Ord(vfMoney));
          dst.IsRTF:=Vars.Objects[idx] = pointer(Ord(vfRTF));
        end;
        if Assigned(G[src.VarNameList[i]]) then begin
          dst.Picture.Assign(G[src.VarNameList[i]]);
        end;
      end;
    end;
  end;

  dst.UserVarValue:=-1;
  p:=SrcRep.UserVarList.IndexOf(dst.UserVar);
  if p <> -1 then begin
    dst.UserVarValue:=integer(SrcRep.UserVarList.Objects[p]);
  end;
end;

function    TARUserRep.GetTemplate:TARTemplate;
begin
  Result:=OrgRep.ARTemplate;
end;

procedure   TARUserRep.SetTemplate(ATemplate:TARTemplate);
begin
  OrgRep.ARTemplate:=ATemplate;
  Invalidate;
end;

function    TARUserRep.GetBandCount(BandName:string):integer;
var IdxBand:integer;
begin
  Result:=0;
  for IdxBand:=0 to SrcRep.BandCount - 1 do begin
    if CT(BandName, SrcRep.Bands[IdxBand].Name) then inc(Result);
  end;
end;

procedure   TARUserRep.Loaded;
var IdxBand, IdxCell:integer;
begin
  inherited Loaded;

  if csDesigning in ComponentState then begin
    oldSrcRep:=SrcRep;
    SrcRep:=OrgRep;
  end else begin
    for IdxBand:=0 to OrgRep.BandCount - 1 do begin
      for IdxCell:=0 to OrgRep.Bands[IdxBand].CellCount - 1 do begin
        OrgRep.Bands[IdxBand].Cells[IdxCell].CreateVarNameList;
      end;
    end;
  end;
end;

procedure   TARUserRep.SaveToFile(FFile :string);
begin
  SrcRep.Save(FFile);
end;

procedure   TARUserRep.LoadFromFile(FFile :string);
begin
  SrcRep.Load(FFile);
  NextTopBandState:=[];
end;

function    TARUserRep.NewGroup(const BandName:string; Value:variant):boolean;
begin
  Result:=not CT(GroupBands.Values[BandName], VarAsType(Value, varString));
end;

function    TARUserRep.AddGroupBands(const BandName:string; Value:variant; ToBand:TARBand = nil; Expanded:boolean = true):TARBand;
var idx, j:integer;
begin
  Result:=nil;
  if not CT(GroupBands.Values[BandName], VarAsType(Value, varString)) then begin

    S[BandName]:=VarAsType(Value, varString);
    Result:=AddBands(BandName, ToBand, Expanded);

    GroupBands.Values[BandName]:=VarAsType(Value, varString);
    idx:=GroupBands.IndexOfName(BandName);
    for j:=idx + 1 to GroupBands.Count - 1 do begin
      GroupBands[j]:='';
    end;
  end;
end;

procedure   TARUserRep.SetDSValues(DS:TDataSet);
var i:integer;
    Picture:TPicture;
begin
  for i:=0 to DS.FieldCount - 1 do begin
    case DS.Fields[i].DataType of
      ftSmallint,
      ftInteger,
      ftWord,
      ftAutoInc,
      ftLargeInt,
      ftGuid,
      ftBCD:        Self.I[DS.Fields[i].FieldName]:=DS.Fields[i].AsInteger;

      ftBoolean:	  if DS.Fields[i].AsBoolean then begin
                      Self.S[DS.Fields[i].FieldName]:='True';
                    end else begin
                      Self.S[DS.Fields[i].FieldName]:='False';
                    end;

      ftString,
      ftMemo,
      ftFmtMemo,
      ftFixedChar,
      ftWideString: Self.S[DS.Fields[i].FieldName]:=DS.Fields[i].AsString;

      ftFloat:      Self.F[DS.Fields[i].FieldName]:=DS.Fields[i].AsFloat;

      ftCurrency:	  Self.C[DS.Fields[i].FieldName]:=DS.Fields[i].AsCurrency;

      ftDate:       Self.D[DS.Fields[i].FieldName]:=DS.Fields[i].AsDateTime;

      ftTime,
      ftDateTime:   Self.T[DS.Fields[i].FieldName]:=DS.Fields[i].AsDateTime;

      ftBlob,
      ftTypedBinary,
      ftGraphic:    begin
                      Picture:=TPicture.Create;
                      try
                        Picture.Assign(DS.Fields[i] as TBlobField);
                        Self.G[DS.Fields[i].FieldName]:=Picture.Graphic;
                      finally
                        Picture.Free;
                      end;
                    end;

      else          begin
                      Self.S[DS.Fields[i].FieldName]:='Unsupported format';
                    end;
//      ftUnknown	Unknown or undetermined
//      ftBytes	Fixed number of bytes (binary storage)
//      ftVarBytes	Variable number of bytes (binary storage)
//      ftParadoxOle	Paradox OLE field
//      ftDBaseOle	dBASE OLE field
//      ftADT	Abstract Data Type field
//      ftArray	Array field
//      ftReference	REF field
//      ftDataSet	DataSet field
//      ftOraBlob	BLOB fields in Oracle 8 tables
//      ftOraClob	CLOB fields in Oracle 8 tables
//      ftVariant	Data of unknown or undetermined type
//      ftInterface	References to interfaces (IUnknown)
//      ftIDispatch	References to IDispatch interfaces
    end;
  end;
end;

{
function    TARUserRep.GetCell(BandName:string; IdxBand:integer; ValName:string):TARCell;
var AIdxBand, IdxBandOrg, IdxCell, IdxVar, Idx:integer;
begin
  Result:=nil;
  for AIdxBand:=0 to SrcRep.BandCount - 1 do begin
    if CT(BandName, SrcRep.Bands[AIdxBand].Name) then begin
      if IdxBand = 0 then begin //band found
        for IdxBandOrg:=0 to OrgRep.BandCount - 1 do begin  //look for template
          if CT(BandName, OrgRep.Bands[IdxBandOrg].Name) then begin //band found
            for IdxCell:=0 to OrgRep.Bands[IdxBandOrg].CellCount - 1 do begin //look for cell
              for IdxVar:=0 to OrgRep.Bands[IdxBandOrg].Cells[IdxCell].VarNameList.Count - 1 do begin //look for vars
                Idx:=OrgRep.Bands[IdxBandOrg].Cells[IdxCell].VarNameList.IndexOf(ValName);
                if Idx >= 0 then begin //var found
                  Result:=SrcRep.Bands[AIdxBand].Cells[IdxCell];
                  exit;
                end;
              end;
            end;
            exit; //nothing found
          end;
        end;
      end;
      dec(IdxBand);
    end;
  end;
end;
}

function    TARUserRep.Cells(ValName:string):TARCell;
var IdxBandOrg, IdxCell, IdxVar, Idx:integer;
begin
  Result:=nil;

  for IdxBandOrg:=0 to OrgRep.BandCount - 1 do begin  //look for template
    if CT(CurrBand.Name, OrgRep.Bands[IdxBandOrg].Name) then begin //band found
      for IdxCell:=0 to OrgRep.Bands[IdxBandOrg].CellCount - 1 do begin //look for cell
        for IdxVar:=0 to OrgRep.Bands[IdxBandOrg].Cells[IdxCell].VarNameList.Count - 1 do begin //look for vars
          Idx:=OrgRep.Bands[IdxBandOrg].Cells[IdxCell].VarNameList.IndexOf(ValName);
          if Idx >= 0 then begin //var found
            Result:=CurrBand.Cells[IdxCell];
            exit;
          end;
        end;
      end;
      exit; //nothing found
    end;
  end;
end;

procedure   TARUserRep.SetFocus;
begin
  inherited;
  
  ClearSel;
  SelAdd(CurrBandIdx, CurrCellIdx);
end;

procedure   TARUserRep.Load(FileName:string);
var IdxBand, IdxCell:integer;
begin
  FPath:=FileName;

//  NextTopBandState:=arbsNone;
  OrgRep.FileSystem:=Self.FileSystem;
  OrgRep.Load(FileName);
  prn.PaperSize    :=OrgRep.PaperSize;
  prn.Orientation  :=OrgRep.Orientation;

  for IdxBand:=0 to OrgRep.BandCount - 1 do begin
    for IdxCell:=0 to OrgRep.Bands[IdxBand].CellCount - 1 do begin
      OrgRep.Bands[IdxBand].Cells[IdxCell].CreateVarNameList;
    end;
  end;

  SrcRep.Font.Assign(OrgRep.Font);
  Repaint;
end;

procedure   TARUserRep.SetPath(APath:string);
begin
  Load(APath);
end;

function    TARUserRep.Cells(IdxBand:integer; IdxCell:integer):TARCell;
begin
  Result:=SrcRep.Cells[IdxBand, IdxCell];
end;

procedure   TARUserRep.PageBreak;
begin
  NextTopBandState:=NextTopBandState + [arbsPageBreak];
end;

procedure   TARUserRep.BeginPara;
begin
  EndPara;

  NextTopBandState:=NextTopBandState + [arbsBeginPara];
end;

procedure   TARUserRep.EndPara;
begin
  if SrcRep.BandCount > 0 then begin
    SrcRep.Bands[SrcRep.BandCount - 1].BandState:=SrcRep.Bands[SrcRep.BandCount - 1].BandState + [arbsEndPara];
  end;
  NextTopBandState:=NextTopBandState - [arbsEndPara];
end;
//////////////////////////////////////////////////////////////////////////////////////
procedure   TARUserRep.MoveToPriorBand;
begin
  CurrBandIdx:=Max(SrcRep.IndexOf(CurrCell.GetTopBand) - 1, 0);
end;

procedure   TARUserRep.MoveToNextBand;
begin
  CurrBandIdx:=Min(SrcRep.IndexOf(CurrCell.GetBottomBand) + 1, SrcRep.BandCount - 1);
end;

end.

