unit SQLIB;
//************************************************************//
//                         AND Interpreter                   //
//  Eugene Hrulev (mrHru) Laxian Key, Omsk, Russia 1999,2001 //
//  Oleg Melnikov (muler) Laxian Key, Omsk, Russia 1999,2001 //
//************************************************************//

interface

uses
  Windows, Classes, Forms, Controls, IBSQL, IBDatabase, SysUtils,
  Graphics;

var       DefaultIBDB:TIBDatabase;
         	ISQLCnt:integer;
const   	MoneyPrecision:integer=100;

function  NewTr(AIBDB:TIBDatabase = nil):TIBTransaction;
function  MR(money:currency):currency;  //money rounder
function  GetUniqueNumber: integer;forward;

type
	TISQL = class
	private
		Tag         :integer;
		Q           :TIBSQL;
    Picture     :TPicture;

		FOwner			:TObject;

		function  GetI (const FieldName:string):integer;
		procedure SetI (const FieldName:string;v:integer);

		function  GetS (const FieldName:string):string;
		procedure SetS (const FieldName:string;v:string);

//		function  GetT (const FieldName:string):string;
//		procedure SetT (const FieldName:string;v:string);

		function  GetF (const FieldName:string):extended;
		procedure SetF (const FieldName:string;v:extended);

		function  GetC (const FieldName:string):currency;
		procedure SetC (const FieldName:string;v:currency);

		function  GetM (const FieldName:string):currency;
		procedure SetM (const FieldName:string;v:currency);

		function  GetD (const FieldName:string):TDateTime;
		procedure SetD (const FieldName:string;v:TDateTime);

		function  GetB (const FieldName:string):boolean;
		procedure SetB (const FieldName:string;v:boolean);

		function  GetV (const FieldName:string):variant;
		procedure SetV (const FieldName:string;v:variant);

		function  GetG (const FieldName:string):TGraphic;
//		function  GetDataSet:TDataSet;

		function	GetSQL:TStrings;

		function	GetEOF	:boolean;
		function	GetBOF	:boolean;
		function	GetEmpty:boolean;

    function	GetTr:TIBTransaction;
    procedure SetTr(ATr:TIBTransaction);

    function	GetFields(const Idx: Integer):TIBXSQLVAR;
		function	GetActive:boolean;
	public
		ss:string;

    AutoTrans :boolean;
    property  Tr:TIBTransaction read GetTr  write SetTr;
    property  Fields[const Idx: Integer]: TIBXSQLVAR  read GetFields;

		procedure Prepare;
		procedure Close;
		procedure Open;
		procedure Exec;

		procedure New;
		procedure Add(s:string);

		procedure Next;

		function  RecordCount:integer;

		function  CheckDataBase:boolean;

		property  I[const FieldName: string]:integer    read GetI write SetI;
		property  S[const FieldName: string]:string     read GetS write SetS;
		property  F[const FieldName: string]:extended   read GetF write SetF;
		property  C[const FieldName: string]:currency   read GetC write SetC;
		property  M[const FieldName: string]:currency   read GetM write SetM;
		property  D[const FieldName: string]:TDateTime  read GetD write SetD;
		property  B[const FieldName: string]:boolean		read GetB write SetB;
		property  V[const FieldName: string]:variant    read GetV write SetV;
		property  G[const FieldName: string]:TGraphic 	read GetG;

//		property  T[const FieldName: string]:string     read GetT write SetT;	//read from BLOB

		property	Active:boolean                       	read GetActive;
		property	EOF:boolean 	                        read GetEOF;
		property	BOF:boolean 	                        read GetBOF;
		property	Empty:boolean                         read GetEmpty;

		property  SQL:TStrings        read GetSQL;
    function  Current:TIBXSQLDA;

		property	Owner:TObject       read FOwner			write	FOwner;

    procedure BatchOutput(OutputObject: TIBBatchOutput);
    procedure BatchInput (InputObject:  TIBBatchInput);

		procedure StartTransaction;
		procedure Commit;
    procedure CommitRetaining;
		procedure Rollback;
    procedure RollbackRetaining;

		constructor Create(AIBDB:TIBDatabase; ATr:TIBTransaction; AAutoTrans:boolean = false);overload;
		constructor Create(AIBDB:TIBDatabase);overload;
		constructor Create(ATr:TIBTransaction);overload;
		constructor Create(AISQL:TISQL);overload;
		constructor Create;overload;
		destructor  Destroy;override;
	end;

implementation

constructor TISQL.Create(AIBDB:TIBDatabase; ATr:TIBTransaction; AAutoTrans:boolean = false);
begin
	inc(ISQLCnt);
	Q         :=TIBSQL.Create(nil);
  Q.Name    :='TIBSQL_' + IntToStr(GetUniqueNumber);
	Tag       :=GetUniqueNumber;
  AutoTrans :=AAutoTrans;

	Q.Database:=AIBDB;
  Q.Transaction:=ATr;
  if AutoTrans then Q.InsertComponent(Atr);
end;

constructor TISQL.Create(AIBDB:TIBDatabase);
begin
  Self.Create(AIBDB, NewTr(AIBDB), true);
end;

constructor TISQL.Create(ATr:TIBTransaction);
begin
  Self.Create(DefaultIBDB, Atr);
end;

constructor TISQL.Create(AISQL:TISQL);
begin
  Self.Create(AISQL.Q.Database, AISQL.Tr);
end;

constructor TISQL.Create;
begin
  Self.Create(DefaultIBDB);
end;

destructor  TISQL.Destroy;
begin
  if Assigned(Picture) then Picture.Free;

	Q.Close;

  if AutoTrans then begin
    if Assigned(Tr) then begin
      if Tr.InTransaction then Tr.Commit;
//      Tr.Free;
    end;
  end;

	Q.Free;
	dec(ISQLCnt);
	inherited;
end;

function  NewTr(AIBDB:TIBDatabase = nil):TIBTransaction;
begin
  Result:=TIBTransaction.Create(nil);
  if Assigned(AIBDB) then begin
    Result.DefaultDatabase:=AIBDB;
  end else begin
    Result.DefaultDatabase:=DefaultIBDB;
  end;

  Result.Params.Add('nowait');
end;

procedure TISQL.Prepare;
begin
  if Q.Open then Q.Close;
  
  if AutoTrans then begin
    if not Tr.InTransaction then Tr.StartTransaction;
  end;
	Q.Prepare;
end;

procedure TISQL.Close;
begin
	Q.Close;
  if AutoTrans then begin
    if Tr.InTransaction then Tr.Commit;
  end;
end;

procedure TISQL.Open;
begin
  if AutoTrans then begin
    if not Tr.InTransaction then Tr.StartTransaction;
  end;

  Q.ExecQuery;
end;

procedure TISQL.Exec;
begin
  if AutoTrans then begin
    if not Tr.InTransaction then Tr.StartTransaction;
  end;

  Q.ExecQuery;
  if AutoTrans then begin
    Tr.Commit;
  end;
end;

procedure TISQL.StartTransaction;
begin
  Q.Transaction.StartTransaction;
end;

procedure TISQL.CommitRetaining;
begin
  Q.Transaction.CommitRetaining;
end;

procedure TISQL.Commit;
begin
  Q.Transaction.Commit;
end;

procedure TISQL.RollbackRetaining;
begin
  Q.Transaction.RollbackRetaining;
end;

procedure TISQL.Rollback;
begin
  Q.Transaction.Rollback;
end;

{$O-}
function  MR(money:currency):currency;  // money rounder
begin
	Result:=(Int(Abs(money)*MoneyPrecision + 0.5)/MoneyPrecision);
	if money < 0 then Result:=-Result;
end;
{$O+}

var
	GuardUniqueNumber :integer;
	UniqueNumber      :integer;

function GetUniqueNumber: integer;
asm
@@1:  MOV     EDX,1
			XCHG    GuardUniqueNumber,EDX
			OR      EDX,EDX
			JZ      @@2

			PUSH    0
			CALL    Sleep
			JMP     @@1

@@2:  MOV     EAX,UniqueNumber
			INC     EAX
			MOV     UniqueNumber,EAX
			MOV     GuardUniqueNumber,EDX
			RET
end;


{
function  TISQL.GetT (const FieldName:string):string;
var	bs:TBlobStream;
		p:PChar;
begin
	bs:=TBlobStream.Create(Q.FieldByName(FieldName). as TMemoField, bmRead);
	GetMem(p, bs.Size + 1);
	try
		bs.Read(p^, bs.Size);
		p[bs.Size]:=#0;
		Result:=p;
	finally
		FreeMem(p);
		bs.Free;
	end;
end;

procedure TISQL.SetT (const FieldName:string;v:string);
var p:PChar;
		len:integer;
begin
	len:=Length(v);
	GetMem(p, len + 1);
	try
		StrPLCopy(p, v, len);
		len:=Length(v);
    p[len]:=#0;
		Q.ParamByName(FieldName).SetBlobData(p, len);
	finally
		FreeMem(p);
	end;
end;
}
function	TISQL.GetActive:boolean;
begin
	Result:=Q.Open;
end;

function	TISQL.GetEOF	:boolean;
begin
	if Q.Open then begin
		Result:=Q.EOF;
	end else begin
		Result:=true;
	end;
end;
function	TISQL.GetBOF	:boolean;
begin
	if Q.Open then begin
		Result:=Q.BOF;
	end else begin
		Result:=true;
	end;
end;
function	TISQL.GetEmpty:boolean;
begin
	Result:=EOF and BOF;
end;

function	TISQL.GetSQL:TStrings;
begin
	Result:=Q.SQL;
end;

function  TISQL.CheckDataBase:boolean;
begin
//  Result:=Assigned(Q.Database) and (Q.Database.Connected);
//  Result:=Q.Database.Connected;
	Result:=true;
end;

procedure TISQL.New;
begin
  if AutoTrans then begin
    if Tr.InTransaction then Tr.Commit;
  end;

	Q.Close;
	Q.SQL.Clear;
	ss:='';
end;

procedure TISQL.Add(s:string);
begin
	Q.SQL.Add(s);
	ss:=ss+s;
end;

procedure TISQL.Next;
begin
	Q.Next;
end;

function  TISQL.RecordCount:integer;
begin
	Result:=Q.RecordCount;
end;

function  TISQL.GetI (const FieldName:string):integer;
begin
	Result:=Q.FieldByName(FieldName).AsInteger;
end;

procedure TISQL.SetI (const FieldName:string;v:integer);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
  Q.Params.ByName(FieldName).AsInteger:=v;
end;

function  TISQL.GetS (const FieldName:string):string;
begin
	Result:=Q.FieldByName(FieldName).AsString;
end;

procedure TISQL.SetS (const FieldName:string;v:string);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsString:=v;
end;

function  TISQL.GetF (const FieldName:string):extended;
begin
	if VarType(Q.FieldByName(FieldName).Value) = varDate then begin
  	Result:=Q.FieldByName(FieldName).AsDateTime;
  end else begin
  	Result:=Q.FieldByName(FieldName).AsDouble;
  end;
end;
procedure TISQL.SetF (const FieldName:string;v:extended);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsDouble:=v;
end;

function  TISQL.GetC (const FieldName:string):currency;
begin
	Result:=Q.FieldByName(FieldName).AsCurrency;
end;
procedure TISQL.SetC (const FieldName:string;v:currency);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsCurrency:=v;
end;

function  TISQL.GetM (const FieldName:string):currency;
begin
	Result:=MR(Q.FieldByName(FieldName).AsCurrency);
end;

procedure TISQL.SetM (const FieldName:string;v:currency);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsCurrency:=v;
end;

function  TISQL.GetD (const FieldName:string):TDateTime;
begin
	Result:=Q.FieldByName(FieldName).AsDateTime;
end;

procedure TISQL.SetD (const FieldName:string;v:TDateTime);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsDateTime:=v;
end;

function  TISQL.GetB (const FieldName:string):boolean;
begin
	Result:=boolean(Q.FieldByName(FieldName).AsInteger);
end;

procedure TISQL.SetB (const FieldName:string;v:boolean);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsInteger:=integer(v);
end;

function  TISQL.GetV (const FieldName:string):variant;
begin
	Result:=Q.FieldByName(FieldName).AsVariant;
end;

procedure TISQL.SetV (const FieldName:string;v:variant);
begin
  if AutoTrans and (not Tr.InTransaction) then Tr.StartTransaction;
	Q.Params.ByName(FieldName).AsVariant:=v;
end;

function  TISQL.GetG (const FieldName:string):TGraphic;
var stream:TMemoryStream;
begin
  if not Assigned(Picture) then Picture:=TPicture.Create;

  stream:=TMemoryStream.Create();
  try
    Q.FieldByName(FieldName).SaveToStream(stream);
    Picture.Graphic.LoadFromStream(stream);
    Result:=Picture.Graphic;
  finally
    stream.Free();
  end;
end;

{
function  TISQL.GetDataSet:TDataSet;
begin
	Result:=Q;
end;
}
function	  TISQL.GetTr:TIBTransaction;
begin
	Result:=Q.Transaction;
end;

procedure   TISQL.SetTr(ATr:TIBTransaction);
begin
  if Q.Transaction.Owner = Q then begin
    Q.Transaction.Free;
  end;
  AutoTrans:=false;
	Q.Transaction:=Atr;
end;

function    TISQL.GetFields(const Idx: Integer):TIBXSQLVAR;
begin
  Result:=Q.Fields[Idx];
end;

function    TISQL.Current: TIBXSQLDA;
begin
  Result := Q.Current;
end;

procedure   TISQL.BatchOutput(OutputObject: TIBBatchOutput);
begin
  Q.BatchOutput(OutputObject);
end;

procedure   TISQL.BatchInput (InputObject:  TIBBatchInput);
begin
  Q.BatchInput (InputObject);
end;

end.

